/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#ifndef _FRUSTRUM_H_
#define _FRUSTRUM_H_
#include "thtypes.h"

class FrustrumObject{
  protected:
    static class Frustrum *ActiveFrustrum;
  public:
    #ifdef __SSE__
    static void* operator new (size_t size);
    static void* operator new[] (size_t size);
    static void operator delete (void *pMem);
    static void operator delete[] (void *pMem);
    #endif

    virtual bool IsInFrustrum()=0;
    static void SetActiveFrustrum(Frustrum *ptr);
};

class FrustrumObjectSphere : public FrustrumObject{
  friend class FrustrumCamera;
  private:
    vec4 v4Position;
    float fRadius;
  public:
    bool IsInFrustrum();
    void SetPosition(vec4 Position);
    void SetRadius(float Radius);
    void Draw();
};

class FrustrumObjectBox : public FrustrumObject{
  friend class FrustrumCamera;
  private:
    vec4 v4Points[8];
    vec4 v4TransPoints[8];
  public:
    bool IsInFrustrum();
    void SetSize(float width, float length, float height);
    void SetSize(const vec4& min, const vec4& max);
    void Transform(mat4a Matrix);
    void Draw();
};

#define TH_PLANE_RIGHT  0
#define TH_PLANE_LEFT   1
#define TH_PLANE_BOTTOM 2
#define TH_PLANE_TOP    3
#define TH_PLANE_FAR    4
#define TH_PLANE_NEAR   5

class Frustrum{
  protected:
    vec4 Planes[6];
  public:
    virtual void Update()=0;
    virtual bool Test(class FrustrumObjectSphere *ptr)=0;
    virtual bool Test(class FrustrumObjectBox *ptr)=0;
    virtual void Points(vec4 *ptr)=0;
    void MovePlane(int num, float value);
};

class FrustrumCamera : public Frustrum{
  public:
    void Update();
    bool Test(class FrustrumObjectSphere *ptr);
    bool Test(class FrustrumObjectBox *ptr);
    void Points(vec4 *ptr);
};

#endif
