/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#ifndef _KAMERAEDITOR_H_
#define _KEMERAEDITOR_H_
#include "gui.h"
#include "shader.h"
#include "vektoren.h"
#include "timer.h"
#include "spline.h"
#include "camera.h"

enum CameraEditorEditState {
  POSITION,DIRECTION,UP
};

#ifdef Button0
#undef Button0
#endif

#ifdef Button1
#undef Button1
#endif

#ifdef Button2
#undef Button2
#endif

#ifdef Button3
#undef Button3
#endif

class CameraEditor {
  private:
    CubicSpline Positions;
    CubicSpline Directions;
    CubicSpline UpVectors;
    vec4 *DrawPositions;
    vec4 *DrawDirections;
    CameraEditorEditState EditState;
    camera *pCam;
    Timer *pTimer;
    Zeitpunkt PlayStart;
    bool bRendering,bPlaying;
    int iRenderingAnz;

    GuiWindow *Window;
    GuiTextInput *InputX;
    GuiTextInput *InputY;
    GuiTextInput *InputZ;
    GuiTextInput *InputT;
    GuiTextInput *InputT2;
    GuiText *TextX,*TextY,*TextZ,*TextT,*TextMode,*TextFehler;
    GuiSelect *Select1;
    GuiButton *ButtonReplace,*ButtonGoto,*ButtonDelete;
    GuiButton *Button0,*Button1,*Button2,*Button3,*ButtonSwitch,*ButtonPlay,*ButtonDraw,*ButtonSave,*ButtonLoad;

    void UpdateInputFields();
    void InsertCurrent();
    bool IsValidNumber(GuiTextInput *ptr);
    void Update();

  public:
    CameraEditor();
    bool Init(Gui *pGui,camera *Cam, Timer *pMainTimer);
    //bool CameraEditorLoadFromFile(const char *filename);
    float UpdateCam();
    void Draw();
    void Play();
    bool LoadFromFile(const char* filename);
    bool SaveToFile(const char* filename);

    inline bool IsPlaying() const {return bPlaying;}

    //Callback Functions
    friend void CameraEditorOnButton0(void *Creator);
    friend void CameraEditorOnButton1(void *Creator);
    friend void CameraEditorOnButton2(void *Creator);
    friend void CameraEditorOnButton3(void *Creator);
    friend void CameraEditorOnButtonSwitch(void *Creator);
    friend void CameraEditorOnButtonPlay(void *Creator);
    friend void CameraEditorOnButtonSave(void *Creator);
    friend void CameraEditorOnButtonLoad(void *Creator);
    friend void CameraEditorOnSelect1Change(void *Creator);
    friend void CameraEditorOnInputXChange(void *Creator);
    friend void CameraEditorOnInputYChange(void *Creator);
    friend void CameraEditorOnInputZChange(void *Creator);
    friend void CameraEditorOnInputTChange(void *Creator);
    friend void CameraEditorOnButtonGoto(void *Creator);
    friend void CameraEditorOnButtonReplace(void *Creator);
    friend void CameraEditorOnButtonDelete(void *Creator);
};

#endif
