/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#ifndef _LANDSCAPE_H_
#define _LANDSCAPE_H_
#include "thtypes.h"
#include "texture.h"
#include "shader.h"
#include "frustrum.h"
#include "vertexbuffer.h"

struct LandQuad  {
  int iRes;
  int iLastRes;
  int iRenderRes;
  float fBlendTime;
  float fHeight;
  float fPosX;
  float fPosY;
  bool bVisible;
  int *iSizesMain[4];
  int *iStartsMain[4];
  int iDivisions[4];
  FrustrumObjectBox BoundingBox;
  bool bBlendLarger;
  float fBlendFaktor;
};

class Landscape {
  private:
    Texture2D *HeightMap;
    int iDivisionDetail;
    vec4 vPosition;
    float fSize;
    float fHeight;
    LandQuad *Quads;
    int iQuadWidth;
    int iSizes[4];
    float fMaxDistance;
    GLuint VertexBufferMain;
    GLuint VertexBufferBorders;

    vec4 *PlacePositions;
    #ifdef __SSE__
    vec4 *PlacePositionsO;
    #endif
    int iVertexSize;

    char *lpcVertexData;
    char *lpcTexCoordData;
    char *lpcNormalData;
    char *lpcSubDivData;

    bool bDataLoaded;

    int VertCount(int faktor);
    void WriteQuad(int size, int faktor, int& iVertNum, int iRes);

  public:
    Landscape(int Detail);
    ~Landscape();
    bool LoadHeightMapTGA(const char *name);
    void CreateQuads(int size);
    void TestDraw(Shader *ptr);
    void DrawBoxes();
    void Draw();
    void TestGenerate(vec4 CamPos, double TimeDiff);
    void SetSize(float value);
    void SetHeight(float value);
    void SetPosition(vec4 Position);
    void SetPosition(float x, float y, float z);
    void SetDistance(float value);
    vec4 GetPosition(float x, float y);
    vec4 GetNormal(float x, float y);
};

#endif
