/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#ifndef _MANAGER_H_
#define _MANAGER_H_
#include "thEngine.h"
#include "xmlshader.h"
#include <vector>

class Texture2DManager{
  private:
    std::vector<Texture2D*> Items;
    bool bError;
    FILE *datei;

  public:
    Texture2DManager();
    ~Texture2DManager();
    Texture2D* LoadTGA(const char *filename, char options);
    Texture2D* CreateEmpty(int width, int height, GLenum DataFormat, GLenum InternalFormat, char options);
    Texture2D* GetTexture(int num);
    inline bool Error() const {return bError;}
};

class ModelManager {
  private:
    std::vector<Model*> Items;
    bool bError;
    FILE *datei;

  public:
    ModelManager();
    ~ModelManager();
    Model* Load3DS(const char *filename, short object_number, char options);
    Model* LoadBillboards3DS(const char *filename, short object_number);
    Model* LoadBillboardsSchrompf(const char *filename,vec4 center);
    Model* LoadSchrompf(const char *filename, char options);
    inline bool Error() const {return bError;}
};

class XmlShaderManager {
  private:
    std::vector<XMLShader*> Items;
    bool bError;
    FILE *datei;

  public:
    XmlShaderManager();
    ~XmlShaderManager();
    XMLShader* LoadXml(const char *filename);
    inline bool Error() const {return bError;}
};

#endif // _MANAGER_H_

