/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#ifndef _MODEL_H_
#define _MODEL_H_
#include "GLee.h"
#include "thtypes.h"
#include "vertexbuffer.h"

#define TH_COLOR              0x01
#define TH_TEXCOORDS          0x02
#define TH_NORMALS            0x04
#define TH_CONNECTIVITY       0x08
#define TH_NORMALS_PER_FACE   0x10
#define TH_NORMALS_PER_VERTEX 0x20
#define TH_TANGENT            0x40
#define TH_BINORMAL           0x80

class Model {
  private:
    char *cName;
    unsigned int iNumVertices;
    unsigned int iNumFaces;
    unsigned int iNumPositions;
    facev4 *Faces;
    facev4con *Connectivity;
    vertexv4 *Vertices;
    vertexposv4 *Positions;
    ggroup *GGroups;
    bool bHasTexcoords;
    bool bHasConnectivity;
    bool bHasTangentBinormal;
    bool bHasColor;
    bool bHasNormals;

    VertexBuffer *vbuffer;

    void GenData(char options);
    unsigned int FindGGroup(unsigned int uiVertNum, unsigned short usGGroupNum);

  public:
    Model();
    ~Model();
    bool LoadModel3DS(const char *filename,short object_number,char options);
    bool LoadBillboards3DS(const char *filename, short object_number);
    bool LoadBillboardsSchrompf(const char *filename,vec4 center);
    bool LoadSchrompf(const char *filename,char options);
    void Draw(char options);
    void DrawShadow(vec4 v4ModelPos, vec4 v4LightPos);
    void SetLocation(char Target,GLint Location);
    inline VertexBuffer* GetVertexBuffer(){return vbuffer;}
};

#endif
