/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#ifndef _SOUND_H_
#define _SOUND_H_
#include <AL/al.h>
#include <vorbis/vorbisfile.h>
#include <vector>
#include "vektoren.h"
#include "camera.h"

class Sound {
  protected:
    ALint aluiState;
    ALuint aluiBuffer1;
    ALuint aluiSource;
    ALenum format;
    ALsizei freq;
    bool bDataLoaded;

  public:
    Sound();
    void SetPitch(float value);
    void SetPosition(float x, float y, float z);
    void SetPosition(float *array);
    void SetPosition(vec4 Position);
    void SetVelocity(float x, float y, float z);
    void SetVelocity(float *array);
    void SetVelocity(vec3 Velocity);
    void SetDirection(float *array);
    void SetDirection(vec3 Direction);
    void SetRolloff(float value);
    void SetRelative(bool value);
    virtual void SetRepeat(bool value);
    void SetVolume(float value);
    void Play();
    void Stop();
    bool Playing();

};

class SoundOgg : public Sound{
  private:
    std::vector<char> DataBuffer;

  public:
    ~SoundOgg();
    bool LoadOgg(const char * filename);
};

class SoundOggStream : public Sound{
  private:
    ALuint aluiBuffer2;
    OggVorbis_File oggFile;
    int Stream(ALuint Buffer);
    bool bRepeat;

  public:
    ~SoundOggStream();
    bool LoadOgg(const char * filename);
    void Close();
    bool Update();
    void SetRepeat(bool value);
};

class Listener : public camera {
  public:
    void Call();
    void SetGain(float value);
    void SetVelocity(float* array);
};
#endif
