/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#ifndef _THENGINE_H_
#define _THENGINE_H_
#include "GLee.h"
#include <GL/glu.h>
#include "timer.h"
#include "my3d.h"
#include "thtypes.h"
#include "texture.h"
#include "model.h"
#include "light.h"
#include "vektoren.h"
#include "tinyxml.h"
#include "manager.h"
#include <vector>
#include <string>
#ifdef __linux
#include <GL/glx.h>
#include <iostream>
#endif

#define th_renderstate_model        0
#define th_renderstate_model_blend  1
#define th_renderstate_text         2
#define th_renderstate_effect       3
#define th_renderstate_lines        4
#define number_renderstates         5

/////////////////////////////////
//Comment out for no Debugging
//#define TH_DEBUG

#ifndef TH_ENGINE_MAIN_FILE
extern Zeitpunkt Runtime;
extern Timer *timer;
extern int iWidth;
extern int iHeight;
extern TiXmlDocument *XmlConfigDoc;
extern double dTimeDiff;
extern bool arbMultisampleSupported;
extern bool bAntialiasing,bIsFullscreen,bShowFps;
extern float fFarClipping;
#ifdef __linux
extern Display*				m_Display;
extern Window					m_XWindow;
#endif
extern class Texture2DManager *TexMngr;
extern class ModelManager *MdlMngr;
extern class XmlShaderManager *ShdrMngr;
#endif

struct FontGlyph {
  int iWidth;
  int iHeight;
  int iTop;
  int iLeft;
  int iRight;
  float fMinTexX,fMaxTexX;
  float fMinTexY,fMaxTexY;
  unsigned char *lpData;
};

struct ThFont {
  int iStartIndex;
  int iAnzChars;
  FontGlyph *Glyphs;
  Texture2D *FontTexture;
  GLuint gluiFontId;
  std::string Name;
  int iMaxHeight;
};

typedef void (*KeyDownFunc)();
typedef void (*MouseMoveFunc)(int,int);
typedef void (*MouseMoveFuncGui)(int,int,int,int);
typedef void (*MouseButtonFunc)(int,int,int,int,int);
typedef void (*MouseWheelFunc)(int);
typedef void (*TextFunc)(char);

void PreDraw();
void Draw();
bool Init();
void DeInit();

void SetTime();

void SetRenderstate(int renderstate);
void ViewOrtho();
void ViewPerspective();
int LoadFont(const char *filename, int size, const char *Name);
int GetFont(const char *Name);
void glPrint(int set, const char *fmt, ...);
int GetMaxFontHeight(int num);
void GetTextSize(int& width, int& height,int set,const char*fmt, ...);
GLvoid KillFont();
#ifdef __linux
void ShowCursor(bool value);
#endif

void ThErrorMessage(const char *Title, const char *Text);
void ThError(const char *Title, const char *Text);

void SetKeyDownFunc(unsigned int Key,KeyDownFunc Func);
void SetKeyUpFunc(unsigned int Key,KeyDownFunc Func);
void SetMouseMoveFuncAbsolute(MouseMoveFunc Func);
void SetMouseMoveFuncRelative(MouseMoveFunc Func);
void SetMouseButtonDownFunc(MouseButtonFunc Func);
void SetMouseButtonUpFunc(MouseButtonFunc Func);
void SetMouseMoveFunc(MouseMoveFuncGui Func);
void SetMouseWheelFunc(MouseWheelFunc Func);
void SetTextFunc(TextFunc Func);
bool KeyIsPressed(unsigned int Key);

void ThGetGLError(const char *title);

//void GetXMLOption(const char *name, int& value);
//void GetXMLOption(const char *name, float& value);
//bool IsXMLOptionAviable(const char *name);

#endif
