/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#ifndef _VALUEEDITOR_H_
#define _VALUEEDITOR_H_
#include "gui.h"
#include "vektoren.h"
#include "shader.h"
#include "tinyxml.h"

struct ValueEditorValue {
  ShaderConstant *Constant;
  int type;
  float fValue;
  float fMin,fMax;
  vec4 v4Value;
  vec4 v4Min,v4Max;
};

struct ThConstantInfoStruct {
  int num;
  TiXmlNode *lpNode;
};

class ValueEditor {
  private:
    std::vector<ValueEditorValue> Values;
    std::vector<ThConstantInfoStruct> Constants;
    TiXmlDocument *XmlConstantFile;

    //GuiElements
    GuiWindow *ValueEditorWindow;
    GuiSliderVertical *Slider0,*Slider1,*Slider2,*Slider3,*Slider4;
    GuiSelect *Select1;
    GuiText *Text1;
    GuiColorView *ColorView1;
    GuiButton *SaveButton;

    bool XmlLoadFloat(TiXmlNode *pChild, float& value);
    void XmlSetColor(TiXmlNode *pChild,vec4& Color);
    void XmlSetFloat(TiXmlNode *pChild,float& value);


  public:
    ValueEditor();
    bool Init(Gui *pGui);
    bool LoadConstants(const char *filename);
    int AddFloat(const char *name, ShaderConstant *ptr,float Min,float Max);
    int AddVec4(const char *name, ShaderConstant *ptr,vec4 Min, vec4 Max);

    //Callback Functions
    friend void ValueEditorSlider0Change(void* Creator);
    friend void ValueEditorSlider1Change(void* Creator);
    friend void ValueEditorSlider2Change(void* Creator);
    friend void ValueEditorSlider3Change(void* Creator);
    friend void ValueEditorSlider4Change(void* Creator);
    friend void ValueEditorSelect1Change(void* Creator);
    friend void ValueEditorSaveXml(void* Creator);
};



#endif
