/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#ifndef _VERTEXBUFFER_H_
#define _VERTEXBUFFER_H_
#include <vector>
#include "GLee.h"
#include "thtypes.h"

#define TH_VB_COLOR         0x00000001
#define TH_VB_NORMAL        0x00000002
#define TH_VB_TEXCOORD0     0x00000004
#define TH_VB_TEXCOORD0_2   0x00000004
#define TH_VB_TEXCOORD0_3   0x00000008
#define TH_VB_TEXCOORD0_4   0x00000010
#define TH_VB_TEXCOORD1     0x00000020
#define TH_VB_TEXCOORD1_2   0x00000020
#define TH_VB_TEXCOORD1_3   0x00000040
#define TH_VB_TEXCOORD1_4   0x00000080
#define TH_VB_TEXCOORD2     0x00000100
#define TH_VB_TEXCOORD2_2   0x00000100
#define TH_VB_TEXCOORD2_3   0x00000200
#define TH_VB_TEXCOORD2_4   0x00000400
#define TH_VB_TEXCOORD3     0x00000800
#define TH_VB_TEXCOORD3_2   0x00000800
#define TH_VB_TEXCOORD3_3   0x00001000
#define TH_VB_TEXCOORD3_4   0x00002000
#define TH_VB_ATTRIBUTE0    0x00004000
#define TH_VB_ATTRIBUTE0_1  0x00004000
#define TH_VB_ATTRIBUTE0_2  0x00008000
#define TH_VB_ATTRIBUTE0_3  0x00010000
#define TH_VB_ATTRIBUTE0_4  0x00020000
#define TH_VB_ATTRIBUTE1    0x00040000
#define TH_VB_ATTRIBUTE1_1  0x00040000
#define TH_VB_ATTRIBUTE1_2  0x00080000
#define TH_VB_ATTRIBUTE1_3  0x00100000
#define TH_VB_ATTRIBUTE1_4  0x00200000
#define TH_VB_ATTRIBUTE2    0x00400000
#define TH_VB_ATTRIBUTE2_1  0x00400000
#define TH_VB_ATTRIBUTE2_2  0x00800000
#define TH_VB_ATTRIBUTE2_3  0x01000000
#define TH_VB_ATTRIBUTE2_4  0x02000000
#define TH_VB_DYNAMIC       0x04000000


typedef std::vector<int> IndexBufferType;

class VertexBuffer {
  private:
    GLuint gluiVertexBuffer;
    GLuint gluiIndexBuffer;
    GLenum glePrimitive;
    GLenum gleIndexType;
    bool bDuplicated;
    bool bDynamic;
    int iInstancedOptions;

    short TexCoordSize[4];
    short AttributeSize[3];
    char *lpcVertexData;
    char *lpcColorData;
    char *lpcNormalData;
    char *lpcTexCoordData[4];
    char *lpcAttributeData[3];
    GLint gliAttributeLocations[3];
    std::vector<float> fBuffer;
    std::vector<IndexBufferType*> IndexBuffers;
    int *IndexBufferSizes;
    int *IndexBufferBaseSizes;
    char **lpcIndexBufferPointer;

    bool bIndexed,bMultiIndexed;
    int iBytesPerVertex;
    int iNumVertices;
    int iNumIndexBuffers;

    int iInstancingBaseSize;
    int iUploadedDataSize;

  public:
    VertexBuffer(int options, GLenum Primitive,GLenum Index);
    VertexBuffer(const VertexBuffer& source);
    ~VertexBuffer();
    void AddVertexData(float *buffer, int len);
    void AddVertexData(vec4 data);
    void AddVertexData(vec3 data);
    void AddVertexData(vec2 data);
    void AddVertexData(float data);
    void AddIndexBuffer();
    void AddIndexData(int Buffer, int Vertex);
    void AddIndexData(int Buffer, int *Array, int len);
    bool UploadData();
    void FreeLocalData();
    void ResetLocalData();
    void FreeGlobalData();
    void ReserveVertexData(int num);
    void ReserveAdditionalVertexData(int num);
    void Draw(int options);
    void DrawRange(int options,int start, int size);
    unsigned int GetDataSize();
    unsigned int GetVerticesInBuffer();
    void GetMinMax(vec4& Min, vec4& Max);
    void SetAttributeLocation(int num, GLint Attribute);
    bool Duplicate(int count);

    void InitInstancedDraw(int options);
    void DeInitInstancedDraw();
    void DrawInstancedFallback(int count);
    void DrawInstanced(int count);
    void DrawMultiple();
    // TODO (Benjamin#2#): Index Array Optimierung\

};

#endif
