/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#ifndef _XMLSHADER_H_
#define _XMLSHADER_H_
#include "shader.h"
#include "scriptparser.h"
#include <string>
#include <vector>

struct BindableUniformInfo {
  int num;
  int size;
  GLenum mode;
};

struct XMLShaderConstantReference {
  std::string name;
  ShaderConstant *ptr;
};

class XMLShader {
  private:
    static ScriptVarVector OptionStack;
    static std::vector<XMLShaderConstantReference> ShaderConstants;
    Shader *theShader;
    ScriptVarVector VarStack;

    bool ProgressShaderSource(const char* code, std::string& source);
    bool GenerateShaderSource(std::string& source, TiXmlNode *father);
    bool ProgressUniforms(TiXmlNode *father,std::string& NameStack,std::vector<int>& InitValues,std::vector<BindableUniformInfo>& UniformBuffers,int& num);
    bool ProgressAttributes(TiXmlNode *father,std::string& NameStack, int& num);

  public:
    XMLShader();
    ~XMLShader();
    static bool LoadOptions(const char *filename);
    static bool GetOption(const char *name, float& value);
    static bool GetOption(const char *name, int& value);
    static void AddOption(const char *name, int value);
    static void AddOption(const char *name, float value);
    static ScriptVar* GetOption(const char *name);
    static ShaderConstant* AddShaderConstantFloat(const char *name);
    static ShaderConstant* AddShaderConstantVec3(const char *name);
    static ShaderConstant* AddShaderConstantVec4(const char *name);
    static ShaderConstant* AddShaderConstantMat4a(const char *name);
    static ShaderConstant* GetShaderConstant(const char *name);

    bool LoadFromFile(const char *lpcFileName);
    Shader *GetShader();

};

#endif
