/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "thEngine.h"
#include "fbo.h"
#include <stdio.h>

const GLenum TextureChannels[]={GL_TEXTURE0,GL_TEXTURE1,GL_TEXTURE2,GL_TEXTURE3,GL_TEXTURE4,GL_TEXTURE5,GL_TEXTURE6,GL_TEXTURE7};

FrameBufferTexture2D::FrameBufferTexture2D(int Width, int Height, int Options, int Format){
  //Init Vars
  lpcErrorMessage=NULL;
  fbo=0;
  DepthBuffer=0;
  StencilBuffer=0;
  ColorBuffer0=0;
  ColorBuffer1=0;
  ColorBuffer2=0;
  iWidth = Width;
  iHeight = Height;

  int anz=0;
  if(Options & TH_COLOR0)
    anz++;
  if(Options & TH_COLOR1)
    anz++;
  if(Options & TH_COLOR2)
    anz++;
  if(Options & TH_COLOR3)
    anz++;

  int MaxColorBuffers;
  glGetIntegerv(GL_MAX_COLOR_ATTACHMENTS_EXT,&MaxColorBuffers);
  if(MaxColorBuffers < anz){
    lpcErrorMessage = new char[50];
    sprintf(lpcErrorMessage,"Nur %i Color Buffer anhängbar\n%i benötigt",MaxColorBuffers,anz);
    return;
  }

  //Frame Buffer Object erstellen
  glGenFramebuffersEXT(1, &fbo);
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fbo);

  //Add Depth Buffer
  if(Options & TH_DEPTH_BUFFER){
    glGenRenderbuffersEXT(1, &DepthBuffer);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, DepthBuffer);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, Width, Height);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, DepthBuffer);
    bDepthBufferTexture=false;
  }

  if(Options & TH_DEPTH_BUFFER_TEXTURE_16){
    glGenTextures(1,&DepthBuffer);
    glBindTexture(GL_TEXTURE_2D, DepthBuffer);
    glTexImage2D(GL_TEXTURE_2D,0,GL_DEPTH_COMPONENT16,Width,Height,0,GL_DEPTH_COMPONENT,GL_UNSIGNED_SHORT,NULL);
    #ifdef TH_FBO_DEBUG
      ThGetGLError("Create Depthtexture");
    #endif
    //Set Texture Parameters
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    #ifdef TH_FBO_DEBUG
      ThGetGLError("Set Depthtexture Parameters");
    #endif
    //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
    //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LESS);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_TEXTURE_2D,DepthBuffer,0);
    #ifdef TH_FBO_DEBUG
      ThGetGLError("Attach Texture to Depth Buffer");
    #endif
    bDepthBufferTexture = true;
  }

  if(Options & TH_DEPTH_BUFFER_TEXTURE_24){
    glGenTextures(1,&DepthBuffer);
    glBindTexture(GL_TEXTURE_2D, DepthBuffer);
    glTexImage2D(GL_TEXTURE_2D,0,GL_DEPTH_COMPONENT24,Width,Height,0,GL_DEPTH_COMPONENT,GL_UNSIGNED_BYTE,NULL);
    //Set Texture Parameters
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
    //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LESS);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_TEXTURE_2D,DepthBuffer,0);
    bDepthBufferTexture = true;
  }

  if(Options & TH_DEPTH_BUFFER_TEXTURE_32){
    glGenTextures(1,&DepthBuffer);
    glBindTexture(GL_TEXTURE_2D, DepthBuffer);
    glTexImage2D(GL_TEXTURE_2D,0,GL_DEPTH_COMPONENT32,Width,Height,0,GL_DEPTH_COMPONENT,GL_UNSIGNED_INT,NULL);
    //Set Texture Parameters
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
    //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LESS);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_TEXTURE_2D,DepthBuffer,0);
    bDepthBufferTexture = true;
  }

  if(anz <= 0){
    glDrawBuffer( GL_NONE );
    glReadBuffer( GL_NONE );
  }

  GLenum format1,format2, format3;
  switch(Format){
    case TH_RGB8:
      format1 = GL_RGB8;
      format2 = GL_RGB;
      format3 = GL_UNSIGNED_BYTE;
      break;
    case TH_RGB16:
      format1 = GL_RGB16;
      format2 = GL_RGB;
      format3 = GL_UNSIGNED_SHORT;
      break;
    case TH_RGBA8:
      format1 = GL_RGBA8;
      format2 = GL_RGBA;
      format3 = GL_UNSIGNED_BYTE;
      break;
    case TH_RGBA16:
      format1 = GL_RGBA16;
      format2 = GL_RGBA;
      format3 = GL_UNSIGNED_SHORT;
      break;
    default:
      lpcErrorMessage = new char[50];
      sprintf(lpcErrorMessage,"Falsches Pixelformat %X",Format);
      return;
  }
  if(Options & TH_COLOR0){
    glGenTextures(1, &ColorBuffer0);
    glBindTexture(GL_TEXTURE_2D, ColorBuffer0);
    //Allocate Sotrage Memory
    glTexImage2D(GL_TEXTURE_2D, 0, format1, Width, Height, 0, format2, format3, NULL);
    //Set Texture Parameters
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    //Attach to Framebuffer
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, ColorBuffer0, 0);
  }

  if(Options & TH_COLOR1){
    glGenTextures(1, &ColorBuffer1);
    glBindTexture(GL_TEXTURE_2D, ColorBuffer1);
    glTexImage2D(GL_TEXTURE_2D, 0, format1, Width, Height, 0, format2, format3, NULL);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_TEXTURE_2D, ColorBuffer1, 0);
  }

  if(Options & TH_COLOR2){
    glGenTextures(1, &ColorBuffer2);
    glBindTexture(GL_TEXTURE_2D, ColorBuffer2);
    glTexImage2D(GL_TEXTURE_2D, 0, format1, Width, Height, 0, format2, format3, NULL);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT2_EXT, GL_TEXTURE_2D, ColorBuffer2, 0);
  }

  Result = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
  if(Result != GL_FRAMEBUFFER_COMPLETE_EXT){
    lpcErrorMessage = new char[200];
    switch(Result){
      case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
        sprintf(lpcErrorMessage,"Fehler: Unvollständiger Anhang");
        break;
      case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
        sprintf(lpcErrorMessage,"Fehler: Framebuffer nicht komplett, fehlender Anhang");
        break;
      case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
        sprintf(lpcErrorMessage,"Fehler: Dimensionen des Framebuffers fehlerhaft");
        break;
      case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
        sprintf(lpcErrorMessage,"Fehler: Format des Framebuffers fehlerhaft");
        break;
      case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
        sprintf(lpcErrorMessage,"Fehler: kein Zeichen-Buffer vorhanden!");
        break;
      case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
        sprintf(lpcErrorMessage,"Fehler: kein Read-Buffer vorhanden!");
        break;
      case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
        sprintf(lpcErrorMessage,"Fehler: Framebuffer wird nicht unterstüzt!");
        break;
    }
  }
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

FrameBufferTexture2D::~FrameBufferTexture2D(){
  if(fbo != 0){
    glDeleteFramebuffersEXT(1, &fbo);
  }
  if(ColorBuffer0 != 0){
    glDeleteTextures(1, &ColorBuffer0);
  }
  if(ColorBuffer1 != 0){
    glDeleteTextures(1,&ColorBuffer1);
  }
  if(ColorBuffer2 != 0){
    glDeleteTextures(1,&ColorBuffer2);
  }
  if(StencilBuffer != 0){
    glDeleteRenderbuffersEXT(1, &StencilBuffer);
  }
  if(DepthBuffer != 0){
    if(bDepthBufferTexture)
      glDeleteTextures(1, &DepthBuffer);
    else
      glDeleteRenderbuffersEXT(1, &DepthBuffer);
  }
  if(lpcErrorMessage != NULL){
    delete []lpcErrorMessage;
  }
}

void FrameBufferTexture2D::Use(){
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glPushAttrib(GL_VIEWPORT_BIT);
  glViewport(0,0,iWidth, iHeight);
}

void FrameBufferTexture2D::Unuse(){
  glPopAttrib();
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

void FrameBufferTexture2D::SwitchTo(){
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glViewport(0,0,iWidth, iHeight);
}

void FrameBufferTexture2D::Bind(short buff){
  switch(buff){
    case -1:
      glBindTexture(GL_TEXTURE_2D, DepthBuffer);
      break;
    case 0:
      glBindTexture(GL_TEXTURE_2D, ColorBuffer0);
      break;
    case 1:
      glBindTexture(GL_TEXTURE_2D, ColorBuffer1);
      break;
    case 2:
      glBindTexture(GL_TEXTURE_2D, ColorBuffer2);
      break;
  }
}

void FrameBufferTexture2D::Bind(short buff,short num){
  if(num < 0)
    num=0;
  if(num > 7)
    num=7;
  glActiveTexture(TextureChannels[num]);
  switch(buff){
    case -1:
      glBindTexture(GL_TEXTURE_2D, DepthBuffer);
      break;
    case 0:
      glBindTexture(GL_TEXTURE_2D, ColorBuffer0);
      break;
    case 1:
      glBindTexture(GL_TEXTURE_2D, ColorBuffer1);
      break;
    case 2:
      glBindTexture(GL_TEXTURE_2D, ColorBuffer2);
      break;
  }
}

void FrameBufferTexture2D::BindToShader(short buff, short num){
  if(num < 0)
    num=0;
  if(num > 16)
    num=16;
  glActiveTextureARB(GL_TEXTURE0_ARB + num);
  switch(buff){
    case -1:
      glBindTexture(GL_TEXTURE_2D, DepthBuffer);
      break;
    case 0:
      glBindTexture(GL_TEXTURE_2D, ColorBuffer0);
      break;
    case 1:
      glBindTexture(GL_TEXTURE_2D, ColorBuffer1);
      break;
    case 2:
      glBindTexture(GL_TEXTURE_2D, ColorBuffer2);
      break;
  }
}

void FrameBufferTexture2D::SetTarget(short num){
  glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT + num);
}

char* FrameBufferTexture2D::Error(){
  return lpcErrorMessage;
}

////////////////////////////////////////////////////

FrameBufferCubeMap::FrameBufferCubeMap(int Width, int Height, int Options, int Format){
  fbo = 0;
  DepthBuffer=0;
  lpcErrorMessage=NULL;
  iWidth = Width;
  iHeight = Height;
  CubeMap=0;

  int MaxColorBuffers;
  glGetIntegerv(GL_MAX_COLOR_ATTACHMENTS_EXT,&MaxColorBuffers);
  if(MaxColorBuffers < 4){
    lpcErrorMessage = new char[50];
    sprintf(lpcErrorMessage,"Nur %i Color Buffer anhängbar\n4 benötigt",MaxColorBuffers);
    return;
  }

  //glEnable(GL_TEXTURE_CUBE_MAP_ARB);

  //Gen Depth Buffer
  glGenRenderbuffersEXT(1, &DepthBuffer);
  glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, DepthBuffer);
  glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, Width, Height);

  //Gen Cube Map
  GLenum format1,format2, format3;
  switch(Format){
    case TH_RGB8:
      format1 = GL_RGB8;
      format2 = GL_RGB;
      format3 = GL_UNSIGNED_BYTE;
      break;
    case TH_RGB16:
      format1 = GL_RGB16;
      format2 = GL_RGB;
      format3 = GL_UNSIGNED_SHORT;
      break;
    case TH_RGBA8:
      format1 = GL_RGBA8;
      format2 = GL_RGBA;
      format3 = GL_UNSIGNED_BYTE;
      break;
    case TH_RGBA16:
      format1 = GL_RGBA16;
      format2 = GL_RGBA;
      format3 = GL_UNSIGNED_SHORT;
      break;
    default:
      lpcErrorMessage = new char[50];
      sprintf(lpcErrorMessage,"Falsches Pixelformat %X",Format);
      return;
  }

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glGenTextures(1, &CubeMap);
  glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, CubeMap);

  glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  for(int i=0;i<6;i++){
    glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB + i, 0, format1, Width, Height, 0, format2, format3, NULL);
  }

  // FBO
  glGenFramebuffersEXT(1, &fbo);
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fbo);

  //Attach Depth Buffer
  glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, DepthBuffer);

  //Attach Color Buffer
  glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB, CubeMap, 0);
  //glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB, CubeMap, 0);
  //glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT2_EXT, GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB, CubeMap, 0);
  //glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT3_EXT, GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB, CubeMap, 0);

  Result = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
  if(Result != GL_FRAMEBUFFER_COMPLETE_EXT){
    lpcErrorMessage = new char[200];
    switch(Result){
      case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
        sprintf(lpcErrorMessage,"Fehler: Unvollständiger Anhang");
        break;
      case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
        sprintf(lpcErrorMessage,"Fehler: Framebuffer nicht komplett, fehlender Anhang");
        break;
      case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
        sprintf(lpcErrorMessage,"Fehler: Dimensionen des Framebuffers fehlerhaft");
        break;
      case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
        sprintf(lpcErrorMessage,"Fehler: Format des Framebuffers fehlerhaft");
        break;
      case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
        sprintf(lpcErrorMessage,"Fehler: kein Zeichen-Buffer vorhanden!");
        break;
      case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
        sprintf(lpcErrorMessage,"Fehler: kein Read-Buffer vorhanden!");
        break;
      case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
        sprintf(lpcErrorMessage,"Fehler: Framebuffer wird nicht unterstüzt!");
        break;
    }
    return;
  }

  //Unbind Framebuffer
  //glDisable(GL_TEXTURE_CUBE_MAP_ARB);
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}


FrameBufferCubeMap::~FrameBufferCubeMap(){
  if(fbo != 0){
    glDeleteFramebuffersEXT(1, &fbo);
  }
  if(CubeMap != 0){
    glDeleteTextures(1, &CubeMap);
  }
  if(DepthBuffer != 0){
    glDeleteRenderbuffersEXT(1, &DepthBuffer);
  }
  if(lpcErrorMessage != NULL){
    delete []lpcErrorMessage;
  }
}

void FrameBufferCubeMap::Use(){
  glPushAttrib(GL_VIEWPORT_BIT);
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glViewport(0,0,iWidth, iHeight);
}

void FrameBufferCubeMap::Unuse(){
  glPopAttrib();
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

void FrameBufferCubeMap::SwitchTo(){
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glViewport(0,0,iWidth, iHeight);
}

/*void FrameBufferCubeMap::SetTarget(short num){
  if(num < 0)
    num=0;
  if(num > 5)
    num=5;
  if(num < 4){
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo1);
    glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT + num);
  }
  else {
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo2);
    glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT + num - 4);
  }
}*/

void FrameBufferCubeMap::Bind(){
  glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, CubeMap);
}

void FrameBufferCubeMap::Bind(short num){
  if(num < 0)
    num=0;
  if(num > 7)
    num=7;
  glActiveTexture(TextureChannels[num]);
  glBindTexture(GL_TEXTURE_CUBE_MAP_ARB,CubeMap);
}

char* FrameBufferCubeMap::Error(){
  return lpcErrorMessage;
}
