/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "thEngine.h"
#include <stdio.h>
#include <vector>
#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_GLYPH_H
#include FT_BITMAP_H
#include <sstream>
#ifdef __linux
#include <stdarg.h>
#endif


std::vector<ThFont*> Fonts;

void KillFont()									// Delete The Font From Memory
{
  for(int i=0;i<Fonts.size();i++){
	glDeleteLists(Fonts[i]->gluiFontId,Fonts[i]->iAnzChars);
	delete []Fonts[i]->Glyphs;
  }
}

void glPrint(int set, const char *fmt, ...)				// Where The Printing Happens
{
	char	text[1024];								// Holds Our String
	va_list	ap;									// Pointer To List Of Arguments
	if (fmt == NULL || set >= Fonts.size() || set < 0)		// If There's No Text or Invalid Font
		return;									// Do Nothing

	va_start(ap, fmt);								// Parses The String For Variables
	    vsprintf(text, fmt, ap);							// And Converts Symbols To Actual Numbers
	va_end(ap);		                        // Results Are Stored In Text

    glPushMatrix();
    Fonts[set]->FontTexture->Bind();
	glPushAttrib(GL_LIST_BIT);				// Pushes The Display List Bits		( NEW )
	glListBase(Fonts[set]->gluiFontId - Fonts[set]->iStartIndex);					// Sets The Base Character to 32	( NEW )
	glCallLists(strlen(text),GL_UNSIGNED_BYTE, text);				// Write The Text To The Screen
	glPopAttrib();
    glPopMatrix();
}

int GetMaxFontHeight(int num){
  if(num < 0 || num >= Fonts.size())
    return -1;
  return Fonts[num]->iMaxHeight;
}

void GetTextSize(int& width, int& height,int set,const char*fmt, ...){
    char text[1024];
    int MaxY,MinY;
    va_list ap;
    if(fmt == NULL || set >= Fonts.size() || set < 0){
      width = -1;
      height = -1;
      return;
    }

    va_start(ap,fmt);
      vsprintf(text,fmt,ap);
    va_end(ap);

    width = 0;

    MaxY = Fonts[set]->Glyphs[text[0] - Fonts[set]->iStartIndex].iTop;
    MinY = Fonts[set]->Glyphs[text[0] - Fonts[set]->iStartIndex].iHeight - Fonts[set]->Glyphs[text[0] - Fonts[set]->iStartIndex].iTop;
    for(int i=0;i<strlen(text);i++){
      int num = text[i] - Fonts[set]->iStartIndex;
      width += Fonts[set]->Glyphs[num].iWidth + Fonts[set]->Glyphs[num].iLeft + Fonts[set]->Glyphs[num].iRight;
      MaxY = (Fonts[set]->Glyphs[num].iTop > MaxY) ? Fonts[set]->Glyphs[num].iTop : MaxY;
      MinY = (Fonts[set]->Glyphs[num].iTop - Fonts[set]->Glyphs[num].iHeight< MinY) ? Fonts[set]->Glyphs[num].iTop - Fonts[set]->Glyphs[num].iHeight: MinY;
    }
    height = MaxY - MinY;
}

void BuildChar(FT_Face& face, ThFont* lpFont , int num){
  unsigned char *buffer;
  char Buchstabe = (char)(num + lpFont->iStartIndex);

  //GLyph Laden
  FT_Load_Glyph( face, FT_Get_Char_Index( face, Buchstabe ), FT_LOAD_DEFAULT );
  FT_Glyph glyph;
  FT_Get_Glyph( face->glyph, &glyph );

  //In ein Bitmap konvertieren
  FT_Glyph_To_Bitmap( &glyph, ft_render_mode_normal, 0, 1 );
  FT_BitmapGlyph bitmap_glyph = (FT_BitmapGlyph)glyph;
  FT_Bitmap& bitmap=bitmap_glyph->bitmap;

  buffer = new unsigned char[2 * bitmap.rows * bitmap.width];
  for(int i=0;i<bitmap.rows * bitmap.width;i++){
    //buffer[i*2] = bitmap.buffer[i];
    buffer[i*2] = 255;
    buffer[i*2+1] = bitmap.buffer[i];
  }

  #ifdef TH_FONT_DEBUG
  FILE *datei;
  datei = fopen("font.log","a");
  int temp;
  fprintf(datei,"Char %c\n",Buchstabe);
  for(int i=0;i<bitmap.rows;i++){
    for(int j=0;j<bitmap.width;j++){
      temp = (bitmap.buffer[i*bitmap.width + j] > 128) ? 1 : 0;
      fprintf(datei,"%i",temp);
    }
    fprintf(datei,"\n");
  }
  #endif

  lpFont->Glyphs[num].lpData = buffer;
  lpFont->Glyphs[num].iWidth = bitmap.width;
  lpFont->Glyphs[num].iHeight = bitmap.rows;
  lpFont->Glyphs[num].iTop = bitmap_glyph->top;
  lpFont->Glyphs[num].iLeft = bitmap_glyph->left;
  lpFont->Glyphs[num].iRight = (face->glyph->advance.x >> 6) - bitmap_glyph->left - bitmap.width;
  #ifdef TH_FONT_DEBUG
  fprintf(datei,"Width: %i\nHeight: %i\nTop: %i\nLeft: %i\nRight: %i\n",lpFont->Glyphs[num].iWidth,lpFont->Glyphs[num].iHeight,lpFont->Glyphs[num].iTop,lpFont->Glyphs[num].iLeft,lpFont->Glyphs[num].iRight);
  fprintf(datei,"\n");
  fclose(datei);
  #endif
}

bool BuildTexture(ThFont *lpFont){
  int MaxWidth,MinY,MaxY,MaxHeight;
  MaxWidth = lpFont->Glyphs[0].iWidth + lpFont->Glyphs[0].iLeft + lpFont->Glyphs[0].iRight;
  MaxY = lpFont->Glyphs[0].iTop;
  MinY = lpFont->Glyphs[0].iHeight - lpFont->Glyphs[0].iTop;
  for(int i=1;i<lpFont->iAnzChars;i++){
    MaxWidth = (lpFont->Glyphs[i].iWidth + lpFont->Glyphs[i].iLeft + lpFont->Glyphs[i].iRight > MaxWidth) ? lpFont->Glyphs[i].iWidth + lpFont->Glyphs[i].iLeft + lpFont->Glyphs[i].iRight : MaxWidth;
    MaxY = (lpFont->Glyphs[i].iTop > MaxY) ? lpFont->Glyphs[i].iTop : MaxY;
    MinY = (lpFont->Glyphs[i].iTop - lpFont->Glyphs[i].iHeight< MinY) ? lpFont->Glyphs[i].iTop - lpFont->Glyphs[i].iHeight: MinY;
  }
  MaxHeight = MaxY - MinY;
  lpFont->iMaxHeight = MaxHeight;

  int size=64;
  int x,y;
  while(size < 2048){
    x=y=0;
    for(int i=0;i<lpFont->iAnzChars;i++){
      if((x + lpFont->Glyphs[i].iWidth + lpFont->Glyphs[i].iLeft + lpFont->Glyphs[i].iRight + 1) > size){
        y+=MaxHeight;
        x=0;
      }
      x+=lpFont->Glyphs[i].iWidth + lpFont->Glyphs[i].iLeft + lpFont->Glyphs[i].iRight + 1;
    }
    if(y > size)
      size*=2;
    else
      break;
  }
  if(size == 2048){
    ThErrorMessage("Font Error","Font Textur zu groß\n");
    return false;
  }

  lpFont->FontTexture = TexMngr->CreateEmpty(size,size,GL_LUMINANCE_ALPHA,GL_RGBA,TH_NEAREST);
  if(lpFont == NULL)
    return false;

  x=y=0;
  for(int i=0;i<lpFont->iAnzChars;i++){
    if((x + lpFont->Glyphs[i].iWidth + lpFont->Glyphs[i].iLeft + lpFont->Glyphs[i].iRight + 1) > size){
      y+=MaxHeight;
      x=0;
    }
    lpFont->FontTexture->Insert(x+lpFont->Glyphs[i].iLeft,y + MaxHeight - lpFont->Glyphs[i].iTop + MinY,lpFont->Glyphs[i].iWidth,lpFont->Glyphs[i].iHeight,lpFont->Glyphs[i].lpData);
    delete lpFont->Glyphs[i].lpData;
    float fStep = 1.0f / (float)size;;
    lpFont->Glyphs[i].fMinTexY = (float)y * fStep + fStep / 2.0f;
    lpFont->Glyphs[i].fMinTexX = (float)x * fStep + fStep / 2.0f;
    x+=lpFont->Glyphs[i].iWidth + lpFont->Glyphs[i].iLeft + lpFont->Glyphs[i].iRight;
    lpFont->Glyphs[i].fMaxTexY = (float)(y+MaxHeight-1) * fStep + fStep / 2.0f;
    lpFont->Glyphs[i].fMaxTexX = (float)(x-1) * fStep + fStep / 2.0f;
    x++;
    glNewList(lpFont->gluiFontId + i, GL_COMPILE);
    float GlyphWidth = lpFont->Glyphs[i].iWidth + lpFont->Glyphs[i].iRight + lpFont->Glyphs[i].iLeft;
    glBegin(GL_QUADS);
      glTexCoord2f(lpFont->Glyphs[i].fMinTexX,lpFont->Glyphs[i].fMinTexY);
      //glVertex2f(0.0,lpFont->Glyphs[i].fMaxTexY - lpFont->Glyphs[i].fMinTexY);
      glVertex2f(0.0,0.0);

      glTexCoord2f(lpFont->Glyphs[i].fMaxTexX,lpFont->Glyphs[i].fMinTexY);
      //glVertex2f(lpFont->Glyphs[i].fMaxTexX - lpFont->Glyphs[i].fMinTexX, lpFont->Glyphs[i].fMaxTexY - lpFont->Glyphs[i].fMinTexY);
      glVertex2f(GlyphWidth,0.0);

      glTexCoord2f(lpFont->Glyphs[i].fMaxTexX,lpFont->Glyphs[i].fMaxTexY);
      //glVertex2f(lpFont->Glyphs[i].fMaxTexX - lpFont->Glyphs[i].fMinTexX, 0.0);
      glVertex2f(GlyphWidth,MaxHeight);

      glTexCoord2f(lpFont->Glyphs[i].fMinTexX,lpFont->Glyphs[i].fMaxTexY);
      //glVertex2f(0.0,0.0);
      glVertex2f(0.0,MaxHeight);

    glEnd();
    glTranslatef(GlyphWidth,0.0,0.0);
    glEndList();
  }

  if(!lpFont->FontTexture->Reupload()){
    ThErrorMessage("Font Error","Failed uploading Font Texture");
    return false;
  }
  return true;
}

int LoadFont(const char *filename, int size, const char *Name){
    ThFont *temp;
    temp = new ThFont;
    std::ostringstream ErrorMessage;

    FT_Library library;
    if (FT_Init_FreeType( &library )){
      ThErrorMessage("Free Type Error","Couldn't Init Free Type Library");
      return -1;
    }

    FT_Face face;
    if (FT_New_Face( library, filename, 0, &face )){
      ErrorMessage << "Konnte Font '" << Name << "' aus '" << filename << "' nicht laden!";
      ThErrorMessage("Free Type Error",ErrorMessage.str().c_str());
      return -1;
    }

    FT_Set_Char_Size( face, size * 64, size * 64, 72, 72);

    temp->iStartIndex = 32;
    temp->iAnzChars = 96;
    temp->Glyphs = new FontGlyph[96];
    temp->gluiFontId = glGenLists(96);
    temp->Name = Name;

    //Building 96 Chars
    for(int i=0;i<96;i++){
      BuildChar(face,temp,i);
    }

    if(!BuildTexture(temp))
      return -1;

    Fonts.push_back(temp);

    //Habe fertig: alles freigeben
    FT_Done_Face(face);
    FT_Done_FreeType(library);
    return Fonts.size();
}

int GetFont(const char *Name){
  for(int i=0;i<Fonts.size();i++)
    if(Fonts[i]->Name == Name)
      return i;
  return -1;
}
