/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "frustrum.h"
#include "vektoren.h"


Frustrum* FrustrumObject::ActiveFrustrum;

void FrustrumObject::SetActiveFrustrum(Frustrum *ptr){
  ActiveFrustrum = ptr;
}

#ifdef __SSE__
void* FrustrumObject::operator new (size_t size){
  void *pMem = th_memalign(16,size);
  return pMem;
}

void* FrustrumObject::operator new[] (size_t size){
  void *pMem = th_memalign(16,size);
  return pMem;
}

void FrustrumObject::operator delete (void *pMem){
  th_free(pMem);
}

void FrustrumObject::operator delete[] (void *pMem){
  th_free(pMem);
}
#endif

void Frustrum::MovePlane(int num,float value){
  if(num < 0 || num > 5)
    return;
  Planes[num].w += value;
}

void FrustrumCamera::Update(){
  mat4a Clip,ModelView,Projection;

  glGetFloatv(GL_MODELVIEW_MATRIX,ModelView.f);
  glGetFloatv(GL_PROJECTION_MATRIX,Projection.f);

  Clip = Mul(ModelView,Projection);

  Planes[0].x = Clip.f[ 3] - Clip.f[ 0];
  Planes[0].y = Clip.f[ 7] - Clip.f[ 4];
  Planes[0].z = Clip.f[11] - Clip.f[ 8];
  Planes[0].w = Clip.f[15] - Clip.f[12];

  Planes[1].x = Clip.f[ 3] + Clip.f[ 0];
  Planes[1].y = Clip.f[ 7] + Clip.f[ 4];
  Planes[1].z = Clip.f[11] + Clip.f[ 8];
  Planes[1].w = Clip.f[15] + Clip.f[12];

  Planes[2].x = Clip.f[ 3] + Clip.f[ 1];
  Planes[2].y = Clip.f[ 7] + Clip.f[ 5];
  Planes[2].z = Clip.f[11] + Clip.f[ 9];
  Planes[2].w = Clip.f[15] + Clip.f[13];

  Planes[3].x = Clip.f[ 3] - Clip.f[ 1];
  Planes[3].y = Clip.f[ 7] - Clip.f[ 5];
  Planes[3].z = Clip.f[11] - Clip.f[ 9];
  Planes[3].w = Clip.f[15] - Clip.f[13];

  Planes[4].x = Clip.f[ 3] - Clip.f[ 2];
  Planes[4].y = Clip.f[ 7] - Clip.f[ 6];
  Planes[4].z = Clip.f[11] - Clip.f[10];
  Planes[4].w = Clip.f[15] - Clip.f[14];

  Planes[5].x = Clip.f[ 3] + Clip.f[ 2];
  Planes[5].y = Clip.f[ 7] + Clip.f[ 6];
  Planes[5].z = Clip.f[11] + Clip.f[10];
  Planes[5].w = Clip.f[15] + Clip.f[14];

  //Normalize Planes
  for(int i=0;i<6;i++)
    Planes[i] = NormalizePlaneV4(Planes[i]);
}

void FrustrumCamera::Points(vec4 *ptr){
  ptr[0] = PlanePoint(Planes[0],Planes[2],Planes[4]);
  ptr[1] = PlanePoint(Planes[0],Planes[3],Planes[4]);
  ptr[2] = PlanePoint(Planes[1],Planes[2],Planes[4]);
  ptr[3] = PlanePoint(Planes[1],Planes[3],Planes[4]);

  ptr[4] = PlanePoint(Planes[0],Planes[2],Planes[5]);
  ptr[5] = PlanePoint(Planes[0],Planes[3],Planes[5]);
  ptr[6] = PlanePoint(Planes[1],Planes[2],Planes[5]);
  ptr[7] = PlanePoint(Planes[1],Planes[3],Planes[5]);
}

bool FrustrumCamera::Test(class FrustrumObjectSphere *ptr){
  for(int i=0;i<6;i++)
    if(DistancePlaneV4(Planes[i],ptr->v4Position)<= -ptr->fRadius)
      return false;
  return true;
}

/*bool FrustrumCamera::Test(class FrustrumObjectBox *ptr){
  bool result;
  for(int j=0;j<8;j++){
    result = true;
    for(int i=0;i<6;i++){
      if(DistancePlaneV4(Planes[i],ptr->v4TransPoints[j]) <= 0){
        result=false;
        break;
      }
    }
    if(result)
      return true;
  }
  return false;
}*/

bool FrustrumCamera::Test(class FrustrumObjectBox *ptr){
  for (int j = 0; j < 6; j++)
  {
    int Count = 0;
    for (int i = 0; i < 8; i++)
      if (DistancePlaneV4(Planes[j],ptr->v4TransPoints[i])>0)
        Count++;
    if (Count == 0)
      return false;
  }
  return true;
}

bool FrustrumObjectSphere::IsInFrustrum(){
  return ActiveFrustrum->Test(this);
}

void FrustrumObjectSphere::SetPosition(vec4 Position){
  v4Position = Position;
}

void FrustrumObjectSphere::SetRadius(float Radius){
  fRadius = Radius;
}

void FrustrumObjectSphere::Draw(){
  vec4 temp;
  glBegin(GL_LINE_STRIP);
    for(float i=0;i<=360.0f;i+=30){
      temp = v4Position;
      temp.x += cos(i/180.0f * M_PI) * fRadius;
      temp.y += sin(i/180.0f * M_PI) * fRadius;
      glVertex3fv(temp.f);
    }
  glEnd();
  glBegin(GL_LINE_STRIP);
    for(float i=0;i<=360.0f;i+=30){
      temp = v4Position;
      temp.x += cos(i/180.0f * M_PI) * fRadius;
      temp.z += sin(i/180.0f * M_PI) * fRadius;
      glVertex3fv(temp.f);
    }
  glEnd();
  glBegin(GL_LINE_STRIP);
    for(float i=0;i<=360.0f;i+=30){
      temp = v4Position;
      temp.y += cos(i/180.0f * M_PI) * fRadius;
      temp.z += sin(i/180.0f * M_PI) * fRadius;
      glVertex3fv(temp.f);
    }
  glEnd();
}

void FrustrumObjectBox::SetSize(float width, float length, float height){
  v4Points[0].x = -(width/2.0f); v4Points[0].y = -(length/2.0f);v4Points[0].z = -(height/2.0f); v4Points[0].w = 1.0f;
  v4Points[1].x = +(width/2.0f); v4Points[1].y = -(length/2.0f);v4Points[1].z = -(height/2.0f); v4Points[1].w = 1.0f;
  v4Points[2].x = +(width/2.0f); v4Points[2].y = +(length/2.0f);v4Points[2].z = -(height/2.0f); v4Points[2].w = 1.0f;
  v4Points[3].x = -(width/2.0f); v4Points[3].y = +(length/2.0f);v4Points[3].z = -(height/2.0f); v4Points[3].w = 1.0f;
  v4Points[4].x = -(width/2.0f); v4Points[4].y = -(length/2.0f);v4Points[4].z = +(height/2.0f); v4Points[4].w = 1.0f;
  v4Points[5].x = +(width/2.0f); v4Points[5].y = -(length/2.0f);v4Points[5].z = +(height/2.0f); v4Points[5].w = 1.0f;
  v4Points[6].x = +(width/2.0f); v4Points[6].y = +(length/2.0f);v4Points[6].z = +(height/2.0f); v4Points[6].w = 1.0f;
  v4Points[7].x = -(width/2.0f); v4Points[7].y = +(length/2.0f);v4Points[7].z = +(height/2.0f); v4Points[7].w = 1.0f;
  for(int i=0;i<8;i++)
    v4TransPoints[i] = v4Points[i];
}

void FrustrumObjectBox::SetSize(const vec4& min, const vec4& max){
  v4Points[0].x = min.x-0.5; v4Points[0].y = min.y-0.5; v4Points[0].z = min.z-0.5; v4Points[0].w=1.0f;
  v4Points[1].x = max.x+0.5; v4Points[1].y = min.y-0.5; v4Points[1].z = min.z-0.5; v4Points[1].w=1.0f;
  v4Points[2].x = max.x+0.5; v4Points[2].y = max.y+0.5; v4Points[2].z = min.z-0.5; v4Points[2].w=1.0f;
  v4Points[3].x = min.x-0.5; v4Points[3].y = max.y+0.5; v4Points[3].z = min.z-0.5; v4Points[3].w=1.0f;
  v4Points[4].x = min.x-0.5; v4Points[4].y = min.y-0.5; v4Points[4].z = max.z+0.5; v4Points[4].w=1.0f;
  v4Points[5].x = max.x+0.5; v4Points[5].y = min.y-0.5; v4Points[5].z = max.z+0.5; v4Points[5].w=1.0f;
  v4Points[6].x = max.x+0.5; v4Points[6].y = max.y+0.5; v4Points[6].z = max.z+0.5; v4Points[6].w=1.0f;
  v4Points[7].x = min.x-0.5; v4Points[7].y = max.y+0.5; v4Points[7].z = max.z+0.5; v4Points[7].w=1.0f;
  for(int i=0;i<8;i++)
    v4TransPoints[i] = v4Points[i];
}

void FrustrumObjectBox::Transform(mat4a Matrix){
  for(int i=0;i<8;i++)
    v4TransPoints[i] = Mul(Matrix,v4Points[i]);
}

void FrustrumObjectBox::Draw(){
  glBegin(GL_LINES);
    glVertex3fv(v4TransPoints[0].f);
    glVertex3fv(v4TransPoints[1].f);

    glVertex3fv(v4TransPoints[1].f);
    glVertex3fv(v4TransPoints[2].f);

    glVertex3fv(v4TransPoints[2].f);
    glVertex3fv(v4TransPoints[3].f);

    glVertex3fv(v4TransPoints[3].f);
    glVertex3fv(v4TransPoints[0].f);

    glVertex3fv(v4TransPoints[4].f);
    glVertex3fv(v4TransPoints[5].f);

    glVertex3fv(v4TransPoints[5].f);
    glVertex3fv(v4TransPoints[6].f);

    glVertex3fv(v4TransPoints[6].f);
    glVertex3fv(v4TransPoints[7].f);

    glVertex3fv(v4TransPoints[7].f);
    glVertex3fv(v4TransPoints[4].f);

    glVertex3fv(v4TransPoints[0].f);
    glVertex3fv(v4TransPoints[4].f);

    glVertex3fv(v4TransPoints[1].f);
    glVertex3fv(v4TransPoints[5].f);

    glVertex3fv(v4TransPoints[2].f);
    glVertex3fv(v4TransPoints[6].f);

    glVertex3fv(v4TransPoints[3].f);
    glVertex3fv(v4TransPoints[7].f);
  glEnd();
}

bool FrustrumObjectBox::IsInFrustrum(){
  return ActiveFrustrum->Test(this);
}
