/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "thEngine.h"
#include "instancing.h"
#include "model.h"

InstancedVertexBuffer::~InstancedVertexBuffer(){
  for(int i=0;i<Mats.size();i++)
    delete Mats[i];
}

int InstancedVertexBuffer::AddInstanceGroup(mat4a *array, int count){
  std::vector<float>* ptr;
  mat4 TempMat;
  bool temp=true;
  ptr = new std::vector<float>;
  ptr->reserve(count * 12);
  for(int i=0;i<count;i++){
    DealignMatrix4(array[i],TempMat.f);
    ptr->insert(ptr->end(),TempMat.f,TempMat.f+12);
  }
  Mats.push_back(ptr);
  return (Mats.size()-1);
}

bool InstancedVertexBuffer::SetInstanceGroup(int num,mat4a *array, int count){
  if(num > Mats.size())
    return false;
  mat4 TempMat;
  Mats[num]->clear();
  Mats[num]->reserve(12 * count);
  for(int i=0;i<count;i++){
    DealignMatrix4(array[i],TempMat.f);
    Mats[num]->insert(Mats[num]->end(),TempMat.f,TempMat.f+12);
  }
  return true;
}

bool InstancedVertexBufferFallback::InitDraw(int options, Shader *ptr){
  int iVertexBufferOptions = 0;
  if(ptr->GetInstanceAttribute() < 0 || ptr->GetInstanceUniform() < 0){
    ThErrorMessage("InstancedVertexBuffer Error","Angegebener Shader ist nicht fr Instancing geignet!");
    return false;
  }
  if(options & TH_TEXCOORDS)
    iVertexBufferOptions |= TH_VB_TEXCOORD0;
  if(options & TH_NORMALS)
    iVertexBufferOptions |= TH_VB_NORMAL;
  if(options & TH_TANGENT)
    iVertexBufferOptions |= TH_VB_ATTRIBUTE0;
  if(options & TH_BINORMAL)
    iVertexBufferOptions |= TH_VB_ATTRIBUTE1;

  iVertexBufferOptions |= TH_VB_ATTRIBUTE2;

  vbuffer->SetAttributeLocation(2,ptr->GetAttributeLocation(ptr->GetInstanceAttribute()));
  vbuffer->InitInstancedDraw(iVertexBufferOptions);
  ShaderInUse = ptr;
  return true;
}

void InstancedVertexBuffer::DeInitDraw(){
  vbuffer->DeInitInstancedDraw();
}

InstancedVertexBufferFallback::InstancedVertexBufferFallback(VertexBuffer *ptr, int MaxInstances){
  vbuffer = new VertexBuffer(*ptr);
  vbuffer->Duplicate(MaxInstances);
  vbuffer->UploadData();
  iMaxInstances = MaxInstances;
}

InstancedVertexBufferFallback::~InstancedVertexBufferFallback(){
  delete vbuffer;
}

void InstancedVertexBufferFallback::Draw(int group){
  if(group > Mats.size())
    return;
  int iAnzahl = Mats[group]->size() / 12 / iMaxInstances;
  int iInstanceCount = Mats[group]->size() / 12;
  if((Mats[group]->size() / 12) % iMaxInstances != 0)
    iAnzahl++;

  for(int i=0;i<iAnzahl;i++){
    int num = iInstanceCount - i*iMaxInstances;
    if(num > iMaxInstances) num = iMaxInstances;
    ShaderInUse->SetUniform(ShaderInUse->GetInstanceUniform(),4,num*3,&(*Mats[group])[i*iMaxInstances*12]);
    vbuffer->DrawInstancedFallback(num);
  }
}

InstancedVertexBufferExtension::InstancedVertexBufferExtension(VertexBuffer *ptr){
  vbuffer = ptr;
  int iMaxSize;
  glGetIntegerv(GL_MAX_BINDABLE_UNIFORM_SIZE_EXT,&iMaxSize);
  iMaxSize = iMaxSize / (12 * sizeof(float));
  iMaxInstances = iMaxSize;
}

bool InstancedVertexBufferExtension::InitDraw(int options, Shader *ptr){
  if(ptr->GetInstanceUniform() < 0){
    ThErrorMessage("InstancedVertexBuffer Error","Angegebener Shader ist nicht fr Instancing geignet!");
    return false;
  }
  //VertexBuffer Initialisieren
  int iVertexBufferOptions = 0;
  if(options & TH_TEXCOORDS)
    iVertexBufferOptions |= TH_VB_TEXCOORD0;
  if(options & TH_NORMALS)
    iVertexBufferOptions |= TH_VB_NORMAL;
  if(options & TH_TANGENT)
    iVertexBufferOptions |= TH_VB_ATTRIBUTE0;
  if(options & TH_BINORMAL)
    iVertexBufferOptions |= TH_VB_ATTRIBUTE1;

  vbuffer->InitInstancedDraw(iVertexBufferOptions);
  ShaderInUse = ptr;
  return true;
}

void InstancedVertexBufferExtension::Draw(int group){
  if(group > Mats.size())
    return;
  int iAnzahl = Mats[group]->size() / 12 / iMaxInstances;
  int iInstanceCount = Mats[group]->size() / 12;
  if((Mats[group]->size() / 12) % iMaxInstances != 0)
    iAnzahl++;

  for(int i=0;i<iAnzahl;i++){
    int num = iInstanceCount - i*iMaxInstances;
    if(num > iMaxInstances) num = iMaxInstances;
    ShaderInUse->UploadData(ShaderInUse->GetInstanceUniform(),0,12*num*sizeof(float),&(*Mats[group])[i*iMaxInstances*12]);
    vbuffer->DrawInstanced(num);
  }
  glBindBuffer(GL_UNIFORM_BUFFER_EXT,0);
}
