/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "kameraeditor.h"
#include <stdio.h>
#include <vector>
#include <sstream>

CameraEditor::CameraEditor(){
  Window = NULL;
  InputX = NULL;
  InputY = NULL;
  InputZ = NULL;
  InputT = NULL;
  Select1 = NULL;
  TextX = TextY = TextZ = TextT = NULL;
  EditState = POSITION;
  pTimer = NULL;
  pCam = NULL;
  bPlaying = false;
  bRendering = false;
  DrawPositions = NULL;
  DrawDirections = NULL;
  iRenderingAnz = -1;
}

void CameraEditorOnButton0(void* Creator){
  CameraEditor* ptr = (CameraEditor*) Creator;
  if(ptr->EditState != POSITION){
    ptr->EditState = POSITION;
    ptr->UpdateInputFields();
    ptr->TextMode->SetCaption("Edit-Mode: Position");
  }
}

void CameraEditorOnButton1(void* Creator){
  CameraEditor* ptr = (CameraEditor*) Creator;
  if(ptr->EditState != DIRECTION){
    ptr->EditState = DIRECTION;
    ptr->UpdateInputFields();
    ptr->TextMode->SetCaption("Edit-Mode: Richtung");
  }
}

void CameraEditorOnButton2(void* Creator){
  CameraEditor* ptr = (CameraEditor*) Creator;
  if(ptr->EditState != UP){
    ptr->EditState = UP;
    ptr->UpdateInputFields();
    ptr->TextMode->SetCaption("Edit-Mode: Up-Vektor");
  }
}

void CameraEditorOnButton3(void* Creator){
  CameraEditor* ptr = (CameraEditor*) Creator;
  ptr->InsertCurrent();
}

void CameraEditorOnButtonSwitch(void* Creator){
  CameraEditor* ptr = (CameraEditor*) Creator;
  if(ptr->bRendering){
    ptr->bRendering = false;
    ptr->ButtonSwitch->SetCaption("Show Lines");
  }
  else{
    ptr->bRendering = true;
    ptr->ButtonSwitch->SetCaption("Hide Lines");
  }
}

void CameraEditorOnButtonPlay(void *Creator){
  CameraEditor* ptr = (CameraEditor*) Creator;
  ptr->Play();
  if(ptr->IsPlaying())
    ptr->ButtonPlay->SetCaption("Stop");
  else
    ptr->ButtonPlay->SetCaption("Play");
}

void CameraEditorOnButtonSave(void *Creator){
  CameraEditor* ptr = (CameraEditor*) Creator;
  ptr->SaveToFile("gfx/cam.dat");
}

void CameraEditorOnButtonLoad(void *Creator){
  CameraEditor* ptr = (CameraEditor*) Creator;
  ptr->LoadFromFile("gfx/cam.dat");
}

void CameraEditorOnSelect1Change(void *Creator){
  CameraEditor* ptr = (CameraEditor*) Creator;
  ptr->InputX->LooseFocus();
  ptr->InputY->LooseFocus();
  ptr->InputZ->LooseFocus();
  ptr->InputT->LooseFocus();
  ptr->UpdateInputFields();
}

void CameraEditorOnInputXChange(void *Creator){
  CameraEditor* ptr = (CameraEditor*) Creator;
  std::ostringstream NewCaption;
  std::stringstream Buffer;
  vec4 position;
  float wert;
  if(ptr->Select1->GetSelected() < 0)
    return;
  if(!ptr->IsValidNumber(ptr->InputX)){
    ptr->TextFehler->SetCaption("X ist keine Zahl");
    return;
  }
  Buffer << ptr->InputX->GetCaption();
  Buffer >> wert;
  switch(ptr->EditState){
    case POSITION:
      ptr->Positions[ptr->Select1->GetSelected()].x = wert;
      position = ptr->Positions[ptr->Select1->GetSelected()];
      NewCaption.precision(4);
      NewCaption << "T:" << position.w << " X:" << position.x << " Y:" << position.y << " Z:" << position.z;
      ptr->Select1->ChangeOption(ptr->Select1->GetSelected(),NewCaption.str().c_str());
      break;
    case DIRECTION:
      ptr->Directions[ptr->Select1->GetSelected()].x = wert;
      break;
    case UP:
      ptr->Directions[ptr->Select1->GetSelected()].x = wert;
      break;
  }
  ptr->Update();
}

void CameraEditorOnInputYChange(void *Creator){
  CameraEditor* ptr = (CameraEditor*) Creator;
  std::ostringstream NewCaption;
  std::stringstream Buffer;
  vec4 position;
  float wert;
  if(ptr->Select1->GetSelected() < 0)
    return;
  if(!ptr->IsValidNumber(ptr->InputY)){
    ptr->TextFehler->SetCaption("Y ist keine Zahl");
    return;
  }
  Buffer << ptr->InputY->GetCaption();
  Buffer >> wert;
  switch(ptr->EditState){
    case POSITION:
      ptr->Positions[ptr->Select1->GetSelected()].y = wert;
      position = ptr->Positions[ptr->Select1->GetSelected()];
      NewCaption.precision(4);
      NewCaption << "T:" << position.w << " X:" << position.x << " Y:" << position.y << " Z:" << position.z;
      ptr->Select1->ChangeOption(ptr->Select1->GetSelected(),NewCaption.str().c_str());
      break;
    case DIRECTION:
      ptr->Directions[ptr->Select1->GetSelected()].y = wert;
      break;
    case UP:
      ptr->Directions[ptr->Select1->GetSelected()].y = wert;
      break;
  }
  ptr->Update();
}

void CameraEditorOnInputZChange(void *Creator){
  CameraEditor* ptr = (CameraEditor*) Creator;
  std::ostringstream NewCaption;
  std::stringstream Buffer;
  vec4 position;
  float wert;
  if(ptr->Select1->GetSelected() < 0)
    return;
  if(!ptr->IsValidNumber(ptr->InputZ)){
    ptr->TextFehler->SetCaption("Z ist keine Zahl");
    return;
  }
  Buffer << ptr->InputZ->GetCaption();
  Buffer >> wert;
  switch(ptr->EditState){
    case POSITION:
      ptr->Positions[ptr->Select1->GetSelected()].z = wert;
      position = ptr->Positions[ptr->Select1->GetSelected()];
      NewCaption.precision(4);
      NewCaption << "T:" << position.w << " X:" << position.x << " Y:" << position.y << " Z:" << position.z;
      ptr->Select1->ChangeOption(ptr->Select1->GetSelected(),NewCaption.str().c_str());
      break;
    case DIRECTION:
      ptr->Directions[ptr->Select1->GetSelected()].z = wert;
      break;
    case UP:
      ptr->Directions[ptr->Select1->GetSelected()].z = wert;
      break;
  }
  ptr->Update();
}

void CameraEditorOnInputTChange(void *Creator){
  CameraEditor* ptr = (CameraEditor*) Creator;
  std::ostringstream NewCaption;
  std::stringstream Buffer;
  vec4 position;
  float wert;
  if(ptr->Select1->GetSelected() < 0)
    return;
  if(!ptr->IsValidNumber(ptr->InputT)){
    ptr->TextFehler->SetCaption("T ist keine Zahl");
    return;
  }
  Buffer << ptr->InputT->GetCaption();
  Buffer >> wert;
  switch(ptr->EditState){
    case POSITION:
      ptr->Positions[ptr->Select1->GetSelected()].w = wert;
      position = ptr->Positions[ptr->Select1->GetSelected()];
      NewCaption.precision(4);
      NewCaption << "T:" << position.w << " X:" << position.x << " Y:" << position.y << " Z:" << position.z;
      ptr->Select1->ChangeOption(ptr->Select1->GetSelected(),NewCaption.str().c_str());
      break;
    case DIRECTION:
      ptr->Directions[ptr->Select1->GetSelected()].w = wert;
      break;
    case UP:
      ptr->Directions[ptr->Select1->GetSelected()].w = wert;
      break;
  }
  ptr->Update();
}

void CameraEditorOnButtonGoto(void *Creator){
  CameraEditor* ptr = (CameraEditor*)Creator;
  int num = ptr->Select1->GetSelected();
  if(num >= 0){
    ptr->pCam->SetFrom(ptr->Positions[num]);
    ptr->pCam->SetTo(ptr->Directions[num]);
    ptr->pCam->SetRotation(ptr->UpVectors[num].x);
  }
}

void CameraEditorOnButtonReplace(void *Creator){
  CameraEditor* ptr = (CameraEditor*)Creator;
  int num = ptr->Select1->GetSelected();
  float t1,t2;
  std::ostringstream NewCaption;
  if(num >= 0){
    NewCaption.precision(4);
    t1 = ptr->Positions[num].w;
    t2 = ptr->Positions[num].w;
    ptr->Positions[num] = ptr->pCam->GetFrom();
    ptr->Positions[num].w = t1;
    ptr->Directions[num] = ptr->pCam->GetTo();
    ptr->Directions[num].w = t2;
    ptr->UpVectors[num].x = ptr->pCam->GetRotation();
    NewCaption << "T:" << ptr->Positions[num].w << " X:" << ptr->Positions[num].x << " Y:" << ptr->Positions[num].y << " Z:" << ptr->Positions[num].z;
    ptr->Select1->ChangeOption(num,NewCaption.str().c_str());
    ptr->Update();
  }
}

void CameraEditorOnButtonDelete(void *Creator){
  CameraEditor* ptr = (CameraEditor*)Creator;
  int num = ptr->Select1->GetSelected();
  if(num >= 0){
    ptr->Select1->RemoveOption(num);
    ptr->Positions.DeleteControlPoint(num);
    ptr->Directions.DeleteControlPoint(num);
    ptr->UpVectors.DeleteControlPoint(num);
    ptr->Update();
  }
}

bool CameraEditor::Init(Gui* pGui,camera *Cam, Timer *pMainTimer){
  pCam = Cam;
  pTimer = pMainTimer;
  PlayStart.SetTimerReference(pTimer);
  Window = pGui->CreateWindow(NULL,300,20,395,230,false,0,"Kamera Editor",this);
  InputX = pGui->CreateTextInput(Window,290,4,100,18,6,7,12,"X",this);
  InputX->SetOnLooseFocus(CameraEditorOnInputXChange);
  InputY = pGui->CreateTextInput(Window,290,24,100,18,6,7,12,"Y",this);
  InputY->SetOnLooseFocus(CameraEditorOnInputYChange);
  InputZ = pGui->CreateTextInput(Window,290,44,100,18,6,7,12,"Z",this);
  InputZ->SetOnLooseFocus(CameraEditorOnInputZChange);
  InputT = pGui->CreateTextInput(Window,290,64,100,18,6,7,12,"T",this);
  InputT->SetOnLooseFocus(CameraEditorOnInputTChange);
  TextX = pGui->CreateText(Window,278,6,1,"X",this);
  TextY = pGui->CreateText(Window,278,26,1,"Y",this);
  TextZ = pGui->CreateText(Window,278,46,1,"Z",this);
  TextT = pGui->CreateText(Window,278,66,1,"T",this);
  ButtonPlay = pGui->CreateButton(Window,280,84,60,18,0,1,"Play",this);
  ButtonPlay->SetOnPress(CameraEditorOnButtonPlay);
  ButtonSwitch = pGui->CreateButton(Window,280,104,80,18,0,1,"Draw Lines",this);
  ButtonSwitch->SetOnPress(CameraEditorOnButtonSwitch);
  Button0 = pGui->CreateButton(Window,4,122,80,20,0,1,"Position",this);
  Button0->SetOnPress(CameraEditorOnButton0);
  Button1 = pGui->CreateButton(Window,86,122,80,20,0,1,"Richtung",this);
  Button1->SetOnPress(CameraEditorOnButton1);
  Button2 = pGui->CreateButton(Window,168,122,80,20,0,1,"Up-Vektor",this);
  Button2->SetOnPress(CameraEditorOnButton2);
  TextMode = pGui->CreateText(Window,252,126,1,"Edit-Mode: Position",this);

  ButtonGoto = pGui->CreateButton(Window,4,144,100,20,0,1,"Goto",this);
  ButtonGoto->SetOnPress(CameraEditorOnButtonGoto);
  ButtonReplace = pGui->CreateButton(Window,106,144,100,20,0,1,"Replace",this);
  ButtonReplace->SetOnPress(CameraEditorOnButtonReplace);
  ButtonDelete = pGui->CreateButton(Window,208,144,100,20,0,1,"Delete",this);
  ButtonDelete->SetOnPress(CameraEditorOnButtonDelete);

  Button3 = pGui->CreateButton(Window,4,166,140,20,0,1,"Aktuelle Einfuegen",this);
  Button3->SetOnPress(CameraEditorOnButton3);
  InputT2 = pGui->CreateTextInput(Window,146,167,100,18,6,7,12,"0.0",this);
  ButtonSave = pGui->CreateButton(Window,330,140,60,20,0,1,"Save",this);
  ButtonSave->SetOnPress(CameraEditorOnButtonSave);
  ButtonLoad = pGui->CreateButton(Window,330,162,60,20,0,1,"Load",this);
  ButtonLoad->SetOnPress(CameraEditorOnButtonLoad);

  TextFehler = pGui->CreateText(Window,6,190,1,"",this);

  ThGuiSelectInit SelectInit;
  SelectInit.MaxElementsVisible = 5;
  SelectInit.Width = 270;
  SelectInit.Space = 1;
  SelectInit.Style = 0;
  SelectInit.ButtonUpStyle = 3;
  SelectInit.ButtonUpHoverStyle = 4;
  SelectInit.ButtonDownStyle = 5;
  SelectInit.ButtonDownHoverStyle = 6;
  SelectInit.SelectStyle = 1;
  SelectInit.SelectHoverStyle = 2;
  SelectInit.SelectSelectedStyle = 3;

  Select1 = pGui->CreateSelect(Window,4,4,SelectInit,this);
  if(Select1 == NULL)
    return false;
  Select1->SetOnChange(CameraEditorOnSelect1Change);
  return true;
}

void CameraEditor::UpdateInputFields(){
  std::ostringstream NewCaption;
  NewCaption.precision(7);
  if(Select1->GetSelected() != -1){
    int num = Select1->GetSelected();
    switch(EditState){
      case POSITION:
        InputY->SetActive(true);
        InputZ->SetActive(true);
        NewCaption.str("");
        NewCaption << Positions[num].x;
        InputX->SetCaption(NewCaption.str().c_str());
        NewCaption.str("");
        NewCaption << Positions[num].y;
        InputY->SetCaption(NewCaption.str().c_str());
        NewCaption.str("");
        NewCaption << Positions[num].z;
        InputZ->SetCaption(NewCaption.str().c_str());
        NewCaption.str("");
        NewCaption << Positions[num].w;
        InputT->SetCaption(NewCaption.str().c_str());
        break;
      case DIRECTION:
        InputY->SetActive(true);
        InputZ->SetActive(true);
        NewCaption.str("");
        NewCaption << Directions[num].x;
        InputX->SetCaption(NewCaption.str().c_str());
        NewCaption.str("");
        NewCaption << Directions[num].y;
        InputY->SetCaption(NewCaption.str().c_str());
        NewCaption.str("");
        NewCaption << Directions[num].z;
        InputZ->SetCaption(NewCaption.str().c_str());
        NewCaption.str("");
        NewCaption << Directions[num].w;
        InputT->SetCaption(NewCaption.str().c_str());
        break;
      case UP:
        InputY->SetActive(false);
        InputZ->SetActive(false);
        NewCaption.str("");
        NewCaption << UpVectors[num].x;
        InputX->SetCaption(NewCaption.str().c_str());
        NewCaption.str("");
        NewCaption << UpVectors[num].w;
        InputT->SetCaption(NewCaption.str().c_str());
        break;
    }
  }
}

void CameraEditor::InsertCurrent(){
  std::stringstream Input;
  std::ostringstream NewCaption;
  vec4 position,direction,up;
  float t;
  int insert_pos;
  Input << InputT2->GetCaption().c_str();
  if(!IsValidNumber(InputT2)){
    TextFehler->SetCaption("Fehler: Ungltige Zahl fuer Zeit");
    return;
  }
  Input >> t;
  if( t < 0.0f){
    TextFehler->SetCaption("Fehler: Zeit ist negativ");
    return;
  }
  position = pCam->GetFrom();
  direction = pCam->GetTo();
  up.x = pCam->GetRotation();
  position.w = t;
  direction.w = t;
  up.w = t;
  NewCaption.precision(4);
  NewCaption << "T:" << t << " X:" << position.x << " Y:" << position.y << " Z:" << position.z;
  if(Positions.size() == 0){
    Positions.AddControlPoint(position);
    Directions.AddControlPoint(direction);
    UpVectors.AddControlPoint(up);
    Select1->AddOption(NewCaption.str().c_str());
    Update();
  }
  else{
    for(insert_pos=0;insert_pos<Positions.size();insert_pos++)
      if(t < Positions[insert_pos].w)
        break;
    if(insert_pos == Positions.size()){
      Positions.AddControlPoint(position);
      Directions.AddControlPoint(direction);
      UpVectors.AddControlPoint(up);
      Select1->AddOption(NewCaption.str().c_str());
      Update();
    }
    else{
      Positions.InsertControlPoint(insert_pos,position);
      Directions.InsertControlPoint(insert_pos,direction);
      UpVectors.InsertControlPoint(insert_pos,up);
      Select1->InsertOption(insert_pos,NewCaption.str().c_str());
      Update();
    }
  }
}

bool CameraEditor::IsValidNumber(GuiTextInput *ptr){
  const std::string& str = ptr->GetCaption();
  for(int i=0;i<str.length();i++){
    if(i==0 && str[i] == '-')
      continue;
    if(str[i] != '.' && (str[i] < '0' || str[i] > '9'))
      return false;
  }
  return true;
}

void CameraEditor::Update(){
  Positions.Update();
  Directions.Update();
  UpVectors.Update();
  if(DrawPositions != NULL)
    delete DrawPositions;
  if(DrawDirections != NULL)
    delete DrawDirections;

  float tdiff = Positions[Positions.size()-1].w;
  float tstep = 0.25f;
  iRenderingAnz = (int)ceil(tdiff / tstep) + 1;

  DrawPositions = new vec4[iRenderingAnz];
  DrawDirections = new vec4[iRenderingAnz];

  int i=0;
  for(float f=0.0f;f<tdiff;f+=tstep){
    DrawPositions[i] = Positions.Get(f);
    DrawPositions[i].w = 1.0f;
    DrawDirections[i] = Directions.Get(f);
    DrawDirections[i].w = 1.0f;
    i++;
  }
}

void CameraEditor::Play(){
  if(!bPlaying){
    PlayStart = pTimer->GetTimeInt64();
    bPlaying = true;
  }
  else bPlaying = false;
}

float CameraEditor::UpdateCam(){
  if(bPlaying){
    Zeitpunkt Jetzt;
    Jetzt.SetTimerReference(pTimer);
    Jetzt = pTimer->GetTimeInt64();
    float tdiff = (Jetzt - PlayStart) / 1000.0;
    if(tdiff > Positions[Positions.size()-1].w){
      tdiff = 0;
      PlayStart = Jetzt;
    }
    pCam->SetFrom(Positions.Get(tdiff));
    pCam->SetTo(Directions.Get(tdiff));
    vec4 up = UpVectors.Get(tdiff);
    pCam->SetRotation(up.x);
    return tdiff;
  }
  return -1.0f;
}

void CameraEditor::Draw(){
   if(bRendering && iRenderingAnz > 1){
     UnloadShader();
     glColor4f(1.0f,0.0f,0.0f,1.0f);
     glLineWidth(1.0f);
     glBegin(GL_LINES);
     for(int i=0;i<iRenderingAnz-1;i++){
       glVertex3fv(DrawPositions[i].f);
       glVertex3fv(DrawPositions[i+1].f);
     }
     glEnd();
     glPointSize(10.0f);
     glBegin(GL_POINTS);
     for(int i=0;i<Positions.size();i++){
       if(Select1->GetSelected() == i){
         glColor3f(0.0f,1.0f,0.0f);
         glVertex3fv(Positions[i].f);
         glColor3f(1.0f,0.0f,0.0f);
       }
       else
         glVertex3fv(Positions[i].f);
     }
     glEnd();
     glColor4f(0.0f,0.0f,1.0f,1.0f);
     glBegin(GL_LINES);
     for(int i=0;i<iRenderingAnz-1;i++){
       glVertex3fv(DrawDirections[i].f);
       glVertex3fv(DrawDirections[i+1].f);
     }
     glEnd();
     glPointSize(10.0f);
     glBegin(GL_POINTS);
     for(int i=0;i<Directions.size();i++){
       if(Select1->GetSelected() == i){
         glColor3f(0.0f,1.0f,0.0f);
         glVertex3fv(Directions[i].f);
         glColor3f(0.0f,0.0f,1.0f);
       }
       else
         glVertex3fv(Directions[i].f);
     }
     glEnd();
   }
}

bool CameraEditor::SaveToFile(const char *filename){
  FILE *datei;
  int size;
  datei = fopen(filename,"wb");
  if(!datei)
    return false;
  size = Positions.size();
  fwrite(&size,sizeof(int),1,datei);
  fwrite(&Positions[0],sizeof(vec4),Positions.size(),datei);
  size = Directions.size();
  fwrite(&size,sizeof(int),1,datei);
  fwrite(&Directions[0],sizeof(vec4),Directions.size(),datei);
  size = UpVectors.size();
  fwrite(&size,sizeof(int),1,datei);
  fwrite(&UpVectors[0],sizeof(vec4),UpVectors.size(),datei);
  fclose(datei);
  return true;
}

bool CameraEditor::LoadFromFile(const char *filename){
  FILE *datei;
  int size;
  std::ostringstream NewCaption;
  datei = fopen(filename,"rb");
  if(!datei)
    return false;
  fread(&size,sizeof(int),1,datei);
  Positions.resize(size);
  fread(&Positions[0],sizeof(vec4),size,datei);
  fread(&size,sizeof(int),1,datei);
  Directions.resize(size);
  fread(&Directions[0],sizeof(vec4),size,datei);
  fread(&size,sizeof(int),1,datei);
  UpVectors.resize(size);
  fread(&UpVectors[0],sizeof(vec4),size,datei);
  fclose(datei);

  Select1->RemoveAllOptions();
  NewCaption.precision(4);
  for(int i=0;i<Positions.size();i++){
    NewCaption.str("");
    NewCaption << "T:" << Positions[i].w << " X:" << Positions[i].x << " Y:" << Positions[i].y << " Z:" << Positions[i].z;
    Select1->AddOption(NewCaption.str().c_str());
  }
  Update();
  return true;
}
