/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "thEngine.h"
#include "landscape.h"
#include "vektoren.h"
#include <stdio.h>
#include <math.h>

Landscape::Landscape(int Detail){
  bDataLoaded=false;
  fSize=1.0f;
  fHeight=1.0f;
  vPosition.x=0; vPosition.y=0; vPosition.z=0; vPosition.w=1.0f;
  iDivisionDetail = Detail;
  Quads = NULL;
  PlacePositions = NULL;
  #ifdef __SSE__
  PlacePositionsO = NULL;
  #endif
}

Landscape::~Landscape(){
  #ifdef __SSE__
  if(PlacePositionsO != NULL)
    delete PlacePositionsO;
  #else
  if(PlacePositions != NULL)
    delete PlacePositions;
  #endif
  delete HeightMap;
}

bool Landscape::LoadHeightMapTGA(const char *name){
  if(bDataLoaded){ //Clear previous Loaded Data
    delete HeightMap;
  }
  char meldung[200];
  HeightMap = new Texture2D();
  if(!HeightMap->LoadTGA(name,TH_LINEAR | TH_CLAMP_S | TH_CLAMP_T)){
    sprintf(meldung,"Konnte Heightmap '%s' nicht laden",name);
    ThErrorMessage("Landscape Error",meldung);
    return false;
  }
  bDataLoaded=true;
  return true;
}

void Landscape::SetHeight(float value){
  fHeight = value;
}

void Landscape::SetPosition(vec4 Position){
  vPosition = Position;
}

void Landscape::SetPosition(float x, float y, float z){
  vPosition.x = x;
  vPosition.y = y;
  vPosition.z = z;
}

void Landscape::SetSize(float value){
  fSize = value;
}

void Landscape::SetDistance(float value){
  fMaxDistance = value;
}

int Landscape::VertCount(int faktor){
  int iDivisions = (iDivisionDetail +2) * faktor;
  return ((iDivisions+1) * 2) * iDivisions;
}

void Landscape::WriteQuad(int size, int faktor, int& iVertexNum, int iRes){
  int x,y,i,j,num,IndexBuffer;
  unsigned int index;
  vec4 Translation,v4Temp,v4Color,n1,n2,n3,n4;
  vec4 p1,p2;
  Translation.z=0.0f;
  Translation.w=1.0f;

  LandQuad *ptr;

  vec4 *Positions,*Normals,*SubDiv;
  vec4 *PositionsO,*NormalsO,*SubDivO;
  vec2 *TexCoords;

  int iDivisions = (iDivisionDetail + 2)*faktor;
  iSizes[iRes] = iDivisions;

  float fStepVertex = (fSize / (float)size) / (float)iDivisions;
  float fStepTex = (1.0f / (float)size) / (float)iDivisions;

  TexCoords = new vec2[(size*iDivisions+1)*(size*iDivisions+1)];

  Positions = new vec4[(size*iDivisions+1)*(size*iDivisions+1)];
  Normals = new vec4[(size*iDivisions+1)*(size*iDivisions+1)];
  SubDiv = new vec4[(size*iDivisions+1)*(size*iDivisions+1)];

  if(iRes == 0){
    PlacePositions = Positions;
    iVertexSize = iDivisions * size + 1;
  }

  for(y=0;y<=size*iDivisions;y++){
    for(x=0;x<=size*iDivisions;x++){
      i=y*(size*iDivisions+1)+x;
      TexCoords[i].x=(float)x * fStepTex;
      TexCoords[i].y=(float)y * fStepTex;
      Positions[i].x=(float)x * fStepVertex;
      Positions[i].y=(float)y * fStepVertex;
      v4Color = HeightMap->GetPixel(TexCoords[i]);
      Positions[i].z= v4Color.z * fHeight;
    }
  }

  v4Color.x = 0; v4Color.y = 0; v4Color.z = 0; v4Color.w = 0;
  for(y=0;y<=size*iDivisions;y++){
    for(x=0;x<=size*iDivisions;x++){
      i = y*(size*iDivisions+1)+x;
      n1 = (y < size*iDivisions) ? SubV4(Positions[i+size*iDivisions+1],Positions[i]) : v4Color;
      n2 = (x < size*iDivisions) ? SubV4(Positions[i+1],Positions[i]) : v4Color;
      n3 = (y > 0) ? SubV4(Positions[i-size*iDivisions-1],Positions[i]) : v4Color;
      n4 = (x > 0) ? SubV4(Positions[i-1],Positions[i]) : v4Color;
      Normals[i] = NormalizeV4(AddV4(AddV4(CrossProductV4(n1,n2),CrossProductV4(n2,n3)),AddV4(CrossProductV4(n3,n4),CrossProductV4(n4,n1))));
    }
  }

  for(y=0;y<=size*iDivisions;y++){
    for(x=0;x<=size*iDivisions;x++){
      i = y*(size*iDivisions+1)+x;
      if(x%2 == 0 && y%2 == 0){
        n1 = (y < size*iDivisions-1) ? SubV4(Positions[i+(size*iDivisions+1)*2],Positions[i]) : v4Color;
        n2 = (x < size*iDivisions-1) ? SubV4(Positions[i+2],Positions[i]) : v4Color;
        n3 = (y > 1) ? SubV4(Positions[i-(size*iDivisions+1)*2],Positions[i]) : v4Color;
        n4 = (x > 1) ? SubV4(Positions[i-2],Positions[i]) : v4Color;
        SubDiv[i] = NormalizeV4(AddV4(AddV4(CrossProductV4(n1,n2),CrossProductV4(n2,n3)),AddV4(CrossProductV4(n3,n4),CrossProductV4(n4,n1))));
        SubDiv[i].w = Positions[i].z;
      }
      else if(y%2 == 0){

        if(x<1){
          SubDiv[i].w = Positions[i+1].z;
        }
        else if(x>size*iDivisions){
          SubDiv[i].w = Positions[i-1].z;
        }
        else{
          SubDiv[i].w = Positions[i+1].z + (Positions[i-1].z-Positions[i+1].z)/2.0;
        }
      }
      else if(x%2 == 0){
        if(y<1){
          SubDiv[i].w = Positions[i+size*iDivisions+1].z;
        }
        else if(y>size*iDivisions){
          SubDiv[i].w = Positions[i-size*iDivisions-1].z;
        }
        else{
          SubDiv[i].w = Positions[i+size*iDivisions+1].z + (Positions[i-size*iDivisions-1].z-Positions[i+size*iDivisions+1].z)/2.0;
        }
      }
      else
        SubDiv[i].w = fHeight+1.0f;
    }
  }
  //system("pause");

  for(y=0;y<=size*iDivisions;y++){
    for(x=0;x<=size*iDivisions;x++){
      i = y*(size*iDivisions+1)+x;
      if(SubDiv[i].w == fHeight+1.0f){
        SubDiv[i] = NormalizeV4(AddV4(SubDiv[i-size*iDivisions-2],SubDiv[i+size*iDivisions+2]));
        SubDiv[i].w = Positions[i-size*iDivisions-2].z + (Positions[i+size*iDivisions+2].z-Positions[i-size*iDivisions-2].z)/2.0;
      }
    }
  }

  for(y=0;y<=size*iDivisions;y++){
    for(x=0;x<=size*iDivisions;x++){
      i = y*(size*iDivisions+1)+x;
      if(x%2==0 && y%2==1){
        p1 = SubDiv[i];
        SubDiv[i] = NormalizeV4(AddV4(SubDiv[i-size*iDivisions-1],SubDiv[i+size*iDivisions+1]));
        SubDiv[i].w = p1.w;
      }
      else if(x%2==1 && y%2==0){
        p1=SubDiv[i];
        SubDiv[i] = NormalizeV4(AddV4(SubDiv[i-1],SubDiv[i+1]));
        SubDiv[i].w = p1.w;
      }
    }
  }

  for(y=0;y<size;y++){
    for(x=0;x<size;x++){
      ptr = &Quads[x+y*size];
      ptr->fPosX = (float)x * (fSize/(float)size);
      ptr->fPosY = (float)y * (fSize/(float)size);
      if(iRes==0)
        ptr->fHeight = 0;
      ptr->iSizesMain[iRes] = new int[iDivisions];
      ptr->iStartsMain[iRes] = new int[iDivisions];
      for(i=0;i<iDivisions;i++){
        ptr->iSizesMain[iRes][i]=(iDivisions+1)*2;
        ptr->iStartsMain[iRes][i]=iVertexNum;
        for(j=0;j<=iDivisions;j++){
          num = (y*iDivisions+i+1)*(size*iDivisions+1)+x*iDivisions+j;
          v4Temp = Positions[num];
          if(v4Temp.z > ptr->fHeight && iRes==0)
            ptr->fHeight = v4Temp.z;
          memcpy(lpcVertexData,v4Temp.f,sizeof(float)*3);
          lpcVertexData+=sizeof(float)*3;

          v4Temp.x = TexCoords[num].x;
          v4Temp.y = TexCoords[num].y;
          if(i==iDivisions-1)
            v4Temp.z = 1.0f;
          else
            v4Temp.z = 0.5f;

          if(j==0)
            v4Temp.w = 0.0f;
          else if(j==iDivisions)
            v4Temp.w = 1.0f;
          else
            v4Temp.w = 0.5f;

          memcpy(lpcTexCoordData,v4Temp.f,sizeof(float)*4);
          lpcTexCoordData += sizeof(float)*4;

          memcpy(lpcNormalData,Normals[num].f,sizeof(float)*3);
          lpcNormalData += sizeof(float)*3;

          memcpy(lpcSubDivData,SubDiv[num].f,sizeof(float)*4);
          lpcSubDivData += sizeof(float)*4;
          iVertexNum++;


          num = (y*iDivisions+i)*(size*iDivisions+1)+x*iDivisions+j;
          v4Temp = Positions[num];
          if(v4Temp.z > ptr->fHeight && iRes==0)
            ptr->fHeight = v4Temp.z;
          memcpy(lpcVertexData,v4Temp.f,sizeof(float)*3);
          lpcVertexData+=sizeof(float)*3;

          v4Temp.x = TexCoords[num].x;
          v4Temp.y = TexCoords[num].y;
          if(i==0)
            v4Temp.z = 0.0f;
          else
            v4Temp.z = 0.5f;

          if(j==0)
            v4Temp.w = 0.0f;
          else if(j==iDivisions)
            v4Temp.w = 1.0f;
          else
            v4Temp.w = 0.5f;

          memcpy(lpcTexCoordData,v4Temp.f,sizeof(float)*4);
          lpcTexCoordData += sizeof(float)*4;

          memcpy(lpcNormalData,Normals[num].f,sizeof(float)*3);
          lpcNormalData += sizeof(float)*3;

          memcpy(lpcSubDivData,SubDiv[num].f,sizeof(float)*4);
          lpcSubDivData += sizeof(float)*4;
          iVertexNum++;
        }
      }
      if(iRes == 0){
        ptr->BoundingBox.SetSize(fSize/(float)size,fSize/(float)size,fHeight);
        Translation.x = ptr->fPosX - (fSize /2) + (fSize/(float)(size*2));
        Translation.y = ptr->fPosY - (fSize /2) + (fSize/(float)(size*2));
        Translation.z = ptr->fHeight/2 + vPosition.z;
        ptr->BoundingBox.Transform(TranslationMatrix(Translation));
        ptr->iLastRes = 5;
        ptr->fBlendFaktor = 0.0f;
      }
    }
  }

  if(iRes != 0){
    delete []Positions;
  }

  delete []Normals;
  delete []SubDiv;
  delete []TexCoords;
}

void Landscape::CreateQuads(int size){
  char *Data;
  int iVertexDataSize,iTexCoordDataSize,iNormalDataSize,iSubDivDataSize,iTotalSize;
  int iVertexNum=0;
  int iQuadSize;
  if(Quads != NULL)
    delete []Quads;
  Quads = new LandQuad[size*size];
  iQuadWidth=size;

  glGenBuffersARB(1,&VertexBufferMain);
  iQuadSize = VertCount(1) + VertCount(2) + VertCount(4) + VertCount(8);
  //iQuadSize = VertCount(2);
  iVertexDataSize = iQuadSize * size * size * 3 * sizeof(float);
  iTexCoordDataSize = iQuadSize * size * size * 4 * sizeof(float);
  iNormalDataSize = iQuadSize * size * size * 3 * sizeof(float);
  iSubDivDataSize = iQuadSize * size * size * 4 * sizeof(float);
  iTotalSize = iVertexDataSize + iTexCoordDataSize + iNormalDataSize + iSubDivDataSize;

  Data = new char[iTotalSize];
  lpcVertexData = Data;
  lpcTexCoordData = lpcVertexData + iVertexDataSize;
  lpcNormalData = lpcTexCoordData + iTexCoordDataSize;
  lpcSubDivData = lpcNormalData + iNormalDataSize;

  WriteQuad(size,8,iVertexNum,0);
  WriteQuad(size,4,iVertexNum,1);
  WriteQuad(size,2,iVertexNum,2);
  WriteQuad(size,1,iVertexNum,3);

  glBindBufferARB(GL_ARRAY_BUFFER_ARB,VertexBufferMain);
  glBufferDataARB(GL_ARRAY_BUFFER_ARB,iTotalSize,Data,GL_STATIC_DRAW_ARB);
  glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);

  delete []Data;

  lpcVertexData=NULL;
  lpcTexCoordData = lpcVertexData + iVertexDataSize;
  lpcNormalData = lpcTexCoordData + iTexCoordDataSize;
  lpcSubDivData = lpcNormalData + iNormalDataSize;

  #ifdef TH_LANDSCAPE_DEBUG
  LandQuad *ptr;
  int x,y,i,iRes=0;
  printf("Size: %i\n",iSizes[iRes]);
  printf("Total Vertices: %i\n",iVertexNum);
  for(y=0;y<size;y++){
    for(x=0;x<size;x++){
      ptr = &Quads[y*size+x];
      printf("Quad %ix%ix\nSizes: ",x,y);
      for(i=0;i<iSizes[iRes];i++)
        printf("%i ",ptr->iSizesMain[iRes][i]);
      printf("\nStarts: ");
      for(i=0;i<iSizes[iRes];i++)
        printf("%i ",ptr->iStartsMain[iRes][i]);
      printf("\n\n");
    }
  }

  MessageBox(NULL,"debug","debug",MB_OK);
  #endif

}

void Landscape::DrawBoxes(){
  for(int i=0;i<iQuadWidth*iQuadWidth;i++){
    if(Quads[i].BoundingBox.IsInFrustrum())
      glColor3f(0.0f,1.0f,0.0f);
    else
      glColor3f(1.0f,0.0f,0.0f);
    Quads[i].BoundingBox.Draw();
  }
}

void Landscape::TestDraw(Shader *ptr){
  int iRes;
  vec4 v4Temp;
  glPushMatrix();
  glTranslatef(vPosition.x-fSize/2,vPosition.y-fSize/2,vPosition.z);
  glBindBufferARB(GL_ARRAY_BUFFER_ARB,VertexBufferMain);
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
  glClientActiveTexture(GL_TEXTURE0);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  glTexCoordPointer(4,GL_FLOAT,0,lpcTexCoordData);
  glClientActiveTexture(GL_TEXTURE1);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  glTexCoordPointer(4,GL_FLOAT,0,lpcSubDivData);
  glVertexPointer(3,GL_FLOAT,0,lpcVertexData);
  glNormalPointer(GL_FLOAT,0,lpcNormalData);
    for(int i=0;i<iQuadWidth*iQuadWidth;i++)
      if(Quads[i].BoundingBox.IsInFrustrum()){
          iRes=Quads[i].iRenderRes;
        if(i>0){
          v4Temp.x = (Quads[i-1].iRenderRes > Quads[i].iRenderRes) ? 1 : 0;
          if(Quads[i-1].iRenderRes == Quads[i].iRenderRes)
            v4Temp.x = (Quads[i].fBlendFaktor + Quads[i-1].fBlendFaktor)/2.0;
        }
        else v4Temp.x = 0;

        if(i<iQuadWidth*iQuadWidth-1){
          v4Temp.y = (Quads[i+1].iRenderRes > Quads[i].iRenderRes) ? 1 : 0;
          if(Quads[i+1].iRenderRes == Quads[i].iRenderRes)
            v4Temp.y = (Quads[i].fBlendFaktor + Quads[i+1].fBlendFaktor)/2.0;
        }
        else v4Temp.y = 0;

        if(i>=iQuadWidth){
          v4Temp.z = (Quads[i-iQuadWidth].iRenderRes > Quads[i].iRenderRes) ? 1 : 0;
          if(Quads[i-iQuadWidth].iRenderRes == Quads[i].iRenderRes)
            v4Temp.z = (Quads[i].fBlendFaktor + Quads[i-iQuadWidth].fBlendFaktor)/2.0;
        }
        else v4Temp.z = 0;

        if(i < iQuadWidth * iQuadWidth - iQuadWidth){
          v4Temp.w = (Quads[i+iQuadWidth].iRenderRes > Quads[i].iRenderRes) ? 1 : 0;
          if(Quads[i+iQuadWidth].iRenderRes == Quads[i].iRenderRes)
            v4Temp.w = (Quads[i].fBlendFaktor + Quads[i+iQuadWidth].fBlendFaktor)/2.0;
        }
        else v4Temp.w = 0;

        ptr->SetUniform(0,Quads[i].fBlendFaktor);
        ptr->SetUniform(1,v4Temp);
      //glColor3f(Quads[i].fBlendFaktor,1.0f-Quads[i].fBlendFaktor,0.0f);
      glMultiDrawArrays(GL_TRIANGLE_STRIP,Quads[i].iStartsMain[iRes],Quads[i].iSizesMain[iRes],iSizes[iRes]);
    }
  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_NORMAL_ARRAY);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  glClientActiveTexture(GL_TEXTURE0);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
  glPopMatrix();
}

void Landscape::TestGenerate(vec4 CamPos, double TimeDiff){
  double IntPart;
  vec4 temp,vPos;
  vPos.x = vPosition.x - fSize / 2.0f;
  vPos.y = vPosition.y - fSize / 2.0f;
  vPos.z = vPosition.z;
  float width = (fSize / (float)iQuadWidth)/2;
  float fDistance;
  int i;
  for(i=0;i<iQuadWidth * iQuadWidth;i++){
    temp.x = Quads[i].fPosX + vPos.x + width;
    temp.y = Quads[i].fPosY + vPos.y + width;
    temp.z = vPos.z + (Quads[i].fHeight / 2);
    temp = SubV4(temp,CamPos);
    temp = MulV4(temp,temp);
    fDistance = sqrt(temp.x + temp.y + temp.z);
    Quads[i].iRes = (int)floor(fDistance / (fSize*fMaxDistance));
    Quads[i].fBlendFaktor = modf(fDistance / (fSize*fMaxDistance),&IntPart);
    Quads[i].bBlendLarger=true;
    if(Quads[i].iRes > 4){
      Quads[i].iRes = 4;
      Quads[i].fBlendFaktor=1.0f;
    }
    if(Quads[i].iRes < 1){
      Quads[i].iRes = 1;
      Quads[i].fBlendFaktor=0.0f;
    }
    Quads[i].iRenderRes=Quads[i].iRes-1;
  }
}

void Landscape::Draw(){
  float width = fSize / (float)iQuadWidth;
  float tex = 1.0f / (float)iQuadWidth;
  float tx,ty,fPosX,fPosY;
  glPushMatrix();

  /*glLineWidth(2.0f);
  glDisable(GL_LIGHTING);

  for(int i=0;i<iQuadWidth * iQuadWidth;i++){
    switch(Quads[i].iRes){
      case 1:
        glColor3f(1.0f,0.0f,0.0f);
        break;
      case 2:
        glColor3f(0.0f,1.0f,0.0f);
        break;
      case 3:
        glColor3f(0.0f,0.0f,1.0f);
        break;
      case 4:
        glColor3f(0.0f,1.0f,1.0f);
        break;
    }
    //glColor3f(1.0f/(float)Quads[i].iRes,0.0f,0.0f);
    if(Quads[i].BoundingBox.IsInFrustrum())
      Quads[i].BoundingBox.Draw();
  }*/

  glTranslatef(vPosition.x - (fSize /2), vPosition.y - (fSize/2), vPosition.z);

  glPopMatrix();
}

vec4 Landscape::GetPosition(float x, float y){
  vec4 res,v1,r1,r2;
  int x1,x2,y1,y2;
  if(x > 1.0f || x < 0.0f || y > 1.0f || y < 0.0f){
    res.set(0.0f);
    return res;
  }

  x1 = (int)floor(x*(float)(iVertexSize-2));
  y1 = (int)floor(y*(float)(iVertexSize-2));
  x2 = x1+1;
  y2 = y1+1;

  if(x2 >= iVertexSize){
    x1--; x2--;
  }

  if(y2 >= iVertexSize){
    y1--; y2--;
  }

  x = fmod(x,1.0f/(float)(iVertexSize-1));
  y = fmod(x,1.0f/(float)(iVertexSize-1));

  if( x + y < 1.0f){
    v1 = PlacePositions[y1 * iVertexSize + x1];
    r1 = SubV4(v1,PlacePositions[y2 * iVertexSize + x1]);
    r2 = SubV4(v1,PlacePositions[y1 * iVertexSize + x2]);
  }
  else {
    x -= 1.0f;
    y -= 1.0f;
    v1 = PlacePositions[y2 * iVertexSize + x2];
    r1 = SubV4(v1,PlacePositions[y2 * iVertexSize + x1]);
    r2 = SubV4(v1,PlacePositions[y1 * iVertexSize + x2]);
  }

  res = AddV4(AddV4(v1,MulV4(r1,x)),MulV4(r2,y));
  res = res + vec4(vPosition.x - (fSize /2), vPosition.y - (fSize/2), vPosition.z,1.0f);
  res.w = 1.0f;
  return res;
}

vec4 Landscape::GetNormal(float x, float y){
  vec4 res,v1,r1,r2;
  int x1,x2,y1,y2;
  if(x > 1.0f || x < 0.0f || y > 1.0f || y < 0.0f){
    res.set(0.0f);
    return res;
  }

  x1 = (int)floor(x*(float)(iVertexSize-2));
  y1 = (int)floor(y*(float)(iVertexSize-2));
  x2 = x1+1;
  y2 = y1+1;

  if(x2 >= iVertexSize){
    x1--; x2--;
  }

  if(y2 >= iVertexSize){
    y1--; y2--;
  }

  x = fmod(x,1.0f/(float)(iVertexSize-1));
  y = fmod(x,1.0f/(float)(iVertexSize-1));

  if( x + y < 1.0f){
    v1 = PlacePositions[y1 * iVertexSize + x1];
    r1 = SubV4(v1,PlacePositions[y2 * iVertexSize + x1]);
    r2 = SubV4(v1,PlacePositions[y1 * iVertexSize + x2]);
  }
  else {
    x -= 1.0f;
    y -= 1.0f;
    v1 = PlacePositions[y2 * iVertexSize + x2];
    r1 = SubV4(v1,PlacePositions[y2 * iVertexSize + x1]);
    r2 = SubV4(v1,PlacePositions[y1 * iVertexSize + x2]);
  }

  res = NormalizeV4(CrossProductV4(r2,r1));
  res.w = 1.0f;
  return res;
}

