/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "GLee.h"
#include "light.h"
#include "vektoren.h"

const GLenum LightNumbers[]={GL_LIGHT0,GL_LIGHT1,GL_LIGHT2,GL_LIGHT3,GL_LIGHT4,GL_LIGHT5,GL_LIGHT6,GL_LIGHT7};

mat4a Light::ViewMatrix;

Light::Light(unsigned short number){
  int i;
  if(number > 7)
    number=7;
  usLightNumber=number;
  for(i=0;i<4;i++){
    fColor[i]=0.0f;
    fAmbient[i]=0.0f;
    fPosition[i]=0.0f;
  }
  fPosition[3]=1.0f;
  bOn=false;
}

void Light::Enable(){
  if(!bOn){
    glEnable(LightNumbers[usLightNumber]);
    bOn=true;
  }
  glLightfv(LightNumbers[usLightNumber],GL_AMBIENT,fAmbient);
  glLightfv(LightNumbers[usLightNumber],GL_DIFFUSE,fColor);
  glLightfv(LightNumbers[usLightNumber],GL_POSITION,fPosition);
}

void Light::Disable(){
  if(bOn){
    glDisable(LightNumbers[usLightNumber]);
    bOn=false;
  }
}

void Light::SetPosition(float x, float y, float z){
  fPosition[0]=x;
  fPosition[1]=y;
  fPosition[2]=z;
  fPosition[3]=1.0f;
}

void Light::SetDirection(float x, float y, float z){
  fPosition[0]=x;
  fPosition[1]=y;
  fPosition[2]=z;
  fPosition[3]=0.0f;
}

void Light::SetPosition(vec3 Position){
  fPosition[0]=Position.x;
  fPosition[1]=Position.y;
  fPosition[2]=Position.z;
  fPosition[3]=1.0f;
}

void Light::SetPosition(vec4 Position){
  fPosition[0]=Position.x;
  fPosition[1]=Position.y;
  fPosition[2]=Position.z;
  fPosition[3]=Position.w;
}

void Light::SetColor(float r, float g, float b, float a){
  fColor[0]=r;
  fColor[1]=g;
  fColor[2]=b;
  fColor[3]=a;
  if(bOn)
    glLightfv(LightNumbers[usLightNumber],GL_DIFFUSE,fColor);
}

void Light::SetAmbient(float r, float g, float b, float a){
  fAmbient[0]=r;
  fAmbient[1]=g;
  fAmbient[2]=b;
  fAmbient[3]=a;
  if(bOn)
    glLightfv(LightNumbers[usLightNumber],GL_AMBIENT,fAmbient);
}

vec4 Light::GetPosV4(){
  vec4 temp;
  temp.x = fPosition[0];
  temp.y = fPosition[1];
  temp.z = fPosition[2];
  temp.w = fPosition[3];
  return temp;
}

void Light::SetViewMatrix(mat4a Matrix){
  ViewMatrix = Matrix;
}

void Light::Update(){
  if(bOn)
    glLightfv(LightNumbers[usLightNumber],GL_POSITION,fPosition);
}

