/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "thEngine.h"
#include "camera.h"
#include "fbo.h"
#include "shader.h"
#include "frustrum.h"
#include "plants.h"
#include "landscape.h"
#include "xmlshader.h"
#include "vertexbuffer.h"
#include "manager.h"
#include "gui.h"
#include "valueeditor.h"
#include "kameraeditor.h"
#include "sound.h"
#include "ois.h"
#include <AL/alc.h>
#include <math.h>
// TODO (Benjamin#1#): Shader Engine Werte einfgen
#include <OISKeyboard.h>


char ThWindowTitle[]="OpenGL Wald Demo by Benjamin Thaut";

ALCdevice* SoundDevice=NULL;
ALCcontext* SoundContext=NULL;
SoundOggStream *BackgroundSound = NULL;

Listener MainCam;
camera LightCam;
float fDistance;
float fRotation;
float fVerzerrung=50.0f;
vec4 CamPos;
vec4 CamDir;

vec4 LightDir;
vec4 LightFrom;
vec4 LightTo;
Light *MyLight=NULL;

mat4a LightModelView;
mat4a LightProjection;
mat4a CameraModelViewInverse;
mat4a ShadowMappingMatrix;

vec4 CamShadowMapPos;

Model *BaumBlaetter=NULL;
Model *BaumStamm=NULL;
Model *BaumStamm2=NULL;
Model *BaumBlaetter2=NULL;
Texture2D *BaumTextur=NULL;
Texture2D *BaumBump=NULL;
Texture2D *BaumBlaetterTextur=NULL;
XMLShader *BaumShader=NULL;
XMLShader *BaumNoTexShader=NULL;
TreeManager *Baeume;

Model *SkyDome=NULL;
XMLShader *SkyShader=NULL;
Texture2D *CloudTexture=NULL;

Model *Cube=NULL;
Texture2D *CubeMap=NULL;

PlantManager *Plants=NULL;
PlantGrasQuadMesh **GrasQuads=NULL;
Model *GrasModel=NULL;
Model *WaterPlane=NULL;
Model *BuschModel=NULL;
Texture2D *GrasTexture=NULL;
XMLShader *GrasShader=NULL;
GLint GrasList;
Texture2D *PlantCoverage=NULL;
int iSmoothGras=0,iWaterResolutionX=0,iWaterResolutionY=0;
float fShadowDistance=200.0f;
float fGrasDistance=10.0f,fWaterHeight=3.0f;

Model *BlumenModel=NULL;
Texture2D *BlumenTextur=NULL;
Texture2D *BlumenTextur2=NULL;

//Shader Constants
ShaderConstant *TimeConstant,*ShadowMappingCamPosConstant,*ShadowMappingMatConstant,*GrasDistanceConstant;
ShaderConstant *ScreenHeightConstant,*ScreenWidthConstant;
//ShaderConstant *SunColorConstant,*SunAmbientConstant,*ShadowColorConstant;
ShaderConstant *SunDirectionConstant;

XMLShader *WaterShader=NULL;
XMLShader *BillboardColorShader=NULL;
XMLShader *BillboardShadowShader=NULL;
Texture2D *LeafColorMap=NULL;
FrameBufferTexture2D *TextureRender=NULL;
FrameBufferTexture2D *TextureRender2=NULL;

XMLShader *TerrainShader=NULL;
XMLShader *TerrainShaderNoTex=NULL;
Landscape *MyTerrain=NULL;
Texture2D *TerrainDetail=NULL;
Texture2D *TerrainColor1=NULL;
Texture2D *TerrainColor2=NULL;
Texture2D *WaterBump=NULL;

FrustrumCamera MainFrustrum;
FrustrumObjectBox **FrustrumObjects=NULL;

//GUI Stuff
bool bGuiActive = false;
Gui *theGui = NULL;
ValueEditor ValueEditor1;
CameraEditor CameraEditor1;

Zeitpunkt StartCalc,StopCalc;
double dZeit;

int XmlTexResX=0,XmlTexResY=0;
vec4 XmlWaterColor;
float XmlReflectionOffset=0,XmlReflectionExp=0,XmlDistraction=0;
float XmlFogStart=0,XmlFogSize=0,XmlFogDensity=0;
int XmlGrasGetShadow=0;

bool bOutput=false;
void Output(){
  bOutput=true;
}

void SaveCam(){
  MainCam.SaveToFile("cam.txt");
}

void LoadCam(){
  MainCam.LoadFromFile("cam.txt");
}

bool bShowBoxes=false;
void SwitchBoxes(){
  bShowBoxes=!bShowBoxes;
}

bool bWireframe=false;
void SwitchWireframe(){
  bWireframe=!bWireframe;
}

bool bCamOn=true;
bool MouseWorks = false;
int Mx = 0, My = 0;

void CamSphere(int x,int y){
  if(!bCamOn || bGuiActive)
    return;
  Mx += x;
  My += y;
  float temp;
  //Position auf der Kugel mit dem Radius fDistancen
  //Erst die Drehung um die Z Achse nach dem X-Wert der Maus
  temp=(1.0f-(float)My/(float)iHeight)*M_PI-M_PI_2; //Volle 360
  CamDir.z = sin(temp);
  CamDir.x = cos(temp);
  CamDir.y = 0.0f;
  //Jetzt Kippen nach Y-Wert der Maus
  temp=(1.0f-(float)Mx/(float)iWidth)*M_PI*2;
  CamDir.y = CamDir.x * sin(temp);
  CamDir.x = CamDir.x * cos(temp);
  //Neue Kamera Position setzten
  CamPos = MainCam.GetFrom();
  MainCam.SetTo(CamPos.x+CamDir.x,CamPos.y+CamDir.y,CamPos.z+CamDir.z);
}

void CheckCam(){
  if(MyTerrain==NULL)
    return;
  vec4 temp;
  if(CamPos.x > 100.0f)
    CamPos.x = 100.0f;
  if(CamPos.y > 100.0f)
    CamPos.y = 100.0f;
  if(CamPos.x < -100.0f)
    CamPos.x = -100.0f;
  if(CamPos.y < -100.0f)
    CamPos.y = -100.0f;
  if(CamPos.z > 100.0f)
    CamPos.z = 100.0f;
  if(CamPos.z < fWaterHeight + 1.0f)
    CamPos.z = fWaterHeight + 1.0f;
  temp = MyTerrain->GetPosition((CamPos.x + 100.0f) / 200.0f,(CamPos.y + 100.0f) / 200.0f);
  if(CamPos.z < temp.z + 1.0f)
    CamPos.z = temp.z + 1.0f;
}

void CamMove(float Move){
  if(!bCamOn)
    return;
  CamPos = MainCam.GetFrom();
  CamPos.x += CamDir.x * dTimeDiff * Move;
  CamPos.y += CamDir.y * dTimeDiff * Move;
  CamPos.z += CamDir.z * dTimeDiff * Move;
  CheckCam();
  MainCam.SetFrom(CamPos.x,CamPos.y,CamPos.z);
  MainCam.SetTo(CamPos.x+CamDir.x,CamPos.y+CamDir.y,CamPos.z+CamDir.z);
}

void CamStrave(float Move){
  vec4 Dir,Up,Right,NewPos;
  Dir = MainCam.GetViewDir();
  Up = MainCam.GetUp();
  Right = CrossProductV4(Dir,Up);
  NewPos = MainCam.GetFrom() + Right * (dTimeDiff * Move);
  CamPos.x = NewPos.x;
  CamPos.y = NewPos.y;
  CamPos.z = NewPos.z;
  CheckCam();
  NewPos.x = CamPos.x;
  NewPos.y = CamPos.y;
  NewPos.z = CamPos.z;
  MainCam.SetFrom(NewPos);
  MainCam.SetTo(NewPos + Dir);
}

void CamRotateLeft(){
  if(!bCamOn || bGuiActive)
    return;
  fRotation = MainCam.GetRotation() - 0.02 * dTimeDiff;
  MainCam.SetRotation(fRotation);
}

void CamRotateRight(){
  if(!bCamOn || bGuiActive)
    return;
  fRotation = MainCam.GetRotation() + 0.02 * dTimeDiff;
  MainCam.SetRotation(fRotation);
}

void CamSwitch(){
  if(!bGuiActive)
    bCamOn=!bCamOn;
}

void ActivateGui(){
  if(bGuiActive){
    if(bIsFullscreen)
      ShowCursor(false);
    bGuiActive = false;
  }
  else {
    if(bIsFullscreen)
      ShowCursor(true);
    bGuiActive = true;
  }
}

void SwitchFps(){
  bShowFps=!bShowFps;
}

void SwitchFreeFlight(){
  CameraEditor1.Play();
}

int iLastMouseX = 0;
int iLastMouseY = 0;

void GuiMouse(int xabs, int yabs, int xrel, int yrel){
  if(bGuiActive && (iLastMouseX != xabs || iLastMouseY != yabs)){
    #ifdef __linux
    theGui->MouseMove(xabs,yabs,xrel,yrel);
    #elif defined WIN32
    theGui->MouseMove(xabs,yabs,xabs - iLastMouseX, yabs - iLastMouseY);
    #endif
    iLastMouseX = xabs;
    iLastMouseY = yabs;
  }
}

void GuiMouseDown(int button, int xabs, int yabs, int xrel, int yrel){
  if(bGuiActive)
    theGui->MouseDown(0,xabs,yabs);
}

void GuiMouseUp(int button, int xabs, int yabs, int xrel, int yrel){
  if(bGuiActive)
    theGui->MouseUp(button,xabs,yabs);
}

void GuiText(char text){
  if(bGuiActive)
    theGui->TextIn(text);
}

void VerzerrungPlus(){
  fVerzerrung += 1.0f;
  if(fVerzerrung > 100.0f)
    fVerzerrung = 100.0f;
}

void VerzerrungMinus(){
  fVerzerrung -= 1.0f;
  if(fVerzerrung < 1.0f)
    fVerzerrung = 1.0f;
}

float LightWinkel=30;
float LightWinkel2=50;

int mode=1;
void SetMode1(){mode=1;}
void SetMode2(){mode=2;}
void SetMode3(){mode=3;}

int iShadowResolution;

double ClippingPlane[]={0.0f,0.0f,-1.0f,0.0f};
//Funktion die aufgrufen wird, wenn das Porgramm startet
bool Init(){
  if(!GLEE_ARB_shader_objects || !GLEE_ARB_shading_language_100){
    ThErrorMessage("Fehler","Kein Shader support");
    return false;
  }

  if(!GLEE_EXT_framebuffer_object){
    ThErrorMessage("Fehler","Kein RenderToTexture support");
    return false;
  }

  //Kamera Initialisieren
  CamPos.x=0;
  CamPos.y=0;
  CamPos.z=3;
  fDistance=3;
  MainCam.SetFrom(0,0,3);
  MainCam.SetTo(0,0,0);
  //Beleuchtung ausschalten (wird nicht gebraucht)
  glDisable(GL_LIGHTING);
  //Funktionen fr Tasten und Maus setzten

  //Zoomfunktionen fr Kamera
  //SetKeyDownFunc(OIS::KC_C,CamSwitch);
  SetKeyDownFunc(OIS::KC_F1,SetMode1);
  SetKeyDownFunc(OIS::KC_F2,SetMode2);
  SetKeyDownFunc(OIS::KC_F3,SetMode3);
  SetKeyDownFunc(OIS::KC_F4,SwitchWireframe);
  SetKeyDownFunc(OIS::KC_F5,SwitchBoxes);
  //SetKeyDownFunc(OIS::KC_F7,Output);
  //SetKeyDownFunc(OIS::KC_NUMPAD0,SaveCam);
  //SetKeyDownFunc(OIS::KC_NUMPAD1,LoadCam);
  SetKeyDownFunc(OIS::KC_F11,SwitchFps);
  SetKeyDownFunc(OIS::KC_F12,ActivateGui);
  //SetKeyDownFunc(OIS::KC_PGUP,VerzerrungPlus);
  //SetKeyDownFunc(OIS::KC_PGDOWN,VerzerrungMinus);
  SetKeyDownFunc(OIS::KC_F6,SwitchFreeFlight);
  SetTextFunc(GuiText);

  LightFrom.x = 0.0f;
  LightFrom.y = 0.0f;
  LightFrom.z = 0.20f;
  LightFrom.w = 0.0f;
  LightTo.x = 0.0f;
  LightTo.y = 0.0f;
  LightTo.z = 1.0f;
  LightTo.w = 0.0f;
  LightDir = NormalizeV4(SubV4(LightTo,LightFrom));

  LightDir.x = LightDir.y = LightDir.z = -1.00f;
  LightDir.z = -0.75f;
  LightDir.w = 1.0f;
  MyLight = new Light(0);
  MyLight->SetDirection(LightDir.x,LightDir.y,LightDir.z);
  MyLight->SetColor(0.75f,0.75f,0.75f,1.0f);
  MyLight->SetAmbient(0.15f,0.15f,0.15f,1.0f);
  MyLight->Enable();

  //Load XML Data
  if(!XMLShader::LoadOptions("config.xml")){
    ThErrorMessage("XML-Fehler","Konnte XML-Optionen nicht laden!");
    return false;
  }

  int iForceFallbackInstancing=0;
  if(!XMLShader::GetOption("ForceFallbackInstancing",iForceFallbackInstancing)){
    ThErrorMessage("XML-Fehler","Konnte option 'ForceFallbackInstancing' nicht finden!");
    return false;
  }

  int iMaxInstances=64, iFallbackInstancing=1;
  if(GLEE_EXT_draw_instanced && GLEE_EXT_bindable_uniform && (bool)iForceFallbackInstancing > 0){
    glGetIntegerv(GL_MAX_BINDABLE_UNIFORM_SIZE_EXT,&iMaxInstances);
    iMaxInstances /= (12 * sizeof(float));
    iFallbackInstancing = 0;
  }
  XMLShader::AddOption("MaxInstances",iMaxInstances);
  XMLShader::AddOption("FallbackInstancing",iFallbackInstancing); //iFallbackInstancing);
  PlantManager::SetFallbackInstancing((bool)iFallbackInstancing);

  TimeConstant = XMLShader::AddShaderConstantFloat("Time");
  ShadowMappingCamPosConstant = XMLShader::AddShaderConstantVec4("ShadowMappingCameraPosition");
  ShadowMappingMatConstant = XMLShader::AddShaderConstantMat4a("ShadowMappingMatrix");
  GrasDistanceConstant = XMLShader::AddShaderConstantFloat("GrasDistance");
  ScreenWidthConstant = XMLShader::AddShaderConstantFloat("ScreenWidth");
  ScreenHeightConstant = XMLShader::AddShaderConstantFloat("ScreenHeight");
  SunDirectionConstant = XMLShader::AddShaderConstantVec3("SunDirection");

  //GuiShaderConstants

  int iShadowBits,iTerrainDetail,iTerrainTiles;
  if(!XMLShader::GetOption("ShadowResolution",iShadowResolution)){
    ThErrorMessage("XML-Fehler","Konnte Option ShadowResolution nicht finden!");
    return false;
  }

  if(!XMLShader::GetOption("ShadowBits",iShadowBits)){
    ThErrorMessage("XML-Fehler","Konnte Option ShadowBits nicht finden!");
    return false;
  }

  if(!XMLShader::GetOption("ShadowDistance",fShadowDistance)){
    ThErrorMessage("XML-Fehler","Konnte Option ShadowDistance nicht finden!");
    return false;
  }

  if(!XMLShader::GetOption("TerrainDetail",iTerrainDetail)){
    ThErrorMessage("XML-Fehler","Konnte Option TerrainDetail nicht finden!");
    return false;
  }

  if(!XMLShader::GetOption("TerrainTiles",iTerrainTiles)){
    ThErrorMessage("XML-Fehler","Konnte Option TerrainTiles nicht finden!");
    return false;
  }

  if(!XMLShader::GetOption("SmoothGras",iSmoothGras)){
    ThErrorMessage("XML-Fehler","Konnte Option SmoothGras nicht finden!");
    return false;
  }

  if(!XMLShader::GetOption("WaterResolutionX",iWaterResolutionX)){
    ThErrorMessage("XML-Fehler","Konnte Option WaterResoultionX nicht finden!");
    return false;
  }

  if(!XMLShader::GetOption("WaterResolutionY",iWaterResolutionY)){
    ThErrorMessage("XML-Fehler","Konnte Option WaterResoultionY nicht finden!");
    return false;
  }

  if(iSmoothGras > 0 && !arbMultisampleSupported)
    iSmoothGras = 0;

  if(!XMLShader::GetOption("GrasDistance",fGrasDistance)){
    ThErrorMessage("XML-Fehler","Konnte Option GrasDistance nicht finden!");
    return false;
  }
  GrasDistanceConstant->Set(&fGrasDistance);

  if(!XMLShader::GetOption("GrasGetShadow",XmlGrasGetShadow)){
    ThErrorMessage("XML-Fehler","Konnte Option 'GrasGetShadow' nicht finden!");
    return false;
  }

  switch(iShadowBits){
    case 16:
      iShadowBits = TH_DEPTH_BUFFER_TEXTURE_16;
      break;
    case 24:
      iShadowBits = TH_DEPTH_BUFFER_TEXTURE_24;
      break;
    case 32:
      iShadowBits = TH_DEPTH_BUFFER_TEXTURE_32;
      break;
    default:
      iShadowBits = TH_DEPTH_BUFFER_TEXTURE_16;
      break;
  }

  TextureRender = new FrameBufferTexture2D(iShadowResolution,iShadowResolution,iShadowBits,TH_RGBA8);
  if(TextureRender->Error() != NULL){
    ThErrorMessage("Frame Buffer Object Error",TextureRender->Error());
    return false;
  }

  TextureRender2 = new FrameBufferTexture2D(iWaterResolutionX,iWaterResolutionY,TH_DEPTH_BUFFER | TH_COLOR0,TH_RGBA8);
  if(TextureRender2->Error() != NULL){
    ThErrorMessage("Frame Buffer Object 2 Error",TextureRender2->Error());
    return false;
  }

  //GUI
  theGui = new Gui();
  if(!theGui->LoadGuiTexture("gfx/guielements.tga"))
    return false;
  if(!theGui->LoadStyles("gfx/styles.xml"))
    return false;
  if(!theGui->LoadGuiShader("shader/gui-shader.xml"))
    return false;

  vec4 TempMin,TempMax;
  TempMin.set(0.0f); TempMax.set(1.0f);

  CameraEditor1.Init(theGui,&MainCam,timer);
  ValueEditor1.Init(theGui);

  if(!ValueEditor1.LoadConstants("shader/ShaderConstants.xml"))
    return false;

  //Bei Bewegung der Maus wird CamSphere(int,int) ausgefhrt
  //Mausbewegungsfunktionen mssen immer "void name(int,int)" entsprechen
  SetMouseMoveFuncRelative(CamSphere);
  SetMouseMoveFunc(GuiMouse);
  SetMouseButtonDownFunc(GuiMouseDown);
  SetMouseButtonUpFunc(GuiMouseUp);

  //Cube = MdlMngr->Load3DS("gfx/cube.3DS",1,TH_NORMALS_PER_FACE | TH_TEXCOORDS);
  GrasModel = MdlMngr->Load3DS("gfx/gras.3DS",1,TH_NORMALS_PER_FACE | TH_TEXCOORDS);
  BlumenModel = MdlMngr->Load3DS("gfx/blumen.3DS",1,TH_NORMALS_PER_FACE | TH_TEXCOORDS);
  WaterPlane = MdlMngr->Load3DS("gfx/plane.3ds",1,TH_NORMALS_PER_FACE | TH_TEXCOORDS);
  BuschModel = MdlMngr->LoadBillboards3DS("gfx/busch.3DS",1);
  vec4 v4Temp = vec4(0.0,0.0,8.0,1.0);
  BaumBlaetter = MdlMngr->LoadBillboardsSchrompf("gfx/LaubBlatt1",v4Temp);
  BaumStamm = MdlMngr->LoadSchrompf("gfx/LaubStamm1",TH_NORMALS_PER_VERTEX | TH_TEXCOORDS | TH_TANGENT | TH_BINORMAL);
  BaumBlaetter2 = MdlMngr->LoadBillboardsSchrompf("gfx/LaubBlatt2",v4Temp);
  BaumStamm2 = MdlMngr->LoadSchrompf("gfx/LaubStamm2",TH_NORMALS_PER_VERTEX | TH_TEXCOORDS | TH_TANGENT | TH_BINORMAL);
  SkyDome = MdlMngr->Load3DS("gfx/skydome.3DS",1,0);
  if(MdlMngr->Error()){
    ThErrorMessage("Modell Fehler","Fehler beim Laden der Modelle\nSiehe engine.log fr nhere Informationen.");
    return false;
  }

  //CubeMap = TexMngr->LoadTGA("gfx/cubemap.tga",TH_LINEAR);
  GrasTexture = TexMngr->LoadTGA("gfx/gras2.tga",TH_LINEAR | TH_MIPMAPS);
  PlantCoverage = TexMngr->LoadTGA("gfx/coverage.tga",TH_NEAREST);
  LeafColorMap = TexMngr->LoadTGA("gfx/busch.tga",TH_LINEAR | TH_MIPMAPS);
  TerrainDetail = TexMngr->LoadTGA("gfx/detail.tga",TH_LINEAR);
  TerrainColor1 = TexMngr->LoadTGA("gfx/gras.tga",TH_LINEAR | TH_MIPMAPS);
  TerrainColor2 = TexMngr->LoadTGA("gfx/rock.tga",TH_LINEAR | TH_MIPMAPS);
  BlumenTextur = TexMngr->LoadTGA("gfx/blumen1.tga",TH_LINEAR | TH_MIPMAPS);
  BlumenTextur2 = TexMngr->LoadTGA("gfx/blumen2.tga",TH_LINEAR | TH_MIPMAPS);
  WaterBump = TexMngr->LoadTGA("gfx/water.tga",TH_LINEAR | TH_MIPMAPS);
  BaumBlaetterTextur = TexMngr->LoadTGA("gfx/BaumBlaetter.tga",TH_LINEAR | TH_MIPMAPS);
  BaumTextur = TexMngr->LoadTGA("gfx/BaumTextur.tga",TH_LINEAR | TH_MIPMAPS);
  BaumBump = TexMngr->LoadTGA("gfx/BaumBump.tga",TH_LINEAR | TH_MIPMAPS);
  CloudTexture = TexMngr->LoadTGA("gfx/clouds.tga",TH_LINEAR | TH_MIPMAPS);
  if(TexMngr->Error()){
    ThErrorMessage("Textur Fehler","Fehler beim Laden der Texturen\nSiehe engine.log fr nhere Informationen.");
    return false;
  }

  TerrainShader = ShdrMngr->LoadXml("shader/Terrain.xml");
  TerrainShaderNoTex = ShdrMngr->LoadXml("shader/TerrainNoTex.xml");
  GrasShader = ShdrMngr->LoadXml("shader/Gras.xml");
  BillboardColorShader = ShdrMngr->LoadXml("shader/BillboardColor.xml");
  WaterShader = ShdrMngr->LoadXml("shader/water.xml");
  BaumShader = ShdrMngr->LoadXml("shader/Baum.xml");
  BaumNoTexShader = ShdrMngr->LoadXml("shader/BaumNoTex.xml");
  BillboardShadowShader = ShdrMngr->LoadXml("shader/BillboardShadow.xml");
  SkyShader = ShdrMngr->LoadXml("shader/SkyShader.xml");
  if(ShdrMngr->Error()){
    ThErrorMessage("Shader Fehler","Fehler beim Laden der Shader\nSiehe engine.log fr nhere Informationen.");
    return false;
  }

  MyTerrain = new Landscape(iTerrainDetail);
  if(!MyTerrain->LoadHeightMapTGA("gfx/heightmap.tga"))
    return false;
  MyTerrain->SetPosition(0.0f,0.0f,0.0f);
  MyTerrain->SetSize(200.0f);
  MyTerrain->SetDistance(0.1);
  MyTerrain->SetHeight(15.0f);
  MyTerrain->CreateQuads(iTerrainTiles);

  //Trees
  Baeume = new TreeManager();
    //Shader
    Baeume->AddShader(BaumShader->GetShader()); // 0
    Baeume->AddShader(BaumNoTexShader->GetShader()); // 1
    Baeume->AddShader(BillboardColorShader->GetShader()); // 2
    Baeume->AddShader(BillboardShadowShader->GetShader()); // 3

    //Texturen
    Baeume->AddTexture(BaumTextur);
    Baeume->AddTexture(BaumBump);
    Baeume->AddTexture(BaumBlaetterTextur);
    Baeume->AddTexture(LeafColorMap);

    //Models
    TreeInfo TempTree;
    //Baum Stamm
    TempTree.vbuffer = BaumStamm->GetVertexBuffer();
    TempTree.ColorShader = 0;
    TempTree.ShadowShader = 1;
    TempTree.Texture = 0;
    TempTree.BumpTexture = 1;
    TempTree.ChildOf = -1;
    TempTree.ColorOptions = TH_VB_NORMAL | TH_VB_TEXCOORD0 | TH_VB_ATTRIBUTE0 | TH_VB_ATTRIBUTE1;
    TempTree.ShadowOptions = 0;
    TempTree.Color.x = 1.0f; TempTree.Color.y = 1.0f; TempTree.Color.z = 0.0f; TempTree.Color.w = 1.0f;
    TempTree.bTextureInShadowNeeded=false;
    Baeume->AddModel(&TempTree);

    //Baum Blaetter
    TempTree.vbuffer = BaumBlaetter->GetVertexBuffer();
    TempTree.ColorShader = 2;
    TempTree.ShadowShader = 3;
    TempTree.Texture = 2;
    TempTree.BumpTexture = -1;
    TempTree.ChildOf = 0;
    TempTree.ColorOptions = TH_VB_NORMAL | TH_VB_TEXCOORD0 | TH_VB_TEXCOORD1;
    TempTree.ShadowOptions = TH_VB_TEXCOORD0 | TH_VB_TEXCOORD1;
    TempTree.Color.x = 1.0f; TempTree.Color.y = 1.0f; TempTree.Color.z = 0.0f; TempTree.Color.w = 1.0f;
    TempTree.bTextureInShadowNeeded=true;
    Baeume->AddModel(&TempTree);

    //Busch
    TempTree.vbuffer = BuschModel->GetVertexBuffer();
    TempTree.ColorShader = 2;
    TempTree.ShadowShader = 3;
    TempTree.Texture = 3;
    TempTree.BumpTexture = -1;
    TempTree.ChildOf = -1;
    TempTree.ColorOptions = TH_VB_NORMAL | TH_VB_TEXCOORD0 | TH_VB_TEXCOORD1;
    TempTree.ShadowOptions = TH_VB_TEXCOORD0 | TH_VB_TEXCOORD1;
    TempTree.Color.x = 1.0f; TempTree.Color.y = 1.0f; TempTree.Color.z = 1.0f; TempTree.Color.w = 1.0f;
    TempTree.bTextureInShadowNeeded=true;
    Baeume->AddModel(&TempTree);

    //Baum Stamm2
    TempTree.vbuffer = BaumStamm2->GetVertexBuffer();
    TempTree.ColorShader = 0;
    TempTree.ShadowShader = 1;
    TempTree.Texture = 0;
    TempTree.BumpTexture = 1;
    TempTree.ChildOf = -1;
    TempTree.ColorOptions = TH_VB_NORMAL | TH_VB_TEXCOORD0 | TH_VB_ATTRIBUTE0 | TH_VB_ATTRIBUTE1;
    TempTree.ShadowOptions = 0;
    TempTree.Color.x = 0.0f; TempTree.Color.y = 1.0f; TempTree.Color.z = 1.0f; TempTree.Color.w = 1.0f;
    TempTree.bTextureInShadowNeeded=false;
    Baeume->AddModel(&TempTree);

    //Baum Blaetter2
    TempTree.vbuffer = BaumBlaetter2->GetVertexBuffer();
    TempTree.ColorShader = 2;
    TempTree.ShadowShader = 3;
    TempTree.Texture = 2;
    TempTree.BumpTexture = -1;
    TempTree.ChildOf = 3;
    TempTree.ColorOptions = TH_VB_NORMAL | TH_VB_TEXCOORD0 | TH_VB_TEXCOORD1;
    TempTree.ShadowOptions = TH_VB_TEXCOORD0 | TH_VB_TEXCOORD1;
    TempTree.Color.x = 0.0f; TempTree.Color.y = 1.0f; TempTree.Color.z = 1.0f; TempTree.Color.w = 1.0f;
    TempTree.bTextureInShadowNeeded=true;
    Baeume->AddModel(&TempTree);

  Baeume->Generate(PlantCoverage,MyTerrain);
  //Trees Ende

  PlantManagerSoft *SoftPlant;
  SoftPlant = new PlantManagerSoft(MyTerrain,16);
  SoftPlant->SetGetShadow(XmlGrasGetShadow);
  SoftPlant->SetDepthTexture(TextureRender);
  SoftPlant->SetGrasDistance(fGrasDistance);
  SoftPlant->AddTexture(GrasTexture);
  SoftPlant->AddTexture(BlumenTextur);
  SoftPlant->AddTexture(BlumenTextur2);
  SoftPlant->AddShader(GrasShader);
  v4Temp.x = 1.0; v4Temp.y = 0.0; v4Temp.z = 0.0;
  SoftPlant->AddModel(GrasModel,v4Temp,0,0);
  v4Temp.x = 0.0; v4Temp.y = 1.0; v4Temp.z = 0.0;
  SoftPlant->AddModel(BlumenModel,v4Temp,1,0);
  v4Temp.x = 0.0; v4Temp.y = 0.0; v4Temp.z = 1.0;
  SoftPlant->AddModel(BlumenModel,v4Temp,2,0);
  SoftPlant->Generate(PlantCoverage);
  Plants = SoftPlant;

  FrustrumObject::SetActiveFrustrum(&MainFrustrum);

  if(iSmoothGras > 0)
    glSampleCoverageARB(0.25,false);

  glAlphaFunc(GL_GREATER,0.75f);

  StartCalc = Runtime;
  StopCalc = Runtime;

  CamSphere(3,4);
  CamMove(0.25);

  //Sound
  SoundDevice = alcOpenDevice(0);
  if(SoundDevice == NULL){
    ThErrorMessage("Sound Fehler","Konnte Sound Device nicht oeffnen!");
    return false;
  }

  SoundContext = alcCreateContext(SoundDevice, 0);
  if(SoundContext == NULL){
    ThErrorMessage("Sound Fehler","Konnte Sound Context nicht erstellen!");
    alcCloseDevice(SoundDevice);
    return false;
  }

  if(!alcMakeContextCurrent(SoundContext)){
    ThErrorMessage("Sound Fehler","Konnte Context nicht aktivieren!");
    alcDestroyContext(SoundContext);
    alcCloseDevice(SoundDevice);
    return false;
  }

  BackgroundSound = new SoundOggStream();
  BackgroundSound->LoadOgg("sfx/background.ogg");
  BackgroundSound->SetRepeat(true);
  BackgroundSound->Update();
  BackgroundSound->Play();

  CameraEditor1.LoadFromFile("gfx/cam.dat");
  CameraEditor1.Play();

  return true;
}

//Funktion die aufgerufen wird, wenn sich das Programm beendet
void DeInit(){
  BackgroundSound->Stop();
  if(MyLight != NULL)
    delete MyLight;
  if(MyTerrain != NULL)
     delete MyTerrain;
  if(TextureRender != NULL)
    delete TextureRender;
  if(Plants != NULL)
    delete Plants;
  if(Baeume != NULL)
    delete Baeume;
  if(theGui != NULL)
    delete theGui;
  if(BackgroundSound != NULL)
    delete BackgroundSound;
  alcMakeContextCurrent(NULL);
  if(SoundContext != NULL)
    alcDestroyContext(SoundContext);
  if(SoundDevice != NULL)
    alcCloseDevice(SoundDevice);
}

float fStep=0.0f;

void PreDraw(){
  if(!bGuiActive && !CameraEditor1.IsPlaying()){
    if(MouseWorks)
      CamSphere(0,0);
    if(KeyIsPressed(OIS::KC_W))
      CamMove(0.025f);
    if(KeyIsPressed(OIS::KC_S))
      CamMove(-0.025f);
    if(KeyIsPressed(OIS::KC_A))
      CamStrave(-0.005f);
    if(KeyIsPressed(OIS::KC_D))
      CamStrave(0.005f);
    if(KeyIsPressed(OIS::KC_Q))
      CamRotateLeft();
    if(KeyIsPressed(OIS::KC_E))
      CamRotateRight();
    if(KeyIsPressed(OIS::KC_LEFT)){
      LightWinkel += dTimeDiff * 0.01;
      if(LightWinkel > 170.0f)
        LightWinkel = 170.0f;
    }
    if(KeyIsPressed(OIS::KC_RIGHT)){
      LightWinkel -= dTimeDiff * 0.01;
      if(LightWinkel < 10.0f)
        LightWinkel = 10.0f;
    }
    if(KeyIsPressed(OIS::KC_UP)){
      LightWinkel2 += dTimeDiff * 0.01;
      if(LightWinkel2 > 170.0f)
        LightWinkel2 = 170.0f;
    }
    if(KeyIsPressed(OIS::KC_DOWN)){
      LightWinkel2 -= dTimeDiff * 0.01;
      if(LightWinkel2 < 10.0f)
        LightWinkel2 = 10.0f;
    }
  }
  if(CameraEditor1.IsPlaying()){
    float t = CameraEditor1.UpdateCam();
    LightWinkel = 30.0f + 80 * (t/30.0f);
    if(LightWinkel > 110.0f)
      LightWinkel = 110.0f;
    LightWinkel2 = 50.0f;
  }
  float temp;
  temp = (float)iWidth;
  ScreenWidthConstant->Set(&temp);
  temp = (float)iHeight;
  ScreenHeightConstant->Set(&temp);
  BackgroundSound->Update();
}

void Draw(){
  vec4 Points[8];
  vec4 Points2[8];
  int VisibleObjects=0;
  fStep+=dTimeDiff/1000.0;
  float time = (float)(timer->GetTimeDouble()/1000.0);
  TimeConstant->Set(&time);
  //if(fStep > 1.0f)
    //fStep -= 1.0f

  MyTerrain->TestGenerate(MainCam.GetFrom(),dTimeDiff);

  LightDir.z = -sin(LightWinkel / 180 * M_PI);
  LightDir.y = cos(LightWinkel / 180 * M_PI);
  //LightDir.z = -1.0f;
  //LightDir.y = 0.0f;
  LightDir.x = -sin(LightWinkel2 / 180 * M_PI);
  LightDir.z = cos(LightWinkel2 / 180 * M_PI) * LightDir.z;
  LightDir = NormalizeV4(LightDir);
  MyLight->SetDirection(LightDir.x,LightDir.y,LightDir.z);
  SunDirectionConstant->Set(&LightDir);


  //Schatten Map Rendern
  TextureRender->Use();
  glClear(GL_DEPTH_BUFFER_BIT);
  glEnable(GL_LIGHTING);
  glDisable(GL_TEXTURE_2D);
  glLoadIdentity();
  MainCam.Call();
  MainFrustrum.Update();
  MainFrustrum.MovePlane(TH_PLANE_FAR,-(fFarClipping-fShadowDistance));
  MainFrustrum.Points(Points);
  MainFrustrum.Points(Points2);


  glLoadIdentity();
  LightCam.SetFrom(LightFrom);
  LightCam.SetTo(AddV4(LightFrom,LightDir));
  //LightCam.SetFrom(0.0f,0.0f,0.0f);
  //LightCam.SetTo(0.0f,0.0f,-1.0f);
  LightCam.Call();
  glPushMatrix();
  //glScalef(1.0f,1.0f,1.0f);
  glGetFloatv(GL_MODELVIEW_MATRIX,LightModelView.f);
  if(bOutput)
    for(int i=0;i<4;i++)
      printf("%f %f %f %f\n",LightModelView.f[i],LightModelView.f[i+1],LightModelView.f[i+2],LightModelView.f[i+3]);
  glPopMatrix();
  vec4 min,max;
  for(int i=0;i<8;i++){
    Points[i] = MulV4(Mul(LightModelView,Points[i]),-1.0f);
    Points[i].z *= -1.0f;
    //Points[i] = Mul(LightModelView,Points[i]);
    if(i==0){
      min = Points[i];
      max = Points[i];
    }
    min.x = (Points[i].x < min.x) ? Points[i].x : min.x;
    max.x = (Points[i].x > max.x) ? Points[i].x : max.x;
    min.y = (Points[i].y < min.y) ? Points[i].y : min.y;
    max.y = (Points[i].y > max.y) ? Points[i].y : max.y;
    min.z = (Points[i].z < min.z) ? Points[i].z : min.z;
    max.z = (Points[i].z > max.z) ? Points[i].z : max.z;
  }

  if(bOutput){
    printf("\nX: Min: %f, Max %f\n",min.x,max.x);
    printf("Y: Min: %f, Max %f\n",min.y,max.y);
    printf("Z: Min: %f, Max %f\n",min.z,max.z);
    printf("Points:\n");
    for(int i=0;i<8;i++)
      printf("%i: %f %f %f\n",i,Points[i].x,Points[i].y,Points[i].z);
    //bOutput=false;
  }

  float TexelSizeX = (max.x - min.x) / (float)iShadowResolution;
  float TexelSizeY = (max.y - min.y) / (float)iShadowResolution;
  double GanzZahl;
  float DiffX = TexelSizeX * modf(min.x / TexelSizeX,&GanzZahl);
  float DiffY = TexelSizeY * modf(min.y / TexelSizeY,&GanzZahl);
  min.x -= DiffX; max.x -= DiffX;
  min.y -= DiffY; max.y -= DiffY;

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(min.x,max.x,min.y,max.y,min.z-100.0f,max.z);
  glMatrixMode(GL_MODELVIEW);

  //glLoadIdentity();
  //MainCam.Call();
  //glGetFloatv(GL_MODELVIEW_MATRIX,LightModelView.f);
  //glLoadIdentity();
  //LightCam.SetFrom(MyLight->GetPosV4());
  //LightCam.SetTo(AddV4(MyLight->GetPosV4(),LightDir));
  //LightCam.SetFrom(Mul(LightModelView,LightFrom));
  //LightCam.SetTo(Mul(LightModelView,LightTo));
  //LightCam.Call();
  //glGetFloatv(GL_MODELVIEW_MATRIX,LightModelView.f);
  //glLoadIdentity();
  //MainCam.Call();
  //glMultMatrixf(LightModelView.f);
  //glLoadIdentity();
  //MainCam.Call();
  MainFrustrum.Update();
  /*MainFrustrum.MovePlane(TH_PLANE_LEFT,10.0f);
  MainFrustrum.MovePlane(TH_PLANE_RIGHT,10.0f);
  MainFrustrum.MovePlane(TH_PLANE_TOP,10.0f);
  MainFrustrum.MovePlane(TH_PLANE_BOTTOM,10.0f);
  MainFrustrum.MovePlane(TH_PLANE_FAR,300.0f);
  MainFrustrum.MovePlane(TH_PLANE_NEAR,300.0f);*/
  //MainFrustrum.Points(Points);
  MyLight->Update();

  glGetFloatv(GL_MODELVIEW_MATRIX,LightModelView.f);
  glGetFloatv(GL_PROJECTION_MATRIX,LightProjection.f);

  CamShadowMapPos = MainCam.GetFrom();
  if(bOutput)
    printf("CamFrom : %f %f %f %f\n",CamShadowMapPos.x,CamShadowMapPos.y,CamShadowMapPos.z,CamShadowMapPos.w);
  CamShadowMapPos.w = 1.0f;
  CamShadowMapPos = Mul(LightModelView,CamShadowMapPos);
  if(bOutput)
    printf("Mul1 : %f %f %f %f\n",CamShadowMapPos.x,CamShadowMapPos.y,CamShadowMapPos.z,CamShadowMapPos.w);
  CamShadowMapPos = Mul(LightProjection,CamShadowMapPos);
  if(bOutput)
    printf("Mul2 : %f %f %f %f\n",CamShadowMapPos.x,CamShadowMapPos.y,CamShadowMapPos.z,CamShadowMapPos.w);
  CamShadowMapPos = DivV4(CamShadowMapPos,CamShadowMapPos.w);
  if(bOutput){
    printf("Div : %f %f %f %f\n",CamShadowMapPos.x,CamShadowMapPos.y,CamShadowMapPos.z,CamShadowMapPos.w);
    bOutput = false;
  }
  CamShadowMapPos.z = fVerzerrung;
  CamShadowMapPos.w = 1.0/(log(fVerzerrung+1)/log(2));
  ShadowMappingCamPosConstant->Set(&CamShadowMapPos);

  TerrainShaderNoTex->GetShader()->Use();
  TerrainShaderNoTex->GetShader()->SetUniform(2,CamShadowMapPos);
  MyTerrain->TestDraw(TerrainShaderNoTex->GetShader());
  UnloadShader();

  glEnable(GL_TEXTURE_2D);
  glEnable(GL_ALPHA_TEST);

  Baeume->Draw(false);

  glDisable(GL_ALPHA_TEST);

  for(int i=0;i<8;i++){
    Points2[i].x *= -1;
    Points2[i].y *= -1;
    Points2[i].z *= -1;
  }

  /*glDisable(GL_LIGHTING);
  glLineWidth(5.0f);
  glColor4f(0.0f,1.0f,0.0f,1.0f);
  glBegin(GL_LINES);
    glVertex3fv(Points2[0].f);
    glVertex3fv(Points2[1].f);

    glVertex3fv(Points2[1].f);
    glVertex3fv(Points2[2].f);

    glVertex3fv(Points2[2].f);
    glVertex3fv(Points2[3].f);

    glVertex3fv(Points2[3].f);
    glVertex3fv(Points2[0].f);


    glVertex3fv(Points2[4].f);
    glVertex3fv(Points2[5].f);

    glVertex3fv(Points2[5].f);
    glVertex3fv(Points2[6].f);

    glVertex3fv(Points2[6].f);
    glVertex3fv(Points2[7].f);

    glVertex3fv(Points2[7].f);
    glVertex3fv(Points2[4].f);

    glVertex3fv(Points2[0].f);
    glVertex3fv(Points2[4].f);

    glVertex3fv(Points2[1].f);
    glVertex3fv(Points2[5].f);

    glVertex3fv(Points2[2].f);
    glVertex3fv(Points2[6].f);

    glVertex3fv(Points2[3].f);
    glVertex3fv(Points2[7].f);
  glEnd();*/

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);

  TextureRender->Unuse();

  //Wasser Textur rendern
  TextureRender2->Use();
  TextureRender2->SetTarget(0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();
  MainCam.Call();

  glEnable(GL_CLIP_PLANE0);
  ClippingPlane[3]=-fWaterHeight;
  glTranslatef(0.0f,0.0f,fWaterHeight*2);
  glClipPlane(GL_CLIP_PLANE0,ClippingPlane);
  glScalef(1.0f,1.0f,-1.0f);

  MyLight->Update();

  //Compute Shadow Mapping Matrix
  glGetFloatv(GL_MODELVIEW_MATRIX,CameraModelViewInverse.f);
  CameraModelViewInverse = Inverse(CameraModelViewInverse);
  ShadowMappingMatrix = Mul(ScaleMatrix(0.5f,0.5f,0.5f),TranslationMatrix(0.5f,0.5f,0.5f));
  ShadowMappingMatrix = Mul(LightProjection,ShadowMappingMatrix);
  ShadowMappingMatrix = Mul(LightModelView,ShadowMappingMatrix);
  ShadowMappingMatrix = Mul(CameraModelViewInverse,ShadowMappingMatrix);
  ShadowMappingMatConstant->Set(&ShadowMappingMatrix);

  TerrainShader->GetShader()->Use();
  TerrainDetail->BindToShader(0);
  TerrainColor1->BindToShader(1);
  TerrainColor2->BindToShader(2);
  TextureRender->BindToShader(-1,3);
  glEnable(GL_LIGHTING);
  MyTerrain->TestDraw(TerrainShader->GetShader());
  UnloadShader();
  glActiveTexture(GL_TEXTURE0);

  glEnable(GL_ALPHA_TEST);
  Plants->Draw(MainCam.GetFrom(),CamShadowMapPos,time);
  TextureRender->BindToShader(-1,1);

  Baeume->Draw(true);

  glDisable(GL_ALPHA_TEST);
  glDisable(GL_LIGHTING);

  SkyShader->GetShader()->Use();
  CloudTexture->BindToShader(0);
  SkyDome->Draw(0);
  UnloadShader();
  /*glPushMatrix();
  glTranslatef(CamPos.x,CamPos.y,-CamPos.z);
  glScalef(48.0f,48.0f,48.0f);
  glTranslatef(0.0f,0.0f,-5.5f);
  glActiveTexture(GL_TEXTURE0);
  CubeMap->Bind();
  glColor3f(1.0f,1.0f,1.0f);
  Cube->Draw(TH_NORMALS | TH_TEXCOORDS);
  glPopMatrix();*/

  glEnable(GL_LIGHTING);

  glDisable(GL_CLIP_PLANE0);
  TextureRender2->Unuse();


  //Normal Rendern
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();

  if(mode == 1){
  if(bAntialiasing)
    glEnable(GL_MULTISAMPLE_ARB);
  glEnable(GL_LIGHTING);
  //glEnable(GL_TEXTURE_2D);
  glDisable(GL_TEXTURE_2D);
  //Koordinaten zurcksetzten
  MainCam.Call();
  MainFrustrum.Update();
  MyLight->Update();

  //Compute Shadow Mapping Matrix
  glGetFloatv(GL_MODELVIEW_MATRIX,CameraModelViewInverse.f);
  CameraModelViewInverse = Inverse(CameraModelViewInverse);
  ShadowMappingMatrix = Mul(ScaleMatrix(0.5f,0.5f,0.5f),TranslationMatrix(0.5f,0.5f,0.5f));
  ShadowMappingMatrix = Mul(LightProjection,ShadowMappingMatrix);
  ShadowMappingMatrix = Mul(LightModelView,ShadowMappingMatrix);
  ShadowMappingMatrix = Mul(CameraModelViewInverse,ShadowMappingMatrix);
  ShadowMappingMatConstant->Set(&ShadowMappingMatrix);

  /*glMatrixMode(GL_TEXTURE);
  glPushMatrix();
  glLoadIdentity();
  glTranslatef(0.5f,0.5f,0.5f);
  glScalef(0.5f,0.5f,0.5f);
  glMultMatrixf(LightProjection.f);
  glMultMatrixf(LightModelView.f);
  glMultMatrixf(CameraModelViewInverse.f);
  glGetFloatv(GL_MODELVIEW_MATRIX,ShadowMappingMatrix.f);
  ShadowMappingMatConstant->Set(&ShadowMappingMatrix);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);*/

  /*glColor3f(1.0f,0.0f,0.0f);
  glDisable(GL_LIGHTING);
  MyTerrain->DrawBoxes();*/

  glColor3f(0.5f,0.5f,0.5f);
  glLineWidth(1.0f);
  TerrainShader->GetShader()->Use();
  //TerrainShader->GetShader()->SetUniform(6,CamShadowMapPos);
  TerrainDetail->BindToShader(0);
  TerrainColor1->BindToShader(1);
  TerrainColor2->BindToShader(2);
  TextureRender->BindToShader(-1,3);
  //TerrainShader->SetUniform(0,blend);
  glEnable(GL_LIGHTING);
  if(bWireframe)
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  MyTerrain->TestDraw(TerrainShader->GetShader());
  UnloadShader();
  glActiveTexture(GL_TEXTURE0);

  glEnable(GL_TEXTURE_2D);

  /*glMatrixMode(GL_TEXTURE);
  glPushMatrix();
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);*/

  glDisable(GL_LIGHTING);

  SkyShader->GetShader()->Use();
  CloudTexture->BindToShader(0);
  SkyDome->Draw(0);
  UnloadShader();

  /*glPushMatrix();
  glTranslatef(CamPos.x,CamPos.y,CamPos.z);
  glScalef(48.0f,48.0f,48.0f);
  glTranslatef(0.0f,0.0f,-5.5f);
  CubeMap->Bind();
  glColor3f(1.0f,1.0f,1.0f);
  Cube->Draw(TH_NORMALS | TH_TEXCOORDS);
  glPopMatrix();*/

  glEnable(GL_LIGHTING);

  if(iSmoothGras <= 0)
    glEnable(GL_ALPHA_TEST);
  else{
    if(!bAntialiasing);
      glEnable(GL_MULTISAMPLE_ARB);
    glEnable(GL_SAMPLE_ALPHA_TO_COVERAGE);
  }
  Plants->Draw(MainCam.GetFrom(),CamShadowMapPos,time);

  TextureRender->BindToShader(-1,1);

  Baeume->Draw(true);

  if(iSmoothGras <= 0)
    glDisable(GL_ALPHA_TEST);
  else{
    if(!bAntialiasing);
      glDisable(GL_MULTISAMPLE_ARB);
    glDisable(GL_SAMPLE_ALPHA_TO_COVERAGE);
  }

  if(bShowBoxes){
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    glLineWidth(3.0f);
    Baeume->DrawBoxes();
  }

  glEnable(GL_TEXTURE_2D);

  WaterShader->GetShader()->Use();
  TextureRender2->BindToShader(0,0);
  TextureRender->BindToShader(-1,2);
  WaterBump->BindToShader(1);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glPushMatrix();
  glTranslatef(0.0f,0.0f,fWaterHeight);
  WaterPlane->Draw(TH_NORMALS | TH_TEXCOORDS);
  glPopMatrix();
  glDisable(GL_BLEND);
  UnloadShader();

  glDisable(GL_TEXTURE_2D);
  glDisable(GL_LIGHTING);
  CameraEditor1.Draw();
  glEnable(GL_TEXTURE_2D);

  if(bWireframe)
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

  if(bAntialiasing)
    glDisable(GL_MULTISAMPLE_ARB);
  }
  if(mode==2){
    glDisable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    TextureRender->Bind(-1);
    glColor3f(1.0f,1.0f,1.0f);
    glBegin(GL_QUADS);
      glTexCoord2f(0.0f,1.0f); glVertex3f(-1.0f, 1.0f,-0.5f);
      glTexCoord2f(1.0f,1.0f); glVertex3f( 1.0f, 1.0f,-0.5f);
      glTexCoord2f(1.0f,0.0f); glVertex3f( 1.0f,-1.0f,-0.5f);
      glTexCoord2f(0.0f,0.0f); glVertex3f(-1.0f,-1.0f,-0.5f);
    glEnd();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
  }
  if(mode==3){
    glDisable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    TextureRender2->Bind(0);
    glColor3f(1.0f,1.0f,1.0f);
    glBegin(GL_QUADS);
      glTexCoord2f(0.0f,1.0f); glVertex3f(-1.0f, 1.0f,-0.5f);
      glTexCoord2f(1.0f,1.0f); glVertex3f( 1.0f, 1.0f,-0.5f);
      glTexCoord2f(1.0f,0.0f); glVertex3f( 1.0f,-1.0f,-0.5f);
      glTexCoord2f(0.0f,0.0f); glVertex3f(-1.0f,-1.0f,-0.5f);
    glEnd();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
  }

  //Draw GUI
  if(bGuiActive){
    glLoadIdentity();
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    theGui->Draw();
    glDisable(GL_BLEND);
    glEnable(GL_DEPTH_TEST);
  }
}
