/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "manager.h"

Texture2DManager::Texture2DManager(){
  bError = false;
}

Texture2DManager::~Texture2DManager(){
  for(int i=0;i<Items.size();i++)
    delete Items[i];
  Items.clear();
}

Texture2D* Texture2DManager::LoadTGA(const char *filename, char options){
  Texture2D *ptr;
  ptr = new Texture2D();
  Items.push_back(ptr);
  datei = fopen("engine.log","a");
  if(!ptr->LoadTGA(filename,options)){
    bError = true;
    fprintf(datei,"Texture2DManager::'%s' konnte nicht geladen werden!\n",filename);
    fclose(datei);
    return NULL;
  }
  fprintf(datei,"Texture2DManager::'%s' wurde erfolgreich geladen!\n",filename);
  fclose(datei);
  return ptr;
}

Texture2D* Texture2DManager::CreateEmpty(int width, int height, GLenum DataFormat, GLenum InternalFormat, char options){
  Texture2D *ptr;
  ptr = new Texture2D();
  Items.push_back(ptr);
  datei = fopen("engine.log","a");
  if(!ptr->CreateEmpty(width,height,DataFormat,InternalFormat,options)){
    bError = true;
    fprintf(datei,"Texture2DManager::Leere Textur (%i,%i) konnte nicht erstellt werden!\n",width,height);
    fclose(datei);
    return NULL;
  }
  fprintf(datei,"Texture2DManager::Leere Textur (%i,%i) erfolgreich erstellt!\n",width,height);
  fclose(datei);
  return ptr;
}

Texture2D* Texture2DManager::GetTexture(int num){
  if(num > Items.size()-1 || num < 0)
    return NULL;
  return Items[num];
}

ModelManager::ModelManager(){
  bError = false;
}

ModelManager::~ModelManager(){
  for(int i=0;i<Items.size();i++)
    delete Items[i];
  Items.clear();
}

Model* ModelManager::Load3DS(const char *filename, short object_number, char options){
  Model *ptr;
  ptr = new Model();
  Items.push_back(ptr);
  datei = fopen("engine.log","a");
  if(!ptr->LoadModel3DS(filename,object_number,options)){
    bError=true;
    fprintf(datei,"ModelManager::'%s' konnte nicht geladen werden!\n",filename);
    fclose(datei);
    delete ptr;
    return NULL;
  }
  fprintf(datei,"ModelManager::'%s' wurde erfolgreich geladen!\n",filename);
  fclose(datei);
  return ptr;
}

Model* ModelManager::LoadBillboards3DS(const char *filename, short object_number){
  Model *ptr;
  ptr = new Model();
  Items.push_back(ptr);
  datei = fopen("engine.log","a");
  if(!ptr->LoadBillboards3DS(filename,object_number)){
    bError=true;
    fprintf(datei,"ModelManager::'%s' konnte nicht geladen werden!\n",filename);
    fclose(datei);
    delete ptr;
    return NULL;
  }
  fprintf(datei,"ModelManager::'%s' wurde erfolgreich geladen!\n",filename);
  fclose(datei);
  return ptr;
}

Model* ModelManager::LoadBillboardsSchrompf(const char *filename, vec4 center){
  Model *ptr;
  ptr = new Model();
  Items.push_back(ptr);
  datei = fopen("engine.log","a");
  if(!ptr->LoadBillboardsSchrompf(filename,center)){
    bError=true;
    fprintf(datei,"ModelManager::'%s' konnte nicht geladen werden!\n",filename);
    fclose(datei);
    delete ptr;
    return NULL;
  }
  fprintf(datei,"ModelManager::'%s' wurde erfolgreich geladen!\n",filename);
  fclose(datei);
  return ptr;
}

Model* ModelManager::LoadSchrompf(const char *filename, char options){
  Model *ptr;
  ptr = new Model();
  Items.push_back(ptr);
  datei = fopen("engine.log","a");
  if(!ptr->LoadSchrompf(filename,options)){
    bError=true;
    fprintf(datei,"ModelManager::'%s' konnte nicht geladen werden!\n",filename);
    fclose(datei);
    delete ptr;
    return NULL;
  }
  fprintf(datei,"ModelManager::'%s' wurde erfolgreich geladen!\n",filename);
  fclose(datei);
  return ptr;
}

XmlShaderManager::XmlShaderManager(){
  bError=false;
}

XmlShaderManager::~XmlShaderManager(){
  for(int i=0;i<Items.size();i++)
    delete Items[i];
  Items.clear();
}

XMLShader* XmlShaderManager::LoadXml(const char *filename){
  XMLShader *ptr;
  ptr = new XMLShader();
  Items.push_back(ptr);
  datei = fopen("engine.log","a");
  if(!ptr->LoadFromFile(filename)){
    bError=true;
    fprintf(datei,"XMLShaderManager::'%s' konnte nicht geladen werden!\n",filename);
    fclose(datei);
    return NULL;
  }
  fprintf(datei,"XMLShaderManager::'%s' wurde erfolgreich geladen!\n",filename);
  fclose(datei);
  return ptr;
}


