/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "thEngine.h"
#include "model.h"
#include "vektoren.h"
#include <stdio.h>
#include <string>

#define TH_EXTRUDE 100.0f

// TODO (Benjamin#3#): 3DS Importer fhig machen Billboards zu importieren

Model::Model(){
  cName=NULL;
  iNumVertices=0;
  iNumFaces=0;
  Faces=NULL;
  Vertices=NULL;
  Positions=NULL;
  GGroups=NULL;
  bHasTexcoords=false;
  bHasColor=false;
  bHasConnectivity=false;
  bHasTangentBinormal=false;
  bHasNormals=false;

  //todo Model auf vertexbuffer umstellen
  vbuffer = NULL;
}

Model::~Model(){
  delete []cName;
  delete []Faces;

  if(Positions != NULL){
    for(int i=0;i<iNumPositions;i++){
      if(Positions[i].v4Normals != NULL)
        delete Positions[i].v4Normals;
      if(Positions[i].usUsedIn != NULL)
        delete Positions[i].usUsedIn;
    }
    delete []Positions;
  }
  if(Vertices != NULL)
    delete []Vertices;

  if(GGroups != NULL){
    for(int i=0;i<32;i++){
      if(GGroups[i].uiFaces != NULL)
        delete GGroups[i].uiFaces;
    }
    delete []GGroups;
  }
}

bool Model::LoadModel3DS(const char *filename,short object_number,char options){
  int i;
  FILE *datei;
  #ifdef MODEL_DEBUG
  FILE *debugtxt;
  #endif
  unsigned short chunk_id;
  unsigned int chunk_length;
  unsigned int uiGGroupFlag;
  unsigned char mybyte;
  unsigned short qty; //wieviel wird gelesen
  unsigned short face_flags; //dummy um face flags zu bersprignen
  unsigned short file_length;
  bool first_mat=true;
  unsigned short curr_mat=0;
  unsigned short temp_pos;
  unsigned short curr_obj=0;
  unsigned short acc_face[32];
  unsigned short temp;
  std::string ErrorMessage;

  #ifdef MODEL_DEBUG
  if((debugtxt=fopen("read3ds.txt","a"))==NULL){return false;}
  #endif
  if((datei=fopen(filename,"rb"))==NULL){
    ErrorMessage = "Konnte Datei ";
    ErrorMessage += filename;
    ErrorMessage += " nicht ffnen";
    ThErrorMessage("Model Error",ErrorMessage.c_str());
    #ifdef MODEL_DEBUG
      fprintf(debugtxt,"Konnte datei %s nicht ffnen",filename);
    #endif
    return false;
  }
  #ifdef MODEL_DEBUG
  fprintf(debugtxt,"------------------------------------------------------------\nLese Objekt %i aus Datei %s\n",object_number,filename);
  #endif
  fread(&chunk_id,2,1,datei); //Start Chunk lesen
  fread(&file_length,4,1,datei); //Lnge der Datei einlesen
  #ifdef MODEL_DEBUG
  fprintf(debugtxt,"Start Chunk gelesen, Dateilnge: %i\n",file_length);
  #endif
  while(ftell(datei)< file_length){ //Loop bis ende der Datei erreicht ist
    fread(&chunk_id,2,1,datei); //Chunk Id lesen
    fread(&chunk_length,4,1,datei); //Chunk lnge lesen
    switch (chunk_id)
    {
      case 0x3d3d: //3D Editor Chunk
        #ifdef MODEL_DEBUG
        fprintf(debugtxt,"3D Editior Chunk gelesen, Lnge %i\n",chunk_length);
        #endif
        break; //Einfach lesen ohne zu berspringen

      case 0x4000: //Object Block
        curr_obj++;
        if(curr_obj!=object_number){ //Wenn dieses Obejekt nicht geladen werden soll
          #ifdef MODEL_DEBUG
          fprintf(debugtxt,"\nObjekt %i bersprungen\n",object_number);
          #endif
          fseek(datei, chunk_length-6, SEEK_CUR); // Object Block berspringen
        }
        else{ //ansonsten Name einlesen
          int i=0;
          mybyte='0';
          cName = new char[100];
          while(mybyte!=0x00){
            fread(&mybyte,1,1,datei);
            cName[i]=mybyte;
            i++;
          }
        #ifdef MODEL_DEBUG
        fprintf(debugtxt,"\nObject %i gelesen, Name: %s\n",object_number,cName);
        #endif
        }
      break;

      case 0x4100: //Muss gelesen werden
      #ifdef MODEL_DEBUG
      fprintf(debugtxt,"\n3D Daten Chunk gelesen\n");
      #endif
      curr_mat=0; first_mat=true;
      break;

      case 0x4110: //Scheitelpunkte einlesen
        fread(&qty,2,1,datei); //Anzahl der Scheitelpunkte
        iNumVertices = qty;
        iNumPositions = qty;
        Vertices = new vertexv4[qty];
        Positions = new vertexposv4[qty];
        for (i=0;i<qty;i++){
          fread(&Positions[i].v4Pos.x,4,1,datei);
          fread(&Positions[i].v4Pos.y,4,1,datei);
          fread(&Positions[i].v4Pos.z,4,1,datei);
          Positions[i].v4Pos.w=1.0f;
          Positions[i].v4Normals=NULL;
          Vertices[i].uiNum = i;
          Positions[i].usTimesUsed=0;
          Positions[i].usUsedIn=NULL;
          Vertices[i].Pos = &Positions[i];
        }
        #ifdef MODEL_DEBUG
        fprintf(debugtxt,"%i Scheitelpunkte eingelesen\n",qty);
        #endif
      break;

      case 0x4120: //Flchen einlesen
        fread(&qty,2,1,datei); //Anzahl der Flchen
        iNumFaces=qty;
        Faces=new facev4[qty];
        for(i=0;i<qty;i++){
          fread(&temp,2,1,datei);
          Faces[i].a = &Vertices[temp];
          fread(&temp,2,1,datei);
          Faces[i].b = &Vertices[temp];
          fread(&temp,2,1,datei);
          Faces[i].c = &Vertices[temp];
          fread (&face_flags,2,1,datei);
        }
        #ifdef MODEL_DEBUG
        fprintf(debugtxt,"%i Flchen eingelesen\n",qty);
        #endif
      break;

      case 0x4140: //Mapping Koordinaten
        fread(&qty,2,1,datei); //Anzahl der Mapping Koordinaten
        if(qty == iNumVertices && (options & TH_TEXCOORDS)){
          #ifdef MODEL_DEBUG
          fprintf(debugtxt,"Found complete Texturecoordinates\n");
          #endif
          bHasTexcoords=true;
          for(i=0;i<qty;i++){
            fread(&Vertices[i].v2Texcoord.x,4,1,datei);
            fread(&Vertices[i].v2Texcoord.y,4,1,datei);
          }
          #ifdef MODEL_DEBUG
          fprintf(debugtxt,"%i Texturkoordinaten eingelesen\n",qty);
          #endif
        }
        else{
          #ifdef MODEL_DEBUG
          if(options & TH_TEXCOORDS)
            fprintf(debugtxt,"Texturecoordinates incomplete, skipping\n");
          else
            fprintf(debugtxt,"No Texturecoordinates needed by user\n");
          #endif
          fseek(datei,chunk_length-8,SEEK_CUR);
          bHasTexcoords=false;
        }
      break;

      case 0x4150: //Glttungsgruppen
        GGroups = new ggroup[32];
        for(i=0;i<32;i++){
          GGroups[i].uiNumFaces=0; //Anzahl der Benuzten Flchen von allen Glttungsgruppen auf 0 setzen
          acc_face[i]=0; //Aktuelle Speicherstelle der Glttungsgruppe auf 0 setzten
        }
        temp_pos=ftell(datei); //Position speichern um spter an selber Stelle erneut zu lesen
        //Zhlen wiviele Flchen jeder Glttungsgruppe angehren
        for(i=0;i<iNumFaces;i++){
          fread(&uiGGroupFlag,4,1,datei);
          for(int c=0;c<32;c++){//Checken ob Bit c von ggroups == 1 und somit Flche i Glttungsgruppe c angehrt
            if(uiGGroupFlag % 2 == 1){
              GGroups[c].uiNumFaces++;
            }
            uiGGroupFlag = uiGGroupFlag/2;
          }
        }
        //Jetzt Flchen einlesen
        fseek(datei,temp_pos,SEEK_SET);
        for(i=0;i<32;i++){//Speicher reservieren
          GGroups[i].uiFaces = new unsigned int[GGroups[i].uiNumFaces];
        }
        for(i=0;i<iNumFaces;i++){
          fread(&uiGGroupFlag,4,1,datei);
          for(int c=0;c<32;c++){//Checken ob Bit c von ggroups == 1 und somit Flche i Glttungsgruppe c angehrt
            if(uiGGroupFlag % 2 == 1){
              GGroups[c].uiFaces[acc_face[c]]=i;
              Faces[i].uiNumGGroup=c;
              acc_face[c]++;
            }
            uiGGroupFlag = uiGGroupFlag/2;
          }
        }
        break;

      default: //Unbekannte Chunks berspringen
        fseek(datei, chunk_length-6, SEEK_CUR);
      break;
    }
  }
 fclose (datei);
 #ifdef MODEL_DEBUG
 fclose (debugtxt);
 #endif

 GenData(options);

 return true; //File erfolgreich geladen
}

void Model::GenData(char options){
  unsigned int i,j,k,curr,uiNewSize;
  bool *bUsed;
  vertexposv4 *TempPositions;
  #ifdef MODEL_DEBUG
  FILE *datei;
  #endif

  //Find double Position Information an eliminate it
  bUsed = new bool[iNumPositions];
  for(i=0;i<iNumPositions;i++)
    bUsed[i]=false;
  uiNewSize=0;
  for(i=0;i<iNumPositions;i++){ //determine size of new array
    if(!bUsed[i]){
      bUsed[i]=true;
      uiNewSize++;
      for(j=i+1;j<iNumPositions;j++){
        if(Positions[i].v4Pos.x == Positions[j].v4Pos.x &&
           Positions[i].v4Pos.y == Positions[j].v4Pos.y &&
           Positions[i].v4Pos.z == Positions[j].v4Pos.z &&
           !bUsed[j])
        {
          bUsed[j]=true;
        }
      }
    }
  }

  TempPositions = new vertexposv4[uiNewSize];
  curr=0;
  for(i=0;i<iNumPositions;i++)
    bUsed[i]=false;
  for(i=0;i<iNumPositions;i++){ //determine size of new array
    if(!bUsed[i]){
      bUsed[i]=true;
      TempPositions[curr] = Positions[i];
      TempPositions[curr].uiNum = curr;
      Vertices[i].Pos = &TempPositions[curr];
      for(j=i+1;j<iNumPositions;j++){
        if(Positions[i].v4Pos.x == Positions[j].v4Pos.x &&
           Positions[i].v4Pos.y == Positions[j].v4Pos.y &&
           Positions[i].v4Pos.z == Positions[j].v4Pos.z &&
           !bUsed[j])
        {
          bUsed[j]=true;
          Vertices[j].Pos = &TempPositions[curr];
        }
      }
      curr++;
    }
  }
  delete []Positions;
  delete []bUsed;
  Positions = TempPositions;
  iNumPositions = uiNewSize;

  //Average Positions for Shadows
  for(i=0;i<iNumFaces;i++){
    Faces[i].v4AveragePos.x = (Faces[i].a->Pos->v4Pos.x + Faces[i].b->Pos->v4Pos.x + Faces[i].c->Pos->v4Pos.x) / 3;
    Faces[i].v4AveragePos.y = (Faces[i].a->Pos->v4Pos.y + Faces[i].b->Pos->v4Pos.y + Faces[i].c->Pos->v4Pos.y) / 3;
    Faces[i].v4AveragePos.z = (Faces[i].a->Pos->v4Pos.z + Faces[i].b->Pos->v4Pos.z + Faces[i].c->Pos->v4Pos.z) / 3;
    Faces[i].v4AveragePos.w = 1.0f;

  }



  //First Step, generate necessary Data
  //Normals per Face or Normals per Vertex need this
  if(options & TH_NORMALS_PER_FACE || options & TH_NORMALS_PER_VERTEX){
    bHasNormals=true;
    for(i=0;i<iNumFaces;i++){
      Faces[i].v4Normal=ComputeNormalV4(Faces[i].a->Pos->v4Pos,Faces[i].b->Pos->v4Pos,Faces[i].c->Pos->v4Pos);
      Faces[i].v4Tangent=TangentV4(Faces[i].a->Pos->v4Pos,Faces[i].b->Pos->v4Pos,Faces[i].c->Pos->v4Pos,Faces[i].a->v2Texcoord,Faces[i].b->v2Texcoord,Faces[i].c->v2Texcoord);
      Faces[i].v4Binormal=BinormalV4(Faces[i].a->Pos->v4Pos,Faces[i].b->Pos->v4Pos,Faces[i].c->Pos->v4Pos,Faces[i].a->v2Texcoord,Faces[i].b->v2Texcoord,Faces[i].c->v2Texcoord);
    }
  }

  //Computing Normals per Vertex
  if(options & TH_NORMALS_PER_VERTEX){
    bUsed = new bool[iNumPositions];
    //Count how many Times each vertex is used in each ggroup
    for(i=0;i<32;i++){ //For every GGroup
      //Reset Bool Array
      for(j=0;j<iNumPositions;j++)
        bUsed[j]=false;
      //Step trough every Face in GGroup
      for(j=0;j<GGroups[i].uiNumFaces;j++){
        // Vertex A
        if(!bUsed[Faces[GGroups[i].uiFaces[j]].a->Pos->uiNum]){
          bUsed[Faces[GGroups[i].uiFaces[j]].a->Pos->uiNum] = true;
          Faces[GGroups[i].uiFaces[j]].a->Pos->usTimesUsed++;
        }
        //Vertex B
        if(!bUsed[Faces[GGroups[i].uiFaces[j]].b->Pos->uiNum]){
          bUsed[Faces[GGroups[i].uiFaces[j]].b->Pos->uiNum] = true;
          Faces[GGroups[i].uiFaces[j]].b->Pos->usTimesUsed++;
        }
        //Vertex C
        if(!bUsed[Faces[GGroups[i].uiFaces[j]].c->Pos->uiNum]){
          bUsed[Faces[GGroups[i].uiFaces[j]].c->Pos->uiNum] = true;
          Faces[GGroups[i].uiFaces[j]].c->Pos->usTimesUsed++;
        }
      }
    }

    //Allocate Memory
    for(i=0;i<iNumPositions;i++){ //For each Position
      Positions[i].v4Normals = new vec4[Positions[i].usTimesUsed];
      if(options & TH_TANGENT)
        Positions[i].v4Tangents = new vec4[Positions[i].usTimesUsed];
      if(options & TH_BINORMAL)
        Positions[i].v4Binormals = new vec4[Positions[i].usTimesUsed];
      //ZeroMemory(Vertices[i].v4Normals,sizeof(vec4)*Vertices[i].usTimesUsed);
      for(j=0;j<Positions[i].usTimesUsed;j++){
        Positions[i].v4Normals[j].x=0.0f;
        Positions[i].v4Normals[j].y=0.0f;
        Positions[i].v4Normals[j].z=0.0f;
        Positions[i].v4Normals[j].w=1.0f;
        if(options & TH_TANGENT){
          Positions[i].v4Tangents[j].x=0.0f;
          Positions[i].v4Tangents[j].y=0.0f;
          Positions[i].v4Tangents[j].z=0.0f;
          Positions[i].v4Tangents[j].w=1.0f;
        }
        if(options & TH_BINORMAL){
          Positions[i].v4Binormals[j].x=0.0f;
          Positions[i].v4Binormals[j].y=0.0f;
          Positions[i].v4Binormals[j].z=0.0f;
          Positions[i].v4Binormals[j].w=1.0f;
        }
      }

      Positions[i].usUsedIn = new unsigned short[Positions[i].usTimesUsed];
      //set invalid ggroups values
      for(j=0;j<Positions[i].usTimesUsed;j++)
        Positions[i].usUsedIn[j]=33;
    }

    //Compute Normals per Vertex and add reference numbers
    for(i=0;i<32;i++){
      //Reset Bool Array
      for(j=0;j<iNumPositions;j++)
        bUsed[j]=false;
      //Step trough every Face in GGroup
      for(j=0;j<GGroups[i].uiNumFaces;j++){
        // Vertex A
        if(!bUsed[Faces[GGroups[i].uiFaces[j]].a->Pos->uiNum]){
          bUsed[Faces[GGroups[i].uiFaces[j]].a->Pos->uiNum] = true;
          //Search for first invalid value to write in
          for(k=0;k<Faces[GGroups[i].uiFaces[j]].a->Pos->usTimesUsed;k++){
            if(Faces[GGroups[i].uiFaces[j]].a->Pos->usUsedIn[k]==33)
              break;
          }
          if(k<Faces[GGroups[i].uiFaces[j]].a->Pos->usTimesUsed)
            Faces[GGroups[i].uiFaces[j]].a->Pos->usUsedIn[k]=i;
          #ifdef MODEL_DEBUG
          else
            MessageBox(NULL,"Mehr Normalen fr Vertex bentigt als Speicher vorhanden\nVertex A","3DS Loader Fehler",MB_OK | MB_ICONERROR);
          #endif
        }
        //add face normal to vertex normal
        Faces[GGroups[i].uiFaces[j]].a->Pos->v4Normals[k].x+=Faces[GGroups[i].uiFaces[j]].v4Normal.x;
        Faces[GGroups[i].uiFaces[j]].a->Pos->v4Normals[k].y+=Faces[GGroups[i].uiFaces[j]].v4Normal.y;
        Faces[GGroups[i].uiFaces[j]].a->Pos->v4Normals[k].z+=Faces[GGroups[i].uiFaces[j]].v4Normal.z;

        if(options & TH_TANGENT){
          Faces[GGroups[i].uiFaces[j]].a->Pos->v4Tangents[k].x+=Faces[GGroups[i].uiFaces[j]].v4Tangent.x;
          Faces[GGroups[i].uiFaces[j]].a->Pos->v4Tangents[k].y+=Faces[GGroups[i].uiFaces[j]].v4Tangent.y;
          Faces[GGroups[i].uiFaces[j]].a->Pos->v4Tangents[k].z+=Faces[GGroups[i].uiFaces[j]].v4Tangent.z;
        }

        if(options & TH_BINORMAL){
          Faces[GGroups[i].uiFaces[j]].a->Pos->v4Binormals[k].x+=Faces[GGroups[i].uiFaces[j]].v4Binormal.x;
          Faces[GGroups[i].uiFaces[j]].a->Pos->v4Binormals[k].y+=Faces[GGroups[i].uiFaces[j]].v4Binormal.y;
          Faces[GGroups[i].uiFaces[j]].a->Pos->v4Binormals[k].z+=Faces[GGroups[i].uiFaces[j]].v4Binormal.z;
        }

        //Vertex B
        if(!bUsed[Faces[GGroups[i].uiFaces[j]].b->Pos->uiNum]){
          bUsed[Faces[GGroups[i].uiFaces[j]].b->Pos->uiNum] = true;
          //Search for first invalid value to write in
          for(k=0;k<Faces[GGroups[i].uiFaces[j]].b->Pos->usTimesUsed;k++){
            if(Faces[GGroups[i].uiFaces[j]].b->Pos->usUsedIn[k]==33)
              break;
          }
          if(k<Faces[GGroups[i].uiFaces[j]].b->Pos->usTimesUsed)
            Faces[GGroups[i].uiFaces[j]].b->Pos->usUsedIn[k]=i;
          #ifdef MODEL_DEBUG
          else
            MessageBox(NULL,"Mehr Normalen fr Vertex bentigt als Speicher vorhanden\nVertex B","3DS Loader Fehler",MB_OK | MB_ICONERROR);
          #endif
        }
        //add face normal to vertex normal
        Faces[GGroups[i].uiFaces[j]].b->Pos->v4Normals[k].x+=Faces[GGroups[i].uiFaces[j]].v4Normal.x;
        Faces[GGroups[i].uiFaces[j]].b->Pos->v4Normals[k].y+=Faces[GGroups[i].uiFaces[j]].v4Normal.y;
        Faces[GGroups[i].uiFaces[j]].b->Pos->v4Normals[k].z+=Faces[GGroups[i].uiFaces[j]].v4Normal.z;

        if(options & TH_TANGENT){
          Faces[GGroups[i].uiFaces[j]].b->Pos->v4Tangents[k].x+=Faces[GGroups[i].uiFaces[j]].v4Tangent.x;
          Faces[GGroups[i].uiFaces[j]].b->Pos->v4Tangents[k].y+=Faces[GGroups[i].uiFaces[j]].v4Tangent.y;
          Faces[GGroups[i].uiFaces[j]].b->Pos->v4Tangents[k].z+=Faces[GGroups[i].uiFaces[j]].v4Tangent.z;
        }

        if(options & TH_BINORMAL){
          Faces[GGroups[i].uiFaces[j]].b->Pos->v4Binormals[k].x+=Faces[GGroups[i].uiFaces[j]].v4Binormal.x;
          Faces[GGroups[i].uiFaces[j]].b->Pos->v4Binormals[k].y+=Faces[GGroups[i].uiFaces[j]].v4Binormal.y;
          Faces[GGroups[i].uiFaces[j]].b->Pos->v4Binormals[k].z+=Faces[GGroups[i].uiFaces[j]].v4Binormal.z;
        }

        //Vertex C
        if(!bUsed[Faces[GGroups[i].uiFaces[j]].c->Pos->uiNum]){
          bUsed[Faces[GGroups[i].uiFaces[j]].c->Pos->uiNum] = true;
          //Search for first invalid value to write in
          for(k=0;k<Faces[GGroups[i].uiFaces[j]].c->Pos->usTimesUsed;k++){
            if(Faces[GGroups[i].uiFaces[j]].c->Pos->usUsedIn[k]==33)
              break;
          }
          if(k<Faces[GGroups[i].uiFaces[j]].c->Pos->usTimesUsed)
            Faces[GGroups[i].uiFaces[j]].c->Pos->usUsedIn[k]=i;
          #ifdef MODEL_DEBUG
          else
            MessageBox(NULL,"Mehr Normalen fr Vertex bentigt als Speicher vorhanden\nVertex C","3DS Loader Fehler",MB_OK | MB_ICONERROR);
          #endif
        }
        //add face normal to vertex normal
        Faces[GGroups[i].uiFaces[j]].c->Pos->v4Normals[k].x+=Faces[GGroups[i].uiFaces[j]].v4Normal.x;
        Faces[GGroups[i].uiFaces[j]].c->Pos->v4Normals[k].y+=Faces[GGroups[i].uiFaces[j]].v4Normal.y;
        Faces[GGroups[i].uiFaces[j]].c->Pos->v4Normals[k].z+=Faces[GGroups[i].uiFaces[j]].v4Normal.z;

        if(options & TH_TANGENT){
          Faces[GGroups[i].uiFaces[j]].c->Pos->v4Tangents[k].x+=Faces[GGroups[i].uiFaces[j]].v4Tangent.x;
          Faces[GGroups[i].uiFaces[j]].c->Pos->v4Tangents[k].y+=Faces[GGroups[i].uiFaces[j]].v4Tangent.y;
          Faces[GGroups[i].uiFaces[j]].c->Pos->v4Tangents[k].z+=Faces[GGroups[i].uiFaces[j]].v4Tangent.z;
        }

        if(options & TH_BINORMAL){
          Faces[GGroups[i].uiFaces[j]].c->Pos->v4Binormals[k].x+=Faces[GGroups[i].uiFaces[j]].v4Binormal.x;
          Faces[GGroups[i].uiFaces[j]].c->Pos->v4Binormals[k].y+=Faces[GGroups[i].uiFaces[j]].v4Binormal.y;
          Faces[GGroups[i].uiFaces[j]].c->Pos->v4Binormals[k].z+=Faces[GGroups[i].uiFaces[j]].v4Binormal.z;
        }
      }
    }

    //Normalize each vertex normal
    for(i=0;i<iNumPositions;i++){
      for(j=0;j<Positions[i].usTimesUsed;j++){
        Positions[i].v4Normals[j] = NormalizeV4(Positions[i].v4Normals[j]);
        if(options & TH_TANGENT)
          Positions[i].v4Tangents[j] = NormalizeV4(Positions[i].v4Tangents[j]);
        if(options & TH_BINORMAL)
          Positions[i].v4Binormals[j] = NormalizeV4(Positions[i].v4Binormals[j]);
      }
    }
    delete []bUsed;

    #ifdef MODEL_DEBUG
    datei = fopen("dumb.txt","w");
    for(i=0;i<iNumFaces;i++){
      fprintf(datei,"---Flaeche %i---\n",i);
      fprintf(datei,"GGroup: %i\n",Faces[i].uiNumGGroup);
      fprintf(datei,"A: %X\nB: %X\nC: %X\n",Faces[i].a->Pos->uiNum,Faces[i].b->Pos->uiNum,Faces[i].c->Pos->uiNum);
    }
    fprintf(datei,"\n\n");
    for(i=0;i<iNumVertices;i++){
      fprintf(datei,"---Vertex %i---\n",i);
      fprintf(datei,"Pos: %i\n",Vertices[i].Pos->uiNum);
    }
    fprintf(datei,"\n\n");
    for(i=0;i<iNumPositions;i++){
      fprintf(datei,"---Position %i---\n",i);
      fprintf(datei,"GGroups:");
      for(j=0;j<Positions[i].usTimesUsed;j++)
        fprintf(datei," %i",Positions[i].usUsedIn[j]);
      fprintf(datei,"\nX:%f Y:%f Z:%f\n",Positions[i].v4Pos.x,Positions[i].v4Pos.y,Positions[i].v4Pos.z);
    }
    fclose(datei);
    #endif
  }

  //Set up Connectivity for Shadows
  if(options & TH_CONNECTIVITY){
    bHasConnectivity=true;
    Connectivity= new facev4con[iNumFaces];
    for(i=0;i<iNumFaces;i++){
      Connectivity[i].con[0]=NULL;
      Connectivity[i].con[1]=NULL;
      Connectivity[i].con[2]=NULL;
    }

    for(i=0;i<iNumFaces;i++){
      //Edge 1
      if(Connectivity[i].con[0] == NULL){
        for(j=0;j<iNumFaces;j++){
          if(j==i)
            continue;
          //1-1
          if(Faces[i].a->Pos == Faces[j].b->Pos && Faces[i].b->Pos == Faces[j].a->Pos){
            Connectivity[i].con[0]=&Connectivity[j];
            Connectivity[j].con[0]=&Connectivity[i];
            break;
          }
          //1-2
          else if(Faces[i].a->Pos == Faces[j].c->Pos && Faces[i].b->Pos == Faces[j].b->Pos){
            Connectivity[i].con[0]=&Connectivity[j];
            Connectivity[j].con[1]=&Connectivity[i];
            break;
          }
          //1-3
          else if(Faces[i].a->Pos == Faces[j].a->Pos && Faces[i].b->Pos == Faces[j].c->Pos){
            Connectivity[i].con[0]=&Connectivity[j];
            Connectivity[j].con[2]=&Connectivity[i];
            break;
          }
        }
      }

      //Edge 2
      if(Connectivity[i].con[1] == NULL){
        for(j=0;j<iNumFaces;j++){
          if(j==i)
            continue;
          //2-1
          if(Faces[i].b->Pos == Faces[j].b->Pos && Faces[i].c->Pos == Faces[j].a->Pos){
            Connectivity[i].con[1]=&Connectivity[j];
            Connectivity[j].con[0]=&Connectivity[i];
            break;
          }
          //2-2
          else if(Faces[i].b->Pos == Faces[j].c->Pos && Faces[i].c->Pos == Faces[j].b->Pos){
            Connectivity[i].con[1]=&Connectivity[j];
            Connectivity[j].con[1]=&Connectivity[i];
            break;
          }
          //2-3
          else if(Faces[i].b->Pos == Faces[j].c->Pos && Faces[i].c->Pos == Faces[j].c->Pos){
            Connectivity[i].con[1]=&Connectivity[j];
            Connectivity[j].con[2]=&Connectivity[i];
            break;
          }
        }
      }

      //Edge 3
      if(Connectivity[i].con[2] == NULL){
        for(j=0;j<iNumFaces;j++){
          if(i==j)
            continue;
          //3-1
          if(Faces[i].c->Pos == Faces[j].b->Pos && Faces[i].a->Pos == Faces[j].a->Pos){
            Connectivity[i].con[2]=&Connectivity[j];
            Connectivity[j].con[0]=&Connectivity[i];
            break;
          }
          //3-2
          else if(Faces[i].c->Pos == Faces[j].c->Pos && Faces[i].a->Pos == Faces[j].b->Pos){
            Connectivity[i].con[2]=&Connectivity[j];
            Connectivity[j].con[1]=&Connectivity[i];
            break;
          }
          //3-3
          else if(Faces[i].c->Pos == Faces[j].a->Pos && Faces[i].a->Pos == Faces[j].c->Pos){
            Connectivity[i].con[2]=&Connectivity[j];
            Connectivity[j].con[2]=&Connectivity[i];
            break;
          }
        }
      }
    }

    #ifdef MODEL_DEBUG
      bool test=false;
      datei = fopen("dumb.txt","a");
      for(i=0;i<iNumFaces;i++){
        if(Connectivity[i].con[0] == NULL || Connectivity[i].con[1] == NULL || Connectivity[i].con[2] == NULL){
          test=true;
          break;
        }
      }
      if(test)
        fprintf(datei,"\nWarning: Keine durchgehenden Verbindungsinformationen!\n\n");
      for(i=0;i<iNumFaces;i++){
        fprintf(datei,"\n---Con %i-%X---\n",i,&Connectivity[i]);
        for(j=0;j<3;j++)
          fprintf(datei,"%i: %8X\n",j,Connectivity[i].con[j]);
      }
      fclose(datei);
    #endif
  }

  //Pass data to Vertexbuffer
  int iFloatsPerVertex=3;
  int iVertexBufferOptions = 0;
  if(options & TH_TEXCOORDS){
    iFloatsPerVertex += 2;
    iVertexBufferOptions |= TH_VB_TEXCOORD0_2;
  }
  if(options & TH_NORMALS_PER_FACE || options & TH_NORMALS_PER_VERTEX){
    iFloatsPerVertex += 3;
    iVertexBufferOptions |= TH_VB_NORMAL;
  }
  if(options & TH_TANGENT){
    iFloatsPerVertex += 3;
    iVertexBufferOptions |= TH_VB_ATTRIBUTE0_3;
  }
  if(options & TH_BINORMAL){
    iFloatsPerVertex += 3;
    iVertexBufferOptions |= TH_VB_ATTRIBUTE1_3;
  }

  vbuffer = new VertexBuffer(iVertexBufferOptions,GL_TRIANGLES,GL_UNSIGNED_SHORT);
  vbuffer->ReserveVertexData(iNumFaces * 3 * iFloatsPerVertex);

  for(i=0;i<iNumFaces;i++){
    for(j=0;j<3;j++){
// TODO (Benjamin#1#): Indexing implementieren

      //Vertex Position
      vbuffer->AddVertexData(Faces[i].vertices[j]->Pos->v4Pos.f,3);

      //Colors (TODO)

      //Normals
      if(options & TH_NORMALS_PER_FACE)
        vbuffer->AddVertexData(Faces[i].v4Normal.f,3);

      if(options & TH_NORMALS_PER_VERTEX){
        k=FindGGroup(Faces[i].vertices[j]->uiNum,Faces[i].uiNumGGroup);
        vbuffer->AddVertexData(Faces[i].vertices[j]->Pos->v4Normals[k].f,3);
      }

      //Texcoords
      if(options & TH_TEXCOORDS)
        vbuffer->AddVertexData(Faces[i].vertices[j]->v2Texcoord);

      //Tangents
      if(options & TH_TANGENT){
        if(options & TH_NORMALS_PER_FACE)
          vbuffer->AddVertexData(Faces[i].v4Tangent.f,3);
        else if(options & TH_NORMALS_PER_VERTEX)
          vbuffer->AddVertexData(Faces[i].vertices[j]->Pos->v4Tangents[k].f,3);
      }

      //Binormals
      if(options & TH_BINORMAL){
        if(options & TH_NORMALS_PER_FACE)
          vbuffer->AddVertexData(Faces[i].v4Binormal.f,3);
        else if(options & TH_NORMALS_PER_VERTEX)
          vbuffer->AddVertexData(Faces[i].vertices[j]->Pos->v4Binormals[k].f,3);
      }
    }
  }

  vbuffer->UploadData();
}

unsigned int Model::FindGGroup(unsigned int uiVertNum, unsigned short usGGroupNum){
  unsigned short i;
  for(i=0;i<Vertices[uiVertNum].Pos->usTimesUsed;i++){
    if(Vertices[uiVertNum].Pos->usUsedIn[i]==usGGroupNum)
      break;
  }
  return i;
}

void Model::Draw(char options){
  int iVertexBufferOptions=0;
  if(options & TH_NORMALS)
    iVertexBufferOptions |= TH_VB_NORMAL;
  if(options & TH_TEXCOORDS)
    iVertexBufferOptions |= TH_VB_TEXCOORD0;
  if(options & TH_TANGENT)
    iVertexBufferOptions |= TH_VB_ATTRIBUTE0;
  if(options & TH_BINORMAL)
    iVertexBufferOptions |= TH_VB_ATTRIBUTE1;

  vbuffer->Draw(iVertexBufferOptions);
}

void Model::DrawShadow(vec4 v4ModelPos, vec4 v4LightPos){
  int i;
  vec4 v4LightDir,v4Temp1,v4Temp2,v4Temp3;
  bool draw;
  if(!bHasConnectivity)
    return;
  v4LightDir = NormalizeV4(v4LightDir);
  v4LightPos.x = v4LightPos.x - v4ModelPos.x;
  v4LightPos.y = v4LightPos.y - v4ModelPos.y;
  v4LightPos.z = v4LightPos.z - v4ModelPos.z;
  for(i=0;i<iNumFaces;i++){
    if(SkalarProductV4(DirectionV4(v4LightPos,Faces[i].v4AveragePos),Faces[i].v4Normal) > 0)
      Connectivity[i].bVisible=true;
    else
      Connectivity[i].bVisible=false;
  }

  glColor4f(1.0f,0.0f,0.0f,1.0f);

  for(i=0;i<iNumFaces;i++){
    if(!Connectivity[i].bVisible)
      continue;

    //Edge 1
    if(Connectivity[i].con[0] != NULL)
      if(!Connectivity[i].con[0]->bVisible)
        draw=true;
      else
        draw=false;
    else
      draw=true;

    if(draw){
      v4Temp1=ExtrudeV4(v4LightPos,Faces[i].a->Pos->v4Pos,TH_EXTRUDE);
      v4Temp2=ExtrudeV4(v4LightPos,Faces[i].b->Pos->v4Pos,TH_EXTRUDE);
      glBegin(GL_QUADS);
        glVertex3f(Faces[i].a->Pos->v4Pos.x,Faces[i].a->Pos->v4Pos.y,Faces[i].a->Pos->v4Pos.z);
        glVertex3f(Faces[i].b->Pos->v4Pos.x,Faces[i].b->Pos->v4Pos.y,Faces[i].b->Pos->v4Pos.z);
        glVertex3f(v4Temp2.x,v4Temp2.y,v4Temp2.z);
        glVertex3f(v4Temp1.x,v4Temp1.y,v4Temp1.z);
      glEnd();
    }

    //Edge 2
    if(Connectivity[i].con[1] != NULL)
      if(!Connectivity[i].con[1]->bVisible)
        draw=true;
      else
        draw=false;
    else
      draw=true;

    if(draw){
      v4Temp1=ExtrudeV4(v4LightPos,Faces[i].b->Pos->v4Pos,TH_EXTRUDE);
      v4Temp2=ExtrudeV4(v4LightPos,Faces[i].c->Pos->v4Pos,TH_EXTRUDE);
      glBegin(GL_QUADS);
        glVertex3f(Faces[i].b->Pos->v4Pos.x,Faces[i].b->Pos->v4Pos.y,Faces[i].b->Pos->v4Pos.z);
        glVertex3f(Faces[i].c->Pos->v4Pos.x,Faces[i].c->Pos->v4Pos.y,Faces[i].c->Pos->v4Pos.z);
        glVertex3f(v4Temp2.x,v4Temp2.y,v4Temp2.z);
        glVertex3f(v4Temp1.x,v4Temp1.y,v4Temp1.z);
      glEnd();
    }

    //Edge 3
    if(Connectivity[i].con[2] != NULL)
      if(!Connectivity[i].con[2]->bVisible)
        draw=true;
      else
        draw=false;
    else
      draw=true;

    if(draw){
      v4Temp1=ExtrudeV4(v4LightPos,Faces[i].c->Pos->v4Pos,TH_EXTRUDE);
      v4Temp2=ExtrudeV4(v4LightPos,Faces[i].a->Pos->v4Pos,TH_EXTRUDE);
      glBegin(GL_QUADS);
        glVertex3f(Faces[i].c->Pos->v4Pos.x,Faces[i].c->Pos->v4Pos.y,Faces[i].c->Pos->v4Pos.z);
        glVertex3f(Faces[i].a->Pos->v4Pos.x,Faces[i].a->Pos->v4Pos.y,Faces[i].a->Pos->v4Pos.z);
        glVertex3f(v4Temp2.x,v4Temp2.y,v4Temp2.z);
        glVertex3f(v4Temp1.x,v4Temp1.y,v4Temp1.z);
      glEnd();
    }
  }

  glBegin(GL_TRIANGLES);
  for(i=0;i<iNumFaces;i++){
    if(Connectivity[i].bVisible)
      continue;

    glVertex3fv(Faces[i].a->Pos->v4Pos.f);
    glVertex3fv(Faces[i].b->Pos->v4Pos.f);
    glVertex3fv(Faces[i].c->Pos->v4Pos.f);

    v4Temp1=ExtrudeV4(v4LightPos,Faces[i].a->Pos->v4Pos,TH_EXTRUDE);
    v4Temp2=ExtrudeV4(v4LightPos,Faces[i].b->Pos->v4Pos,TH_EXTRUDE);
    v4Temp3=ExtrudeV4(v4LightPos,Faces[i].c->Pos->v4Pos,TH_EXTRUDE);

    glVertex3fv(v4Temp3.f);
    glVertex3fv(v4Temp2.f);
    glVertex3fv(v4Temp1.f);
  }
  glEnd();
}

void Model::SetLocation(char Target, GLint Location){
  switch(Target){
    case TH_TANGENT:
      vbuffer->SetAttributeLocation(0,Location);
      break;
    case TH_BINORMAL:
      vbuffer->SetAttributeLocation(1,Location);
      break;
  }
}

bool Model::LoadBillboards3DS(const char *filename, short object_number){
  int i,j;
  FILE *datei;
  #ifdef MODEL_DEBUG
  FILE *debugtxt;
  #endif
  unsigned short chunk_id;
  unsigned int chunk_length;
  unsigned int uiGGroupFlag;
  unsigned char mybyte;
  unsigned short qty; //wieviel wird gelesen
  unsigned short face_flags; //dummy um face flags zu bersprignen
  unsigned short file_length;
  bool first_mat=true;
  unsigned short curr_mat=0;
  unsigned short temp_pos;
  unsigned short curr_obj=0;
  unsigned short acc_face[32];
  unsigned short temp;
  mat4a MeshMatrix,NormalM;
  mat4 TempMat;

  #ifdef MODEL_DEBUG
  debugtxt = fopen("billboard.log","w");
  #endif

  if((datei=fopen(filename,"rb"))==NULL){
    return false;
  }
  fread(&chunk_id,2,1,datei); //Start Chunk lesen
  fread(&file_length,4,1,datei); //Lnge der Datei einlesen
  while(ftell(datei)< file_length){ //Loop bis ende der Datei erreicht ist
    fread(&chunk_id,2,1,datei); //Chunk Id lesen
    fread(&chunk_length,4,1,datei); //Chunk lnge lesen
    switch (chunk_id)
    {
      case 0x3d3d: //3D Editor Chunk
        break; //Einfach lesen ohne zu berspringen

      case 0x4000: //Object Block
        curr_obj++;
        if(curr_obj!=object_number){ //Wenn dieses Obejekt nicht geladen werden soll
          fseek(datei, chunk_length-6, SEEK_CUR); // Object Block berspringen
        }
        else{ //ansonsten Name einlesen
          int i=0;
          mybyte='0';
          //cName = new char[100];
          while(mybyte!=0x00){
            fread(&mybyte,1,1,datei);
            //cName[i]=mybyte;
            i++;
          }
        }
      break;

      case 0x4100: //Muss gelesen werden
      curr_mat=0; first_mat=true;
      break;

      case 0x4110: //Scheitelpunkte einlesen
        fread(&qty,2,1,datei); //Anzahl der Scheitelpunkte
        iNumVertices = qty;
        iNumPositions = qty;
        Vertices = new vertexv4[qty];
        Positions = new vertexposv4[qty];
        for (i=0;i<qty;i++){
          fread(&Positions[i].v4Pos.x,4,1,datei);
          fread(&Positions[i].v4Pos.y,4,1,datei);
          fread(&Positions[i].v4Pos.z,4,1,datei);
          Positions[i].v4Pos.w=1.0f;
          Positions[i].v4Normals=NULL;
          Vertices[i].uiNum = i;
          Positions[i].usTimesUsed=0;
          Positions[i].usUsedIn=NULL;
          Vertices[i].Pos = &Positions[i];
        }
      break;

      case 0x4120: //Flchen einlesen
        fread(&qty,2,1,datei); //Anzahl der Flchen
        iNumFaces=qty;
        Faces=new facev4[qty];
        for(i=0;i<qty;i++){
          fread(&temp,2,1,datei);
          Faces[i].a = &Vertices[temp];
          fread(&temp,2,1,datei);
          Faces[i].b = &Vertices[temp];
          fread(&temp,2,1,datei);
          Faces[i].c = &Vertices[temp];
          fread (&face_flags,2,1,datei);
        }
      break;

      case 0x4140: //Mapping Koordinaten
        fread(&qty,2,1,datei); //Anzahl der Mapping Koordinaten
        if(qty == iNumVertices){
          bHasTexcoords=true;
          for(i=0;i<qty;i++){
            fread(&Vertices[i].v2Texcoord.x,4,1,datei);
            fread(&Vertices[i].v2Texcoord.y,4,1,datei);
          }
        }
        else{
          fseek(datei,chunk_length-8,SEEK_CUR);
          bHasTexcoords=false;
        }
      break;

      case 0x4160: //Mesh Matrix
        MeshMatrix.set(0);
        fread(&MeshMatrix.f[0],4,3,datei);
        fread(&MeshMatrix.f[4],4,3,datei);
        fread(&MeshMatrix.f[8],4,3,datei);
        fread(&MeshMatrix.f[12],4,3,datei);
        //MeshMatrix = Transponse(MeshMatrix);
        //MeshMatrix.f[15]=1.0f;
        MeshMatrix = Right2Left(MeshMatrix);
        NormalM = NormalMatrix(MeshMatrix);
        #ifdef MODEL_DEBUG
        fprintf(debugtxt,"MeshMatrix:\n");
        fprintf(debugtxt,"%f %f %f %f\n",MeshMatrix.f[0],MeshMatrix.f[4],MeshMatrix.f[8],MeshMatrix.f[12]);
        fprintf(debugtxt,"%f %f %f %f\n",MeshMatrix.f[1],MeshMatrix.f[5],MeshMatrix.f[9],MeshMatrix.f[13]);
        fprintf(debugtxt,"%f %f %f %f\n",MeshMatrix.f[2],MeshMatrix.f[6],MeshMatrix.f[10],MeshMatrix.f[14]);
        fprintf(debugtxt,"%f %f %f %f\n",MeshMatrix.f[3],MeshMatrix.f[7],MeshMatrix.f[11],MeshMatrix.f[15]);
        fprintf(debugtxt,"\nNormalMatrix\n");
        fprintf(debugtxt,"%f %f %f %f\n",NormalM.f[0],NormalM.f[4],NormalM.f[8],NormalM.f[12]);
        fprintf(debugtxt,"%f %f %f %f\n",NormalM.f[1],NormalM.f[5],NormalM.f[9],NormalM.f[13]);
        fprintf(debugtxt,"%f %f %f %f\n",NormalM.f[2],NormalM.f[6],NormalM.f[10],NormalM.f[14]);
        fprintf(debugtxt,"%f %f %f %f\n",NormalM.f[3],NormalM.f[7],NormalM.f[11],NormalM.f[15]);
        #endif
        //TempMat.f[12]=0; TempMat.f[13]=0; TempMat.f[14]=0; TempMat.f[15]=1;
      break;


      default: //Unbekannte Chunks berspringen
        fseek(datei, chunk_length-6, SEEK_CUR);
      break;
    }
  }
 fclose (datei);

 if(!bHasTexcoords || Faces == NULL || Positions == NULL || Vertices == NULL){
   if(Faces != NULL){
     delete []Faces;
     Faces = NULL;
   }
   if(Vertices != NULL){
     delete []Vertices;
     Vertices = NULL;
   }
   if(Positions != NULL){
     delete []Positions;
     Positions = NULL;
   }
   return false;
 }

 #ifdef MODEL_DEBUG
 for(i=0;i<iNumFaces;i++)
   fprintf(debugtxt,"Face %i\nPoints: %i %i %i\n",i,Faces[i].vertices[0]->uiNum,Faces[i].vertices[1]->uiNum,Faces[i].vertices[2]->uiNum);
 fprintf(debugtxt,"\n\n");
 for(i=0;i<iNumPositions;i++)
   fprintf(debugtxt,"Position %i\n%f %f %f\n",i,Positions[i].v4Pos.x,Positions[i].v4Pos.y,Positions[i].v4Pos.z);
 #endif

 //Vertexbuffer erstellen
 vbuffer = new VertexBuffer(TH_VB_NORMAL | TH_VB_TEXCOORD0_4,GL_QUADS,GL_UNSIGNED_SHORT);
 vbuffer->ReserveVertexData(iNumFaces * 3);

 vec4 v4Temp,v4Temp2,v4BillboardPosition,v4Normal;
 vec4 BillboardPositions[4];
 vec2 BillboardTexcoords[4];
 bool found;
 bool *bUsed = new bool[iNumFaces];
 for(i=0;i<iNumFaces;i++)
   bUsed[i] = false;

 for(i=0;i<iNumFaces;i++){
   if(bUsed[i])
     continue;
   for(j=0;j<iNumFaces;j++){
     if(i==j || bUsed[j])
       continue;
     found = false;
     if(Faces[i].vertices[0] == Faces[j].vertices[2] &&
        Faces[i].vertices[2] == Faces[j].vertices[0])
     {
        BillboardPositions[0] = Faces[i].vertices[1]->Pos->v4Pos;
        BillboardPositions[1] = Faces[i].vertices[2]->Pos->v4Pos;
        BillboardPositions[2] = Faces[j].vertices[1]->Pos->v4Pos;
        BillboardPositions[3] = Faces[j].vertices[2]->Pos->v4Pos;

        BillboardTexcoords[0] = Faces[i].vertices[1]->v2Texcoord;
        BillboardTexcoords[1] = Faces[i].vertices[2]->v2Texcoord;
        BillboardTexcoords[2] = Faces[j].vertices[1]->v2Texcoord;
        BillboardTexcoords[3] = Faces[j].vertices[2]->v2Texcoord;
        found=true;
     }

     if(found){
        bUsed[j]=true;
        bUsed[i]=true;

        v4Temp = (BillboardPositions[0] + BillboardPositions[1]) * 0.5;
        v4Temp2 = (BillboardPositions[2] + BillboardPositions[3]) * 0.5;
        v4BillboardPosition = (v4Temp + v4Temp2) * 0.5;
        v4BillboardPosition.w = 1.0f;

        //Matrices not needed??
        v4Normal = NormalizeV4(v4BillboardPosition);
        v4Temp =  v4BillboardPosition;
        #ifdef MODEL_DEBUG
        fprintf(debugtxt,"\nBillboard:\nPosition: %f %f %f\nNormal: %f %f %f\n",v4Temp.x,v4Temp.y,v4Temp.z,v4Normal.x,v4Normal.y,v4Normal.z);
        #endif
        for(int i=0;i<4;i++){
          //Vertex Position
          vbuffer->AddVertexData(v4Temp.f,3);
          //Normal
          v4Temp2.set(0.0f);
          v4Temp2.w = 1.0f;
          v4Temp2 = MeshMatrix * v4Temp2;
          v4Temp2 = NormalizeV4(v4Temp2 - v4Temp);
          vbuffer->AddVertexData(v4Temp2.f,3);
          //vbuffer->AddVertexData(v4Normal.f,3);
          //Texcoord
          v4Temp2 = BillboardPositions[i] - v4BillboardPosition;
          v4Temp2.w = v4Temp2.z;
          v4Temp2.z = v4Temp2.x;
          v4Temp2.x = BillboardTexcoords[i].x;
          v4Temp2.y = BillboardTexcoords[i].y;
          vbuffer->AddVertexData(v4Temp2);
        }
     }
   }
 }
 delete []bUsed;
 if(Faces != NULL){
   delete []Faces;
   Faces = NULL;
 }
 if(Vertices != NULL){
   delete []Vertices;
   Vertices = NULL;
 }
 if(Positions != NULL){
   delete []Positions;
   Positions = NULL;
 }
 #ifdef MODEL_DEBUG
 fclose(debugtxt);
 #endif

 if(!vbuffer->UploadData())
   return false;
 return true;
}

bool Model::LoadBillboardsSchrompf(const char *filename,vec4 center){
  float array[20];
  int array2[3];
  vec4 v4Temp;
  float fTemp;
  int size;
  int test;
  FILE *datei;
  if((datei=fopen(filename,"rb"))==NULL)
    return false;

  if(vbuffer == NULL){
    vbuffer = new VertexBuffer(TH_VB_NORMAL | TH_VB_TEXCOORD0_4,GL_TRIANGLES,GL_UNSIGNED_SHORT);
    vbuffer->AddIndexBuffer();
  }

  fread(&size,sizeof(int),1,datei);
  vbuffer->ReserveAdditionalVertexData(size);

  for(int i=0;i<size;i++){
    //Positionen
    fread(&v4Temp,sizeof(float),3,datei);
    fTemp = v4Temp.z;
    v4Temp.z = v4Temp.y;
    v4Temp.y = fTemp;
    vbuffer->AddVertexData(v4Temp.f,3);
    v4Temp.w = 1.0f;
    v4Temp = NormalizeV4(center - v4Temp);

    //Normalen
    fread(&array,sizeof(float),3,datei);
    vbuffer->AddVertexData(v4Temp.f,3);

    //Nutzlose daten
    fread(&array,sizeof(float),4,datei);

    //Texcoords
    fread(&array,sizeof(float),2,datei);
    array[1] = array[1] * -1.0f + 1.0f;
    vbuffer->AddVertexData(array,2);

    //Nutzlose Daten
    fread(&array,sizeof(float),3,datei);

    //Entfaltungsdaten
    fread(&array,sizeof(float),4,datei);
    vbuffer->AddVertexData(array,2);
  }
  //Test integer
  fread(&test,sizeof(int),1,datei);

  fread(&size,sizeof(int),1,datei);
  for(int i=0;i<size;i++){
    fread(array2,sizeof(int),3,datei);
    vbuffer->AddIndexData(0,array2,3);
  }

  fclose(datei);

  return vbuffer->UploadData();
}

bool Model::LoadSchrompf(const char *filename, char options){
  float array[20];
  int array2[3];
  vec4 v4Normal,v4Tangent,v4Binormal;
  int size;
  int test;
  float BinormalLength,fTemp;
  FILE *datei;
  if((datei=fopen(filename,"rb"))==NULL)
    return false;

  int iVBoptions = 0;
  if(options & TH_TEXCOORDS)
    iVBoptions |= TH_VB_TEXCOORD0_2;
  if(options & TH_NORMALS_PER_FACE || options & TH_NORMALS_PER_VERTEX)
    iVBoptions |= TH_VB_NORMAL;
  if(options & TH_TANGENT)
    iVBoptions |= TH_VB_ATTRIBUTE0_3;
  if(options & TH_BINORMAL)
    iVBoptions |= TH_VB_ATTRIBUTE1_3;


  if(vbuffer == NULL){
    vbuffer = new VertexBuffer(iVBoptions,GL_TRIANGLES,GL_UNSIGNED_SHORT);
    vbuffer->AddIndexBuffer();
  }

  fread(&size,sizeof(int),1,datei);
  vbuffer->ReserveAdditionalVertexData(size);

  for(int i=0;i<size;i++){
    //Positionen
    fread(&array,sizeof(float),3,datei);
    fTemp    = array[2];
    array[2] = array[1];
    array[1] = fTemp;
    vbuffer->AddVertexData(array,3);

    //Normalen
    fread(v4Normal.f,sizeof(float),3,datei);
    fTemp = v4Normal.z;
    v4Normal.z = v4Normal.y;
    v4Normal.y = fTemp;
    v4Normal = MulV4(v4Normal,-1.0f);
    if(options & TH_NORMALS_PER_FACE || options & TH_NORMALS_PER_VERTEX)
      vbuffer->AddVertexData(v4Normal.f,3);

    //Tangente
    fread(v4Tangent.f,sizeof(float),3,datei);
    fTemp = v4Tangent.z;
    v4Tangent.z = v4Tangent.y;
    v4Tangent.y = fTemp;
    v4Tangent = MulV4(v4Tangent,-1.0f);

    //Binormale
    fread(&BinormalLength,sizeof(float),1,datei);

    //Texcoords
    fread(&array,sizeof(float),2,datei);
    if(options & TH_TEXCOORDS){
      array[1] = array[1] * -1.0f + 1.0f;
      vbuffer->AddVertexData(array,2);
    }

    //Richtige Reihenfolge
    if(options & TH_TANGENT)
      vbuffer->AddVertexData(v4Tangent.f,3);
    if(options & TH_BINORMAL){
      v4Binormal = MulV4(NormalizeV4(CrossProductV4(v4Normal,v4Tangent)),-BinormalLength);
      //v4Binormal = CrossProductV4(v4Normal,v4Tangent);
      vbuffer->AddVertexData(v4Binormal.f,3);
    }

    //Nutzlose Daten
    fread(&array,sizeof(float),7,datei);
  }
  //Test integer
  fread(&test,sizeof(int),1,datei);

  fread(&size,sizeof(int),1,datei);
  for(int i=0;i<size;i++){
    fread(array2,sizeof(int),3,datei);
    vbuffer->AddIndexData(0,array2,3);
  }

  fclose(datei);

  return vbuffer->UploadData();
}
