/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "thEngine.h"
#include <fstream>
#include <iostream>
#include "ois.h"

//////////////////////////////// OS Nuetral Headers ////////////////
#include "OISInputManager.h"
#include "OISException.h"
#include "OISKeyboard.h"
#include "OISMouse.h"
//#include "OISJoyStick.h"
#include "OISEvents.h"

#include <iostream>
#include <vector>
#include <sstream>

////////////////////////////////////Needed Windows Headers////////////
#if defined OIS_WIN32_PLATFORM
#  define WIN32_LEAN_AND_MEAN
#  include <windows.h>
//#  include "resource.h"
//////////////////////////////////////////////////////////////////////
////////////////////////////////////Needed Linux Headers//////////////
#elif defined OIS_LINUX_PLATFORM
#  include <X11/Xlib.h>
#endif
//////////////////////////////////////////////////////////////////////

extern KeyDownFunc KeysDown[256];
extern KeyDownFunc KeysUp[256];
extern bool KeyStates[256];
extern MouseMoveFunc OnMouseMoveAbsolute;
extern MouseMoveFunc OnMouseMoveRelative;
extern MouseMoveFuncGui OnMouseMove;
extern MouseButtonFunc OnMouseButtonDown;
extern MouseButtonFunc OnMouseButtonUp;
extern MouseWheelFunc OnMouseWheel;
extern TextFunc OnTextIn;
extern bool KeyESC;

using namespace OIS;

InputManager *g_InputManager = 0; //Input Manager
Keyboard *g_kb  = 0;		//Keyboard Device
Mouse	 *g_m   = 0;		//Mouse Device

//////////// Common Event handler class ////////
class EventHandler : public KeyListener, public MouseListener
{
public:
	EventHandler() {}
	~EventHandler() {}
	bool keyPressed( const KeyEvent &arg ) {
	    if(arg.key == KC_ESCAPE)
	      KeyESC = true;
	    if(arg.key >= 0 && arg.key <= 255){
          if(KeysDown[arg.key] != NULL)
            KeysDown[arg.key]();
          KeyStates[arg.key]=true;
	    }
	    if(OnTextIn != NULL){
          if(arg.key == 28)
            OnTextIn('\n');
          else
	        OnTextIn((char)arg.text);
	    }
	    #ifdef TH_OIS_DEBUG
		std::cout << "\nKeyPressed {" << arg.key
			<< ", " << ((Keyboard*)(arg.device))->getAsString(arg.key)
			<< "} || Character (" << (char)arg.text << ")" << std::endl;
        #endif
		return true;
	}
	bool keyReleased( const KeyEvent &arg ) {
	    if(arg.key >= 0 && arg.key <= 255){
	      KeyStates[arg.key]=false;
	      if(KeysUp[arg.key] != NULL)
	        KeysUp[arg.key]();
        }
		return true;
	}
	bool mouseMoved( const MouseEvent &arg ) {
		const OIS::MouseState& s = arg.state;
		if(OnMouseMoveAbsolute != NULL)
		  #ifdef __linux
		  OnMouseMoveAbsolute(s.X.abs+6,s.Y.abs+4);
		  #elif defined WIN32
		  OnMouseMoveAbsolute(s.X.abs,s.Y.abs);
		  #endif
        if(OnMouseMoveRelative != NULL)
          OnMouseMoveRelative(s.X.rel,s.Y.rel);
        if(OnMouseMove != NULL)
          #ifdef __linux
          OnMouseMove(s.X.abs+6,s.Y.abs+4,s.X.rel,s.Y.rel);
          #elif defined WIN32
          OnMouseMove(s.X.abs,s.Y.abs,s.X.rel,s.Y.rel);
          #endif
        if(OnMouseWheel != NULL)
          OnMouseWheel(s.Z.rel);
        #ifdef TH_OIS_DEBUG
		std::cout << "\nMouseMoved: Abs("
				  << s.X.abs << ", " << s.Y.abs << ", " << s.Z.abs << ") Rel("
				  << s.X.rel << ", " << s.Y.rel << ", " << s.Z.rel << ")";
        #endif
		return true;
	}
	bool mousePressed( const MouseEvent &arg, MouseButtonID id ) {
		const OIS::MouseState& s = arg.state;
		if(OnMouseButtonDown != NULL)
		  #ifdef __linux
		  OnMouseButtonDown(id,s.X.abs+6,s.Y.abs+4,s.X.rel,s.Y.rel);
		  #elif defined WIN32
		  OnMouseButtonDown(id,s.X.abs,s.Y.abs,s.X.rel,s.Y.rel);
		  #endif
		#ifdef TH_OIS_DEBUG
		std::cout << "\nMouse button #" << id << " pressed. Abs("
				  << s.X.abs << ", " << s.Y.abs << ", " << s.Z.abs << ") Rel("
				  << s.X.rel << ", " << s.Y.rel << ", " << s.Z.rel << ")";
        #endif
		return true;
	}
	bool mouseReleased( const MouseEvent &arg, MouseButtonID id ) {
		const OIS::MouseState& s = arg.state;
		if(OnMouseButtonUp != NULL)
		  #ifdef __linux
		  OnMouseButtonUp(id,s.X.abs+6,s.Y.abs+4,s.X.rel,s.Y.rel);
		  #elif defined WIN32
		  OnMouseButtonUp(id,s.X.abs,s.Y.abs,s.X.rel,s.Y.rel);
		  #endif
		#ifdef TH_OIS_DEBUG
		std::cout << "\nMouse button #" << id << " released. Abs("
				  << s.X.abs << ", " << s.Y.abs << ", " << s.Z.abs << ") Rel("
				  << s.X.rel << ", " << s.Y.rel << ", " << s.Z.rel << ")";
        #endif
		return true;
	}
};

//Create a global instance
EventHandler handler;



#if defined OIS_WIN32_PLATFORM
bool SetupOIS(HWND hWnd){
#elif defined OIS_LINUX_PLATFORM
bool SetupOIS(Window xWin,Display *xDisp){
#endif
  ParamList pl;
  std::ostringstream wnd;

  //Windows Stuff
  #if defined OIS_WIN32_PLATFORM
    wnd << (size_t)hWnd;
    pl.insert(std::make_pair( std::string("WINDOW"), wnd.str() ));
	pl.insert(std::make_pair(std::string("w32_mouse"), std::string("DISCL_FOREGROUND" )));
	pl.insert(std::make_pair(std::string("w32_mouse"), std::string("DISCL_NONEXCLUSIVE")));
  //Linux Stuff
  #elif defined OIS_LINUX_PLATFORM
	//Select what events we want to listen to locally
	XSelectInput(xDisp, xWin, StructureNotifyMask);
	XEvent evtent;
	do
	{
		XNextEvent(xDisp, &evtent);
	} while(evtent.type != MapNotify);
	wnd << xWin;

	pl.insert(std::make_pair(std::string("WINDOW"), wnd.str()));
	pl.insert(std::make_pair(std::string("x11_keyboard_grab"),std::string("false")));
	if(bIsFullscreen){
	  pl.insert(std::make_pair(std::string("x11_mouse_grab"), std::string("true")));
	  pl.insert(std::make_pair(std::string("x11_mouse_hide"), std::string("true")));
	}
	else{
	  pl.insert(std::make_pair(std::string("x11_mouse_grab"), std::string("false")));
	  pl.insert(std::make_pair(std::string("x11_mouse_hide"), std::string("false")));
	}

  #endif

  try {
	//This never returns null.. it will raise an exception on errors
	g_InputManager = InputManager::createInputSystem(pl);

	unsigned int v = g_InputManager->getVersionNumber();
	std::ofstream LogFile;
	LogFile.open("engine.log",std::ios::app);
	LogFile << "OIS Version: " << (v>>16 ) << "." << ((v>>8) & 0x000000FF) << "." << (v & 0x000000FF)
		<< "\n\tRelease Name: " << g_InputManager->getVersionName()
		<< "\n\tPlatform: " << g_InputManager->inputSystemName()
		<< "\n\tNumber of Mice: " << g_InputManager->numMice()
		<< "\n\tNumber of Keyboards: " << g_InputManager->numKeyboards()
		<< "\n\tNumber of Joys/Pads = " << g_InputManager->numJoySticks() << "\n\t";

	if( g_InputManager->numKeyboards() > 0 )
	{
		g_kb = (Keyboard*)g_InputManager->createInputObject( OISKeyboard, true );
		LogFile << "Created buffered keyboard\n\t";
		g_kb->setEventCallback( &handler );
	}

	if( g_InputManager->numMice() > 0 )
	{
		g_m = (Mouse*)g_InputManager->createInputObject( OISMouse, true );
		const OIS::MouseState& TempState = g_m->getMouseState();
		TempState.width = iWidth;
		TempState.height = iHeight;
		LogFile << "Created buffered mouse\n\t";
		g_m->setEventCallback( &handler );

	}

	LogFile << "\n";
	LogFile.close();
  }
  catch( const Exception &ex )
  {
	#if defined OIS_WIN32_PLATFORM
      MessageBox( NULL, ex.eText, "An OIS exception during Startup has occured!", MB_OK |
	  MB_ICONERROR | MB_TASKMODAL);
	#else
	  std::cout << "\nOIS Exception during Startup Caught!\n" << "\t" << ex.eText << "[Line "
      << ex.eLine << " in " << ex.eFile << "]\nExiting App";
    #endif
    return false;
  }
  return true;
}

bool OISCheckInput(){
  try {
    //Keyboard
    if( g_kb )
      g_kb->capture();

    //Mouse
    if( g_m )
      g_m->capture();
  }
  catch( const Exception &ex )
  {
	#if defined OIS_WIN32_PLATFORM
      MessageBox( NULL, ex.eText, "An OIS exception has occured!", MB_OK |
	  MB_ICONERROR);
	#else
	  std::cout << "\nOIS Exception Caught!\n" << "\t" << ex.eText << "[Line "
      << ex.eLine << " in " << ex.eFile << "]\n";
    #endif
    return false;
  }
  return true;
}

void OISSetWindowSize(int x, int y){
  if(g_m){
    const MouseState &ms = g_m->getMouseState();
    ms.width = x;
    ms.height = y;
  }
}

