/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "plants.h"
#include "math.h"
#include "thEngine.h"

Shader* PlantScrub::ColorShader;
Shader* PlantScrub::ShadowShader;

void PlantScrub::SetColorShader(Shader *ptr){
  ColorShader = ptr;
}

void PlantScrub::SetShadowShader(Shader *ptr){
  ShadowShader = ptr;
}

PlantScrub::PlantScrub(float Radius, int RandomSeed, Texture2D *ptr){
  FrustrumObjectSphere *BoundingSphere;
  vec4 TempPos,TempNormal,Middle;
  Textur = ptr;
  fRadius = Radius;
  BoundingSphere = new FrustrumObjectSphere;
  BoundingSphere->SetRadius(fRadius*3.0f);
  TempPos = v4Position;
  TempPos.z += fRadius*3.0f;
  BoundingSphere->SetPosition(TempPos);


  vec2 TexCoord1[4];
  vec2 TexCoord2[4];
  NumLeafes=64;
  float LeafSize=0.5f;
  char *lpcTemp;
  lpcVertexData = new char[10 * sizeof(float) * NumLeafes * 4];
  lpcTexCoordData1 = lpcVertexData + (3 * sizeof(float) * NumLeafes * 4);
  lpcTexCoordData2 = lpcTexCoordData1 + (2 * sizeof(float) * NumLeafes * 4);
  lpcNormalData = lpcTexCoordData2 + (2 * sizeof(float) * NumLeafes * 4);

  TexCoord1[0].x = 0.0f; TexCoord1[0].y = 0.0f; TexCoord2[0].x = -LeafSize; TexCoord2[0].y = -LeafSize;
  TexCoord1[1].x = 0.0f; TexCoord1[1].y = 1.0f; TexCoord2[1].x = -LeafSize; TexCoord2[1].y =  LeafSize;
  TexCoord1[2].x = 1.0f; TexCoord1[2].y = 1.0f; TexCoord2[2].x =  LeafSize; TexCoord2[2].y =  LeafSize;
  TexCoord1[3].x = 1.0f; TexCoord1[3].y = 0.0f; TexCoord2[3].x =  LeafSize; TexCoord2[3].y = -LeafSize;

  Middle.x = Middle.y = 0.0f;
  Middle.y = fRadius;
  Middle.w = 1.0f;

  for(int i=0;i<NumLeafes;i++){
    for(int j=0;j<4;j++){
      memcpy(lpcTexCoordData1,&TexCoord1[j],sizeof(float)*2);
      lpcTexCoordData1 += sizeof(float)*2;

      memcpy(lpcTexCoordData2,&TexCoord2[j],sizeof(float)*2);
      lpcTexCoordData2 += sizeof(float)*2;
    }
  }

  lpcTemp = lpcVertexData;
  srand(RandomSeed);
  float step,xvalue;
  for(int i=0;i<8;i++){
    xvalue=(float)i/7.0f;
    step = (1-pow(0.025,1-xvalue))*(1.5-pow(0.1,xvalue));
    for(int j=0;j<8;j++){
      TempPos.x = sin(M_PI * 2.0f * (float)j/7.0f + 25.0f * (float)i) * fRadius * step + float(rand()%200-100)/400.0f;
      TempPos.y = cos(M_PI * 2.0f * (float)j/7.0f + 25.0f * (float)i) * fRadius * step + float(rand()%200-100)/400.0f;
      TempPos.z = xvalue * fRadius * 2.0f + float(rand()%200-100)/400.0f;
      TempPos.w = 1.0f;
      TempNormal = NormalizeV4(SubV4(TempPos,Middle));
      for(int k=0;k<4;k++){
        memcpy(lpcTemp,&TempPos,3 * sizeof(float));
        lpcTemp += 3 * sizeof(float);
        memcpy(lpcNormalData,&TempNormal,3 * sizeof(float));
        lpcNormalData += 3 * sizeof(float);
      }
    }
  }

  glGenBuffersARB(1,&gluiVertexBuffer);
  glBindBufferARB(GL_ARRAY_BUFFER_ARB,gluiVertexBuffer);
  printf("upload data\n");
  glBufferDataARB(GL_ARRAY_BUFFER_ARB, NumLeafes * 4 * 10 * sizeof(float), lpcVertexData, GL_STATIC_DRAW_ARB);

  ThGetGLError("Upload Bush Data");

  glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
  delete lpcVertexData;

  lpcVertexData = NULL;
  lpcTexCoordData1 = lpcVertexData + 3 * sizeof(float) * NumLeafes * 4;
  lpcTexCoordData2 = lpcTexCoordData1 + 2 * sizeof(float) * NumLeafes * 4;
  lpcNormalData = lpcTexCoordData2 + 2 * sizeof(float) * NumLeafes * 4;

  BoundingObject = BoundingSphere;
  Recalc();
}

PlantScrub::~PlantScrub(){
  glDeleteBuffersARB(1,&gluiVertexBuffer);
  delete BoundingObject;
}

void PlantScrub::Recalc(){
  vec4 TempPos;
  FrustrumObjectSphere *BoundingSphere;
  mat4a trans,rot;
  trans = TranslationMatrix(v4Position);
  rot = RotationMatrixXYZ(v3Rotation);
  Matrix = Mul(trans,rot);
  BoundingSphere = (FrustrumObjectSphere*)BoundingObject;
  TempPos = v4Position;
  TempPos.z += fRadius*3.0f;
  BoundingSphere->SetPosition(v4Position);
}

void PlantScrub::Draw(char mode){
  if(!BoundingObject->IsInFrustrum())
    return;
  glPushMatrix();
  glMultMatrixf(Matrix.f);
  switch(mode){
    case TH_DRAW_MODE_COLOR:
      Textur->Bind(0);
      ColorShader->Use();
      ColorShader->SetUniform(1,fTime);
      glBindBuffer(GL_ARRAY_BUFFER_ARB,gluiVertexBuffer);
      glEnableClientState(GL_VERTEX_ARRAY);
      glEnableClientState(GL_NORMAL_ARRAY);
      glVertexPointer(3,GL_FLOAT,0,lpcVertexData);
      glNormalPointer(GL_FLOAT,0,lpcNormalData);
      glClientActiveTexture(GL_TEXTURE0);
      glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      glTexCoordPointer(2,GL_FLOAT,0,lpcTexCoordData1);
      glClientActiveTexture(GL_TEXTURE1);
      glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      glTexCoordPointer(2,GL_FLOAT,0,lpcTexCoordData2);

      glDrawArrays(GL_QUADS,0,NumLeafes*4);

      glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      glClientActiveTexture(GL_TEXTURE0);
      glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      glDisableClientState(GL_NORMAL_ARRAY);
      glDisableClientState(GL_VERTEX_ARRAY);
      glBindBuffer(GL_ARRAY_BUFFER_ARB,0);
      UnloadShader();
      break;
    case TH_DRAW_MODE_SHADOW:
      Textur->Bind(0);
      //ShadowShader->Use();
      ShadowShader->SetUniform(1,fTime);
      glBindBuffer(GL_ARRAY_BUFFER_ARB,gluiVertexBuffer);
      glEnableClientState(GL_VERTEX_ARRAY);
      glEnableClientState(GL_NORMAL_ARRAY);
      glVertexPointer(3,GL_FLOAT,0,lpcVertexData);
      glNormalPointer(GL_FLOAT,0,lpcNormalData);
      glClientActiveTexture(GL_TEXTURE0);
      glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      glTexCoordPointer(2,GL_FLOAT,0,lpcTexCoordData1);
      glClientActiveTexture(GL_TEXTURE1);
      glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      glTexCoordPointer(2,GL_FLOAT,0,lpcTexCoordData2);

      glDrawArrays(GL_QUADS,0,NumLeafes*4);

      glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      glClientActiveTexture(GL_TEXTURE0);
      glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      glDisableClientState(GL_NORMAL_ARRAY);
      glDisableClientState(GL_VERTEX_ARRAY);
      glBindBuffer(GL_ARRAY_BUFFER_ARB,0);
      //UnloadShader();
      break;
  }
  glPopMatrix();
}

void PlantScrub::DrawBounding(){
  ((FrustrumObjectSphere*)BoundingObject)->Draw();
}

void PlantScrub::SetTime(float value){
  fTime = value;
}


PlantGrasQuad::PlantGrasQuad(Landscape *Terrain, float x, float y, float size, Texture2D *coverage,vec4 TargetColor){
  vec4 v,n,color;
  bool first=true;
  int TexSizeX,TexSizeY,TexStartX,TexStartY;
  TexStartX = (int)((float)coverage->GetWidth() * x);
  TexStartY = (int)((float)coverage->GetHeight() * y);
  TexSizeX = (int)((float)coverage->GetWidth() * size);
  TexSizeY = (int)((float)coverage->GetHeight() * size);
  TexSizeX = (TexStartX + TexSizeX > coverage->GetWidth()) ? coverage->GetWidth() - TexStartX : TexSizeX;
  TexSizeY = (TexStartY + TexSizeY > coverage->GetHeight()) ? coverage->GetHeight() - TexStartY : TexSizeY;

  for(int iy=0;iy<TexSizeY;iy++){
    for(int ix=0;ix<TexSizeX;ix++){
      color = coverage->GetPixel(TexStartX + ix, TexStartY + iy);
      if(color.x == TargetColor.x && color.y == TargetColor.y && color.z == TargetColor.z){
        v = Terrain->GetPosition((float)(TexStartX + ix)/(float)coverage->GetWidth(),(float)(TexStartY + iy)/(float)coverage->GetHeight());
        n = Terrain->GetNormal((float)(TexStartX + ix)/(float)coverage->GetWidth(),(float)(TexStartY + iy)/(float)coverage->GetHeight());
        v4Positions.push_back(v);
        v4Normals.push_back(n);
      }
    }
  }

  v = Terrain->GetPosition(x,y);
  n = Terrain->GetPosition(x+size,y+size);
  v4Position = v+(n-v)*0.5;
}

float PlantGrasQuad::fGrasDistance;
vec4 PlantGrasQuad::CamPos;

float PlantGrasQuad::Distance(){
  vec4 Point;
  Point = v4Position - CamPos;
  Point = Point * Point;
  return sqrt(Point.x + Point.y + Point.z);
}

PlantGrasQuadMesh::PlantGrasQuadMesh(InstancedVertexBuffer *vbuffer, Landscape *Terrain, float x, float y, float size, Texture2D *coverage, vec4 TargetColor) : PlantGrasQuad(Terrain,x,y,size,coverage,TargetColor){
  vec4 v,n,vx,vy;
  mat4a mat,res;
  float winkel;
  mat.set(0);
  mat.f[15]=1.0f;
  InstancedVBuffer = vbuffer;
  for(int i=0;i<v4Positions.size();i++){
    v = v4Positions[i];
    n = v4Normals[i];
    winkel = (float)(rand()%10000)/10000.0f;
    vx.x = cos(winkel); vx.y = sin(winkel); vx.z = 0.0f; vx.w = 1.0f;
    vy = NormalizeV4(CrossProductV4(vx,n));
    vx = NormalizeV4(CrossProductV4(vy,n));
    mat.f[0] = vx.x; mat.f[1] = vy.x; mat.f[2] = n.x;
    mat.f[4] = vx.y; mat.f[5] = vy.y; mat.f[6] = n.y;
    mat.f[8] = vx.z; mat.f[9] = vy.z; mat.f[10] = n.z;
    res = Mul(mat,TranslationMatrix(v));
    Matrizen.push_back(res);
  }
  iInstanceGroup = InstancedVBuffer->AddInstanceGroup(&Matrizen[0],Matrizen.size());
  v4Normals.clear();
}

void PlantGrasQuadMesh::Draw(char mode){
  InstancedVBuffer->Draw(iInstanceGroup);
}

// TODO (Benjamin#1#): Pflanzen-Manager fertigstellen (dann Programm wieder lauffhig)

bool PlantManager::bFallbackInstancing = true;

void PlantManager::SetFallbackInstancing(bool value){
  bFallbackInstancing = value;
}

PlantManager::PlantManager(Landscape *ter, int Divisions){
  Terrain = ter;
  float fStep = 1.0f / (float)Divisions;
  vec4 min,max,res;
  FrustrumObjectBox *ptr;
  iDivisions = Divisions;
  fGrasDistance = 30;
  iGetShadow = 0;
  DepthTexture = NULL;

  FrustrumObjects = new FrustrumObjectBox[iDivisions * iDivisions];
  FrustrumVisible = new bool[iDivisions * iDivisions];
  for(int y=0;y<Divisions;y++){
    for(int x=0;x<Divisions;x++){
      res = Terrain->GetPosition(fStep * (float)x,fStep * (float)y);
      min = res;
      max = res;

      res = Terrain->GetPosition(fStep * (float)(x+1),fStep * (float)y);
      min.x = (res.x < min.x) ? res.x : min.x;
      min.y = (res.y < min.y) ? res.y : min.y;
      min.z = (res.z < min.z) ? res.z : min.z;
      max.x = (res.x > max.x) ? res.x : max.x;
      max.y = (res.y > max.y) ? res.y : max.y;
      max.z = (res.z > max.z) ? res.z : max.z;

      res = Terrain->GetPosition(fStep * (float)x, fStep * (float)(y+1));
      min.x = (res.x < min.x) ? res.x : min.x;
      min.y = (res.y < min.y) ? res.y : min.y;
      min.z = (res.z < min.z) ? res.z : min.z;
      max.x = (res.x > max.x) ? res.x : max.x;
      max.y = (res.y > max.y) ? res.y : max.y;
      max.z = (res.z > max.z) ? res.z : max.z;

      res = Terrain->GetPosition(fStep * (float)(x+1), fStep * (float)(y+1));
      min.x = (res.x < min.x) ? res.x : min.x;
      min.y = (res.y < min.y) ? res.y : min.y;
      min.z = (res.z < min.z) ? res.z : min.z;
      max.x = (res.x > max.x) ? res.x : max.x;
      max.y = (res.y > max.y) ? res.y : max.y;
      max.z = (res.z > max.z) ? res.z : max.z;

      ptr = &FrustrumObjects[y*iDivisions + x];
      max.z+=2.0f;
      min.z-=1.0f;
      ptr->SetSize(max.x - min.x, max.y - min.y, max.z - min.z);
      min.w = 1.0f;
      ptr->Transform(TranslationMatrix(min+(max-min)*0.5));
    }
  }
}

PlantManager::~PlantManager(){
  delete []FrustrumObjects;
  delete []FrustrumVisible;
  for(int i=0;i<Models.size();i++)
    delete Models[i];
}

void PlantManager::AddTexture(Texture2D *ptr){
  Textures.push_back(ptr);
}

void PlantManager::AddShader(XMLShader *ptr){
  Shader.push_back(ptr);
}

void PlantManager::AddModel(Model *ptr, vec4 Color, int Texture, int ShaderNum){
  InstancedVertexBuffer *temp;
  if(!bFallbackInstancing)
    temp = new InstancedVertexBufferExtension(ptr->GetVertexBuffer());
  else
    temp = new InstancedVertexBufferFallback(ptr->GetVertexBuffer(),64);
  Models.push_back(temp);
  Colors.push_back(Color);
  ModelTextureUsage.push_back(Texture);
  ModelShaderUsage.push_back(ShaderNum);
}

PlantManagerSoft::PlantManagerSoft(Landscape *ter, int Divisions) : PlantManager(ter,Divisions){
  Plants = NULL;
}

PlantManagerSoft::~PlantManagerSoft(){
  if(Plants != NULL){
    for(int i=0;i<Models.size() * iDivisions * iDivisions;i++)
      delete Plants[i];
    delete []Plants;
  }
}

void PlantManagerSoft::SetGrasDistance(float value){
  fGrasDistance = value;
  PlantGrasQuadMesh::SetGrasDistance(fGrasDistance);
}

void PlantManagerSoft::Generate(Texture2D *coverage){
  float fx,fy,fsize;
  if(Models.size() <= 0)
    return;
  Plants = new PlantGrasQuadMesh*[Models.size() * iDivisions * iDivisions];
  fsize = 1.0f/(float)iDivisions;
  for(int y=0;y<iDivisions;y++){
    for(int x=0;x<iDivisions;x++){
      for(int i=0;i<Models.size();i++){
        fx = fsize * (float)x;
        fy = fsize * (float)y;
        Plants[i*iDivisions*iDivisions+y*iDivisions+x] = new PlantGrasQuadMesh(Models[i],Terrain,fx,fy,fsize,coverage,Colors[i]);
      }
    }
  }
}

void PlantManagerSoft::Draw(vec4 CamPos,vec4 CamShadowMapPos, float fTime){
  float fDistance;
  int iLastShader = -1;
  int iLastTexture = -1;
  int iNum;
  if(Models.size() <= 0 || Shader.size() <= 0 || Textures.size() <= 0)
    return;

  //Sichtbarkeit Bestimmen
  for(int i=0;i<iDivisions * iDivisions;i++)
    FrustrumVisible[i] = FrustrumObjects[i].IsInFrustrum();

  if(iGetShadow > 0 && DepthTexture != NULL)
    DepthTexture->Bind(-1,1);

  PlantGrasQuadMesh::SetCamPos(CamPos);

  //Alle Modelle Zeichnen
  for(int i=0;i<Models.size();i++){
    //Anderer Shader?
    if(ModelShaderUsage[i] != iLastShader){
      iLastShader = ModelShaderUsage[i];
      Shader[iLastShader]->GetShader()->Use();
    }
    if(!Models[i]->InitDraw(TH_NORMALS | TH_TEXCOORDS,Shader[iLastShader]->GetShader()))
      return;
    //Andere Textur?
    if(ModelTextureUsage[i] != iLastTexture){
      iLastTexture = ModelTextureUsage[i];
      Textures[iLastTexture]->Bind(0);
    }
    //Zeichnen
    for(int j=0;j<iDivisions * iDivisions;j++){
      if(FrustrumVisible[j]){
        iNum = i * iDivisions * iDivisions + j;
        fDistance = Plants[iNum]->Distance();
        if(fDistance < fGrasDistance - 10.0f){
          glColor4f(1.0f,1.0f,1.0f,1.0f);
          Plants[iNum]->Draw(0);
        }
      }
    }

    Models[i]->DeInitDraw();
  }

  UnloadShader();
}

//Tree Manager
TreeManager::~TreeManager(){
  int i=0;
  for(i=0;i<Models.size();i++){
    if(Models[i].Frustrums != NULL)
      delete []Models[i].Frustrums;
    if(Models[i].ModelMatrices != NULL)
      delete []Models[i].ModelMatrices;
    if(Models[i].bVisible != NULL)
      delete []Models[i].bVisible;
  }
}

int TreeManager::AddShader(Shader *ptr){
  Shaders.push_back(ptr);
  return Shaders.size()-1;
}

int TreeManager::AddTexture(Texture2D *ptr){
  Textures.push_back(ptr);
  return Textures.size()-1;
}

int TreeManager::AddModel(TreeInfo *ptr){
  ptr->ModelMatrices = NULL;
  ptr->Frustrums = NULL;
  ptr->Count = 0;
  ptr->bVisible = NULL;
  if(ptr->ChildOf < 0){
    ptr->vbuffer->GetMinMax(ptr->min,ptr->max);
    Models.push_back(*ptr);
    return Models.size()-1;
  }
  else{
    if(ptr->ChildOf >= Models.size())
      return -1;
    vec4 min,max;
    ptr->vbuffer->GetMinMax(min,max);
    max.z += 1.0f;
    ptr->max = max;
    ptr->min = min;
    Models[ptr->ChildOf].max.x = (max.x > Models[ptr->ChildOf].max.x) ? max.x : Models[ptr->ChildOf].max.x;
    Models[ptr->ChildOf].max.y = (max.y > Models[ptr->ChildOf].max.y) ? max.y : Models[ptr->ChildOf].max.y;
    Models[ptr->ChildOf].max.z = (max.z > Models[ptr->ChildOf].max.z) ? max.z : Models[ptr->ChildOf].max.z;
    Models[ptr->ChildOf].min.x = (min.x < Models[ptr->ChildOf].min.x) ? min.x : Models[ptr->ChildOf].min.x;
    Models[ptr->ChildOf].min.y = (min.y < Models[ptr->ChildOf].min.y) ? min.y : Models[ptr->ChildOf].min.y;
    Models[ptr->ChildOf].min.z = (min.z < Models[ptr->ChildOf].min.z) ? min.z : Models[ptr->ChildOf].min.z;
    Models.push_back(*ptr);
    return Models.size()-1;
  }
  return -1;
}

void TreeManager::Generate(Texture2D *coverage,Landscape *Terrain){
  std::vector<mat4a> Matrizen;
  int TexSizeX,TexSizeY;
  float winkel;
  mat4a res,mat;
  vec4 p,color;
  if(Models.size() <= 0)
    return;
  TexSizeX = coverage->GetWidth();
  TexSizeY = coverage->GetHeight();
  mat = IdentityM4();
  for(int i=0;i<Models.size();i++){
    if(Models[i].ChildOf >= 0)
      continue;
    Matrizen.clear();
    for(int x=0;x<TexSizeX;x++){
      for(int y=0;y<TexSizeY;y++){
        color = coverage->GetPixel(x,y);
        if(color != Models[i].Color)
          continue;
        p = Terrain->GetPosition((float)x/(float)(TexSizeX-1),(float)y/(float)(TexSizeY-1));
        winkel = (float)(rand()%10000)/10000.0f;
        res = Mul(RotationMatrixXYZ(0.0f,0.0f,winkel),TranslationMatrix(p));
        Matrizen.push_back(res);
      }
    }
    Models[i].ModelMatrices = new mat4a[Matrizen.size()];
    Models[i].Frustrums = new FrustrumObjectBox[Matrizen.size()];
    Models[i].bVisible = new bool[Matrizen.size()];
    memcpy(Models[i].ModelMatrices,&Matrizen[0],sizeof(mat4a)*Matrizen.size());
    Models[i].Count = Matrizen.size();
    for(int j=0;j<Models[i].Count;j++){
      Models[i].Frustrums[j].SetSize(Models[i].min,Models[i].max);
      Models[i].Frustrums[j].Transform(Models[i].ModelMatrices[j]);
    }
  }
}

void TreeManager::Draw(bool Color){
  for(int m=0;m<Models.size();m++){
    if(Models[m].ChildOf < 0)
      for(int i=0;i<Models[m].Count;i++)
        Models[m].bVisible[i] = Models[m].Frustrums[i].IsInFrustrum();
  }
  for(int s=0;s<Shaders.size();s++){
    Shaders[s]->Use();
    for(int m=0;m<Models.size();m++){
      int m2 = (Models[m].ChildOf < 0) ? m : Models[m].ChildOf;
      if((Models[m].ColorShader != s && Color) || (Models[m].ShadowShader != s && !Color) || Models[m2].Count == 0)
        continue;
      if((Color || Models[m].bTextureInShadowNeeded) && Models[m].Texture >= 0)
        Textures[Models[m].Texture]->BindToShader(0);
      if(Color && Models[m].BumpTexture >= 0)
        Textures[Models[m].BumpTexture]->BindToShader(2);

      int options = (Color) ? Models[m].ColorOptions : Models[m].ShadowOptions;
      if(options & TH_VB_ATTRIBUTE0)
        Models[m].vbuffer->SetAttributeLocation(0,Shaders[s]->GetAttributeLocation(0));
      if(options & TH_VB_ATTRIBUTE1)
        Models[m].vbuffer->SetAttributeLocation(1,Shaders[s]->GetAttributeLocation(1));
      if(options & TH_VB_ATTRIBUTE2)
        Models[m].vbuffer->SetAttributeLocation(2,Shaders[s]->GetAttributeLocation(2));
      Models[m].vbuffer->InitInstancedDraw(options);

      for(int i=0;i<Models[m2].Count;i++){
        if(!Models[m2].bVisible[i])
          continue;
        glPushMatrix();
        glMultMatrixf(Models[m2].ModelMatrices[i].f);
        Models[m].vbuffer->DrawMultiple();
        glPopMatrix();
      }
      Models[m].vbuffer->DeInitInstancedDraw();
    }
  }
  UnloadShader();
}

void TreeManager::DrawBoxes(){
  for(int i=0;i<Models.size();i++){
    if(Models[i].ChildOf < 0){
      for(int j=0;j<Models[i].Count;j++){
        if(Models[i].bVisible[j])
          glColor3f(0.0f,1.0f,0.0f);
        else
          glColor3f(1.0f,0.0f,0.0f);
        Models[i].Frustrums[j].Draw();
      }
    }
  }
}
