/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "thEngine.h"
#include "scriptparser.h"

bool IsLetter(char c){
  if((c < 'a' || c > 'z') && (c < 'A' || c > 'Z'))
    return false;
  return true;
}

bool IsDigit(char c){
  if(c < 48 || c > 57)
    return false;
  return true;
}

ScriptVar* GetVar(const char *name,ScriptVarVector& Stack){
  for(int i=0;i<Stack.size();i++){
    if(strcmp(name,Stack[i].name)==0)
      return &Stack[i];
  }
  return NULL;
}

ScriptVar* AddVar(const char *name,ScriptVarVector& Stack){
  ScriptVar TempVar;
  for(int i=0;i<Stack.size();i++){
    if(strcmp(name,Stack[i].name)==0)
      return NULL;
  }
  TempVar.name = new char[strlen(name)+1];
  strcpy(TempVar.name,name);
  Stack.push_back(TempVar);
  return &Stack[Stack.size()-1];
}

#ifdef TH_SCRIPTPARSER_DEBUG
void PrintChain(ScriptVarChain *current){
  while(current != NULL){
    switch(current->Var.type){
      case TH_INT:
        printf("Int %i\n",current->Var.iData);
        break;
      case TH_FLOAT:
        printf("Float %f\n",current->Var.fData);
        break;
      case TH_OPERATOR:
        switch(current->Var.iData){
          case TH_OP_ADD:
            printf("Op +\n");
            break;
          case TH_OP_SUB:
            printf("Op -\n");
            break;
          case TH_OP_MUL:
            printf("Op *\n");
            break;
          case TH_OP_DIV:
            printf("Op /\n");
            break;
          case TH_OP_E:
            printf("Op ==\n");
            break;
          case TH_OP_NE:
            printf("Op !=\n");
            break;
          case TH_OP_B:
            printf("Op >\n");
            break;
          case TH_OP_BE:
            printf("Op >=\n");
            break;
          case TH_OP_S:
            printf("Op <\n");
            break;
          case TH_OP_SE:
            printf("Op <=\n");
            break;
          case TH_OP_NOT:
            printf("Op !\n");
            break;
          case TH_OP_AND:
            printf("Op &&\n");
            break;
          case TH_OP_OR:
            printf("Op ||\n");
            break;
          default:
            printf("OP other\n");
            break;
        }
        break;
      default:
        printf("Unkown type\n");
        break;
    }
    current = current->next;
  }
}
#endif

bool ParseExpression(const char *expr, ScriptVarVector *OptionStack, ScriptVarVector *VarStack, ScriptVar& result){
  char meldung[200],type;
  const char *ptr = expr;
  int len,i,IntValue;
  float FloatValue;
  char *TempString;
  ScriptVar TempVar, *lpTempVar;
  ScriptVarChain *first=NULL,*current,*del;
  while(*ptr != 0){
    //Leerzeichen berspringen
    while(*ptr == ' ')
      ptr++;
    if(*ptr == 0)
      break;

    //Variable
    if(IsLetter(*ptr) || *ptr=='_'){
      //Lnge des Variablennamens

      for(len=0;IsLetter(ptr[len]) || IsDigit(ptr[len]) || ptr[len]=='_';len++);

      //Variablennamen kopieren
      TempString = new char[len+1];
      for(i=0;i<len;i++)
        TempString[i]=ptr[i];
      TempString[len]=0;

      //Variable suchen
      lpTempVar = NULL;
      if(VarStack != NULL)
        lpTempVar = GetVar(TempString,*VarStack);
      if(OptionStack != NULL && lpTempVar == NULL)
        lpTempVar = GetVar(TempString,*OptionStack);
      if(lpTempVar == NULL){
        sprintf(meldung,"Konnte Variable '%s' nicht finden",TempString);
        delete TempString;
        ThErrorMessage("Script Error",meldung);
        return false;
      }
      delete TempString;

      //Variable hinzufgen
      if(first==NULL){
        first = new ScriptVarChain;
        first->next = NULL;
        first->prev = NULL;
        first->Var = *lpTempVar;
        current = first;
      }
      else{
        current->next = new ScriptVarChain;
        current->next->next=NULL;
        current->next->prev=current;
        current->next->Var = *lpTempVar;
        current = current->next;
      }
      //Eingelesenes berspringen
      ptr += len;
    }
    //Zahl
    else if(IsDigit(*ptr) || *ptr=='.'){
      //Laenge der Zahl
      for(len=0;IsDigit(ptr[len]) || ptr[len]=='.';len++);

      //Zahl kopieren
      TempString = new char[len+1];
      for(i=0;i<len;i++)
        TempString[i]=ptr[i];
      TempString[len]=0;

      //Int oder Float?
      IntValue = atoi(TempString);
      FloatValue = atof(TempString);
      delete []TempString;

      if((float)IntValue == FloatValue){
        TempVar.type = TH_INT;
        TempVar.iData = IntValue;
      }
      else{
        TempVar.type = TH_FLOAT;
        TempVar.fData = FloatValue;
      }

      //Variable hinzufgen
      if(first==NULL){
        first = new ScriptVarChain;
        first->next = NULL;
        first->prev = NULL;
        first->Var = TempVar;
        current = first;
      }
      else{
        current->next = new ScriptVarChain;
        current->next->next=NULL;
        current->next->prev=current;
        current->next->Var = TempVar;
        current = current->next;
      }

      //Eingelesenes berspringen
      ptr += len;
    }
    //Klammer ( )
    else if(*ptr == '('){
      //Lnge der Klammer
      for(len=1,i=1;i>0;len++){
        if(ptr[len]=='(')
          i++;
        else if(ptr[len]==')')
          i--;
        else if(ptr[len]==0)
          break;
      }
      if(i != 0){
        ThErrorMessage("Script Error","Nicht geschlossene Klammer");
        return false;
      }
      #ifdef TH_SCRIPTPARSER_DEBUG
      printf("Laenge der Klammer %i\n",len);
      system("pause");
      #endif

      //Klammer kopieren
      TempString = new char[len-1];
      for(i=1;i<len-1;i++)
        TempString[i-1]=ptr[i];
      TempString[len-2]=0;
      #ifdef TH_SCRIPTPARSER_DEBUG
      printf("Klammerinhalt '%s'\n",TempString);
      system("pause");
      #endif

      //Rekursion
      if(!ParseExpression(TempString,OptionStack,VarStack,TempVar))
        return false;

      //Hinzufgen
      delete []TempString;
      if(first==NULL){
        first = new ScriptVarChain;
        first->next = NULL;
        first->prev = NULL;
        first->Var = TempVar;
        current = first;
      }
      else{
        current->next = new ScriptVarChain;
        current->next->next=NULL;
        current->next->prev=current;
        current->next->Var = TempVar;
        current = current->next;
      }

      //Eingelesenes berspringen
      ptr += len;
    }
    //Operatoren
    else {
      TempVar.type = TH_OPERATOR;
      if(*ptr=='+')
        TempVar.iData = TH_OP_ADD;
      else if(*ptr=='-')
        TempVar.iData = TH_OP_SUB;
      else if(*ptr=='*')
        TempVar.iData = TH_OP_MUL;
      else if(*ptr=='/')
        TempVar.iData = TH_OP_DIV;
      else if(ptr[0]=='&' && ptr[1]=='&'){
        TempVar.iData = TH_OP_AND;
        ptr++;
      }
      else if(ptr[0]=='|' && ptr[1]=='|'){
        TempVar.iData = TH_OP_OR;
        ptr++;
      }
      else if(ptr[0]=='>' && ptr[1]=='='){
        TempVar.iData = TH_OP_BE;
        ptr++;
      }
      else if(*ptr == '>')
        TempVar.iData = TH_OP_B;
      else if(ptr[0]=='<' && ptr[1]=='='){
        TempVar.iData = TH_OP_SE;
        ptr++;
      }
      else if(*ptr == '<')
        TempVar.iData = TH_OP_S;
      else if(ptr[0]=='=' && ptr[1]=='='){
        TempVar.iData = TH_OP_E;
        ptr++;
      }
      else if(ptr[0]=='!' && ptr[1]=='='){
        TempVar.iData = TH_OP_NE;
        ptr++;
      }
      else if(*ptr == '!')
        TempVar.iData = TH_OP_NOT;
      else{
        sprintf(meldung,"Unbekannter Operator '%c'-%i",*ptr,*ptr);
        ThErrorMessage("Script Error",meldung);
        return false;
      }
      ptr++;

      //Variable Hinzufgen
      if(first==NULL){
        first = new ScriptVarChain;
        first->next = NULL;
        first->prev = NULL;
        first->Var = TempVar;
        current = first;
      }
      else{
        current->next = new ScriptVarChain;
        current->next->next=NULL;
        current->next->prev=current;
        current->next->Var = TempVar;
        current = current->next;
      }
    }
  }

  //Operatoren umsetzten
  //Nur eine Variable
  if((first->Var.type == TH_INT || first->Var.type == TH_FLOAT) && first->next == NULL){
    result = first->Var;
    return true;
  }
  //Minuszeichen am beginn
  else if(first->Var.type == TH_OPERATOR && first->Var.iData == TH_OP_SUB){
    if(first->next == NULL){
      ThErrorMessage("Script Error","Nur Operator - vorhanden");
      return false;
    }
    else if(first->next->Var.type == TH_OPERATOR){
      ThErrorMessage("Script Error","2 Aufeinanderfolgende Operatoren");
      return false;
    }
    else{
      if(first->next->Var.type == TH_INT)
        first->next->Var.iData *= -1;
      else
        first->next->Var.fData *= -1.0f;

      first = first->next;
      delete first->prev;
      first->prev == NULL;
    }
  }

  #ifdef TH_SCRIPTPARSER_DEBUG
  PrintChain(first);
  system("pause");
  #endif

  // ! Operator
  current = first;
  while(current != NULL){
    if(current->Var.type == TH_OPERATOR && current->Var.iData == TH_OP_NOT){
      if(current->next == NULL){
        ThErrorMessage("Script Error","! Operator am Ende einer Expression");
        return false;
      }
      if(current->next->Var.type == TH_OPERATOR){
        ThErrorMessage("Script Error","Anderer Operator folgt auf !-Operator");
        return false;
      }
      if(current->next->Var.type == TH_INT)
        IntValue = (current->next->Var.iData > 0) ? 0 : 1;
      else
        IntValue = (current->next->Var.fData > 0) ? 0 : 1;

      del = current->next;
      current->next = current->next->next;
      if(current->next != NULL)
        current->next->prev = current;
      delete del;

      current->Var.type = TH_INT;
      current->Var.iData = IntValue;
    }
    current = current->next;
  }

  #ifdef TH_SCRIPTPARSER_DEBUG
  PrintChain(first);
  system("pause");
  #endif

  // * / Operatoren
  current = first;
  while(current != NULL){
    if(current->Var.type == TH_OPERATOR && (current->Var.iData == TH_OP_MUL || current->Var.iData == TH_OP_DIV)){
      if(current->prev == NULL || current->next == NULL){
        ThErrorMessage("Script Error","* / Operator am Begin oder Ende einer Expression");
        return false;
      }
      if(current->prev->Var.type == TH_OPERATOR || current->next->Var.type == TH_OPERATOR){
        ThErrorMessage("Script Error","* / folgt auf anderen Operator");
        return false;
      }
      if(current->prev->Var.type == TH_FLOAT && current->next->Var.type == TH_FLOAT){
        if(current->Var.iData == TH_OP_MUL)
          FloatValue = current->prev->Var.fData * current->next->Var.fData;
        else
          FloatValue = current->prev->Var.fData / current->next->Var.fData;
        type = TH_FLOAT;
      }
      else if(current->prev->Var.type == TH_FLOAT){
        if(current->Var.iData == TH_OP_MUL)
          FloatValue = current->prev->Var.fData * (float)current->next->Var.iData;
        else
          FloatValue = current->prev->Var.fData / (float)current->next->Var.iData;
        type = TH_FLOAT;
      }
      else if(current->next->Var.type == TH_FLOAT){
        if(current->Var.iData == TH_OP_MUL)
          FloatValue = (float)current->prev->Var.iData * current->next->Var.fData;
        else
          FloatValue = (float)current->prev->Var.iData / current->next->Var.fData;
        type = TH_FLOAT;
      }
      else {
        if(current->Var.iData == TH_OP_MUL)
          IntValue = current->prev->Var.iData * current->next->Var.iData;
        else
          IntValue = current->prev->Var.iData / current->next->Var.iData;
        type = TH_INT;
      }

      //2 Elemente aus Kette lschen
      current->prev->next = current->next->next;
      if(current->next->next != NULL)
        current->next->next->prev = current->prev;
      del = current;
      current = current->prev;
      delete del->next;
      delete del;

      //Ergebnis setzten
      current->Var.type = type;
      if(type == TH_FLOAT)
        current->Var.fData = FloatValue;
      else
        current->Var.iData = IntValue;
      current = current->next;
    }
    else
      current=current->next;
  }

  #ifdef TH_SCRIPTPARSER_DEBUG
  PrintChain(first);
  system("pause");
  #endif

  // + - Operatoren
  current = first;
  while(current != NULL){
    if(current->Var.type == TH_OPERATOR && (current->Var.iData == TH_OP_ADD || current->Var.iData == TH_OP_SUB)){
      if(current->prev == NULL || current->next == NULL){
        ThErrorMessage("Script Error","+ - Operator am Begin oder Ende einer Expression");
        return false;
      }
      if(current->prev->Var.type == TH_OPERATOR || current->next->Var.type == TH_OPERATOR){
        ThErrorMessage("Script Error","+ - folgt auf anderen Operator");
        return false;
      }
      if(current->prev->Var.type == TH_FLOAT && current->next->Var.type == TH_FLOAT){
        if(current->Var.iData == TH_OP_ADD)
          FloatValue = current->prev->Var.fData + current->next->Var.fData;
        else
          FloatValue = current->prev->Var.fData - current->next->Var.fData;
        type = TH_FLOAT;
      }
      else if(current->prev->Var.type == TH_FLOAT){
        if(current->Var.iData == TH_OP_ADD)
          FloatValue = current->prev->Var.fData + (float)current->next->Var.iData;
        else
          FloatValue = current->prev->Var.fData - (float)current->next->Var.iData;
        type = TH_FLOAT;
      }
      else if(current->next->Var.type == TH_FLOAT){
        if(current->Var.iData == TH_OP_ADD)
          FloatValue = (float)current->prev->Var.iData + current->next->Var.fData;
        else
          FloatValue = (float)current->prev->Var.iData - current->next->Var.fData;
        type = TH_FLOAT;
      }
      else {
        if(current->Var.iData == TH_OP_ADD)
          IntValue = current->prev->Var.iData + current->next->Var.iData;
        else
          IntValue = current->prev->Var.iData - current->next->Var.iData;
        type = TH_INT;
      }

      //2 Elemente aus Kette lschen
      current->prev->next = current->next->next;
      if(current->next->next != NULL)
        current->next->next->prev = current->prev;
      del = current;
      current = current->prev;
      delete del->next;
      delete del;

      //Ergebnis setzten
      current->Var.type = type;
      if(type == TH_FLOAT)
        current->Var.fData = FloatValue;
      else
        current->Var.iData = IntValue;
      current = current->next;
    }
    else
      current=current->next;
  }

  #ifdef TH_SCRIPTPARSER_DEBUG
  PrintChain(first);
  system("pause");
  #endif

  // Vergleichsoperatoren != == < > <= >=
  current = first;
  while(current != NULL){
    if(current->Var.type == TH_OPERATOR && (current->Var.iData == TH_OP_E || current->Var.iData == TH_OP_NE || current->Var.iData == TH_OP_S || current->Var.iData == TH_OP_SE || current->Var.iData == TH_OP_B || current->Var.iData == TH_OP_BE)){
      if(current->prev == NULL || current->next == NULL){
        ThErrorMessage("Script Error","Vergleichsoperator am Begin oder Ende einer Expression");
        return false;
      }
      if(current->prev->Var.type == TH_OPERATOR || current->next->Var.type == TH_OPERATOR){
        ThErrorMessage("Script Error","Vergleichsoperator folgt auf anderen Operator");
        return false;
      }
      if(current->prev->Var.type == TH_FLOAT && current->next->Var.type == TH_FLOAT){
        switch(current->Var.iData){
          case TH_OP_NE:
            IntValue = (current->prev->Var.fData != current->next->Var.fData) ? 1 : 0;
            break;
          case TH_OP_E:
            IntValue = (current->prev->Var.fData == current->next->Var.fData) ? 1 : 0;
            break;
          case TH_OP_B:
            IntValue = (current->prev->Var.fData > current->next->Var.fData) ? 1 : 0;
            break;
          case TH_OP_BE:
            IntValue = (current->prev->Var.fData >= current->next->Var.fData) ? 1 : 0;
            break;
          case TH_OP_S:
            IntValue = (current->prev->Var.fData < current->next->Var.fData) ? 1 : 0;
            break;
          case TH_OP_SE:
            IntValue = (current->prev->Var.fData <= current->next->Var.fData) ? 1 : 0;
            break;
        }
      }
      else if(current->prev->Var.type == TH_FLOAT){
        switch(current->Var.iData){
          case TH_OP_NE:
            IntValue = (current->prev->Var.fData != (float)current->next->Var.iData) ? 1 : 0;
            break;
          case TH_OP_E:
            IntValue = (current->prev->Var.fData == (float)current->next->Var.iData) ? 1 : 0;
            break;
          case TH_OP_B:
            IntValue = (current->prev->Var.fData > (float)current->next->Var.iData) ? 1 : 0;
            break;
          case TH_OP_BE:
            IntValue = (current->prev->Var.fData >= (float)current->next->Var.iData) ? 1 : 0;
            break;
          case TH_OP_S:
            IntValue = (current->prev->Var.fData < (float)current->next->Var.iData) ? 1 : 0;
            break;
          case TH_OP_SE:
            IntValue = (current->prev->Var.fData <= (float)current->next->Var.iData) ? 1 : 0;
            break;
        }
      }
      else if(current->next->Var.type == TH_FLOAT){
        switch(current->Var.iData){
          case TH_OP_NE:
            IntValue = ((float)current->prev->Var.iData != current->next->Var.fData) ? 1 : 0;
            break;
          case TH_OP_E:
            IntValue = ((float)current->prev->Var.iData == current->next->Var.fData) ? 1 : 0;
            break;
          case TH_OP_B:
            IntValue = ((float)current->prev->Var.iData > current->next->Var.fData) ? 1 : 0;
            break;
          case TH_OP_BE:
            IntValue = ((float)current->prev->Var.iData >= current->next->Var.fData) ? 1 : 0;
            break;
          case TH_OP_S:
            IntValue = ((float)current->prev->Var.iData < current->next->Var.fData) ? 1 : 0;
            break;
          case TH_OP_SE:
            IntValue = ((float)current->prev->Var.iData <= current->next->Var.fData) ? 1 : 0;
            break;
        }
      }
      else {
        switch(current->Var.iData){
          case TH_OP_NE:
            IntValue = (current->prev->Var.iData != current->next->Var.iData) ? 1 : 0;
            break;
          case TH_OP_E:
            IntValue = (current->prev->Var.iData == current->next->Var.iData) ? 1 : 0;
            break;
          case TH_OP_B:
            IntValue = (current->prev->Var.iData > current->next->Var.iData) ? 1 : 0;
            break;
          case TH_OP_BE:
            IntValue = (current->prev->Var.iData >= current->next->Var.iData) ? 1 : 0;
            break;
          case TH_OP_S:
            IntValue = (current->prev->Var.iData < current->next->Var.iData) ? 1 : 0;
            break;
          case TH_OP_SE:
            IntValue = (current->prev->Var.iData <= current->next->Var.iData) ? 1 : 0;
            break;
        }
      }

      //2 Elemente aus Kette lschen
      current->prev->next = current->next->next;
      if(current->next->next != NULL)
        current->next->next->prev = current->prev;
      del = current;
      current = current->prev;
      delete del->next;
      delete del;

      //Ergebnis setzten
      current->Var.type = TH_INT;
      current->Var.iData = IntValue;
      current = current->next;
    }
    else
      current=current->next;
  }

  #ifdef TH_SCRIPTPARSER_DEBUG
  PrintChain(first);
  system("pause");
  #endif

  //Logische Operatoren
  current = first;
  while(current != NULL){
    if(current->Var.type == TH_OPERATOR && (current->Var.iData == TH_OP_AND || current->Var.iData == TH_OP_OR)){
      if(current->prev == NULL || current->next == NULL){
        ThErrorMessage("Script Error","&& || Operator am Begin oder Ende einer Expression");
        return false;
      }
      if(current->prev->Var.type == TH_OPERATOR || current->next->Var.type == TH_OPERATOR){
        ThErrorMessage("Script Error","&& || folgt auf anderen Operator");
        return false;
      }
      if(current->prev->Var.type == TH_FLOAT && current->next->Var.type == TH_FLOAT){
        if(current->Var.iData == TH_OP_AND)
          IntValue = (current->prev->Var.fData > 0 && current->next->Var.fData > 0) ? 1 : 0;
        else
          IntValue = (current->prev->Var.fData > 0 || current->next->Var.fData > 0) ? 1 : 0;
      }
      else if(current->prev->Var.type == TH_FLOAT){
        if(current->Var.iData == TH_OP_AND)
          IntValue = (current->prev->Var.fData > 0 && current->next->Var.iData > 0) ? 1 : 0;
        else
          IntValue = (current->prev->Var.fData > 0 || current->next->Var.iData > 0) ? 1 : 0;
      }
      else if(current->next->Var.type == TH_FLOAT){
        if(current->Var.iData == TH_OP_AND)
          IntValue = (current->prev->Var.iData > 0 && current->next->Var.fData > 0) ? 1 : 0;
        else
          IntValue = (current->prev->Var.iData > 0 || current->next->Var.fData > 0) ? 1 : 0;
      }
      else {
        if(current->Var.iData == TH_OP_AND)
          IntValue = (current->prev->Var.iData > 0 && current->next->Var.iData > 0) ? 1 : 0;
        else
          IntValue = (current->prev->Var.iData > 0 || current->next->Var.iData > 0) ? 1 : 0;
      }

      //2 Elemente aus Kette lschen
      current->prev->next = current->next->next;
      if(current->next->next != NULL)
        current->next->next->prev = current->prev;
      del = current;
      current = current->prev;
      delete del->next;
      delete del;

      //Ergebnis setzten
      current->Var.type = TH_INT;
      current->Var.iData = IntValue;
      current = current->next;
    }
    else
      current=current->next;
  }

  #ifdef TH_SCRIPTPARSER_DEBUG
  PrintChain(first);
  system("pause");
  #endif

  result = first->Var;
  return true;
}

bool ParseAssigment(const char *expr, ScriptVarVector *OptionStack, ScriptVarVector *VarStack){
  int i,len,offset=0,gleich_pos=-1;
  char *VarName,*Expression,meldung[200];
  ScriptVar *ptr,TempVar;
  bool new_var=false;

  //Neue Variable erstellen?
  if(strncmp("var ",expr,4)==0){
    for(i=0;expr[i]!=0;i++)
      if(expr[i]=='=')
        break;

    if(expr[i]!=0){
      gleich_pos = i;
      offset=4;
      new_var=true;
      if(VarStack == NULL){
        ThErrorMessage("Script Error","Auftrag fr neue Variable ohne angabe eines Variablen-Stacks!");
        return false;
      }
    }
  }

  len=i-offset;
  VarName = new char[len+1];

  for(i=0;i<len;i++)
    VarName[i]=expr[i+offset];
  VarName[len]=0;

  //Wieviele Leerzeichen am Anfang
  for(i=0;i<len;i++){
    if(VarName[i]!=' ')
       break;
  }
  offset=i;

  //Leerzeichen entfernen
  for(i=0;i+offset<=len;i++)
    VarName[i]=VarName[i+offset];

  //Leerzeichen am Ende entfernen
  for(i=len-offset-1;i>=0;i--){
    if(VarName[i] != ' ')
      break;
    else
      VarName[i]=0;
  }

  if(!IsLetter(VarName[0]) && VarName[0] != '_'){
    ThErrorMessage("Script Error","Variablennamen mssen mit Buchstaben oder einem '_' beginnen");
    delete []VarName;
    return false;
  }

  for(i=0;i<strlen(VarName);i++)
    if(!IsLetter(VarName[i]) && !IsDigit(VarName[i]) && VarName[i] != '_'){
      ThErrorMessage("Script Error","Variablennamen drfen nur Buchstaben Zahlen und '_' enthalten");
      delete []VarName;
      return false;
    }

  if(new_var){
    ptr=NULL;
    if(OptionStack != NULL)
      ptr = GetVar(VarName,*OptionStack);
    if(ptr == NULL)
      ptr = GetVar(VarName,*VarStack);
    if(ptr != NULL){
      sprintf(meldung,"Variable mit dem namen '%s' ist bereits definiert!",VarName);
      ThErrorMessage("Script Error",meldung);
      delete []VarName;
      return false;
    }

    ptr=AddVar(VarName,*VarStack);
    if(ptr == NULL){
      sprintf(meldung,"Fehler beim erstellen der Variable '%s'",VarName);
      ThErrorMessage("Script Error",meldung);
      delete []VarName;
      return false;
    }

    ptr->type = TH_INT;
    ptr->iData = 0;
  }
  else {
    ptr = NULL;
    if(VarStack != NULL)
      ptr = GetVar(VarName,*VarStack);
    if(ptr == NULL && OptionStack != NULL)
      ptr = GetVar(VarName,*OptionStack);

    if(ptr == NULL){
      sprintf(meldung,"Variable mit dem Namen '%s' nicht gefunden",VarName);
      ThErrorMessage("Script Error",meldung);
      delete []VarName;
      return false;
    }
  }

  //Gleichheitszeichen gefunden?
  if(gleich_pos != -1){
    len = strlen(expr)-gleich_pos-1;
    Expression = new char[len+1];
    for(i=0;i<len;i++)
      Expression[i]=expr[gleich_pos+1+i];
    Expression[len]=0;

    if(!ParseExpression(Expression,OptionStack,VarStack,TempVar)){
      delete []Expression;
      return false;
    }

    delete []Expression;
    ptr->type = TempVar.type;
    ptr->iData = TempVar.iData;
  }

  delete []VarName;
  return true;
}
