/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "thEngine.h"
#include "texture.h"
#include "vektoren.h"
#include <stdio.h>
#include <math.h>

const GLenum TextureChannels[]={GL_TEXTURE0,GL_TEXTURE1,GL_TEXTURE2,GL_TEXTURE3,GL_TEXTURE4,GL_TEXTURE5,GL_TEXTURE6,GL_TEXTURE7};

Texture2D::Texture2D(){
  lpData = NULL;
  uiBitsPerPixel=0;
  uiWidth=0;
  uiHeight=0;
  gluiTextureId=0;
}

Texture2D::~Texture2D(){
  if(lpData != NULL)
    delete []lpData;
  if(gluiTextureId != 0)
    glDeleteTextures(1,&gluiTextureId);
}

bool Texture2D::LoadTGA(const char *filename,char options){
  GLubyte     TGAheader[12]={0,0,2,0,0,0,0,0,0,0,0,0};	// Uncompressed TGA Header
  GLubyte     TGAheadercompressed[12]={0,0,10,0,0,0,0,0,0,0,0,0}; //Compressed TGA Header
  GLubyte     TGAcompare[12];								// Used To Compare TGA Header
  GLubyte     header[6];									// First 6 Useful Bytes From The Header
  GLuint      bytesPerPixel;								// Holds Number Of Bytes Per Pixel Used In The TGA File
  GLuint      imageSize;									// Used To Store The Image Size When Setting Aside Ram
  GLuint      temp;										// Temporary Variable
  GLuint      type=GL_RGBA;								// Set The Default GL Mode To RBGA (32 BPP)
  GLuint      currentbyte=0;
  bool        compressed=false;

  FILE *file = fopen(filename, "rb");						// Open The TGA File

  if(file==NULL ||										// Does File Even Exist?
     fread(TGAcompare,1,sizeof(TGAcompare),file)!=sizeof(TGAcompare) ||	// Are There 12 Bytes To Read?
     memcmp(TGAheader,TGAcompare,sizeof(TGAheader))!=0               &&
     memcmp(TGAheadercompressed,TGAcompare,sizeof(TGAheader))!=0)	// Does The Header Match What We Want?
  {
    if (file == NULL){									// Did The File Even Exist? *Added Jim Strong*
      return false;
    }									// Return False
    else
    {
      fclose(file);									// If Anything Failed, Close The File
      return false;									// Return False
    }
  }
  if(TGAcompare[2]==10)
    compressed=true;

  if(fread(header,1,sizeof(header),file)!=sizeof(header)){
    fclose(file);
    return false;
  }

  uiWidth  = header[1] * 256 + header[0];			// Determine The TGA Width	(highbyte*256+lowbyte)
  uiHeight = header[3] * 256 + header[2];			// Determine The TGA Height	(highbyte*256+lowbyte)

  if(uiWidth==0 || uiHeight==0 || (header[4]!=24 && header[4]!=32))					// Is The TGA 24 or 32 Bit?
  {
    fclose(file);										// If Anything Failed, Close The File
    return false;										// Return False
  }

  uiBitsPerPixel = header[4];							// Grab The TGA's Bits Per Pixel (24 or 32)
  bytesPerPixel	= uiBitsPerPixel/8;						// Divide By 8 To Get The Bytes Per Pixel
  imageSize = uiWidth*uiHeight*bytesPerPixel;	// Calculate The Memory Required For The TGA Data

  lpData=new GLubyte[imageSize];		// Reserve Memory To Hold The TGA Data

  if(!compressed){
    if(fread(lpData, 1, imageSize, file)!=imageSize)	// Does The Image Size Match The Memory Reserved?
    {
      fclose(file);										// Close The File
      return false;										// Return False
    }

    for(GLuint i=0; i<int(imageSize); i+=bytesPerPixel)		// Loop Through The Image Data
    {														// Swaps The 1st And 3rd Bytes ('R'ed and 'B'lue)
      temp=lpData[i];							// Temporarily Store The Value At Image Data 'i'
      lpData[i] = lpData[i + 2];	// Set The 1st Byte To The Value Of The 3rd Byte
      lpData[i + 2] = temp;					// Set The 3rd Byte To The Value In 'temp' (1st Byte Value)
    }
  }
  else { //Loading compressed TGA file
    GLubyte * colorbuffer = new GLubyte[bytesPerPixel];
    GLubyte mybyte;
    while(currentbyte < imageSize){
      if(fread(&mybyte,1,1,file)!=1)
        return false;

      if(mybyte < 128){ // Non RLE Data
        mybyte++;
        for(short counter=0;counter < mybyte; counter++){
          if(fread(colorbuffer,1,bytesPerPixel,file)!=bytesPerPixel)
            return false;
          lpData[currentbyte]=colorbuffer[2]; // Rot
          lpData[currentbyte+1]=colorbuffer[1]; // Grün
          lpData[currentbyte+2]=colorbuffer[0]; // Blau
          if(bytesPerPixel == 4){
           lpData[currentbyte+3]=colorbuffer[3]; // Alpha
          }
          currentbyte += bytesPerPixel;
        }
      }
      else { //RLE Data
        mybyte -=127;
        if(fread(colorbuffer,1,bytesPerPixel,file)!=bytesPerPixel)
          return false; //Versuche Pixel zu lesen
        for(short counter=0;counter < mybyte; counter++){//Selbe Farbe `mybyte` mal in den Speicher schreiben
          lpData[currentbyte]=colorbuffer[2]; //Rot
          lpData[currentbyte+1]=colorbuffer[1]; //Grün
          lpData[currentbyte+2]=colorbuffer[0]; //Blau
          if(bytesPerPixel == 4)
            lpData[currentbyte+3]=colorbuffer[3]; // Alpha
          currentbyte += bytesPerPixel;
        }
      }
    }
  }

  fclose (file);											// Close The File

  // Build A Texture From The Data
  glGenTextures(1, &gluiTextureId);					// Generate OpenGL texture IDs

  glBindTexture(GL_TEXTURE_2D, gluiTextureId);			// Bind Our Texture

  if(options & TH_MIPMAPS){
    glTexParameterf(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
    if(options & TH_LINEAR){
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
    else{
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);	// Linear Filtered
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);	// Linear Filtered
    }
  }
  else{
    if(options & TH_LINEAR){
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
    else{
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);	// Linear Filtered
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);	// Linear Filtered
    }
  }

  if(options & TH_CLAMP_S)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  else
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);

  if(options & TH_CLAMP_T)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  else
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

  if(uiBitsPerPixel==24)									// Was The TGA 24 Bits
    type=GL_RGB;										// If So Set The 'type' To GL_RGB

  glTexImage2D(GL_TEXTURE_2D, 0, type, uiWidth, uiHeight, 0, type, GL_UNSIGNED_BYTE, lpData);
  return true;
}

bool Texture2D::CreateEmpty(int width, int height, GLenum DataFormat, GLenum InternalFormat, char options){
  switch(DataFormat){
    case GL_LUMINANCE:
      uiBitsPerPixel = 8;
      break;
    case GL_LUMINANCE_ALPHA:
      uiBitsPerPixel = 16;
      break;
    case GL_RGB:
      uiBitsPerPixel = 24;
      break;
    case GL_RGBA:
      uiBitsPerPixel = 32;
      break;
    default:
      ThErrorMessage("Texture2D Error","Invalid DataFormat");
      return false;
      break;
  }

  gleDataFormat = DataFormat;
  gleInternalFormat = InternalFormat;

  lpData = new GLubyte[uiBitsPerPixel / 8 * width * height];
  memset(lpData,0,uiBitsPerPixel / 8 * width * height);

  uiWidth = width;
  uiHeight = height;

  // Build A Texture From The Data
  glGenTextures(1, &gluiTextureId);					// Generate OpenGL texture IDs

  glBindTexture(GL_TEXTURE_2D, gluiTextureId);			// Bind Our Texture

  if(options & TH_MIPMAPS){
    glTexParameterf(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
    if(options & TH_LINEAR){
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
    else{
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);	// Linear Filtered
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);	// Linear Filtered
    }
  }
  else{
    if(options & TH_LINEAR){
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
    else{
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);	// Linear Filtered
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);	// Linear Filtered
    }
  }

  if(options & TH_CLAMP_S)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  else
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);

  if(options & TH_CLAMP_T)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  else
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

  glTexImage2D(GL_TEXTURE_2D, 0, InternalFormat, uiWidth, uiHeight, 0, DataFormat, GL_UNSIGNED_BYTE, lpData);
  return true;
}

bool Texture2D::Insert(int iXto, int iYto, int iSrcWidth, int iSrcHeight, unsigned char *buffer){
  int size = uiBitsPerPixel / 8;
  if(iXto + iSrcWidth > uiWidth || iYto + iSrcHeight > uiHeight || lpData == NULL){
    printf("Insert Failed: %i %i %i %i\n",iXto,iYto,iSrcWidth,iSrcHeight);
    return false;
  }

  for(int y=0;y<iSrcHeight;y++){
    for (int x=0;x<iSrcWidth;x++){
      for(int i=0;i<size;i++){
        lpData[((y+iYto)*uiWidth + x + iXto)*size+i] = buffer[(y*iSrcWidth+x)*size+i];
      }
    }
  }
  return true;
}

bool Texture2D::Reupload(){
  if(lpData == NULL)
    return false;
  glTexSubImage2D(GL_TEXTURE_2D,0,0,0,uiWidth,uiHeight,gleDataFormat,GL_UNSIGNED_BYTE,lpData);
  return true;
}

void Texture2D::Bind(short num){
  if(num < 0)
    num=0;
  if(num > 7)
    num=7;
  glActiveTexture(TextureChannels[num]);
  glBindTexture(GL_TEXTURE_2D,gluiTextureId);
}

void Texture2D::Bind(){
  glBindTexture(GL_TEXTURE_2D,gluiTextureId);
}

void Texture2D::BindToShader(int num){
  glActiveTextureARB(GL_TEXTURE0_ARB + num);
  glBindTexture(GL_TEXTURE_2D,gluiTextureId);
}

vec4 Texture2D::GetPixel(float x, float y){
  vec4 color;
  color.x=0;color.y=0;color.z=0; color.w=0.0f;
  if(lpData==NULL)
    return color;
  int x1,x2,y1,y2;
  vec4 colors[4];
  int uiBytesPerPixel = uiBitsPerPixel/8;
  if(x > 1.0f) x=1.0f;
  if(x < 0.0f) x=0.0f;
  if(y > 1.0f) y=1.0f;
  if(y < 0.0f) y=1.0f;

  x1=(int)floor(x*(float)(uiWidth-1));
  x2=(int)ceil(x*(float)(uiWidth-1));
  y1=(int)floor(y*(float)(uiHeight-1));
  y2=(int)ceil(y*(float)(uiHeight-1));

  x1 = (x1 < 0) ? 0 : x1;
  x2 = (x2 < 0) ? 0 : x2;
  y1 = (y1 < 0) ? 0 : y1;
  y2 = (y2 < 0) ? 0 : y2;

  x1 = (x1 >= uiWidth) ? (uiWidth-1) : x1;
  x2 = (x2 >= uiWidth) ? (uiWidth-1) : x2;
  y1 = (y1 >= uiHeight) ? (uiHeight-1) : y1;
  y2 = (y2 >= uiHeight) ? (uiHeight-1) : y2;

  if(x1 != x2)
    x = 1.0f - (x - ((float)x1 / (float)(uiWidth-1))) / (((float)x2 - (float)x1)/(float)(uiWidth-1));
  else
    x = 0.0f;

  if(y1 != y2)
    y = 1.0f - (y - ((float)y1 / (float)(uiHeight-1))) / (((float)y2 - (float)y1)/(float)(uiHeight-1));
  else
    y = 0.0f;

  color.w = (uiBitsPerPixel > 24) ? 0.0f : 1.0f;

  color.x += ((float)lpData[y1*uiWidth*uiBytesPerPixel+x1*uiBytesPerPixel] / 255.0f) * (x * y);
  color.y += ((float)lpData[y1*uiWidth*uiBytesPerPixel+x1*uiBytesPerPixel+1] / 255.0f) * (x * y);
  color.z += ((float)lpData[y1*uiWidth*uiBytesPerPixel+x1*uiBytesPerPixel+2] / 255.0f) * (x * y);
  if(uiBitsPerPixel > 24)
    color.w += ((float)lpData[y1*uiWidth*uiBytesPerPixel+x1*uiBytesPerPixel+3] / 255.0f) * (x * y);

  color.x += ((float)lpData[y1*uiWidth*uiBytesPerPixel+x2*uiBytesPerPixel] / 255.0f) * ((1.0f - x) * y);
  color.y += ((float)lpData[y1*uiWidth*uiBytesPerPixel+x2*uiBytesPerPixel+1] / 255.0f) * ((1.0f - x) * y);
  color.z += ((float)lpData[y1*uiWidth*uiBytesPerPixel+x2*uiBytesPerPixel+2] / 255.0f) * ((1.0f - x) * y);
  if(uiBitsPerPixel > 24)
    color.w += ((float)lpData[y1*uiWidth*uiBytesPerPixel+x2*uiBytesPerPixel+3] / 255.0f) * ((1.0f - x) * y);

  color.x += ((float)lpData[y2*uiWidth*uiBytesPerPixel+x1*uiBytesPerPixel] / 255.0f) * (x * (1.0f - y));
  color.y += ((float)lpData[y2*uiWidth*uiBytesPerPixel+x1*uiBytesPerPixel+1] / 255.0f) * (x * (1.0f - y));
  color.z += ((float)lpData[y2*uiWidth*uiBytesPerPixel+x1*uiBytesPerPixel+2] / 255.0f) * (x * (1.0f - y));
  if(uiBitsPerPixel > 24)
    color.w += ((float)lpData[y2*uiWidth*uiBytesPerPixel+x1*uiBytesPerPixel+3] / 255.0f) * (x * (1.0f - y));

  color.x += ((float)lpData[y2*uiWidth*uiBytesPerPixel+x2*uiBytesPerPixel] / 255.0f) * ((1.0f - x) * (1.0f - y));
  color.y += ((float)lpData[y2*uiWidth*uiBytesPerPixel+x2*uiBytesPerPixel+1] / 255.0f) * ((1.0f - x) * (1.0f - y));
  color.z += ((float)lpData[y2*uiWidth*uiBytesPerPixel+x2*uiBytesPerPixel+2] / 255.0f) * ((1.0f - x) * (1.0f - y));
  if(uiBitsPerPixel > 24)
    color.w += ((float)lpData[y2*uiWidth*uiBytesPerPixel+x2*uiBytesPerPixel+3] / 255.0f) * ((1.0f - x) * (1.0f - y));

  return color;
}

vec4 Texture2D::GetPixel(vec2 Coords){
  return GetPixel(Coords.x,Coords.y);
}

vec4 Texture2D::GetPixel(int x, int y){
  vec4 color;
  int uiBytesPerPixel = uiBitsPerPixel/8;
  color.set(0);
  if(x > uiWidth || y > uiHeight)
    return color;
  color.x += (float)lpData[y*uiWidth*uiBytesPerPixel+x*uiBytesPerPixel] / 255.0f;
  color.y += (float)lpData[y*uiWidth*uiBytesPerPixel+x*uiBytesPerPixel+1] / 255.0f;
  color.z += (float)lpData[y*uiWidth*uiBytesPerPixel+x*uiBytesPerPixel+2] / 255.0f;
  color.w = (uiBitsPerPixel != 32) ? 1.0f : (float)lpData[y*uiWidth*uiBytesPerPixel+x*uiBytesPerPixel+3] / 255.0f;
  return color;
}

unsigned int Texture2D::GetHeight(){
  return uiHeight;
}

unsigned int Texture2D::GetWidth(){
  return uiWidth;
}
