/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#define TH_ENGINE_MAIN_FILE
////////////Includes/////////////
#include "thEngine.h"
#include "ois.h"
/////////////////////////////////
#ifdef TH_DEBUG
#include <stdio.h>
#endif
/////////////////////////////////
#ifdef __linux
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/xf86vmode.h>
#include <X11/keysym.h>
#endif
#include <string>
#include <iostream>
#include <fstream>

#ifdef WIN32
LRESULT CALLBACK WindowProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
#endif

extern char ThWindowTitle[];

Texture2DManager *TexMngr;
ModelManager *MdlMngr;
XmlShaderManager *ShdrMngr;

int ThErrorLength=0;
char *ThErrorString;
bool bThError=false;

Zeitpunkt Runtime;
Zeitpunkt LastFrame;
int iFrames=0;
float fps;
double dTimeSinceLastFPS=0;
double dTimeDiff;
float fFrametime=0;
Timer *timer;

float fFarClipping=500.0f;

FILE *debugtxt;

#ifdef TH_DEBUG
double AverageFps=0;
double FpsAnz=0;
#endif

#ifdef __linux
Cursor cursor;		//A blank cursor
#endif

int curr_renderstate=th_renderstate_model;

//Window specific Stuff
#ifdef __linux
Display*				m_Display;
int						m_XScreenID;
Window					m_XWindow;
GLXContext				m_GLXContext;
XSetWindowAttributes	m_XSetAttr;
XF86VidModeModeInfo		m_XF86DeskMode;
#elif defined WIN32
HINSTANCE hInstance=NULL;
HWND hWnd=NULL;
HDC hDC=NULL;
HGLRC hRC=NULL;
#endif

bool arbMultisampleSupported = false;
int	arbMultisampleFormat = 0;

int iWidth;
int iHeight;

bool bIsVisible=false;
bool bIsWorking=true;
bool bIsFullscreen=false;
bool bShowFps=false;

KeyDownFunc KeysDown[256];
KeyDownFunc KeysUp[256];
bool KeyStates[256];
MouseMoveFunc OnMouseMoveAbsolute=NULL;
MouseMoveFunc OnMouseMoveRelative=NULL;
MouseMoveFuncGui OnMouseMove=NULL;
MouseButtonFunc OnMouseButtonUp=NULL;
MouseButtonFunc OnMouseButtonDown=NULL;
MouseWheelFunc OnMouseWheel=NULL;
TextFunc OnTextIn=NULL;
bool KeyESC=false;

TiXmlDocument *XmlConfigDoc;
const char * LoadBasicXml();
int XmlResX=0,XmlResY=0,XmlColorBits=0,XmlSamples=0,XmlDepthBits=0;
bool XmlFullscreen=false;
bool XmlKeepCurrent=false;
bool bAntialiasing=false;
bool bVsync=false;

void SetKeyDownFunc(unsigned int Key,KeyDownFunc Func){
  if((Key >= 0) && (Key <= 255))
    KeysDown[Key]=Func;
}

void SetKeyUpFunc(unsigned int Key, KeyDownFunc Func){
  if((Key >= 0) && (Key <= 255))
    KeysUp[Key]=Func;
}

bool KeyIsPressed(unsigned int Key){
  return KeyStates[Key];
}

void SetMouseMoveFuncAbsolute(MouseMoveFunc Func){
  OnMouseMoveAbsolute = Func;
}

void SetMouseMoveFuncRelative(MouseMoveFunc Func){
  OnMouseMoveRelative = Func;
}

void SetMouseMoveFunc(MouseMoveFuncGui Func){
  OnMouseMove = Func;
}

void SetMouseWheelFunc(MouseWheelFunc Func){
  OnMouseWheel = Func;
}

void SetMouseButtonDownFunc(MouseButtonFunc Func){
  OnMouseButtonDown = Func;
}

void SetMouseButtonUpFunc(MouseButtonFunc Func){
  OnMouseButtonUp = Func;
}

void SetTextFunc(TextFunc Func){
  OnTextIn = Func;
}

void SetTime(){
  Runtime.SetTimerReference(timer);
  Runtime = timer->GetTimeInt64();
  LastFrame = Runtime;
}

bool StdInit(){
  bool error;
  timer = new Timer(error);
  if(error){
    ThErrorMessage("Timer Error","Kein Performance-Timer verfügbar!");
    return false;
  }
  Runtime.SetTimerReference(timer);
  Runtime = timer->GetTimeInt64();
  LastFrame = Runtime;

  #ifdef WIN32
  if(GLEE_WGL_EXT_swap_control){
    if(bVsync)
      wglSwapIntervalEXT(1);
    else
      wglSwapIntervalEXT(0);
  }
  #else
  if(GLEE_GLX_SGI_swap_control){
    if(bVsync)
      glXSwapIntervalSGI(1);
    else
      glXSwapIntervalSGI(0);
  }
  XSelectInput(m_Display,m_XWindow,PointerMotionMask | KeyPressMask | KeyReleaseMask);
  //XSelectInput(m_Display,m_XWindow,KeyPressMask | KeyReleaseMask);
  #endif

  //Manager Erzeugen
  TexMngr = new Texture2DManager();
  MdlMngr = new ModelManager();
  ShdrMngr = new XmlShaderManager();

  if (!LoadFont("gfx/arial.ttf",18,"FPS Font")){ //Load Standard Font
    ThErrorMessage("Font Error","Konnte Engine Font nicht laden!");
    return false;
  }
  glShadeModel(GL_SMOOTH);							// Enable Smooth Shading
  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);				// Black Background
  glClearDepth(1.0f);									// Depth Buffer Setup
  glEnable(GL_DEPTH_TEST);							// Enables Depth Testing
  glEnable(GL_TEXTURE_2D);							// Enable 2D Texture Mapping
  glDepthFunc(GL_LESS);								// The Type Of Depth Testing To Do
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	// Really Nice Perspective Calculations
  glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  //glBlendFunc(GL_SRC_ALPHA,GL_ONE);
  glDisable(GL_BLEND);
  glEnable(GL_COLOR_MATERIAL);
  if(bAntialiasing)
    glEnable(GL_MULTISAMPLE_ARB);
  else
    glDisable(GL_MULTISAMPLE_ARB);
  return true;
}

void StdDeInit(){
  //Manager zerstören
  delete ShdrMngr;
  delete MdlMngr;
  delete TexMngr;
}

void ShowFps(){
    SetRenderstate(th_renderstate_text);
    glActiveTexture(GL_TEXTURE0);
    glDisable(GL_LIGHTING);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glDisable(GL_DEPTH_TEST);
    glColor3f(1.0f,1.0f,0.0f);
    glTranslatef(iWidth-100,6,0);
    glPrint(0,"FPS:%.1f",fps);
    glTranslatef(0,16,0);
    glPrint(0,"%.3fms",fFrametime);
    glEnable(GL_DEPTH_TEST);
    SetRenderstate(th_renderstate_model);
}

void TerminateApplication(){
	//PostMessage (hWnd, WM_QUIT, 0, 0);
	bIsWorking = false;
}

void RecalcAspect (int width, int height){
	glViewport (0, 0, (GLsizei)(width), (GLsizei)(height));
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	gluPerspective (45.0f, (GLfloat)(width)/(GLfloat)(height),0.1f, fFarClipping);
	glMatrixMode (GL_MODELVIEW);
	iWidth = width;
	iHeight = height;
	OISSetWindowSize(iWidth,iHeight);
}

//OS API Specific Stuff
#ifdef __linux
void ShowCursor(bool value){
  if(!value)
    XDefineCursor(m_Display, m_XWindow, cursor);
  else
	XUndefineCursor(m_Display, m_XWindow);
}

BOOL DestroyMainWindow (){
	if(m_GLXContext)
	{
		if(!glXMakeCurrent(m_Display, None, NULL))
		{
			ThErrorMessage("X Error","Couldn't release Rendering Context");
		}
		glXDestroyContext(m_Display, m_GLXContext);
		m_GLXContext = NULL;
	}

	if(bIsFullscreen)
	{
		XF86VidModeSwitchToMode(m_Display, m_XScreenID, &m_XF86DeskMode);
		XF86VidModeSetViewPort(m_Display, m_XScreenID, 0, 0);
	}
	XCloseDisplay(m_Display);

	return true;														// Return True
}

bool CreateMainWindow (char *title, int width, int height, int ColorBits, int DepthBits, bool fullscreen){
	int attrListDbl[] = {GLX_RGBA,GLX_DEPTH_SIZE,DepthBits,GLX_DOUBLEBUFFER,None};
	int attrListMultisample[] = {GLX_RGBA,GLX_DEPTH_SIZE,DepthBits,GLX_DOUBLEBUFFER,GLX_SAMPLE_BUFFERS,1,GLX_SAMPLES,XmlSamples,None};

	XVisualInfo *vi;
	Colormap cmap;
	int dpyWidth, dpyHeight;
	int i;
	int vidModeMajorVersion, vidModeMinorVersion;
	XF86VidModeModeInfo **modes;
	int modeNum;
	int bestMode=0;
	Atom wmDelete;

    // Open a connection to the X-server.
	m_Display = XOpenDisplay(0);

    if(m_Display == NULL)
	{
		ThErrorMessage("X Error","Failed to open Display");
		return false;
	}

	m_XScreenID = DefaultScreen(m_Display);

	//  Query Videomode version
	if(!XF86VidModeQueryVersion(m_Display, &vidModeMajorVersion,
							&vidModeMinorVersion))
	{
		ThErrorMessage("X Error","Couldn't query Video Mode Version");
		return false;
	}

	// More videomode queries
	if(!XF86VidModeGetAllModeLines(m_Display, m_XScreenID, &modeNum, &modes))
	{
		ThErrorMessage("X Error","Couldn't query Video Settings");
		return false;
	}

	m_XF86DeskMode = *modes[0];

	// Loop through all videomodes to find the most appropriate
	for (i = 0; i < modeNum; i++)
	{
		if ((modes[i]->hdisplay == width) && (modes[i]->vdisplay == height))
		{
			bestMode = i;
		}
	}

	// Get a X11-visual
	if(XmlSamples > 0){
	  vi = glXChooseVisual(m_Display, m_XScreenID, attrListMultisample);
	    if(NULL == vi)
	    {
	      ThErrorMessage("X Error","Multisampling not supported");
          return false;
	    }
	    else
	      arbMultisampleSupported=true;
	}
	else{
	  vi = glXChooseVisual(m_Display, m_XScreenID, attrListDbl);
	  if(NULL == vi)
	  {
	      ThErrorMessage("X Error","Doublebuffer not aviable");
          return false;
	  }
	}

	// Get a GLX-context
	m_GLXContext = glXCreateContext(m_Display, vi, 0, GL_TRUE);

	// Bail out if failed
	if(m_GLXContext == NULL)
	{
		ThErrorMessage("X Error","Couldn't get Render Context");
		return false;
	}

	cmap = XCreateColormap(m_Display, RootWindow(m_Display, vi->screen),vi->visual, AllocNone);
	m_XSetAttr.colormap = cmap;
	m_XSetAttr.border_pixel = 0;

	if(fullscreen)
	{
		XF86VidModeSwitchToMode(m_Display, m_XScreenID, modes[bestMode]);
		XF86VidModeSetViewPort(m_Display, m_XScreenID, 0, 0);
		dpyWidth = modes[bestMode]->hdisplay;
		dpyHeight = modes[bestMode]->vdisplay;
		XFree(modes);
		m_XSetAttr.override_redirect = True;
		m_XSetAttr.event_mask = ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask |
								StructureNotifyMask;
		m_XWindow = XCreateWindow(m_Display, RootWindow(m_Display, vi->screen),
								0, 0, dpyWidth, dpyHeight, 0, vi->depth, InputOutput, vi->visual,
								CWBorderPixel | CWColormap | CWEventMask | CWOverrideRedirect,
								&m_XSetAttr);
		//XWarpPointer(m_Display, None, m_XWindow, 0, 0, 0, 0, 0, 0);
		XMapRaised(m_Display, m_XWindow);
		//XGrabKeyboard(m_Display, m_XWindow, True, GrabModeAsync,GrabModeAsync, CurrentTime);
		//XGrabPointer(m_Display, m_XWindow, True, ButtonPressMask,
		//GrabModeAsync, GrabModeAsync, m_XWindow, None, CurrentTime);
		bIsFullscreen = true;
	}
	else
	{
		m_XSetAttr.event_mask = ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask |
								StructureNotifyMask;
		m_XWindow = XCreateWindow(m_Display, RootWindow(m_Display, vi->screen),
								  0, 0, width, height, 0, vi->depth, InputOutput, vi->visual,
								  CWBorderPixel | CWColormap | CWEventMask, &m_XSetAttr);
		wmDelete = XInternAtom(m_Display, "WM_DELETE_WINDOW", True);
		XSetWMProtocols(m_Display, m_XWindow, &wmDelete, 1);

		XSetStandardProperties(m_Display, m_XWindow,
		title, None, None, NULL, 0, NULL);
		XMapRaised(m_Display, m_XWindow);
	}

	if(!SetupOIS(m_XWindow,m_Display))
	  return false;

	// Switch to the new GLX-context
	if(!glXMakeCurrent(m_Display, m_XWindow, m_GLXContext))
	{
		ThErrorMessage("X Error","Couldn't switch to new rendering Context");
		return false;
	}

    //Create a blank cursor:
	Pixmap bm_no;
	XColor black, dummy;
	Colormap colormap;
	static char no_data[] = { 0,0,0,0,0,0,0,0 };

	colormap = DefaultColormap( m_Display, DefaultScreen(m_Display) );
	XAllocNamedColor( m_Display, colormap, "black", &black, &dummy );
	bm_no = XCreateBitmapFromData( m_Display, m_XWindow, no_data, 8, 8 );
	cursor = XCreatePixmapCursor( m_Display, bm_no, bm_no, &black, &black, 0, 0 );

	if(fullscreen)
	  ShowCursor(false);

	RecalcAspect (width, height);
	return true;
}

void ProcessEvents()
{
	XEvent event;
	int temp_key;

	while(XPending(m_Display) > 0)
	{
		XNextEvent(m_Display, &event);

		switch(event.type)
		{
			case Expose:
				if (event.xexpose.count != 0)
					break;
				break;
			case ConfigureNotify:
				if ((event.xconfigure.width != iWidth) ||
					(event.xconfigure.height != iHeight))
				{
					iWidth = event.xconfigure.width;
					iHeight = event.xconfigure.height;
					RecalcAspect(iWidth,iHeight);
				}
				break;

			case ClientMessage:
                {
                    std::string result;

                    result = std::string(XGetAtomName(m_Display, event.xclient.message_type));

                    if (result == "WM_PROTOCOLS")
                    {
                        bIsWorking = false;
                    }
                    break;
                }
			default:
				break;
		}
	}
}
#endif

#ifdef WIN32
BOOL RegisterWindowClass (){																		// TRUE If Successful
	// Register A Window Class
	WNDCLASSEX windowClass;												// Window Class
	ZeroMemory (&windowClass, sizeof (WNDCLASSEX));						// Make Sure Memory Is Cleared
	windowClass.cbSize			= sizeof (WNDCLASSEX);					// Size Of The windowClass Structure
	windowClass.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;	// Redraws The Window For Any Movement / Resizing
	windowClass.lpfnWndProc		= (WNDPROC)(WindowProc);				// WindowProc Handles Messages
	windowClass.hInstance		= hInstance;				            // Set The Instance
	windowClass.hIcon			= LoadIcon(NULL, IDI_WINLOGO);          // Load default Icon
	windowClass.hbrBackground	= NULL;			                        // No Background required
	windowClass.hCursor			= LoadCursor(NULL, IDC_ARROW);			// Load The Arrow Pointer
	windowClass.lpszClassName	= "ThEngine";				            // Sets The Applications Classname
	if (RegisterClassEx (&windowClass) == 0)							// Did Registering The Class Fail?
	{
		//Should Never Happen
		ThErrorMessage("Window Error","RegisterClassEx Failed!");
		return FALSE;													// Return False (Failure)
	}
	return TRUE;														// Return True (Success)
}

bool InitMultisample(HINSTANCE hInstance,HWND hWnd,PIXELFORMATDESCRIPTOR pfd,int ColorBits,int DepthBits){
  char meldung[200];
  if(!GLEE_WGL_ARB_multisample){
    ThErrorMessage("Init Fehler","Kein Support für Multisampling vorhanden!");
    return false;
  }

  HDC hDC = GetDC(hWnd);

	int pixelFormat;
	bool valid;
	UINT numFormats;
	float fAttributes[] = {0,0};

	int iAttributes[] = { WGL_DRAW_TO_WINDOW_ARB,GL_TRUE,
		WGL_SUPPORT_OPENGL_ARB,GL_TRUE,
		WGL_ACCELERATION_ARB,WGL_FULL_ACCELERATION_ARB,
		WGL_COLOR_BITS_ARB,ColorBits-8,
		WGL_ALPHA_BITS_ARB,8,
		WGL_DEPTH_BITS_ARB,DepthBits,
		WGL_STENCIL_BITS_ARB,0,
		WGL_DOUBLE_BUFFER_ARB,GL_TRUE,
		WGL_SAMPLE_BUFFERS_ARB,GL_TRUE,
		WGL_SAMPLES_ARB, XmlSamples,
		0,0};

	// First We Check To See If We Can Get A Pixel Format For 4 Samples
	valid = wglChoosePixelFormatARB(hDC,iAttributes,fAttributes,1,&pixelFormat,&numFormats);

	// if We Returned True, And Our Format Count Is Greater Than 1
	if (valid && numFormats >= 1)
	{
		arbMultisampleSupported	= true;
		arbMultisampleFormat	= pixelFormat;
		return arbMultisampleSupported;
	}

    //Hat nicht geklappt
    sprintf(meldung,"Multisampling mit %i Samples wird nicht unterstüzt!",XmlSamples);
    ThErrorMessage("Init Error",meldung);

	// Return The Valid Format
	return arbMultisampleSupported;
}

BOOL DestroyMainWindow (){
	if (bIsFullscreen)										            // Is Window In Fullscreen Mode
	{
		ChangeDisplaySettings (NULL,0);									// Switch Back To Desktop Resolution
		ShowCursor (TRUE);												// Show The Cursor
	}

	if (hWnd != 0)												        // Does The Window Have A Handle?
	{
		if (hDC != 0)											        // Does The Window Have A Device Context?
		{
			wglMakeCurrent (hDC, 0);							        // Set The Current Active Rendering Context To Zero
			if (hRC != 0)										        // Does The Window Have A Rendering Context?
			{
				wglDeleteContext (hRC);							        // Release The Rendering Context
				hRC = 0;										        // Zero The Rendering Context

			}
			ReleaseDC (hWnd,hDC);						                // Release The Device Context
			hDC = 0;											        // Zero The Device Context
		}
		DestroyWindow (hWnd);									        // Destroy The Window
		hWnd = 0;												        // Zero The Window Handle
	}

	return true;														// Return True
}

bool CreateMainWindow (char *title, int width, int height, int ColorBits, int DepthBits, bool fullscreen){
	DWORD windowStyle = WS_OVERLAPPEDWINDOW;
	DWORD windowExtendedStyle = WS_EX_APPWINDOW;
    DEVMODE dmScreenSettings;
    iWidth = width;
    iHeight = height;

	PIXELFORMATDESCRIPTOR pfd =
	{
		sizeof (PIXELFORMATDESCRIPTOR),
		1,
		PFD_DRAW_TO_WINDOW |
		PFD_SUPPORT_OPENGL |
		PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,
		ColorBits, //Color Depth
		0, 0, 0, 0, 0, 0,
		0,
		0,
		0,
		0, 0, 0, 0,
		DepthBits, //Depth Buffer
		0, //Stencil Buffer
		0, //Auxiliary Buffer
		PFD_MAIN_PLANE,
		0,
		0, 0, 0
	};

	GLuint PixelFormat;

	if (fullscreen)
	{
	  ZeroMemory (&dmScreenSettings, sizeof (DEVMODE));
	  dmScreenSettings.dmSize				= sizeof (DEVMODE);
	  if(XmlKeepCurrent){
	    EnumDisplaySettings(NULL,ENUM_CURRENT_SETTINGS,&dmScreenSettings); //Retrieve current Settings
	    width = dmScreenSettings.dmPelsWidth;
	    iWidth = width;
	    height = dmScreenSettings.dmPelsHeight;
	    iHeight = height;
	    ColorBits = dmScreenSettings.dmBitsPerPel;
	    XmlColorBits = ColorBits;
	  }
	  else{
	    //Change Display Size
	    dmScreenSettings.dmPelsWidth		= width;
	    dmScreenSettings.dmPelsHeight		= height;
	    dmScreenSettings.dmBitsPerPel		= ColorBits;
	    dmScreenSettings.dmFields			= DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
	  }
	}

	RECT windowRect = {0, 0, width, height};

	if(fullscreen)
	{
	  if (ChangeDisplaySettings (&dmScreenSettings, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
      {
        MessageBox (HWND_DESKTOP, "Vollbildmodus nicht verfügbar!", "Error", MB_OK | MB_ICONEXCLAMATION);
        bIsFullscreen = false;
      }
      else
      {
        ShowCursor (FALSE);
        windowStyle = WS_POPUP;
        windowExtendedStyle |= WS_EX_TOPMOST;
        bIsFullscreen = true;
      }
	}
	else
	{
		AdjustWindowRectEx (&windowRect, windowStyle, 0, windowExtendedStyle);
	}

	hWnd = CreateWindowEx (windowExtendedStyle,					// Extended Style
								   "ThEngine",	                        // Class Name
								   title,					            // Window Title
								   windowStyle,							// Window Style
								   0, 0,								// Window X,Y Position
								   windowRect.right - windowRect.left,	// Window Width
								   windowRect.bottom - windowRect.top,	// Window Height
								   HWND_DESKTOP,						// Desktop Is Window's Parent
								   0,									// No Menu
								   hInstance,
								   NULL);

	if (hWnd == 0)
      return false;

	hDC = GetDC (hWnd);
	if (hDC == 0)
	{
		DestroyWindow (hWnd);
		hWnd = 0;
		return false;
	}

    if(!arbMultisampleSupported){
      PixelFormat = ChoosePixelFormat (hDC, &pfd);				        // Find A Compatible Pixel Format
	  if (PixelFormat == 0)
  	  {
		ReleaseDC (hWnd, hDC);							                // Release Our Device Context
		hDC = 0;												// Zero The Device Context
		DestroyWindow (hWnd);									        // Destroy The Window
		hWnd = 0;												// Zero The Window Handle
		return false;													// Return False
	  }
    }
    else
      PixelFormat = arbMultisampleFormat;

	if (SetPixelFormat (hDC, PixelFormat, &pfd) == FALSE)		        // Try To Set The Pixel Format
	{
		ReleaseDC (hWnd, hDC);							                // Release Our Device Context
		hDC = 0;												        // Zero The Device Context
		DestroyWindow (hWnd);									        // Destroy The Window
		hWnd = 0;												        // Zero The Window Handle
		return false;
	}

	hRC = wglCreateContext (hDC);						        // Try To Get A Rendering Context
	if (hRC == 0)
	{
		ReleaseDC (hWnd, hDC);							                // Release Our Device Context
		hDC = 0;												        // Zero The Device Context
		DestroyWindow (hWnd);									        // Destroy The Window
		hWnd = 0;												        // Zero The Window Handle
		return false;
	}

	// Make The Rendering Context Our Current Rendering Context
	if (wglMakeCurrent (hDC, hRC) == FALSE)
	{
		wglDeleteContext (hRC);									        // Delete The Rendering Context
		hRC = 0;												// Zero The Rendering Context
		ReleaseDC (hWnd, hDC);							                // Release Our Device Context
		hDC = 0;												// Zero The Device Context
		DestroyWindow (hWnd);									        // Destroy The Window
		hWnd = 0;												// Zero The Window Handle
		return false;
	}

	if(!arbMultisampleSupported && XmlSamples > 0){
	  if(InitMultisample(hInstance,hWnd,pfd,ColorBits,DepthBits)){
        DestroyMainWindow();
	    return CreateMainWindow(title,width,height,ColorBits,DepthBits,fullscreen);
	  }
	}

	ShowWindow (hWnd, SW_NORMAL);
	bIsVisible = true;

	if(!SetupOIS(hWnd))
	  return false;

    if(fullscreen){
      ShowCursor(false);
    }

	RecalcAspect (width, height);
	return true;
}

LRESULT CALLBACK WindowProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

	switch (uMsg)														// Evaluate Window Message
	{
		case WM_SYSCOMMAND:												// Intercept System Commands
		{
			switch (wParam)												// Check System Calls
			{
				case SC_SCREENSAVE:										// Screensaver Trying To Start?
				case SC_MONITORPOWER:									// Monitor Trying To Enter Powersave?
				return 0;												// Prevent From Happening
			}
			break;														// Exit
		}
		return 0;														// Return														// Return

		case WM_CLOSE:													// Closing The Window
			TerminateApplication();								// Terminate The Application
		return 0;														// Return

		case WM_SIZE:													// Size Action Has Taken Place
			switch (wParam)												// Evaluate Size Action
			{
				case SIZE_MINIMIZED:									// Was Window Minimized?
					bIsVisible = false;							// Set isVisible To False
				return 0;												// Return

				case SIZE_MAXIMIZED:									// Was Window Maximized?
					bIsVisible = true;							// Set isVisible To True
					RecalcAspect (LOWORD (lParam), HIWORD (lParam));		// Reshape Window - LoWord=Width, HiWord=Height
				return 0;												// Return

				case SIZE_RESTORED:										// Was Window Restored?
					bIsVisible = true;							// Set isVisible To True
					RecalcAspect (LOWORD (lParam), HIWORD (lParam));		// Reshape Window - LoWord=Width, HiWord=Height
				return 0;

				default:
				    RecalcAspect(LOWORD (lParam), HIWORD (lParam));
                return 0;											// Return
			}
		break;															// Break

	}

	return DefWindowProc (hWnd, uMsg, wParam, lParam);					// Pass Unhandled Messages To DefWindowProc
}
#endif

#ifdef __linux
int main(int argc, char **argv){
#elif defined WIN32
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow){
	MSG	  msg;											// Window Message Structur
#endif
  	for(int i=0;i<256;i++){
      KeysDown[i]=NULL;
      KeysUp[i]=NULL;
      KeyStates[i]=false;
  	}

	#ifdef TH_DEBUG
	debugtxt=fopen("debug.txt","w");
	fprintf(debugtxt,"Engine started in debug mode\n");
	fclose(debugtxt);
	#endif

	debugtxt=fopen("engine.log","w");
	fprintf(debugtxt,"Engine Started\n");
	fclose(debugtxt);

	const char * XmlErgebnis = LoadBasicXml();
	if(XmlErgebnis != NULL){
	  #ifdef __linux
	  std::cout << "XML-Error\n" << XmlErgebnis << "\n";
	  #endif
	  ThError("XML-Error",XmlErgebnis);
	  return -1;
	}

	#ifdef WIN32
    // Register A Class For Our Window To Use
	if (RegisterWindowClass () == false)					// Did Registering A Class Fail?
	{
		// Failure
		//MessageBox (HWND_DESKTOP, "Error Registering Window Class!", "Error", MB_OK | MB_ICONEXCLAMATION);
		ThErrorMessage("Window Error","Error Registering Window Class");
		return -1;														// Terminate Application
	}
	#endif

	bIsWorking = true;
	//Create Main Window
    #ifdef __linux
    if(!CreateMainWindow(ThWindowTitle,XmlResX,XmlResY,XmlColorBits,XmlDepthBits,XmlFullscreen)){
	  ThErrorMessage("Kritischer Fehler","Konnte Hauptfenster nicht erstellen\nProgramm wird beendet");
	  TerminateApplication();
	}
	#elif defined WIN32
	#ifdef __SSE__
	if(!IsProcessorFeaturePresent(6)){
       //MessageBox(HWND_DESKTOP, "Error: No SSE aviable!","Error",MB_OK | MB_ICONERROR);
       ThErrorMessage("SSE Error","No SSE aviable!");
       return -1;
	}

	if(!CreateMainWindow(ThWindowTitle,XmlResX,XmlResY,XmlColorBits,XmlDepthBits,XmlFullscreen)){
    #else
	if(!CreateMainWindow(ThWindowTitle,XmlResX,XmlResY,XmlColorBits,XmlDepthBits,XmlFullscreen)){
    #endif
	  //MessageBox(NULL,"Konnte Hauptfenster nicht erstellen\nProgramm wird beendet","Kritischer Fehler",MB_OK);
	  ThErrorMessage("Kritischer Fehler","Konnte Hauptfenster nicht erstellen\nProgramm wird beendet");
	  TerminateApplication();
	}
	#endif
	else if(!StdInit()){
	  //MessageBox(NULL,"Standard Initialisierung fehlgeschlagen","Kritsicher Fehler",MB_OK);
	  ThErrorMessage("Kritischer Fehler","Standard Initialisierung fehlgeschlagen");
	  TerminateApplication();
	}
	else if(!Init()){
	  //MessageBox(NULL,"Engine Initialisierung fehlgeschalgen","Kritsicher fehler",MB_OK);
	  ThErrorMessage("Kritischer Fehler","Engine Initialisierung fehlgeschlagen");
	  TerminateApplication();
	}
	else{
      SetTime();
      debugtxt=fopen("engine.log","a");
	  fprintf(debugtxt,"Init Finished\n");
	  fclose(debugtxt);

	  while(bIsWorking){
        OISCheckInput();
        if(KeyESC)
          bIsWorking=false;
	    #ifdef __linux
        ProcessEvents();
        #elif defined WIN32
        if(PeekMessage (&msg, hWnd, 0, 0, PM_REMOVE) != 0){
	      if(msg.message == WM_QUIT)
	        bIsWorking = false;
          else
            DispatchMessage (&msg);
	    }
	    else {
        #endif
          Runtime = timer->GetTimeInt64();
          dTimeDiff = Runtime - LastFrame;
          //if(dTimeDiff < 5)
            //usleep(20);
          dTimeSinceLastFPS+=dTimeDiff;
          if(dTimeSinceLastFPS > 250){
            fps = static_cast<double>(iFrames)*(1000/dTimeSinceLastFPS);
            fFrametime = dTimeSinceLastFPS/static_cast<double>(iFrames);
            #ifdef TH_DEBUG
            AverageFps = (AverageFps * FpsAnz + fps)/ (FpsAnz+1.0);
            FpsAnz+=1.0;
            #endif
            dTimeSinceLastFPS=0;
            iFrames=0;
          }
          PreDraw();
	      Draw();
	      if(bShowFps)
	        ShowFps();
	      glFlush();
	      #ifdef __linux
	      glXSwapBuffers(m_Display, m_XWindow);
	      #elif defined WIN32
	      SwapBuffers(hDC);
	      #endif
	      iFrames++;
	      LastFrame = Runtime;
        #ifdef WIN32
	    }
	    #endif
	  }
	}

    if(ThErrorLength != 0){
      #ifdef __linux
      std::cout << "Fehlermeldungen:" << std::endl << ThErrorString << std::endl;
      #elif defined WIN32
      MessageBox(NULL,ThErrorString,"Fehlermeldungen",MB_OK | MB_ICONERROR);
      #endif
      std::ofstream LogFile;
	  LogFile.open("engine.log",std::ios::app);
	  LogFile << "Fehler Aufgetreten:\n" << ThErrorString << "\n";
	  LogFile.close();
    }

    KillFont();
	DeInit();
	StdDeInit();

	DestroyMainWindow();

    #ifdef TH_DEBUG
	debugtxt=fopen("debug.txt","a");
	fprintf(debugtxt,"Average FPS: %.2f\n",static_cast<float>(AverageFps));
	fprintf(debugtxt,"Error-Log: %s",ThErrorString);
	fclose(debugtxt);
	#endif

	#ifdef WIN32
	UnregisterClass ("ThEngine", hInstance);
	#endif
	return 0;
}

void ViewOrtho(){
    glFlush();
    glViewport(0,0,iWidth,iHeight);
	glMatrixMode(GL_PROJECTION);								// Select Projection
	glPushMatrix();												// Push The Matrix
	glLoadIdentity();											// Reset The Matrix
	glOrtho( 0, iWidth , iHeight , 0, -1, 1 );							// Select Ortho Mode (640x480)
	glMatrixMode(GL_MODELVIEW);									// Select Modelview Matrix
	glPushMatrix();												// Push The Matrix
	glLoadIdentity();											// Reset The Matrix
}

void ViewPerspective(){
    glFlush();
    glViewport(0,0,iWidth,iHeight);
	glMatrixMode( GL_PROJECTION );								// Select Projection
	glPopMatrix();												// Pop The Matrix
	glMatrixMode( GL_MODELVIEW );								// Select Modelview
	glPopMatrix();												// Pop The Matrix
}

void SetRenderstate(int renderstate){
  if(curr_renderstate != renderstate && renderstate < number_renderstates){
    switch(renderstate){
      case th_renderstate_model:
        switch(curr_renderstate){
          case th_renderstate_model_blend:
            glDisable(GL_BLEND);
            break;
          case th_renderstate_text:
            glDisable(GL_BLEND);
            ViewPerspective();
            break;
          case th_renderstate_effect:
            ViewPerspective();
            break;
          case th_renderstate_lines:
            ViewPerspective();
            glEnable(GL_TEXTURE_2D);
            glDisable(GL_LINE_SMOOTH);
            glDisable(GL_BLEND);
            break;
        }
        break;
      case th_renderstate_model_blend:
        switch(curr_renderstate){
          case th_renderstate_model:
            glEnable(GL_BLEND);
            break;
          case th_renderstate_text:
            glEnable(GL_BLEND);
            ViewOrtho();
            break;
          case th_renderstate_effect:
            ViewOrtho();
            break;
          case th_renderstate_lines:
            ViewPerspective();
            glEnable(GL_TEXTURE_2D);
            glDisable(GL_LINE_SMOOTH);
            break;
        }
        break;
      case th_renderstate_text:
        switch(curr_renderstate){
          case th_renderstate_model:
            glEnable(GL_BLEND);
            glDisable(GL_LIGHTING);
            ViewOrtho();
            break;
          case th_renderstate_model_blend:
            ViewOrtho();
            break;
          case th_renderstate_effect:
            glEnable(GL_BLEND);
            break;
          case th_renderstate_lines:
            glEnable(GL_TEXTURE_2D);
            glDisable(GL_LINE_SMOOTH);
            break;
        }
        break;
      case th_renderstate_effect:
        switch(curr_renderstate){
          case th_renderstate_model:
            ViewOrtho();
            break;
          case th_renderstate_model_blend:
            glDisable(GL_BLEND);
            ViewOrtho();
            break;
          case th_renderstate_text:
            glDisable(GL_BLEND);
            break;
          case th_renderstate_lines:
            glDisable(GL_BLEND);
            glEnable(GL_TEXTURE_2D);
            glDisable(GL_LINE_SMOOTH);
            break;
        }
        break;
      case th_renderstate_lines:
        glEnable(GL_LINE_SMOOTH);
        glDisable(GL_TEXTURE_2D);
        switch(curr_renderstate){
          case th_renderstate_model:
            ViewOrtho();
            glEnable(GL_BLEND);
            break;
          case th_renderstate_model_blend:
            ViewOrtho();
            break;
          case th_renderstate_effect:
            glEnable(GL_BLEND);
            break;
        }
        break;
    }
  curr_renderstate = renderstate;
  }
}

const char ErrorInvalidValue[]="Ein Config-Element hat einen Fehlerhaften inhalt";
const char ErrorNoElement[]="Nicht verarbeitbare Information im Config-Tag";
const char ErrorMissingInformation[]="Nicht alle benötigten Informationen im Config-Tag vorhanden!";
const char ErrorNoConfigFile[]="Konnte config.xml nicht finden!";

const char * LoadBasicXml(){
  FILE *ConfigFile;
  ConfigFile = fopen("config.xml","r");
  if(ConfigFile == NULL)
    return ErrorNoConfigFile;
  else
    fclose(ConfigFile);
  XmlConfigDoc = new TiXmlDocument("config.xml");
  bool bLoadOk = XmlConfigDoc->LoadFile();
  if(!bLoadOk)
    return XmlConfigDoc->ErrorDesc();

  TiXmlNode* pChild;
  TiXmlNode* pChild2;
  TiXmlNode* pChild3;
  TiXmlText* pText;
  int t,t2,t3;

  for(pChild = XmlConfigDoc->FirstChild(); pChild != 0; pChild = pChild->NextSibling()){
    t = pChild->Type();
	switch(t){
	  case TiXmlNode::ELEMENT:
        if(strcmp("config",pChild->Value())==0){
          for(pChild2 = pChild->FirstChild(); pChild2 != 0; pChild2 = pChild2->NextSibling()){
            t2 = pChild2->Type();
            if(t2==TiXmlNode::ELEMENT){
              if(strcmp("resolutionx",pChild2->Value())==0){
                pChild3 = pChild2->FirstChild();
                t3 = pChild3->Type();
                if(t3 != TiXmlNode::TEXT)
                  return ErrorInvalidValue;
                else
                  XmlResX = atoi(pChild3->Value());
              }
              else if(strcmp("resolutiony",pChild2->Value())==0){
                pChild3 = pChild2->FirstChild();
                t3 = pChild3->Type();
                if(t3 != TiXmlNode::TEXT)
                  return ErrorInvalidValue;
                else
                  XmlResY=atoi(pChild3->Value());
              }
              else if(strcmp("DepthBits",pChild2->Value())==0){
                pChild3 = pChild2->FirstChild();
                t3 = pChild3->Type();
                if(t3 != TiXmlNode::TEXT)
                  return ErrorInvalidValue;
                else
                  XmlDepthBits=atoi(pChild3->Value());
              }
              else if(strcmp("ColorBits",pChild2->Value())==0){
                pChild3 = pChild2->FirstChild();
                if(pChild3->Type() != TiXmlNode::TEXT)
                  return ErrorInvalidValue;
                else
                  XmlColorBits=atoi(pChild3->Value());
              }
              else if(strcmp("fullscreen",pChild2->Value())==0){
                pChild3 = pChild2->FirstChild();
                t3 = pChild3->Type();
                if(t3 != TiXmlNode::TEXT)
                  return ErrorInvalidValue;
                else{
                  XmlFullscreen=(bool)atoi(pChild3->Value());
                  if(atoi(pChild3->Value()) > 1)
                    XmlKeepCurrent=true;
                }
              }
              else if(strcmp("MultisampleSamples",pChild2->Value())==0){
                pChild3 = pChild2->FirstChild();
                if(pChild3->Type() != TiXmlNode::TEXT)
                  return ErrorInvalidValue;
                else
                  XmlSamples = atoi(pChild3->Value());
              }
              else if(strcmp("antialiasing",pChild2->Value())==0){
                pChild3 = pChild2->FirstChild();
                if(pChild3->Type() != TiXmlNode::TEXT)
                  return ErrorInvalidValue;
                else
                  bAntialiasing = (bool)atoi(pChild3->Value());
              }
              else if(strcmp("vsync",pChild2->Value())==0){
                pChild3 = pChild2->FirstChild();
                if(pChild3->Type() != TiXmlNode::TEXT)
                  return ErrorInvalidValue;
                else
                  bVsync = (bool)atoi(pChild3->Value());
              }
              else
                return ErrorNoElement;
            }
          }
        }
	    break;
	}
  }

  if(XmlResX==0 || XmlResY == 0 || XmlColorBits == 0 || XmlDepthBits == 0)
    return ErrorMissingInformation;

  return NULL;
}

void ThErrorMessage(const char *Title,const char *Text){
  int i;
  bThError=true;
  int newlength = ThErrorLength + strlen(Text) + strlen(Title) + 3;
  char *buff = new char[newlength];

  for(i=0;i<ThErrorLength;i++){
    buff[i]=ThErrorString[i];
  }

  for(i=0;i<strlen(Title);i++){
    buff[i+ThErrorLength]=Title[i];
  }
  buff[ThErrorLength+i]='\n';

  for(i=0;i<strlen(Text);i++){
    buff[i+ThErrorLength+strlen(Title)+1]=Text[i];
  }

  buff[i+ThErrorLength+strlen(Title)+1]='\n';
  buff[i+ThErrorLength+strlen(Title)+2]='\0';

  if(ThErrorLength != 0)
    delete []ThErrorString;
  ThErrorString = buff;
  ThErrorLength = newlength-1;
}

void ThError(const char *Title, const char *Text){
  #ifdef WIN32
  MessageBox(NULL,Text,Title,MB_OK | MB_ICONERROR);
  #elif defined __linux
  std::string Cmd;
  Cmd = "zenity --error --text=\"";
  Cmd += Text;
  Cmd += "\"";
  if(system(Cmd.c_str())!=127)
    return;

  Cmd = "notify-send --urgency=critical --category=error \"";
  Cmd += Title;
  Cmd += "\" \"";
  Cmd += Text;
  Cmd += "\"";
  if(system(Cmd.c_str())!=127)
    return;

  Cmd = "xmessage \"";
  Cmd += Text;
  Cmd += "\"";
  system(Cmd.c_str());
  #endif
}

void ThGetGLError(const char *title){
  #ifdef TH_DEBUG
  GLenum Error = glGetError();
  printf("glGetError - %s\n",title);
  while(Error != GL_NO_ERROR){
    switch(Error){
      case GL_INVALID_ENUM:
        printf("GL_INVALID_ENUM\n");
        break;
      case GL_INVALID_VALUE:
        printf("GL_INVALID_VALUE\n");
        break;
      case GL_INVALID_OPERATION:
        printf("GL_INVALID_OPERATION\n");
        break;
      case GL_STACK_OVERFLOW:
        printf("GL_STACK_OVERFLOW\n");
        break;
      case GL_STACK_UNDERFLOW:
        printf("GL_STACK_UNDERFLOW\n");
        break;
      case GL_OUT_OF_MEMORY:
        printf("GL_OUT_OF_MEMORY\n");
        break;
      case GL_NO_ERROR:
        printf("GL_NO_ERROR\n");
        break;
      default:
        printf("Unknown Error: %X\n",Error);
    }
  }
  #endif
}
