/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "timer.h"

Timer::Timer(bool &error)								// Initialize Our Timer (Get It Ready)
{
  #ifdef WIN32
  if (!QueryPerformanceFrequency((LARGE_INTEGER *) &ui64Frequency))
  {
	// No Performace Counter Available
    error = true;
  }
  else
  {
	// Performance Counter Is Available
	QueryPerformanceCounter((LARGE_INTEGER *) &ui64Start);
	dResolution = ((double)1.0f)/((double)ui64Frequency);
	ui64Elapsed = ui64Start;
	error = false;
  }
  #else
  clock_gettime(CLOCK_PROCESS_CPUTIME_ID,&timestruct);
  ui64Start = (uint64)timestruct.tv_nsec + (uint64)timestruct.tv_sec * 1000000000L;
  ui64Elapsed = ui64Start;
  error = false;
  #endif
}

float Timer::GetTimeFloat()
{
  uint64 ui64Time;
  #ifdef WIN32
  QueryPerformanceCounter((LARGE_INTEGER *) &ui64Time);
  return ( (float) ( ui64Time - ui64Start) * (float)dResolution)*1000.0f;
  #else
  clock_gettime(CLOCK_PROCESS_CPUTIME_ID,&timestruct);
  ui64Time = (uint64)timestruct.tv_nsec + (uint64)timestruct.tv_sec * 1000000000L;
  return (float)((ui64Time - ui64Start) / 1000000L);
  #endif
}

double Timer::GetTimeDouble()
{
  uint64 ui64Time;
  #ifdef WIN32
  QueryPerformanceCounter((LARGE_INTEGER *) &ui64Time);
  return ( (double) ( ui64Time - ui64Start) * dResolution) * 1000;
  #else
  clock_gettime(CLOCK_PROCESS_CPUTIME_ID,&timestruct);
  ui64Time = (uint64)timestruct.tv_nsec + (uint64)timestruct.tv_sec * 1000000000L;
  return (double)((ui64Time - ui64Start) / 1000000L);
  #endif
}

uint64 Timer::GetTimeInt64(){
  uint64 ui64Time;
  #ifdef WIN32
  QueryPerformanceCounter((LARGE_INTEGER *) &ui64Time);
  #else
  clock_gettime(CLOCK_PROCESS_CPUTIME_ID,&timestruct);
  ui64Time = (uint64)timestruct.tv_nsec + (uint64)timestruct.tv_sec * 1000000000L;
  #endif
  return ui64Time;
}

double Timer::GetResolution(){
  return dResolution;
}

Zeitpunkt::Zeitpunkt(){
  ui64Time=0;
  TimerReference = NULL;
}

Zeitpunkt::Zeitpunkt(Timer *lpTimer){
  TimerReference = lpTimer;
  ui64Time=0;
}

Zeitpunkt::Zeitpunkt(const Zeitpunkt& source){
  TimerReference = source.TimerReference;
  ui64Time = source.ui64Time;
}

Zeitpunkt& Zeitpunkt::operator = (const uint64& src){
  ui64Time = src;
  return *this;
}

double Zeitpunkt::operator - (const Zeitpunkt& op2) const{
  #ifdef WIN32
  double dResolution=TimerReference->GetResolution();
  return static_cast<double>(ui64Time - op2.ui64Time) * dResolution * 1000;
  #else
  return static_cast<double>(ui64Time - op2.ui64Time) / 1000000.0;
  #endif
}

bool Zeitpunkt::operator > (const Zeitpunkt& op2) const
{
  return (ui64Time > op2.ui64Time);
}

void Zeitpunkt::SetTimerReference(Timer *lpTimer){
  TimerReference = lpTimer;
}
