/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "thEngine.h"
#include "valueeditor.h"
#include <sstream>
#include <vector>

ValueEditor::ValueEditor(){
  XmlConstantFile = NULL;
  Slider0 = NULL;
  Slider1 = NULL;
  Slider2 = NULL;
  Slider3 = NULL;
  Slider4 = NULL;
  Select1 = NULL;
  Text1 = NULL;
  ColorView1 = NULL;
  SaveButton = NULL;
}

void ValueEditorSlider0Change(void *Creator){
  ValueEditor* ptr = (ValueEditor*) Creator;
  int num = ptr->Select1->GetSelected();
  if(num != -1){
    ptr->Values[num].fValue = ptr->Slider0->GetValue();
    if(ptr->Values[num].Constant != NULL)
      ptr->Values[num].Constant->Set(&ptr->Values[num].fValue);
    ptr->Text1->PrintToCaption("Wert: %f",ptr->Values[num].fValue);
  }
}

void ValueEditorSlider1Change(void *Creator){
  ValueEditor* ptr = (ValueEditor*) Creator;
  int num = ptr->Select1->GetSelected();
  if(num != -1){
    ptr->ColorView1->SetRed(ptr->Slider1->GetValue());
    ptr->Values[num].v4Value = ptr->ColorView1->GetColor();
    if(ptr->Values[num].Constant != NULL)
      ptr->Values[num].Constant->Set(&ptr->Values[num].v4Value);
  }
}
void ValueEditorSlider2Change(void* Creator){
  ValueEditor* ptr = (ValueEditor*) Creator;
  int num = ptr->Select1->GetSelected();
  if(num != -1){
    ptr->ColorView1->SetGreen(ptr->Slider2->GetValue());
    ptr->Values[num].v4Value = ptr->ColorView1->GetColor();
    if(ptr->Values[num].Constant != NULL)
      ptr->Values[num].Constant->Set(&ptr->Values[num].v4Value);
  }
}
void ValueEditorSlider3Change(void* Creator){
  ValueEditor* ptr = (ValueEditor*) Creator;
  int num = ptr->Select1->GetSelected();
  if(num != -1){
    ptr->ColorView1->SetBlue(ptr->Slider3->GetValue());
    ptr->Values[num].v4Value = ptr->ColorView1->GetColor();
    if(ptr->Values[num].Constant != NULL)
      ptr->Values[num].Constant->Set(&ptr->Values[num].v4Value);
  }
}
void ValueEditorSlider4Change(void* Creator){
  ValueEditor* ptr = (ValueEditor*) Creator;
  int num = ptr->Select1->GetSelected();
  if(num != -1){
    ptr->ColorView1->SetAlpha(ptr->Slider4->GetValue());
    ptr->Values[num].v4Value = ptr->ColorView1->GetColor();
    if(ptr->Values[num].Constant != NULL)
      ptr->Values[num].Constant->Set(&ptr->Values[num].v4Value);
  }
}

void ValueEditorSelect1Change(void* Creator){
  ValueEditor* ptr = (ValueEditor*) Creator;
  ptr->Slider0->SetActive(false);
  ptr->Slider1->SetActive(false);
  ptr->Slider2->SetActive(false);
  ptr->Slider3->SetActive(false);
  ptr->Slider4->SetActive(false);
  ptr->Text1->SetActive(false);
  ptr->ColorView1->SetActive(false);
  int num = ptr->Select1->GetSelected();
  if(num < 0)
    return;
  switch(ptr->Values[num].type){
    case 0: //float
      ptr->Slider0->SetActive(true);
      ptr->Slider0->SetMinMax(ptr->Values[num].fMin,ptr->Values[num].fMax);
      ptr->Slider0->SetValue(ptr->Values[num].fValue);
      ptr->Text1->SetActive(true);
      ptr->Text1->PrintToCaption("Wert: %f",ptr->Values[num].fValue);
      break;

    case 1: //vec4
      ptr->Slider1->SetActive(true);
      ptr->Slider1->SetMinMax(ptr->Values[num].v4Min.x,ptr->Values[num].v4Max.x);
      ptr->Slider1->SetValue(ptr->Values[num].v4Value.x);
      ptr->Slider2->SetActive(true);
      ptr->Slider2->SetMinMax(ptr->Values[num].v4Min.y,ptr->Values[num].v4Max.y);
      ptr->Slider2->SetValue(ptr->Values[num].v4Value.y);
      ptr->Slider3->SetActive(true);
      ptr->Slider3->SetMinMax(ptr->Values[num].v4Min.z,ptr->Values[num].v4Max.z);
      ptr->Slider3->SetValue(ptr->Values[num].v4Value.z);
      ptr->Slider4->SetActive(true);
      ptr->Slider4->SetMinMax(ptr->Values[num].v4Min.w,ptr->Values[num].v4Max.w);
      ptr->Slider4->SetValue(ptr->Values[num].v4Value.w);
      ptr->ColorView1->SetActive(true);
      ptr->ColorView1->SetColor(ptr->Values[num].v4Value);
      break;

    default:
      break;
  }
}

void ValueEditorSaveXml(void *Creator){
  ValueEditor* ptr = (ValueEditor*) Creator;
  for(int i=0;i<ptr->Constants.size();i++){
    switch(ptr->Values[ptr->Constants[i].num].type){
      case 0:
        ptr->XmlSetFloat(ptr->Constants[i].lpNode,ptr->Values[ptr->Constants[i].num].fValue);
        break;
      case 1:
        ptr->XmlSetColor(ptr->Constants[i].lpNode,ptr->Values[ptr->Constants[i].num].v4Value);
        break;
    }
  }

  if(!ptr->XmlConstantFile->SaveFile())
    ThErrorMessage("Fehler Beim XML speichern",ptr->XmlConstantFile->ErrorDesc());
}

bool ValueEditor::Init(Gui *pGui){
  ValueEditorWindow = pGui->CreateWindow(NULL,20,20,350,162,false,0,"Value Editor",this);
  Slider0 = pGui->CreateSliderVertical(ValueEditorWindow,204,2,100,0.0f,1.0f,0,0,1,"V",this);
  Slider0->SetOnChange(ValueEditorSlider0Change);
  Slider0->SetActive(false);
  Slider1 = pGui->CreateSliderVertical(ValueEditorWindow,204,2,100,0.0f,1.0f,0,0,1,"R",this);
  Slider1->SetOnChange(ValueEditorSlider1Change);
  Slider1->SetActive(false);
  Slider2 = pGui->CreateSliderVertical(ValueEditorWindow,224,2,100,0.0f,1.0f,0,0,1,"G",this);
  Slider2->SetOnChange(ValueEditorSlider2Change);
  Slider2->SetActive(false);
  Slider3 = pGui->CreateSliderVertical(ValueEditorWindow,244,2,100,0.0f,1.0f,0,0,1,"B",this);
  Slider3->SetOnChange(ValueEditorSlider3Change);
  Slider3->SetActive(false);
  Slider4 = pGui->CreateSliderVertical(ValueEditorWindow,264,2,100,0.0f,1.0f,0,0,1,"A",this);
  Slider4->SetOnChange(ValueEditorSlider4Change);
  Slider4->SetActive(false);
  ColorView1 = pGui->CreateColorView(ValueEditorWindow,284,30,2,this);
  ColorView1->SetActive(false);
  Text1 = pGui->CreateText(ValueEditorWindow,224,8,1,"Value",this);
  Text1->SetActive(false);

  ThGuiSelectInit SelectInit;
  SelectInit.MaxElementsVisible = 5;
  SelectInit.Width = 200;
  SelectInit.Space = 1;
  SelectInit.Style = 0;
  SelectInit.ButtonUpStyle = 3;
  SelectInit.ButtonUpHoverStyle = 4;
  SelectInit.ButtonDownStyle = 5;
  SelectInit.ButtonDownHoverStyle = 6;
  SelectInit.SelectStyle = 1;
  SelectInit.SelectHoverStyle = 2;
  SelectInit.SelectSelectedStyle = 3;

  Select1 = pGui->CreateSelect(ValueEditorWindow,4,4,SelectInit,this);
  if(Select1 == NULL)
    return false;
  Select1->SetOnChange(ValueEditorSelect1Change);
  SaveButton = pGui->CreateButton(ValueEditorWindow,4,120,200,20,0,1,"Aenderungen speichern",this);
  SaveButton->SetOnPress(ValueEditorSaveXml);

  return true;
}

int ValueEditor::AddFloat(const char *name, ShaderConstant *ptr, float Min, float Max){
  ValueEditorValue TempValue;
  TempValue.type = 0;
  TempValue.fValue = Min;
  TempValue.fMin = Min;
  TempValue.fMax = Max;
  TempValue.Constant = ptr;
  Values.push_back(TempValue);
  Select1->AddOption(name);
  return Values.size()-1;
}

int ValueEditor::AddVec4(const char *name, ShaderConstant *ptr, vec4 Min, vec4 Max){
  ValueEditorValue TempValue;
  TempValue.type = 1;
  TempValue.v4Value = Min;
  TempValue.v4Min = Min;
  TempValue.v4Max = Max;
  TempValue.Constant = ptr;
  Values.push_back(TempValue);
  Select1->AddOption(name);
  return Values.size()-1;
}

bool ValueEditor::XmlLoadFloat(TiXmlNode *pChild, float& value){
  std::ostringstream ErrorMessage;
  TiXmlElement *pElement;
  TiXmlAttribute *pAttrib;
  bool bv=false;
  if(pChild->Type() != TiXmlNode::ELEMENT){
    ErrorMessage << "Node '" << pChild->Value() << "' ist kein gültiges Element für die Floateinlesung";
    ThErrorMessage("Value Editor Error",ErrorMessage.str().c_str());
    return false;
  }
  pElement = pChild->ToElement();
  for(pAttrib = pElement->FirstAttribute();pAttrib != 0; pAttrib = pAttrib->Next()){
    if(strcmp("v",pAttrib->Name())==0){
      value = atof(pAttrib->Value());
      bv = true;
    }
    else{
      ErrorMessage << "Unbekanntes Attribut '" << pAttrib->Name() << "' in einem " << pChild->Value() << "-Element";
      ThErrorMessage("Value Editor Error",ErrorMessage.str().c_str());
      return false;
    }
  }

  if(!bv){
    ErrorMessage << "Value (v) Attribut für Farbeinlesung ein einem " << pChild->Value() << "-Tag ist nicht vorhanden!";
    ThErrorMessage("Value Editor Error",ErrorMessage.str().c_str());
    return false;
  }

  return true;
}

bool ValueEditor::LoadConstants(const char *filename){
  std::ostringstream ErrorMessage;
  //Filename vaild?
  FILE *datei;
  datei = fopen(filename,"r");
  if(!datei){
    ErrorMessage << "Konnte Datei '" << filename << "' nicht öffnen!";
    ThErrorMessage("Value Editor Error",ErrorMessage.str().c_str());
    return false;
  }
  fclose(datei);

  XmlConstantFile = new TiXmlDocument(filename);
  bool bLoadOk = XmlConstantFile->LoadFile();
  if(!bLoadOk){
    ErrorMessage << "Fehler beim Laden von '" << filename << "': " << XmlConstantFile->ErrorDesc();
    ThErrorMessage("Value Editor Error",ErrorMessage.str().c_str());
    return false;
  }

  TiXmlNode *pChild,*pChild2,*pChild3;
  TiXmlElement *pElement;
  TiXmlAttribute *pAttrib;
  bool bname=false,bvalue=false;
  ValueEditorValue TempValue,*lpTempValue;
  ShaderConstant *lpTempConstant;
  ThConstantInfoStruct TempConstantInfo;
  std::string intern,display;


  pChild = XmlConstantFile->FirstChild()->NextSibling();

  if(pChild->Type() != TiXmlNode::ELEMENT){
    ErrorMessage << "Ungültige Style Datei '" << filename << "': constants-Tag nicht am Begin der Datei gefunden";
    ThErrorMessage("Value Editor Error",ErrorMessage.str().c_str());
    return false;
  }
  else if(strcmp(pChild->Value(),"constants")!=0){
    ErrorMessage << "Falscher Tag am Begin der Datei '" << filename << "' gefunden: '" << pChild->Value() << "'";
    ThErrorMessage("Value Editor Error",ErrorMessage.str().c_str());
    return false;
  }

  for(pChild = pChild->FirstChild(); pChild != 0; pChild = pChild->NextSibling()){
    if(pChild->Type() == TiXmlNode::ELEMENT){
      if(strcmp("vec4",pChild->Value())==0){
        bname=bvalue=false;
        TempValue.v4Min.set(0.0f); TempValue.v4Max.set(1.0f);
        for(pChild2 = pChild->FirstChild(); pChild2 != 0; pChild2 = pChild2->NextSibling()){
          if(pChild2->Type() == TiXmlNode::ELEMENT){
            if(strcmp("name",pChild2->Value())==0){
              bname = true;
              bool bintern=false,bdisplay=false;
              pElement = pChild2->ToElement();
              for(pAttrib = pElement->FirstAttribute(); pAttrib != 0; pAttrib = pAttrib->Next()){
                if(strcmp("intern",pAttrib->Name())==0){
                  bintern=true;
                  intern = pAttrib->Value();
                }
                else if(strcmp("display",pAttrib->Name())==0){
                  bdisplay=true;
                  display = pAttrib->Value();
                }
                else {
                  ErrorMessage << "Unbekanntes Attribut '" << pAttrib->Name() << "' in einem " << pChild2->Value() << "-Tag";
                  ThErrorMessage("Value Editor Error",ErrorMessage.str().c_str());
                  return false;
                }
              }

              if(!bintern){
                ThErrorMessage("Value Editor Error","intern Attribut eines Name-Tags ist nicht vorhanden!");
                return false;
              }

              if(!bdisplay){
                ThErrorMessage("Value Editor Error","display Attribut eines Name-Tags ist nicht vorhanden!");
                return false;
              }
            }
            else if(strcmp("value",pChild2->Value())==0){
              if(!XmlLoadColor(pChild2,TempValue.v4Value))
                return false;
              bvalue=true;
              TempConstantInfo.lpNode = pChild2;
            }
            else if(strcmp("clampfrom",pChild2->Value())==0){
              if(!XmlLoadColor(pChild2,TempValue.v4Min))
                return false;
            }
            else if(strcmp("clampto",pChild2->Value())==0){
              if(!XmlLoadColor(pChild2,TempValue.v4Max))
                return false;
            }
            else {
              ErrorMessage << "Unbekanntes Element '" << pChild2->Value() << "' in einem vec4-Tag";
              ThErrorMessage("Value Editor Error",ErrorMessage.str().c_str());
              return false;
            }
          }
        }

        if(!bname){
          ThErrorMessage("Value Editor Error","name-Tag innerhalb eines vec4-Tags ist nicht vorhanden!");
          return false;
        }
        if(!bvalue){
          ThErrorMessage("Value Editor Error","value-Tag innerhalb eines vec4-Tags ist nicht vorhanden!");
          return false;
        }

        //Add to Arrays
        lpTempConstant = XMLShader::AddShaderConstantVec4(intern.c_str());
        lpTempConstant->Set(&TempValue.v4Value);
        TempConstantInfo.num = AddVec4(display.c_str(),lpTempConstant,TempValue.v4Min,TempValue.v4Max);
        Values[TempConstantInfo.num].v4Value = TempValue.v4Value;
        Constants.push_back(TempConstantInfo);
      }
      else if(strcmp("float",pChild->Value())==0){
        bname=bvalue=false;
        TempValue.fMin=0.0f; TempValue.fMax=1.0f;
        for(pChild2 = pChild->FirstChild(); pChild2 != 0; pChild2 = pChild2->NextSibling()){
          if(pChild2->Type() == TiXmlNode::ELEMENT){
            if(strcmp("name",pChild2->Value())==0){
              bname = true;
              bool bintern=false,bdisplay=false;
              pElement = pChild2->ToElement();
              for(pAttrib = pElement->FirstAttribute(); pAttrib != 0; pAttrib = pAttrib->Next()){
                if(strcmp("intern",pAttrib->Name())==0){
                  bintern=true;
                  intern = pAttrib->Value();
                }
                else if(strcmp("display",pAttrib->Name())==0){
                  bdisplay=true;
                  display = pAttrib->Value();
                }
                else {
                  ErrorMessage << "Unbekanntes Attribut '" << pAttrib->Name() << "' in einem " << pChild2->Value() << "-Tag";
                  ThErrorMessage("Value Editor Error",ErrorMessage.str().c_str());
                  return false;
                }
              }

              if(!bintern){
                ThErrorMessage("Value Editor Error","intern Attribut eines Name-Tags ist nicht vorhanden!");
                return false;
              }

              if(!bdisplay){
                ThErrorMessage("Value Editor Error","display Attribut eines Name-Tags ist nicht vorhanden!");
                return false;
              }
            }
            else if(strcmp("value",pChild2->Value())==0){
              if(!XmlLoadFloat(pChild2,TempValue.fValue))
                return false;
              bvalue=true;
              TempConstantInfo.lpNode = pChild2;
            }
            else if(strcmp("clampfrom",pChild2->Value())==0){
              if(!XmlLoadFloat(pChild2,TempValue.fMin))
                return false;
            }
            else if(strcmp("clampto",pChild2->Value())==0){
              if(!XmlLoadFloat(pChild2,TempValue.fMax))
                return false;
            }
            else {
              ErrorMessage << "Unbekanntes Element '" << pChild2->Value() << "' in einem float-Tag";
              ThErrorMessage("Value Editor Error",ErrorMessage.str().c_str());
              return false;
            }
          }
        }

        if(!bname){
          ThErrorMessage("Value Editor Error","name-Tag innerhalb eines vec4-Tags ist nicht vorhanden!");
          return false;
        }
        if(!bvalue){
          ThErrorMessage("Value Editor Error","value-Tag innerhalb eines vec4-Tags ist nicht vorhanden!");
          return false;
        }

        //Add to Arrays
        lpTempConstant = XMLShader::AddShaderConstantFloat(intern.c_str());
        lpTempConstant->Set(&TempValue.fValue);
        TempConstantInfo.num = AddFloat(display.c_str(),lpTempConstant,TempValue.fMin,TempValue.fMax);
        Values[TempConstantInfo.num].fValue = TempValue.fValue;
        Constants.push_back(TempConstantInfo);
      }
    }
  }
  return true;
}

void ValueEditor::XmlSetColor(TiXmlNode *pChild,vec4& Color){
  if(pChild->Type() != TiXmlNode::ELEMENT)
    return;

  TiXmlElement *pElement = pChild->ToElement();
  TiXmlAttribute *pAttrib;
  for(pAttrib = pElement->FirstAttribute();pAttrib != 0; pAttrib = pAttrib->Next()){
    if(strcmp("r",pAttrib->Name())==0){
      pAttrib->SetDoubleValue(Color.x);
    }
    else if(strcmp("g",pAttrib->Name())==0){
      pAttrib->SetDoubleValue(Color.y);
    }
    else if(strcmp("b",pAttrib->Name())==0){
      pAttrib->SetDoubleValue(Color.z);
    }
    else if(strcmp("a",pAttrib->Name())==0){
      pAttrib->SetDoubleValue(Color.w);
    }
  }
}

void ValueEditor::XmlSetFloat(TiXmlNode *pChild,float& value){
  if(pChild->Type() != TiXmlNode::ELEMENT)
    return;

  TiXmlElement *pElement = pChild->ToElement();
  TiXmlAttribute *pAttrib;
  for(pAttrib = pElement->FirstAttribute();pAttrib != 0; pAttrib = pAttrib->Next()){
    if(strcmp("v",pAttrib->Name())==0){
      pAttrib->SetDoubleValue(value);
    }
  }
}
