/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "vektoren.h"

//Vektorklasse
#ifdef __SSE__
void* vec4::operator new (size_t size){
  void *pMem = th_memalign(16,size);
  return pMem;
}

void* vec4::operator new[] (size_t size){
  void *pMem = th_memalign(16,size);
  return pMem;
}

void vec4::operator delete (void *pMem){
  th_free(pMem);
}

void vec4::operator delete[] (void *pMem){
  th_free(pMem);
}
#endif

///////////// Matrix Klasse //////////////////////
#ifdef __SSE__
void* mat4a::operator new (size_t size){
  void *pMem = th_memalign(16,size);
  return pMem;
}

void* mat4a::operator new[] (size_t size){
  void *pMem = th_memalign(16,size);
  return pMem;
}

void mat4a::operator delete (void *pMem){
  th_free(pMem);
}

void mat4a::operator delete[] (void *pMem){
  th_free(pMem);
}
#endif


mat4a mat4a::operator * (const mat4a& m) const {
  #ifdef __SSE__
  vec4 x,y,z,w;
  mat4a result;
  //1 Spalte
  for(int i=0;i<4;i++){
    x.m = _mm_set_ps(this->f[i*4],this->f[i*4],this->f[i*4],this->f[i*4]);
    y.m = _mm_set_ps(this->f[i*4+1],this->f[i*4+1],this->f[i*4+1],this->f[i*4+1]);
    z.m = _mm_set_ps(this->f[i*4+2],this->f[i*4+2],this->f[i*4+2],this->f[i*4+2]);
    w.m = _mm_set_ps(this->f[i*4+3],this->f[i*4+3],this->f[i*4+3],this->f[i*4+3]);
    x.m = _mm_mul_ps(m.m[0],x.m);
    y.m = _mm_mul_ps(m.m[1],y.m);
    z.m = _mm_mul_ps(m.m[2],z.m);
    w.m = _mm_mul_ps(m.m[3],w.m);
    result.m[i] = _mm_add_ps(x.m,y.m);
    result.m[i] = _mm_add_ps(result.m[i],z.m);
    result.m[i] = _mm_add_ps(result.m[i],w.m);
  }
  return result;
  #else
  mat4a result;
  for(int i=0;i<4;i++){
    result.f[i*4]   = m.f[0] * this->f[i*4] + m.f[4] * this->f[i*4+1] + m.f[ 8] * this->f[i*4+2] + m.f[12] * this->f[i*4+3];
    result.f[i*4+1] = m.f[1] * this->f[i*4] + m.f[5] * this->f[i*4+1] + m.f[ 9] * this->f[i*4+2] + m.f[13] * this->f[i*4+3];
    result.f[i*4+2] = m.f[2] * this->f[i*4] + m.f[6] * this->f[i*4+1] + m.f[10] * this->f[i*4+2] + m.f[14] * this->f[i*4+3];
    result.f[i*4+3] = m.f[3] * this->f[i*4] + m.f[7] * this->f[i*4+1] + m.f[11] * this->f[i*4+2] + m.f[15] * this->f[i*4+3];
  }
  return result;
  #endif
}

vec4 mat4a::operator * (const vec4& v) const {
#ifdef __SSE__
  vec4 v1,v2,v3,v4,res;
  v1.m = _mm_set_ps(v.x,v.x,v.x,v.x);
  v2.m = _mm_set_ps(v.y,v.y,v.y,v.y);
  v3.m = _mm_set_ps(v.z,v.z,v.z,v.z);
  v4.m = _mm_set_ps(v.w,v.w,v.w,v.w);
  v1.m = _mm_mul_ps(v1.m,this->m[0]);
  v2.m = _mm_mul_ps(v2.m,this->m[1]);
  v3.m = _mm_mul_ps(v3.m,this->m[2]);
  v4.m = _mm_mul_ps(v4.m,this->m[3]);
  res.m = _mm_add_ps(v1.m,v2.m);
  res.m = _mm_add_ps(res.m,v3.m);
  res.m = _mm_add_ps(res.m,v4.m);
  return res;
#else
  vec4 temp;
  temp.x = v.x * this->f[0] + v.y * this->f[4] + v.z * this->f[8]  + v.w * this->f[12];
  temp.y = v.x * this->f[1] + v.y * this->f[5] + v.z * this->f[9]  + v.w * this->f[13];
  temp.z = v.x * this->f[2] + v.y * this->f[6] + v.z * this->f[10] + v.w * this->f[14];
  temp.w = v.x * this->f[3] + v.y * this->f[7] + v.z * this->f[11] + v.w * this->f[15];
  return temp;
#endif
}

///////////// Direction V4 ///////////////////////
#ifdef __SSE__
vec4 DirectionV4(const vec4& a, const vec4& b){
  float length;
  vec4 v1,v2;
  v2.m = _mm_sub_ps(b.m,a.m);
  v1.m = _mm_mul_ps(v2.m,v2.m);
  length = sqrt(v1.x+v1.y+v1.z);
  if(length != 0){
    v1.m = _mm_set_ps(length,length,length,length);
    v2.m = _mm_div_ps(v2.m,v1.m);
  }
  return v2;
}
#else
vec4 DirectionV4(const vec4& a, const vec4& b){
  vec4 res;
  res.x = b.x - a.x;
  res.y = b.y - a.y;
  res.z = b.z - a.z;
  res=NormalizeV4(res);
  return res;
}
#endif

/////////// Extrude V4 ////////////////////
#ifdef __SSE__
vec4 ExtrudeV4(const vec4& v1, const vec4& v2, const float& fDistance){
  float length;
  vec4 a,b;
  a.m = _mm_sub_ps(v2.m,v1.m);
  b.m = _mm_mul_ps(a.m,a.m);
  length = sqrt(b.x+b.y+b.z);
  if(length != 0){
    b.m = _mm_set_ps(length,length,length,length);
    a.m = _mm_div_ps(a.m,b.m);
  }
  b.x = a.x * fDistance;
  b.y = a.y * fDistance;
  b.z = a.z * fDistance;
  return b;
}
#else
vec4 ExtrudeV4(const vec4& v1, const vec4& v2, const float& fDistance){
  vec4 a,b;
  a.x = v2.x - v1.x;
  a.y = v2.y - v1.y;
  a.z = v2.z - v1.z;
  a=NormalizeV4(a);
  b.x = a.x * fDistance;
  b.y = a.y * fDistance;
  b.z = a.z * fDistance;
  return b;
}
#endif

///////////// Compute Normal V4 ////////////
#ifdef __SSE__
vec4 ComputeNormalV4(const vec4& t1, const vec4& t2, const vec4& t3){
  vec4 normal,v1,v2,v3;
  float length;
  //v2.m = _mm_sub_ps(t1.m,t2.m);
  //v3.m = _mm_sub_ps(t1.m,t3.m);
  v2 = t1 - t2;
  v3 = t1 - t3;
  //normal
  normal.x = v2.y * v3.z - v2.z * v3.y;
  normal.y = v2.z * v3.x - v2.x * v3.z;
  normal.z = v2.x * v3.y - v2.y * v3.x;

  v1.m = _mm_mul_ps(normal.m,normal.m);
  length = sqrt(v1.x + v1.y + v1.z);
  if(length != 0){
    v1.m = _mm_set_ps(length,length,length,length);
    normal.m = _mm_div_ps(normal.m,v1.m);
  }
  normal.w = 1.0f;
  return normal;
}
#else
vec4 ComputeNormalV4(const vec4& t1, const vec4& t2, const vec4& t3){
  vec4 normal,v2,v3;
  //V1
  v2.x=t1.x-t2.x; v2.y=t1.y-t2.y; v2.z=t1.z-t2.z;
  //V2
  v3.x=t1.x-t3.x; v3.y=t1.y-t3.y; v3.z=t1.z-t3.z;
  //normal
  normal.x = v2.y * v3.z - v2.z * v3.y;
  normal.y = v2.z * v3.x - v2.x * v3.z;
  normal.z = v2.x * v3.y - v2.y * v3.x;
  normal.w = 1.0f;
  normal = NormalizeV4(normal);
  return normal;
}
#endif

vec4 CrossProductV4(const vec4& v1, const vec4& v2){
  vec4 out;
  out.x = v1.y * v2.z - v1.z * v2.y;
  out.y = v1.z * v2.x - v1.x * v2.z;
  out.z = v1.x * v2.y - v1.y * v2.x;
  out.w = 1.0f;
  return out;
}

mat4a AlignMatrix4(float *matrix){
  mat4a m;
  int x,y;
  for(y=0;y<4;y++)
    for(x=0;x<4;x++)
      m.f[x*4+y]=matrix[y*4+x];
  return m;
}

void DealignMatrix4(const mat4a& src, float *dst){
  int x,y;
  for(y=0;y<4;y++)
    for(x=0;x<4;x++)
      dst[x*4+y]=src.f[y*4+x];
}

////////// MulMat4a ///////////
#ifdef __SSE__
vec4 Mul(const mat4a& matrix,const vec4& vektor){
  vec4 v1,v2,v3,v4,res;
  v1.m = _mm_set_ps(vektor.x,vektor.x,vektor.x,vektor.x);
  v2.m = _mm_set_ps(vektor.y,vektor.y,vektor.y,vektor.y);
  v3.m = _mm_set_ps(vektor.z,vektor.z,vektor.z,vektor.z);
  v4.m = _mm_set_ps(vektor.w,vektor.w,vektor.w,vektor.w);
  v1.m = _mm_mul_ps(v1.m,matrix.m[0]);
  v2.m = _mm_mul_ps(v2.m,matrix.m[1]);
  v3.m = _mm_mul_ps(v3.m,matrix.m[2]);
  v4.m = _mm_mul_ps(v4.m,matrix.m[3]);
  res.m = _mm_add_ps(v1.m,v2.m);
  res.m = _mm_add_ps(res.m,v3.m);
  res.m = _mm_add_ps(res.m,v4.m);
  return res;
}
#else
vec4 Mul(const mat4a& matrix,const vec4& vektor){
  vec4 temp;
  temp.x = vektor.x * matrix.f[0] + vektor.y * matrix.f[4] + vektor.z * matrix.f[8]  + vektor.w * matrix.f[12];
  temp.y = vektor.x * matrix.f[1] + vektor.y * matrix.f[5] + vektor.z * matrix.f[9]  + vektor.w * matrix.f[13];
  temp.z = vektor.x * matrix.f[2] + vektor.y * matrix.f[6] + vektor.z * matrix.f[10] + vektor.w * matrix.f[14];
  temp.w = vektor.x * matrix.f[3] + vektor.y * matrix.f[7] + vektor.z * matrix.f[11] + vektor.w * matrix.f[15];
  return temp;
}
#endif

#ifdef __SSE__
mat4a Mul(const mat4a& m2, const mat4a& m1){
  vec4 x,y,z,w;
  mat4a result;
  //1 Spalte
  for(int i=0;i<4;i++){
    x.m = _mm_set_ps(m2.f[i*4],m2.f[i*4],m2.f[i*4],m2.f[i*4]);
    y.m = _mm_set_ps(m2.f[i*4+1],m2.f[i*4+1],m2.f[i*4+1],m2.f[i*4+1]);
    z.m = _mm_set_ps(m2.f[i*4+2],m2.f[i*4+2],m2.f[i*4+2],m2.f[i*4+2]);
    w.m = _mm_set_ps(m2.f[i*4+3],m2.f[i*4+3],m2.f[i*4+3],m2.f[i*4+3]);
    x.m = _mm_mul_ps(m1.m[0],x.m);
    y.m = _mm_mul_ps(m1.m[1],y.m);
    z.m = _mm_mul_ps(m1.m[2],z.m);
    w.m = _mm_mul_ps(m1.m[3],w.m);
    result.m[i] = _mm_add_ps(x.m,y.m);
    result.m[i] = _mm_add_ps(result.m[i],z.m);
    result.m[i] = _mm_add_ps(result.m[i],w.m);
  }
  return result;
}
#else
mat4a Mul(const mat4a& m2, const mat4a& m1){
  mat4a result;
  for(int i=0;i<4;i++){
    result.f[i*4]   = m1.f[0] * m2.f[i*4] + m1.f[4] * m2.f[i*4+1] + m1.f[ 8] * m2.f[i*4+2] + m1.f[12] * m2.f[i*4+3];
    result.f[i*4+1] = m1.f[1] * m2.f[i*4] + m1.f[5] * m2.f[i*4+1] + m1.f[ 9] * m2.f[i*4+2] + m1.f[13] * m2.f[i*4+3];
    result.f[i*4+2] = m1.f[2] * m2.f[i*4] + m1.f[6] * m2.f[i*4+1] + m1.f[10] * m2.f[i*4+2] + m1.f[14] * m2.f[i*4+3];
    result.f[i*4+3] = m1.f[3] * m2.f[i*4] + m1.f[7] * m2.f[i*4+1] + m1.f[11] * m2.f[i*4+2] + m1.f[15] * m2.f[i*4+3];
  }
  return result;
}
#endif

const mat4a RotationMatrixXYZ(const vec3& v3Rotation){
  mat4a result;
  float A,B,C,D,E,F,AD,BD;

  A       = cos(v3Rotation.x/(-180.0f) * M_PI);
  B       = sin(v3Rotation.x/(-180.0f) * M_PI);
  C       = cos(v3Rotation.y/(-180.0f) * M_PI);
  D       = sin(v3Rotation.y/(-180.0f) * M_PI);
  E       = cos(v3Rotation.z/(-180.0f) * M_PI);
  F       = sin(v3Rotation.z/(-180.0f) * M_PI);
  AD      =   A * D;
  BD      =   B * D;
  result.f[0]  =   C * E;
  result.f[4]  =  -C * F;
  result.f[8]  =   D;
  result.f[1]  =  BD * E + A * F;
  result.f[5]  = -BD * F + A * E;
  result.f[9]  =  -B * C;
  result.f[2]  = -AD * E + B * F;
  result.f[6]  =  AD * F + B * E;
  result.f[10] =   A * C;
  result.f[3]  =  result.f[7] = result.f[11] = result.f[12] = result.f[13] = result.f[14] = 0;
  result.f[15] =  1;
  return result;
}

const mat4a RotationMatrixXYZ(float x, float y, float z){
  mat4a result;
  float A,B,C,D,E,F,AD,BD;

  A       = cos(x/(-180.0f) * M_PI);
  B       = sin(x/(-180.0f) * M_PI);
  C       = cos(y/(-180.0f) * M_PI);
  D       = sin(y/(-180.0f) * M_PI);
  E       = cos(z/(-180.0f) * M_PI);
  F       = sin(z/(-180.0f) * M_PI);
  AD      =   A * D;
  BD      =   B * D;
  result.f[0]  =   C * E;
  result.f[4]  =  -C * F;
  result.f[8]  =   D;
  result.f[1]  =  BD * E + A * F;
  result.f[5]  = -BD * F + A * E;
  result.f[9]  =  -B * C;
  result.f[2]  = -AD * E + B * F;
  result.f[6]  =  AD * F + B * E;
  result.f[10] =   A * C;
  result.f[3]  =  result.f[7] = result.f[11] = result.f[12] = result.f[13] = result.f[14] = 0;
  result.f[15] =  1;
  return result;
}

const mat4a TranslationMatrix(const vec4& Translation){
  mat4a result;
  result.f[12] = Translation.x;
  result.f[13] = Translation.y;
  result.f[14] = Translation.z;
  result.f[15] = Translation.w;
  result.f[0] = result.f[5] = result.f[10] = 1.0f;
  result.f[4] = result.f[8] = result.f[1] = result.f[9] = result.f[2] = result.f[6] = result.f[3] = result.f[7] = result.f[11] = 0.0f;
  return result;
}

const mat4a TranslationMatrix(float x, float y, float z){
  mat4a result;
  result.f[12] = x;
  result.f[13] = y;
  result.f[14] = z;
  result.f[0] = result.f[5] = result.f[10] = result.f[15] = 1.0f;
  result.f[4] = result.f[8] = result.f[1] = result.f[9] = result.f[2] = result.f[6] = result.f[3] = result.f[7] = result.f[11] = 0.0f;
  return result;
}

const mat4a ScaleMatrix(const vec3& v3Scale){
  mat4a result;
  result.f[0] = v3Scale.x;
  result.f[5] = v3Scale.y;
  result.f[10] = v3Scale.z;
  result.f[15] = 1.0f;
  result.f[1] = result.f[2] = result.f[3] = result.f[4] = result.f[6] = result.f[7] = result.f[8] = result.f[9] = result.f[11] = result.f[12] = result.f[13] = result.f[14] = 0.0f;
  return result;
}

const mat4a ScaleMatrix(float x, float y, float z){
  mat4a result;
  result.f[0] = x;
  result.f[5] = y;
  result.f[10] = z;
  result.f[15] = 1.0f;
  result.f[1] = result.f[2] = result.f[3] = result.f[4] = result.f[6] = result.f[7] = result.f[8] = result.f[9] = result.f[11] = result.f[12] = result.f[13] = result.f[14] = 0.0f;
  return result;
}

//Inverse Matrix Functions
mat3a IdentityM3(){
  mat3a mat;
  mat.f[0]=1.0f; mat.f[1]=0.0f; mat.f[2]=0.0f;
  mat.f[3]=0.0f; mat.f[4]=1.0f; mat.f[5]=0.0f;
  mat.f[6]=0.0f; mat.f[7]=0.0f; mat.f[8]=1.0f;
  return mat;
}

mat4a IdentityM4(){
  mat4a mat;
  mat.f[ 0]=1.0f; mat.f[ 1]=0.0f; mat.f[ 2]=0.0f; mat.f[ 3]=0.0f;
  mat.f[ 4]=0.0f; mat.f[ 5]=1.0f; mat.f[ 6]=0.0f; mat.f[ 7]=0.0f;
  mat.f[ 8]=0.0f; mat.f[ 9]=0.0f; mat.f[10]=1.0f; mat.f[11]=0.0f;
  mat.f[12]=0.0f; mat.f[13]=0.0f; mat.f[14]=0.0f; mat.f[15]=1.0f;
  return mat;
}

float Det(const mat3a& mat){
  float det = mat.f[0] * ( mat.f[4]*mat.f[8] - mat.f[7]*mat.f[5] )
            - mat.f[1] * ( mat.f[3]*mat.f[8] - mat.f[6]*mat.f[5] )
            + mat.f[2] * ( mat.f[3]*mat.f[7] - mat.f[6]*mat.f[4] );
  return det;
}

mat3a Inverse(const mat3a& ma)
     {
     float det = Det(ma);
     mat3a mr;
     if ( fabs( det ) < 0.0005 )
       return IdentityM3();

     mr.f[0] =    ma.f[4]*ma.f[8] - ma.f[5]*ma.f[7]   / det;
     mr.f[1] = -( ma.f[1]*ma.f[8] - ma.f[7]*ma.f[2] ) / det;
     mr.f[2] =    ma.f[1]*ma.f[5] - ma.f[4]*ma.f[2]   / det;
     mr.f[3] = -( ma.f[3]*ma.f[8] - ma.f[5]*ma.f[6] ) / det;
     mr.f[4] =    ma.f[0]*ma.f[8] - ma.f[6]*ma.f[2]   / det;
     mr.f[5] = -( ma.f[0]*ma.f[5] - ma.f[3]*ma.f[2] ) / det;
     mr.f[6] =    ma.f[3]*ma.f[7] - ma.f[6]*ma.f[4]   / det;
     mr.f[7] = -( ma.f[0]*ma.f[7] - ma.f[6]*ma.f[1] ) / det;
     mr.f[8] =    ma.f[0]*ma.f[4] - ma.f[1]*ma.f[3]   / det;
     return mr;
}

mat3a Submat( const mat4a& mr, int i, int j ) {
  mat3a mb;
  int di, dj, si, sj;
  // loop through 3x3 submatrix
  for( di = 0; di < 3; di ++ ) {
    for( dj = 0; dj < 3; dj ++ ) {
	  // map 3x3 element (destination) to 4x4 element (source)
	  si = di + ( ( di >= i ) ? 1 : 0 );
	  sj = dj + ( ( dj >= j ) ? 1 : 0 );
	  // copy element
	  mb.f[di * 3 + dj] = mr.f[si * 4 + sj];
    }
  }
  return mb;
}

float Det( const mat4a& mr )
{
  float  det, result = 0, i = 1;
  mat3a msub3;
  int     n;
  for ( n = 0; n < 4; n++, i *= -1 )
  {
    msub3 = Submat( mr, 0, n );
    det     = Det( msub3 );
    result += mr.f[n] * det * i;
  }
  return result;
}

mat4a Inverse( const mat4a& ma ){
  mat4a mr;
  float  mdet = Det( ma );
  mat3a mtemp;
  int  i, j, sign;
  if ( fabs( mdet ) < 0.0005 )
	return IdentityM4();
  for ( i = 0; i < 4; i++ )
    for ( j = 0; j < 4; j++ ){
      sign = 1 - ( (i +j) % 2 ) * 2;
      mtemp = Submat( ma, i, j );
      mr.f[i+j*4] = ( Det( mtemp ) * sign ) / mdet;
    }
  return mr;
}

mat4a Transponse(const mat4a& m){
  mat4a mr;
  mr.f[0] = m.f[0];
  mr.f[1] = m.f[4];
  mr.f[2] = m.f[8];
  mr.f[3] = m.f[12];
  mr.f[4] = m.f[1];
  mr.f[5] = m.f[5];
  mr.f[6] = m.f[9];
  mr.f[7] = m.f[13];
  mr.f[8] = m.f[2];
  mr.f[9] = m.f[6];
  mr.f[10] = m.f[10];
  mr.f[11] = m.f[14];
  mr.f[12] = m.f[3];
  mr.f[13] = m.f[7];
  mr.f[14] = m.f[11];
  mr.f[15] = m.f[15];
  return mr;
}

mat4a NormalMatrix(const mat4a& m){
  mat4a mr = m;
  mr.f[3] = 0;
  mr.f[7] = 0;
  mr.f[11] = 0;
  mr.f[12] = 0;
  mr.f[13] = 0;
  mr.f[14] = 0;
  mr.f[15] = 1.0f;
  mr = Inverse(mr);
  mr = Transponse(mr);
  return mr;
}

mat4a Right2Left(const mat4a& m){
  mat4a mr = m;
  for(int i=0;i<4;i++){
    mr.f[4+i] = m.f[8+i];// * -1.0f;
    mr.f[8+i] = m.f[4+i] * -1.0f;
  }
  return mr;
}

//////////// UpVektorV4 ////////////////////////
vec4 UpVektorV4(float X,float Y,float Z,float Radians){
  float RotY,RotZ,Distance;
  vec4 Up;
  //Z Rotation Berechnen
  Distance = sqrt(X * X + Y * Y);
  if(Distance != 0){
    if(Y >= 0)
      RotZ = acos( X / Distance);
    else
      RotZ = 2 * M_PI - acos( X / Distance);
    X = Distance;
    Y = 0;
  }
  else {
    RotZ=0;
  }
  //Y Rotation Berechnen
  Distance = sqrt( X * X + Z * Z);
  if(Distance != 0)
    RotY = acos( X / Distance);
  else
    RotY = 0;
  //Up Point berechnen
  X = 0;
  Y = sin(Radians);
  Z = cos(Radians);
  //Um Y Achse drehen
  if(RotY != 0){
    X = sin(RotY) * Z;
    Z = cos(RotY) * Z;
  }
  //Um Z Achse drehen
  if(RotZ != 0){
    Distance = sqrt(X * X + Y * Y);
    if(Distance != 0){
      if(Y >= 0)
        RotZ += acos(X / Distance);
      else
        RotZ += 2 * M_PI - acos(X / Distance);
      X = cos(RotZ) * Distance;
      Y = sin(RotZ) * Distance;
    }
  }
  Up.x = X;
  Up.y = Y;
  Up.z = Z;
  Up = NormalizeV4(Up);
  Up.w = 1.0f;
  return Up;
}

//Quaternion Functions
#ifdef __SSE__
vec4 NormalizeQ(const vec4& q){
  float length;
  vec4 temp,res;
  temp.m = _mm_mul_ps(q.m,q.m);
  length = sqrt(temp.x + temp.y + temp.z + temp.w);
  if(length != 0){
    temp.m = _mm_set_ps(length,length,length,length);
    res.m = _mm_div_ps(q.m,temp.m);
  }
  return res;
}
#else
vec4 NormalizeQ(const vec4& q){
  float length;
  vec4 res;
  length = sqrt(q.x * q.x + q.y * q.y + q.z * q.z + q.w * q.w);
  if(length != 0){
    res.x = q.x / length;
    res.y = q.y / length;
    res.z = q.z / length;
    res.w = q.w / length;
  }
  return res;
}
#endif

vec4 CreateRotationQ(const vec4& axis, float angle){
  float temp;
  vec4 res;
  angle /= 2;
  temp = sin(angle);
  res.x = axis.x * temp;
  res.y = axis.y * temp;
  res.z = axis.z * temp;
  res.w = cos(angle);
  return res;
}

vec4 MulQ(const vec4& q1, const vec4& q2){
  vec4 res;
  res.w = q1.w * q2.w - q1.x * q2.x - q1.y * q2.y - q1.z * q2.z;
  res.x = q1.w * q2.x + q1.x * q2.w - q1.y * q2.z - q1.z * q2.y;
  res.y = q1.w * q2.y + q1.y * q2.w + q1.z * q2.x - q1.x * q2.z;
  res.z = q1.w * q2.z * q1.z * q2.w + q1.x * q2.y - q1.y * q2.x;
  return res;
}

vec4 InverseQ(const vec4& q){
  vec4 res;
  res.x = q.x * -1;
  res.y = q.y * -1;
  res.z = q.z * -1;
  res.w = q.w;
  return res;
}

mat4a QToMat4a(const vec4& q){
  float mat[16];
  float xx = q.x * q.x;
  float xy = q.x * q.y;
  float xz = q.x * q.z;
  float xw = q.x * q.w;
  float yy = q.y * q.y;
  float yz = q.y * q.z;
  float yw = q.y * q.w;
  float zz = q.z * q.z;
  float zw = q.z * q.w;
  mat[0]  = 1 - 2 * ( yy + zz );
  mat[1]  =     2 * ( xy - zw );
  mat[2]  =     2 * ( xz + yw );
  mat[4]  =     2 * ( xy + zw );
  mat[5]  = 1 - 2 * ( xx + zz );
  mat[6]  =     2 * ( yz - xw );
  mat[8]  =     2 * ( xz - yw );
  mat[9]  =     2 * ( yz + xw );
  mat[10] = 1 - 2 * ( xx + yy );
  mat[3]  = mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0;
  mat[15] = 1;
  return AlignMatrix4(mat);
}

vec4 TangentV4(vec4 v1, vec4 v2, vec4 v3, vec2 t1, vec2 t2, vec2 t3){
  vec4 t;
  float div;
  v2.x = v2.x - v1.x;
  v2.y = v2.y - v1.y;
  v2.z = v2.z - v1.z;
  v3.x = v3.x - v1.x;
  v3.y = v3.y - v1.y;
  v3.z = v3.z - v1.z;
  //T
  div = (t2.x - t1.x) * (t3.y - t1.y) - (t3.x - t1.x) * (t2.y - t1.y);
  div = fabs(div);
  t.x = ((t3.y - t1.y) * v2.x - (t2.y - t1.y) * v3.x)/div;
  t.y = ((t3.y - t1.y) * v2.y - (t2.y - t1.y) * v3.y)/div;
  t.z = ((t3.y - t1.y) * v2.z - (t2.y - t1.y) * v3.z)/div;
  t.w = 1.0f;
  return t;
}

vec4 BinormalV4(vec4 v1, vec4 v2, vec4 v3, vec2 t1, vec2 t2, vec2 t3){
  vec4 b;
  float div;
  v2.x = v2.x - v1.x;
  v2.y = v2.y - v1.y;
  v2.z = v2.z - v1.z;
  v3.x = v3.x - v1.x;
  v3.y = v3.y - v1.y;
  v3.z = v3.z - v1.z;
  //B
  div = (t2.x - t1.x) * (t3.y - t1.y) - (t3.x - t1.x) * (t2.y - t1.y);
  div = fabs(div);
  b.x = ((-1)*(t3.x - t1.x) * v2.x + (t2.x - t1.x) * v3.x)/div;
  b.y = ((-1)*(t3.x - t1.x) * v2.y + (t2.x - t1.x) * v3.y)/div;
  b.z = ((-1)*(t3.x - t1.x) * v2.z + (t2.x - t1.x) * v3.z)/div;
  b.w = 1.0f;
  return b;
}

//////////// Plane Functions ///////////////////

#ifdef __SSE__
vec4 NormalizePlaneV4(const vec4& Plane){
  float length;
  vec4 temp,res;
  temp = Plane * Plane;
  length=sqrt(temp.f[0]+temp.f[1]+temp.f[2]);
  if(length != 0){
    temp.set(length);
    res= Plane / temp;
  }
  return res;
}
#else
vec4 NormalizePlaneV4(const vec4& Plane){
  float length = sqrt(Plane.x * Plane.x + Plane.y * Plane.y + Plane.z * Plane.z);
  vec4 res;
  if(length != 0){
    res.x = Plane.x / length;
    res.y = Plane.y / length;
    res.z = Plane.z / length;
    res.w = Plane.w / length;
  }
  return res;
}
#endif

float DistancePlaneV4(const vec4& Plane,const vec4& v){
  return Plane.x * v.x + Plane.y * v.y + Plane.z * v.z + Plane.w;
}

vec4 PlanePoint(const vec4& p1, const vec4& p2, const vec4& p3){
  float d;
  vec4 result;
  d = p1.x*p2.y*p3.z + p1.y*p2.z*p3.x + p1.z*p2.x*p3.y - p3.x*p2.y*p1.z - p3.y*p2.z*p1.x - p3.z*p2.x*p1.y;
  if(d!=0){
    result.x = p1.w*p2.y*p3.z + p1.y*p2.z*p3.w + p1.z*p2.w*p3.y - p3.w*p2.y*p1.z - p3.y*p2.z*p1.w - p3.z*p2.w*p1.y;
    result.y = p1.x*p2.w*p3.z + p1.w*p2.z*p3.x + p1.z*p2.x*p3.w - p3.x*p2.w*p1.z - p3.w*p2.z*p1.x - p3.z*p2.x*p1.w;
    result.z = p1.x*p2.y*p3.w + p1.y*p2.w*p3.x + p1.w*p2.x*p3.y - p3.x*p2.y*p1.w - p3.y*p2.w*p1.x - p3.w*p2.x*p1.y;
    result.w = 1.0f;

    result.x /= d;
    result.y /= d;
    result.z /= d;
  }

  return result;
}

//////////// Old Functions /////////////////////
vertex UpVektor(float X,float Y,float Z,float Radians){
  float RotY,RotZ,Distance;
  vertex Up;
  //Z Rotation Berechnen
  Distance = sqrt(X * X + Y * Y);
  if(Distance != 0){
    if(Y >= 0)
      RotZ = acos( X / Distance);
    else
      RotZ = 2 * M_PI - acos( X / Distance);
    X = Distance;
    Y = 0;
  }
  else {
    RotZ=0;
  }
  //Y Rotation Berechnen
  Distance = sqrt( X * X + Z * Z);
  if(Distance != 0)
    RotY = acos( X / Distance);
  else
    RotY = 0;
  //Up Point berechnen
  X = 0;
  Y = sin(Radians);
  Z = cos(Radians);
  //Um Y Achse drehen
  if(RotY != 0){
    X = sin(RotY) * Z;
    Z = cos(RotY) * Z;
  }
  //Um Z Achse drehen
  if(RotZ != 0){
    Distance = sqrt(X * X + Y * Y);
    if(Distance != 0){
      if(Y >= 0)
        RotZ += acos(X / Distance);
      else
        RotZ += 2 * M_PI - acos(X / Distance);
      X = cos(RotZ) * Distance;
      Y = sin(RotZ) * Distance;
    }
  }
  Up.x = X;
  Up.y = Y;
  Up.z = Z;
  Up = normalize(Up);
  return Up;
}

vertex normalize(vertex vektor){
  float length;
  length=sqrt(vektor.x*vektor.x+vektor.y*vektor.y+vektor.z*vektor.z);
  if(length != 0){
    vektor.x=vektor.x/length;
    vektor.y=vektor.y/length;
    vektor.z=vektor.z/length;
  }
  return vektor;
}

