/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "vertexbuffer.h"

VertexBuffer::VertexBuffer(int options,GLenum Primitive,GLenum Index){
  int offset=3;
  int temp;

  glePrimitive = Primitive;
  gleIndexType = Index;
  bDuplicated = false;
  if(options & TH_VB_DYNAMIC)
    bDynamic = true;
  else
    bDynamic = false;

  if(gleIndexType != GL_UNSIGNED_SHORT && gleIndexType != GL_UNSIGNED_INT)
    gleIndexType = GL_UNSIGNED_SHORT;
  IndexBufferSizes=NULL;
  gluiIndexBuffer = 0;
  gluiVertexBuffer = 0;
  iNumIndexBuffers = 0;
  iUploadedDataSize = 0;

  for(int i=0;i<4;i++)
    TexCoordSize[i]=0;

  //Vertex Data
  lpcVertexData = 0;
  lpcColorData = 0;
  lpcNormalData = 0;
  for(int i=0;i<4;i++)
    lpcTexCoordData[i]=0;


  //Color Data
  if(TH_VB_COLOR & options){
    lpcColorData = lpcVertexData + offset * sizeof(float);
    offset+=4;
  }

  //Normal Data
  if(TH_VB_NORMAL & options){
    lpcNormalData = lpcVertexData + offset * sizeof(float);
    offset+=3;
  }

  //TexCoordData
  for(int i=0;i<4;i++){
    temp = TH_VB_TEXCOORD0_2 << (i*3);
    for(int j=2;j<5;j++){
      if(temp & options){
        lpcTexCoordData[i] = lpcVertexData + offset * sizeof(float);
        offset += j;
        TexCoordSize[i]=j;
        break;
      }
      temp = temp << 1;
    }
  }

  //Attribute Data
  for(int i=0;i<3;i++){
    temp = TH_VB_ATTRIBUTE0_1 << (i*4);
    for(int j=1;j<5;j++){
      if(temp & options){
        lpcAttributeData[i] = lpcVertexData + offset * sizeof(float);
        offset += j;
        AttributeSize[i]=j;
        break;
      }
      temp = temp << 1;
    }
  }

  iBytesPerVertex = offset * sizeof(float);
  IndexBufferBaseSizes = NULL;
  IndexBufferSizes = NULL;
  lpcIndexBufferPointer = NULL;
}

VertexBuffer::VertexBuffer(const VertexBuffer& source){
  gluiVertexBuffer = 0;
  gluiIndexBuffer = 0;
  bDuplicated = source.bDuplicated;
  bDynamic = source.bDynamic;
  glePrimitive = source.glePrimitive;
  gleIndexType = source.gleIndexType;
  iUploadedDataSize = source.iUploadedDataSize;

  memcpy(TexCoordSize,source.TexCoordSize,sizeof(short)*4);
  memcpy(AttributeSize,source.AttributeSize,sizeof(short)*3);
  lpcVertexData = source.lpcVertexData;
  lpcColorData = source.lpcColorData;
  lpcNormalData = source.lpcNormalData;
  memcpy(lpcTexCoordData,source.lpcTexCoordData,sizeof(char *)*4);
  memcpy(lpcAttributeData,source.lpcAttributeData,sizeof(char *)*3);
  memcpy(gliAttributeLocations,source.gliAttributeLocations,sizeof(GLint)*3);
  fBuffer = source.fBuffer;
  IndexBuffers.reserve(source.IndexBuffers.size());
  for(int i=0;i<IndexBuffers.size();i++){
    IndexBuffers[i] = new IndexBufferType;
    IndexBuffers[i]->reserve(source.IndexBuffers[i]->size());
    memcpy(&(*IndexBuffers[i])[0],&(*source.IndexBuffers[i])[0],sizeof(int)*source.IndexBuffers[i]->size());
  }

  bIndexed = source.bIndexed;
  bMultiIndexed = source.bMultiIndexed;
  iBytesPerVertex = source.iBytesPerVertex;
  iNumVertices = source.iNumVertices;
  iNumIndexBuffers = source.iNumIndexBuffers;

  if(iNumIndexBuffers > 0 && bIndexed){
    IndexBufferSizes = new int[iNumIndexBuffers];
    memcpy(IndexBufferSizes,source.IndexBufferSizes,sizeof(int)*iNumIndexBuffers);

    IndexBufferBaseSizes = new int[iNumIndexBuffers];
    memcpy(IndexBufferBaseSizes,source.IndexBufferBaseSizes,sizeof(int)*iNumIndexBuffers);

    lpcIndexBufferPointer = new char*[iNumIndexBuffers];
    memcpy(lpcIndexBufferPointer,source.lpcIndexBufferPointer,sizeof(char*)*iNumIndexBuffers);
  }
  else{
    IndexBufferSizes = NULL;
    IndexBufferBaseSizes = NULL;
    lpcIndexBufferPointer = NULL;
  }
}

VertexBuffer::~VertexBuffer(){
  fBuffer.clear();
  for(int i=0;i<IndexBuffers.size();i++){
    delete IndexBuffers[i];
  }
  IndexBuffers.clear();
  if(IndexBufferSizes != NULL)
    delete []IndexBufferSizes;
  if(gluiIndexBuffer != 0)
    glDeleteBuffers(1,&gluiIndexBuffer);
  if(gluiVertexBuffer != 0)
    glDeleteBuffers(1,&gluiIndexBuffer);
}

void VertexBuffer::AddVertexData(float *buffer, int len){
  for(int i=0;i<len;i++)
    fBuffer.push_back(buffer[i]);
}

void VertexBuffer::AddVertexData(vec4 data){
  fBuffer.push_back(data.x);
  fBuffer.push_back(data.y);
  fBuffer.push_back(data.z);
  fBuffer.push_back(data.w);
}

void VertexBuffer::AddVertexData(vec3 data){
  fBuffer.push_back(data.x);
  fBuffer.push_back(data.y);
  fBuffer.push_back(data.z);
}

void VertexBuffer::AddVertexData(vec2 data){
  fBuffer.push_back(data.x);
  fBuffer.push_back(data.y);
}

void VertexBuffer::AddVertexData(float data){
  fBuffer.push_back(data);
}

void VertexBuffer::AddIndexBuffer(){
  IndexBufferType *ptr;
  ptr = new IndexBufferType;
  IndexBuffers.push_back(ptr);
}

void VertexBuffer::AddIndexData(int Buffer, int Data){
  IndexBuffers[Buffer]->push_back(Data);
}

void VertexBuffer::AddIndexData(int Buffer, int *Array, int len){
  for(int i=0;i<len;i++)
    IndexBuffers[Buffer]->push_back(Array[i]);
}

bool VertexBuffer::UploadData(){
  int BufferSize;
  unsigned int uiTemp;
  unsigned short usTemp;
  char *Buffer,*lpcTemp,*lpcTemp2;
  if(IndexBufferSizes != NULL)
    delete IndexBufferSizes;

  if(gluiIndexBuffer != 0){
    glDeleteBuffers(1,&gluiIndexBuffer);
    gluiIndexBuffer=0;
  }

  if(gluiVertexBuffer != 0 && !bDynamic){
    glDeleteBuffers(1,&gluiVertexBuffer);
    gluiVertexBuffer=0;
  }

  if(IndexBuffers.size() <= 0){
    IndexBufferSizes = NULL;
    bIndexed = false;
  }
  else {
      bIndexed = true;
      //Mehrere Indexdaten?
      if(IndexBuffers.size() > 1)
        bMultiIndexed=true;
      else
        bMultiIndexed=false;
      iNumIndexBuffers = IndexBuffers.size();
      //Array fr Gre fr Indexbuffer anlegen
      IndexBufferSizes = new int[IndexBuffers.size()];
      BufferSize=0;
      for(int i=0;i<IndexBuffers.size();i++){
        IndexBufferSizes[i] = IndexBuffers[i]->size();
        BufferSize += IndexBuffers[i]->size();
      }

      //16 oder 32 Bit
      if(gleIndexType == GL_UNSIGNED_SHORT)
      BufferSize *= 2;
      else if(gleIndexType == GL_UNSIGNED_INT)
        BufferSize *= 4;

      //Temporren Buffer anelgen und fllen
      Buffer = new char[BufferSize];
      lpcIndexBufferPointer = new char*[BufferSize];
      lpcTemp = Buffer;
      lpcTemp2 = NULL;
      //16-Bit Indicies
      if(gleIndexType == GL_UNSIGNED_SHORT){
        for(int i=0;i<IndexBuffers.size();i++){
          lpcIndexBufferPointer[i] = lpcTemp2;
          for(int j=0;j<IndexBuffers[i]->size();j++){
            usTemp = (unsigned short)(*IndexBuffers[i])[j];
            memcpy(lpcTemp,&usTemp,2);
            lpcTemp += 2;
            lpcTemp2 += 2;
          }
        }
      }
      //32-Bit Indicies
      else if(gleIndexType == GL_UNSIGNED_INT){
        for(int i=0;i<IndexBuffers.size();i++){
          lpcIndexBufferPointer[i] = lpcTemp2;
          for(int j=0;j<IndexBuffers[i]->size();j++){
            uiTemp = (unsigned int)(*IndexBuffers[i])[j];
            memcpy(lpcTemp,&uiTemp,4);
            lpcTemp += 4;
            lpcTemp2 += 4;
          }
        }
      }

      //Indexbuffer erzeugen
      glGenBuffers(1,&gluiIndexBuffer);
      //Daten bertragen
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,gluiIndexBuffer);
      glBufferData(GL_ELEMENT_ARRAY_BUFFER,BufferSize,Buffer,GL_STATIC_DRAW);
      //Buffer lsen
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0);

      //Speicher freigeben
      delete []Buffer;
  }

  //Gre des Datenbuffers ermitteln
  if(fBuffer.size() % (iBytesPerVertex / sizeof(float))!=0)
    return false;

  iNumVertices = fBuffer.size() / (iBytesPerVertex / sizeof(float));

  BufferSize = iBytesPerVertex * iNumVertices;

  //Vertex Buffer erzeugen
  if(bDynamic){
    if(BufferSize > iUploadedDataSize || gluiVertexBuffer == 0){
      glDeleteBuffers(1,&gluiVertexBuffer);
      glGenBuffers(1,&gluiVertexBuffer);
      glBindBuffer(GL_ARRAY_BUFFER,gluiVertexBuffer);
      glBufferData(GL_ARRAY_BUFFER,BufferSize,&fBuffer[0],GL_DYNAMIC_DRAW);
      iUploadedDataSize = BufferSize;
      glBindBuffer(GL_ARRAY_BUFFER,0);
    }
    else{
      glBindBuffer(GL_ARRAY_BUFFER,gluiVertexBuffer);
      glBufferSubData(GL_ARRAY_BUFFER,0,BufferSize,&fBuffer[0]);
      glBindBuffer(GL_ARRAY_BUFFER,0);
    }
  }
  else {
    glGenBuffers(1,&gluiVertexBuffer);
    glBindBuffer(GL_ARRAY_BUFFER,gluiVertexBuffer);
    //Daten hochladen
    glBufferData(GL_ARRAY_BUFFER,BufferSize,&fBuffer[0],GL_STATIC_DRAW);
    iUploadedDataSize = BufferSize;
    //Buffer freigeben
    glBindBuffer(GL_ARRAY_BUFFER,0);
  }

  return true;
}

void VertexBuffer::ReserveVertexData(int num){
  fBuffer.reserve(num * (iBytesPerVertex / sizeof(float)));
}

void VertexBuffer::ReserveAdditionalVertexData(int num){
  fBuffer.reserve(fBuffer.size() + num * (iBytesPerVertex / sizeof(float)));
}

void VertexBuffer::Draw(int options){
  if(gluiVertexBuffer==0){
    printf("Vertex Buffer Warning: Data not yet uploaded!\n");
    return;
  }
  glBindBuffer(GL_ARRAY_BUFFER,gluiVertexBuffer);

  //Position
  glEnableClientState(GL_VERTEX_ARRAY);
  glVertexPointer(3,GL_FLOAT,iBytesPerVertex,lpcVertexData);

  if(TH_VB_COLOR & options && lpcColorData != 0){
    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(4,GL_FLOAT,iBytesPerVertex,lpcColorData);
  }

  if(TH_VB_NORMAL & options && lpcNormalData != 0){
    glNormalPointer(GL_FLOAT,iBytesPerVertex,lpcNormalData);
    glEnableClientState(GL_NORMAL_ARRAY);
  }

  if(TH_VB_TEXCOORD0 & options && lpcTexCoordData[0] != 0){
    glClientActiveTexture(GL_TEXTURE0);
    glTexCoordPointer(TexCoordSize[0],GL_FLOAT,iBytesPerVertex,lpcTexCoordData[0]);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  if(TH_VB_TEXCOORD1 & options && lpcTexCoordData[1] != 0){
    glClientActiveTexture(GL_TEXTURE1);
    glTexCoordPointer(TexCoordSize[1],GL_FLOAT,iBytesPerVertex,lpcTexCoordData[1]);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  if(TH_VB_TEXCOORD2 & options && lpcTexCoordData[2] != 0){
    glClientActiveTexture(GL_TEXTURE2);
    glTexCoordPointer(TexCoordSize[2],GL_FLOAT,iBytesPerVertex,lpcTexCoordData[2]);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  if(TH_VB_TEXCOORD3 & options && lpcTexCoordData[3] != 0){
    glClientActiveTexture(GL_TEXTURE3);
    glTexCoordPointer(TexCoordSize[3],GL_FLOAT,iBytesPerVertex,lpcTexCoordData[3]);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  if(TH_VB_ATTRIBUTE0 & options && lpcAttributeData[0] != 0){
    glEnableVertexAttribArray(gliAttributeLocations[0]);
    glVertexAttribPointer(gliAttributeLocations[0],AttributeSize[0],GL_FLOAT,false,iBytesPerVertex,lpcAttributeData[0]);
  }

  if(TH_VB_ATTRIBUTE1 & options && lpcAttributeData[1] != 0){
    glEnableVertexAttribArray(gliAttributeLocations[1]);
    glVertexAttribPointer(gliAttributeLocations[1],AttributeSize[1],GL_FLOAT,false,iBytesPerVertex,lpcAttributeData[1]);
  }

  if(TH_VB_ATTRIBUTE2 & options && lpcAttributeData[2] != 0){
    glEnableVertexAttribArray(gliAttributeLocations[2]);
    glVertexAttribPointer(gliAttributeLocations[2],AttributeSize[2],GL_FLOAT,false,iBytesPerVertex,lpcAttributeData[2]);
  }

  if(bIndexed && gluiIndexBuffer != 0){
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,gluiIndexBuffer);
    if(bMultiIndexed){
      glMultiDrawElements(glePrimitive,IndexBufferSizes,gleIndexType,(const GLvoid**)lpcIndexBufferPointer,iNumIndexBuffers);
    }
    else{
      glDrawElements(glePrimitive,IndexBufferSizes[0],gleIndexType,NULL);
    }
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0);
  }
  else{
    glDrawArrays(glePrimitive,0,iNumVertices);
  }

  //Client States deaktivieren
  glDisableClientState(GL_VERTEX_ARRAY);
  if(TH_VB_COLOR & options && lpcColorData != 0)
    glDisableClientState(GL_COLOR_ARRAY);
  if(TH_VB_NORMAL & options && lpcNormalData != 0)
    glDisableClientState(GL_NORMAL_ARRAY);
  if(TH_VB_TEXCOORD3 & options && lpcTexCoordData[3] != 0){
    glActiveTexture(GL_TEXTURE3);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }
  if(TH_VB_TEXCOORD2 & options && lpcTexCoordData[2] != 0){
    glActiveTexture(GL_TEXTURE2);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }
  if(TH_VB_TEXCOORD1 & options && lpcTexCoordData[1] != 0){
    glActiveTexture(GL_TEXTURE1);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }
  if(TH_VB_TEXCOORD0 & options && lpcTexCoordData[0] != 0){
    glActiveTexture(GL_TEXTURE0);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }
  if(TH_VB_ATTRIBUTE0 & options && lpcAttributeData[0] != 0)
    glDisableVertexAttribArray(gliAttributeLocations[0]);
  if(TH_VB_ATTRIBUTE1 & options && lpcAttributeData[1] != 0)
    glDisableVertexAttribArray(gliAttributeLocations[1]);
  if(TH_VB_ATTRIBUTE2 & options && lpcAttributeData[2] != 0)
    glDisableVertexAttribArray(gliAttributeLocations[2]);
}

void VertexBuffer::DrawRange(int options,int start, int size){
  if(gluiVertexBuffer==0){
    printf("Vertex Buffer Warning: Data not yet uploaded!\n");
    return;
  }
  glBindBuffer(GL_ARRAY_BUFFER,gluiVertexBuffer);

  //Position
  glEnableClientState(GL_VERTEX_ARRAY);
  glVertexPointer(3,GL_FLOAT,iBytesPerVertex,lpcVertexData);

  if(TH_VB_COLOR & options && lpcColorData != 0){
    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(4,GL_FLOAT,iBytesPerVertex,lpcColorData);
  }

  if(TH_VB_NORMAL & options && lpcNormalData != 0){
    glNormalPointer(GL_FLOAT,iBytesPerVertex,lpcNormalData);
    glEnableClientState(GL_NORMAL_ARRAY);
  }

  if(TH_VB_TEXCOORD0 & options && lpcTexCoordData[0] != 0){
    glClientActiveTexture(GL_TEXTURE0);
    glTexCoordPointer(TexCoordSize[0],GL_FLOAT,iBytesPerVertex,lpcTexCoordData[0]);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  if(TH_VB_TEXCOORD1 & options && lpcTexCoordData[1] != 0){
    glClientActiveTexture(GL_TEXTURE1);
    glTexCoordPointer(TexCoordSize[1],GL_FLOAT,iBytesPerVertex,lpcTexCoordData[1]);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  if(TH_VB_TEXCOORD2 & options && lpcTexCoordData[2] != 0){
    glClientActiveTexture(GL_TEXTURE2);
    glTexCoordPointer(TexCoordSize[2],GL_FLOAT,iBytesPerVertex,lpcTexCoordData[2]);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  if(TH_VB_TEXCOORD3 & options && lpcTexCoordData[3] != 0){
    glClientActiveTexture(GL_TEXTURE3);
    glTexCoordPointer(TexCoordSize[3],GL_FLOAT,iBytesPerVertex,lpcTexCoordData[3]);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  if(TH_VB_ATTRIBUTE0 & options && lpcAttributeData[0] != 0){
    glEnableVertexAttribArray(gliAttributeLocations[0]);
    glVertexAttribPointer(gliAttributeLocations[0],AttributeSize[0],GL_FLOAT,false,iBytesPerVertex,lpcAttributeData[0]);
  }

  if(TH_VB_ATTRIBUTE1 & options && lpcAttributeData[1] != 0){
    glEnableVertexAttribArray(gliAttributeLocations[1]);
    glVertexAttribPointer(gliAttributeLocations[1],AttributeSize[1],GL_FLOAT,false,iBytesPerVertex,lpcAttributeData[1]);
  }

  if(TH_VB_ATTRIBUTE2 & options && lpcAttributeData[2] != 0){
    glEnableVertexAttribArray(gliAttributeLocations[2]);
    glVertexAttribPointer(gliAttributeLocations[2],AttributeSize[2],GL_FLOAT,false,iBytesPerVertex,lpcAttributeData[2]);
  }

  if(bIndexed && gluiIndexBuffer != 0){
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,gluiIndexBuffer);
    if(bMultiIndexed){
      //glMultiDrawElements(glePrimitive,IndexBufferSizes,gleIndexType,(const GLvoid**)lpcIndexBufferPointer,iNumIndexBuffers);
    }
    else{
      glDrawRangeElements(glePrimitive,start,IndexBufferSizes[0],size,gleIndexType,NULL);
    }
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0);
  }
  else{
    glDrawArrays(glePrimitive,start,size);
  }

  //Client States deaktivieren
  glDisableClientState(GL_VERTEX_ARRAY);
  if(TH_VB_COLOR & options && lpcColorData != 0)
    glDisableClientState(GL_COLOR_ARRAY);
  if(TH_VB_NORMAL & options && lpcNormalData != 0)
    glDisableClientState(GL_NORMAL_ARRAY);
  if(TH_VB_TEXCOORD3 & options && lpcTexCoordData[3] != 0){
    glActiveTexture(GL_TEXTURE3);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }
  if(TH_VB_TEXCOORD2 & options && lpcTexCoordData[2] != 0){
    glActiveTexture(GL_TEXTURE2);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }
  if(TH_VB_TEXCOORD1 & options && lpcTexCoordData[1] != 0){
    glActiveTexture(GL_TEXTURE1);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }
  if(TH_VB_TEXCOORD0 & options && lpcTexCoordData[0] != 0){
    glActiveTexture(GL_TEXTURE0);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }
  if(TH_VB_ATTRIBUTE0 & options && lpcAttributeData[0] != 0)
    glDisableVertexAttribArray(gliAttributeLocations[0]);
  if(TH_VB_ATTRIBUTE1 & options && lpcAttributeData[1] != 0)
    glDisableVertexAttribArray(gliAttributeLocations[1]);
  if(TH_VB_ATTRIBUTE2 & options && lpcAttributeData[2] != 0)
    glDisableVertexAttribArray(gliAttributeLocations[2]);
}

void VertexBuffer::DrawMultiple(){
  if(bIndexed && gluiIndexBuffer != 0){
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,gluiIndexBuffer);
    if(bMultiIndexed){
      glMultiDrawElements(glePrimitive,IndexBufferSizes,gleIndexType,(const GLvoid**)lpcIndexBufferPointer,iNumIndexBuffers);
    }
    else{
      glDrawElements(glePrimitive,IndexBufferSizes[0],gleIndexType,NULL);
    }
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0);
  }
  else{
    glDrawArrays(glePrimitive,0,iNumVertices);
  }
}

void VertexBuffer::FreeLocalData(){
  for(int i=0;i<IndexBuffers.size();i++)
    delete IndexBuffers[i];
  IndexBuffers.clear();
  fBuffer.clear();
}

void VertexBuffer::ResetLocalData(){
  fBuffer.resize(0);
}

void VertexBuffer::FreeGlobalData(){
  if(gluiVertexBuffer != 0){
    glDeleteBuffers(1,&gluiVertexBuffer);
    gluiVertexBuffer=0;
  }

  if(gluiIndexBuffer != 0){
    glDeleteBuffers(1,&gluiIndexBuffer);
    gluiIndexBuffer=0;
  }
}

unsigned int VertexBuffer::GetDataSize(){
  unsigned int size=0;
  if(gluiVertexBuffer == 0)
    return 0;
  size += iBytesPerVertex * iNumVertices;
  for(int i=0;i<IndexBuffers.size();i++){
    if(gleIndexType == GL_UNSIGNED_SHORT)
      size += 2 * IndexBufferSizes[i];
    else if(gleIndexType == GL_UNSIGNED_INT)
      size += 4 * IndexBufferSizes[i];
  }
  return size;
}

unsigned int VertexBuffer::GetVerticesInBuffer(){
  return (unsigned int)(fBuffer.size() / (iBytesPerVertex / 4));
}

void VertexBuffer::SetAttributeLocation(int num, GLint Attribute){
  if (num >= 0 && num <= 2)
    gliAttributeLocations[num]=Attribute;
}

#include <stdio.h>

bool VertexBuffer::Duplicate(int count){
    if(bDuplicated)
      return false;
    bDuplicated = true;

    float *TempArray;
    int i,j;
    int step = iBytesPerVertex / sizeof(float) + 1;
    TempArray = new float[iNumVertices*step];
    j=0;

    for(i=0;i<iNumVertices*step;i++){
      if((i+1) % step == 0){
        TempArray[i]=0.0f;
        //printf("a: %f\n",0.0f);
      }
      else{
        //printf("c: %f ",fBuffer[j]);
        TempArray[i]=fBuffer[j];
        j++;
      }
    }

    fBuffer.clear();
    fBuffer.reserve(iNumVertices*step*count);

    for(i=0;i<count;i++){
      for(j=0;j<iNumVertices*step;j++){
        if((j+1) % step == 0)
          TempArray[j]=(float)i;
      }
      fBuffer.insert(fBuffer.end(),TempArray,TempArray + (iNumVertices * step));
    }

    //Index Buffer vervielfltigen
    if(IndexBuffers.size() > 0){
      IndexBufferBaseSizes = new int[IndexBuffers.size()];
      for(i=0;i<IndexBuffers.size();i++){
        IndexBuffers[i]->reserve(IndexBuffers[i]->size() * count);
        for(j=0;j<count-1;j++){
          for(int z=0;z<IndexBufferSizes[i];z++)
            IndexBuffers[i]->push_back((*IndexBuffers[i])[z]+iNumVertices);
        }
        IndexBufferBaseSizes[i] = IndexBufferSizes[i];
        IndexBufferSizes[i] *= count;
      }
    }

    lpcAttributeData[2] = lpcVertexData + (step-1) * sizeof(float);
    AttributeSize[2]=1;
    iBytesPerVertex += sizeof(float);
    iInstancingBaseSize = iNumVertices;
    iNumVertices *= count;
    delete []TempArray;
}

void VertexBuffer::InitInstancedDraw(int options){
  iInstancedOptions = options;
  glBindBuffer(GL_ARRAY_BUFFER,gluiVertexBuffer);

  //Position
  glEnableClientState(GL_VERTEX_ARRAY);
  glVertexPointer(3,GL_FLOAT,iBytesPerVertex,lpcVertexData);

  if(TH_VB_COLOR & options && lpcColorData != 0){
    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(4,GL_FLOAT,iBytesPerVertex,lpcColorData);
  }

  if(TH_VB_NORMAL & options && lpcNormalData != 0){
    glNormalPointer(GL_FLOAT,iBytesPerVertex,lpcNormalData);
    glEnableClientState(GL_NORMAL_ARRAY);
  }

  if(TH_VB_TEXCOORD0 & options && lpcTexCoordData[0] != 0){
    glClientActiveTexture(GL_TEXTURE0);
    glTexCoordPointer(TexCoordSize[0],GL_FLOAT,iBytesPerVertex,lpcTexCoordData[0]);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  if(TH_VB_TEXCOORD1 & options && lpcTexCoordData[1] != 0){
    glClientActiveTexture(GL_TEXTURE1);
    glTexCoordPointer(TexCoordSize[1],GL_FLOAT,iBytesPerVertex,lpcTexCoordData[1]);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  if(TH_VB_TEXCOORD2 & options && lpcTexCoordData[2] != 0){
    glClientActiveTexture(GL_TEXTURE2);
    glTexCoordPointer(TexCoordSize[2],GL_FLOAT,iBytesPerVertex,lpcTexCoordData[2]);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  if(TH_VB_TEXCOORD3 & options && lpcTexCoordData[3] != 0){
    glClientActiveTexture(GL_TEXTURE3);
    glTexCoordPointer(TexCoordSize[3],GL_FLOAT,iBytesPerVertex,lpcTexCoordData[3]);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  if(TH_VB_ATTRIBUTE0 & options && lpcAttributeData[0] != 0){
    glEnableVertexAttribArray(gliAttributeLocations[0]);
    glVertexAttribPointer(gliAttributeLocations[0],AttributeSize[0],GL_FLOAT,false,iBytesPerVertex,lpcAttributeData[0]);
  }

  if(TH_VB_ATTRIBUTE1 & options && lpcAttributeData[1] != 0){
    glEnableVertexAttribArray(gliAttributeLocations[1]);
    glVertexAttribPointer(gliAttributeLocations[1],AttributeSize[1],GL_FLOAT,false,iBytesPerVertex,lpcAttributeData[1]);
  }

  if(TH_VB_ATTRIBUTE2 & options && lpcAttributeData[2] != 0){
    glEnableVertexAttribArray(gliAttributeLocations[2]);
    glVertexAttribPointer(gliAttributeLocations[2],AttributeSize[2],GL_FLOAT,false,iBytesPerVertex,lpcAttributeData[2]);
  }
}

void VertexBuffer::DeInitInstancedDraw(){
  int options = iInstancedOptions;

  //Client States deaktivieren
  glDisableClientState(GL_VERTEX_ARRAY);
  if(TH_VB_COLOR & options && lpcColorData != 0)
    glDisableClientState(GL_COLOR_ARRAY);
  if(TH_VB_NORMAL & options && lpcNormalData != 0)
    glDisableClientState(GL_NORMAL_ARRAY);
  if(TH_VB_TEXCOORD3 & options && lpcTexCoordData[3] != 0){
    glActiveTexture(GL_TEXTURE3);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }
  if(TH_VB_TEXCOORD2 & options && lpcTexCoordData[2] != 0){
    glActiveTexture(GL_TEXTURE2);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }
  if(TH_VB_TEXCOORD1 & options && lpcTexCoordData[1] != 0){
    glActiveTexture(GL_TEXTURE1);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }
  if(TH_VB_TEXCOORD0 & options && lpcTexCoordData[0] != 0){
    glActiveTexture(GL_TEXTURE0);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }
  if(TH_VB_ATTRIBUTE0 & options && lpcAttributeData[0] != 0)
    glDisableVertexAttribArray(gliAttributeLocations[0]);
  if(TH_VB_ATTRIBUTE1 & options && lpcAttributeData[1] != 0)
    glDisableVertexAttribArray(gliAttributeLocations[1]);
  if(TH_VB_ATTRIBUTE2 & options && lpcAttributeData[2] != 0)
    glDisableVertexAttribArray(gliAttributeLocations[2]);

  iInstancedOptions = 0;
}

void VertexBuffer::DrawInstancedFallback(int count){
  if(count * iInstancingBaseSize > iNumVertices)
    return;

  if(bIndexed && gluiIndexBuffer != 0){
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,gluiIndexBuffer);
    if(!bMultiIndexed)
      glDrawElements(glePrimitive,IndexBufferBaseSizes[0] * count,gleIndexType,NULL);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0);
  }
  else{
    glDrawArrays(glePrimitive,0,iInstancingBaseSize * count);
  }
}

void VertexBuffer::DrawInstanced(int count){
  if(bIndexed && gluiIndexBuffer != 0){
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,gluiIndexBuffer);
    if(!bMultiIndexed)
      glDrawElementsInstancedEXT(glePrimitive,IndexBufferSizes[0] * count,gleIndexType,NULL,count);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0);
  }
  else
    glDrawArraysInstancedEXT(glePrimitive,0,iNumVertices,count);
}

void VertexBuffer::GetMinMax(vec4& Min, vec4& Max){
  int stride = iBytesPerVertex / sizeof(float);
  memcpy(Min.f,&fBuffer[0],sizeof(float)*3);
  memcpy(Max.f,&fBuffer[0],sizeof(float)*3);
  for(int i=1;i<iNumVertices;i++){
    vec4 *position = (vec4*)&fBuffer[i*stride];
    Min.x = (position->x < Min.x) ? position->x : Min.x;
    Min.y = (position->y < Min.y) ? position->y : Min.y;
    Min.z = (position->z < Min.z) ? position->z : Min.z;

    Max.x = (position->x > Max.x) ? position->x : Max.x;
    Max.y = (position->y > Max.y) ? position->y : Max.y;
    Max.z = (position->z > Max.z) ? position->z : Max.z;
  }
}
