/*************************************************************
Copyright (c) 2008

Benjamin Thaut. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, this list of conditions
     and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
     and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY BENJAMIN THAUT "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

web: www.benjamin-thaut.de
*************************************************************/
#include "thEngine.h"
#include "xmlshader.h"
#include <string>
#include <sstream>

ScriptVarVector XMLShader::OptionStack;
std::vector<XMLShaderConstantReference> XMLShader::ShaderConstants;

XMLShader::XMLShader(){
  theShader = NULL;
}

XMLShader::~XMLShader(){
  if(theShader != NULL)
    delete theShader;
}

bool XMLShader::GetOption(const char *name, float& value){
  for(int i=0;i<OptionStack.size();i++){
    if(strcmp(name,OptionStack[i].name)==0){
      if(OptionStack[i].type == TH_FLOAT)
        value = OptionStack[i].fData;
      else
        value = (float)OptionStack[i].iData;
      return true;
    }
  }
  return false;
}

bool XMLShader::GetOption(const char *name, int& value){
  for(int i=0;i<OptionStack.size();i++){
    if(strcmp(name,OptionStack[i].name)==0){
      if(OptionStack[i].type == TH_INT)
        value = OptionStack[i].iData;
      else
        value = (int)OptionStack[i].fData;
      return true;
    }
  }
  return false;
}

ScriptVar* XMLShader::GetOption(const char *name){
  for(int i=0;i<OptionStack.size();i++){
    if(strcmp(name,OptionStack[i].name)==0)
      return &OptionStack[i];
  }
  return NULL;
}

void XMLShader::AddOption(const char *name, int value){
  ScriptVar *ptr = GetOption(name);
  ScriptVar TempVar;
  if(ptr == NULL){
    TempVar.name = new char[strlen(name)+1];
    strcpy(TempVar.name,name);
    TempVar.type = TH_INT;
    TempVar.iData = value;
    OptionStack.push_back(TempVar);
  }
  else {
    ptr->type = TH_INT;
    ptr->iData = value;
  }
}

void XMLShader::AddOption(const char *name, float value){
  ScriptVar *ptr = GetOption(name);
  ScriptVar TempVar;
  if(ptr == NULL){
    TempVar.name = new char[strlen(name)+1];
    strcpy(TempVar.name,name);
    TempVar.type = TH_FLOAT;
    TempVar.fData = value;
    OptionStack.push_back(TempVar);
  }
  else {
    ptr->type = TH_FLOAT;
    ptr->fData = value;
  }
}

bool XMLShader::LoadOptions(const char *filename){
  char meldung[200];
  FILE *datei;
  TiXmlDocument *XmlFile;
  float FloatVar;
  int IntVar;
  datei = fopen(filename,"r");
  if(!datei){
    sprintf(meldung,"Konnte %s nicht ffnen!",filename);
    ThErrorMessage("XML-Shader Error",meldung);
    return false;
  }
  fclose(datei);

  XmlFile = new TiXmlDocument(filename);
  if(!XmlFile->LoadFile()){
    ThErrorMessage("XML-Shader Error",XmlFile->ErrorDesc());
    delete XmlFile;
    return false;
  }

  TiXmlNode* pChild;
  TiXmlNode* pChild2;
  TiXmlNode* pChild3;
  TiXmlText* pText;
  int t,t2;

  for(pChild = XmlFile->FirstChild(); pChild != 0; pChild = pChild->NextSibling()){
    t=pChild->Type();
    switch(t){
      case TiXmlNode::ELEMENT:
        if(strcmp("options",pChild->Value())==0){
          for(pChild2 = pChild->FirstChild(); pChild2 != 0; pChild2 = pChild2->NextSibling()){
            t2 = pChild2->Type();
            switch(t2){
              case TiXmlNode::ELEMENT:
                pChild3 = pChild2->FirstChild();
                if(pChild3->Type() != TiXmlNode::TEXT){
                  sprintf(meldung,"Ungltiger Wert fr %s in <options>",pChild2->Value());
                  ThErrorMessage("XML-Shader Error",meldung);
                  delete XmlFile;
                  return false;
                }
                FloatVar = atof(pChild3->Value());
                IntVar = atoi(pChild3->Value());
                if((float)IntVar == FloatVar)
                  AddOption(pChild2->Value(),IntVar);
                else
                  AddOption(pChild2->Value(),FloatVar);
                break;
              default:
                break;
            }
          }
        }
        break;
      default:
        break;
    }
  }
  return true;
}

ShaderConstant* XMLShader::AddShaderConstantFloat(const char *name){
  XMLShaderConstantReference temp;
  temp.ptr = new ShaderConstantFloat();
  temp.name = name;
  ShaderConstants.push_back(temp);
  return temp.ptr;
}

ShaderConstant* XMLShader::AddShaderConstantVec3(const char *name){
  XMLShaderConstantReference temp;
  temp.ptr = new ShaderConstantVec3();
  temp.name = name;
  ShaderConstants.push_back(temp);
  return temp.ptr;
}

ShaderConstant* XMLShader::AddShaderConstantVec4(const char *name){
  XMLShaderConstantReference temp;
  temp.ptr = new ShaderConstantVec4();
  temp.name = name;
  ShaderConstants.push_back(temp);
  return temp.ptr;
}

ShaderConstant* XMLShader::AddShaderConstantMat4a(const char *name){
  XMLShaderConstantReference temp;
  temp.ptr = new ShaderConstantMat4a();
  temp.name = name;
  ShaderConstants.push_back(temp);
  return temp.ptr;
}

ShaderConstant* XMLShader::GetShaderConstant(const char *name){
  for(int i=0;i<ShaderConstants.size();i++)
    if(ShaderConstants[i].name == name)
      return ShaderConstants[i].ptr;
  return NULL;
}


bool XMLShader::ProgressShaderSource(const char* code, std::string& source){
  char Ergebnis[20];
  std::string NewSource=code;
  std::string Expression;
  std::string::size_type start,stop;
  ScriptVar TempVar;

  while(true){
    start=NewSource.find('%');
    if(start == std::string::npos)
      break;
    stop=NewSource.find('%',start+1);
    if(stop == std::string::npos){
      ThErrorMessage("XMLShader Error","Nichtgeschlossenes % im Shader Source");
      return false;
    }
    Expression = NewSource.substr(start+1,stop-start-1);
    if(!ParseExpression(Expression.c_str(),&OptionStack,&VarStack,TempVar)){
      ThErrorMessage("XMLShader Error","Fehlerhafte Expression im Shader Source!");
      return false;
    }
    if(TempVar.type == TH_FLOAT)
      sprintf(Ergebnis,"%f",TempVar.fData);
    else
      sprintf(Ergebnis,"%i",TempVar.iData);
    NewSource.replace(start,stop-start+1,Ergebnis,strlen(Ergebnis));
  }

  source += NewSource;
  return true;
}

bool XMLShader::GenerateShaderSource(std::string& source, TiXmlNode *father){
  TiXmlNode *pChild;
  TiXmlNode *pChild2;
  TiXmlAttribute *pAttrib;
  TiXmlElement *pElement;
  ScriptVar TempVar;
  std::string sInit,sCondition,sExpr;

  for(pChild = father->FirstChild();pChild != 0;pChild = pChild->NextSibling()){
    if(pChild->Type()==TiXmlNode::ELEMENT){
      //If-Tag
      if(strcmp("if",pChild->Value())==0){
        pElement = pChild->ToElement();
        pAttrib = pElement->FirstAttribute();
        if(strcmp("expr",pAttrib->Name())!=0){
          ThErrorMessage("XMLShader Error","Kein expr-Attribut fr if tag gefunden");
          return false;
        }
        else{
          if(!ParseExpression(pAttrib->Value(),&OptionStack,&VarStack,TempVar)){
            ThErrorMessage("XMLShader Error","Fehler in If-Expression");
            return false;
          }
          //Rekursion
          if(TempVar.iData == 1)
            if(!GenerateShaderSource(source,pChild))
              return false;
        }
      }
      //Source-Tag
      else if(strcmp("source",pChild->Value())==0){
        for(pChild2 = pChild->FirstChild();pChild2 != 0;pChild2 = pChild2->NextSibling()){
          if(pChild2->Type() == TiXmlNode::TEXT){
            if(!ProgressShaderSource(pChild2->Value(),source))
              return false;
          }
        }
      }
      //Expression-Tag
      else if(strcmp("expr",pChild->Value())==0){
        for(pChild2 = pChild->FirstChild();pChild2 != 0;pChild2 = pChild2->NextSibling()){
          if(pChild2->Type() == TiXmlNode::TEXT){
            if(!ParseAssigment(pChild2->Value(),&OptionStack,&VarStack)){
              ThErrorMessage("XMLShader Error","Fehler bei der Ausfhrung eines expression-tags");
              return false;
            }
          }
        }
      }
      //While-Tag
      else if(strcmp("while",pChild->Value())==0){
        pElement = pChild->ToElement();
        pAttrib = pElement->FirstAttribute();
        if(strcmp("expr",pAttrib->Name())!=0){
          ThErrorMessage("XMLShader Error","Kein expr-Attribut fr While-Tag gefunden!");
          return false;
        }
        else{
          while(true){
            if(!ParseExpression(pAttrib->Value(),&OptionStack,&VarStack,TempVar)){
              ThErrorMessage("XMLShader Error","Fehler in While-Expression");
              return false;
            }
            if(TempVar.iData != 1)
              break;
            if(!GenerateShaderSource(source,pChild))
              return false;
          }
        }
      }
      //For-Tag
      else if(strcmp("for",pChild->Value())==0){
        pElement = pChild->ToElement();
        sInit.clear(); sCondition.clear(); sExpr.clear();
        for(pAttrib=pElement->FirstAttribute(); pAttrib != 0; pAttrib = pAttrib->Next()){
          if(strcmp("init",pAttrib->Name())==0)
            sInit = pAttrib->Value();
          else if(strcmp("condition",pAttrib->Name())==0)
            sCondition = pAttrib->Value();
          else if(strcmp("expr",pAttrib->Name())==0)
            sExpr = pAttrib->Value();
        }
        if(sInit.empty()){
          ThErrorMessage("XMLShader Error","Kein init-Attribut fr for-tag");
          return false;
        }
        if(sCondition.empty()){
          ThErrorMessage("XMLShader Error","Kein Condition-Attribut fr for-tag");
          return false;
        }
        if(sExpr.empty()){
          ThErrorMessage("XMLShader Error","Kein expr-Attribut fr for-tag");
          return false;
        }
        if(!ParseAssigment(sInit.c_str(),&OptionStack,&VarStack)){
          ThErrorMessage("XMLShader Error","Fehler in init einer For-Schleife");
          return false;
        }
        while(true){
          if(!ParseExpression(sCondition.c_str(),&OptionStack,&VarStack,TempVar)){
            ThErrorMessage("XMLShader Error","Fehler in Condition einer For-Schleife");
            return false;
          }
          if(TempVar.iData == 0)
            break;
          if(!GenerateShaderSource(source,pChild))
            return false;
          if(!ParseAssigment(sExpr.c_str(),&OptionStack,&VarStack)){
            ThErrorMessage("XMLShader Error","Fehler in Expression einer For-Schleife");
            return false;
          }
        }
      }
    }
  }
  return true;
}

bool XMLShader::ProgressUniforms(TiXmlNode *father,std::string& NameStack,std::vector<int>& InitValues,std::vector<BindableUniformInfo>& UniformBuffers,int& num){
  TiXmlNode *pChild;
  TiXmlAttribute *pAttrib;
  TiXmlElement *pElement;
  BindableUniformInfo TempUniformBuffer;
  std::string sName,sInit,sType,sSize,sMode,sSpecial,sConstant;
  ScriptVar TempVar;
  ShaderConstant *TempShaderConstant;
  std::ostringstream ErrorMessage;

  for(pChild = father->FirstChild(); pChild != 0; pChild=pChild->NextSibling()){
    if(pChild->Type() == TiXmlNode::ELEMENT){
      //If-Tag
      if(strcmp("if",pChild->Value())==0){
        pElement = pChild->ToElement();
        pAttrib = pElement->FirstAttribute();
        if(strcmp("expr",pAttrib->Name())!=0){
          ThErrorMessage("XMLShader Error","Kein expr-Attribut fr if tag gefunden");
          return false;
        }
        else{
          if(!ParseExpression(pAttrib->Value(),&OptionStack,&VarStack,TempVar)){
            ThErrorMessage("XMLShader Error","Fehler in If-Expression");
            return false;
          }
          //Rekursion
          if(TempVar.iData == 1)
            if(!ProgressUniforms(pChild,NameStack,InitValues,UniformBuffers,num))
              return false;
        }
      }
    }
    //Uniforms
    if(strcmp("uniform",pChild->Value())==0){
      pElement = pChild->ToElement();
      sName.clear(); sInit.clear(); sType.clear(); sSize.clear(); sMode.clear(); sSpecial.clear(); sConstant.clear();
      for(pAttrib = pElement->FirstAttribute(); pAttrib != 0; pAttrib=pAttrib->Next()){
        if(strcmp("name",pAttrib->Name())==0)
          sName = pAttrib->Value();
        else if(strcmp("init",pAttrib->Name())==0)
          sInit = pAttrib->Value();
        else if(strcmp("type",pAttrib->Name())==0)
          sType = pAttrib->Value();
        else if(strcmp("size",pAttrib->Name())==0)
          sSize = pAttrib->Value();
        else if(strcmp("mode",pAttrib->Name())==0)
          sMode = pAttrib->Value();
        else if(strcmp("special",pAttrib->Name())==0)
          sSpecial = pAttrib->Value();
        else if(strcmp("constant",pAttrib->Name())==0)
          sConstant = pAttrib->Value();
        else {
          ErrorMessage << "Unbekanntes Attribut '" << pAttrib->Name() << "' fr uniform";
          ThErrorMessage("XMLShader Error",ErrorMessage.str().c_str());
          return false;
        }
      }
      if(sName.empty()){
        ThErrorMessage("XMLShader Error","Kein Name fr Uniform");
        return false;
      }
      if(!NameStack.empty())
        NameStack+="\n";
      NameStack+=sName;
      if(!sInit.empty())
        InitValues.push_back(atoi(sInit.c_str()));
      else
        InitValues.push_back(-1);
      //Bindable Uniforms
      if(!sType.empty()){
        if(sType != "bindable"){
          ThErrorMessage("XMLShader Error","Unbekannter Uniform Typ");
          return false;
        }
        else {
          if(!sInit.empty()){
            ThErrorMessage("XMLShader Error","Bindable Uniform mit Initialisierung ist nicht zulssig");
            return false;
          }
          if(!ParseExpression(sSize.c_str(),&OptionStack,&VarStack,TempVar)){
            ThErrorMessage("XMLShader Error","Fehler in Bindable Uniform Size Expression");
            return false;
          }
          if(TempVar.type != TH_INT || TempVar.iData <= 0){
            ThErrorMessage("XMLShader Error","Ungltige Gre fr Bindable Uniform");
            return false;
          }

          TempUniformBuffer.mode = GL_STATIC_DRAW;
          if(!sMode.empty()){
            if(sMode == "static")
              TempUniformBuffer.mode = GL_STATIC_DRAW;
            else if(sMode == "dynamic")
              TempUniformBuffer.mode = GL_DYNAMIC_DRAW;
            else{
              ErrorMessage << "Ungltiger Mode fr Bindable Uniform: '" << sMode.c_str() << "'";
              ThErrorMessage("XMLShader Error",ErrorMessage.str().c_str());
              return false;
            }
          }
          TempUniformBuffer.size = TempVar.iData;
          TempUniformBuffer.num = InitValues.size()-1;
          UniformBuffers.push_back(TempUniformBuffer);
        }
      }
      //Instancing
      if(!sSpecial.empty()){
        if(sSpecial == "instancing")
          theShader->SetInstanceUniform(num);
      }
      //Constants
      if(!sConstant.empty()){
        TempShaderConstant = GetShaderConstant(sConstant.c_str());
        if(TempShaderConstant == NULL){
          ErrorMessage << "Unbekannte default Konstante '" << sConstant.c_str() << "'";
          ThErrorMessage("XMLShader Error",ErrorMessage.str().c_str());
          return false;
        }
        theShader->AddShaderConstant(num,TempShaderConstant);
      }
      //Uniform Ende
      num++;
    }
  }
  return true;
}

bool XMLShader::ProgressAttributes(TiXmlNode *father,std::string& NameStack, int& num){
  TiXmlNode *pChild;
  TiXmlAttribute *pAttrib;
  TiXmlElement *pElement;
  std::string sName,sInit;
  ScriptVar TempVar;

  for(pChild = father->FirstChild(); pChild != 0; pChild=pChild->NextSibling()){
    if(pChild->Type() == TiXmlNode::ELEMENT){
      //If-Tag
      if(strcmp("if",pChild->Value())==0){
        pElement = pChild->ToElement();
        pAttrib = pElement->FirstAttribute();
        if(strcmp("expr",pAttrib->Name())!=0){
          ThErrorMessage("XMLShader Error","Kein expr-Attribut fr if tag gefunden");
          return false;
        }
        else{
          if(!ParseExpression(pAttrib->Value(),&OptionStack,&VarStack,TempVar)){
            ThErrorMessage("XMLShader Error","Fehler in If-Expression");
            return false;
          }
          //Rekursion
          if(TempVar.iData == 1)
            if(!ProgressAttributes(pChild,NameStack,num))
              return false;
        }
      }
    }
    //Uniforms
    if(strcmp("attribute",pChild->Value())==0){
      pElement = pChild->ToElement();
      sName.clear(); sInit.clear();
      for(pAttrib = pElement->FirstAttribute(); pAttrib != 0; pAttrib=pAttrib->Next()){
        if(strcmp("name",pAttrib->Name())==0)
          sName = pAttrib->Value();
        else if(strcmp("special",pAttrib->Name())==0){
          if(strcmp("instancing",pAttrib->Value())==0)
            theShader->SetInstanceAttribute(num);
        }
      }
      if(sName.empty()){
        ThErrorMessage("XMLShader Error","Kein Name fr Attribut");
        return false;
      }
      if(!NameStack.empty())
        NameStack+="\n";
      NameStack+=sName;
      num++;
    }
  }
  return true;
}

bool XMLShader::LoadFromFile(const char *lpcFileName){
  char meldung[200];
  FILE *datei;
  std::string VertexShaderSource;
  std::string GeometryShaderSource;
  std::string FragmentShaderSource;
  std::string Uniforms;
  std::string Attributes;
  std::vector<int> UniformsInit;
  std::vector<BindableUniformInfo> UniformBuffers;
  int iGSin=0,iGSout=0,iGSnum=0;
  ScriptVar TempVar;
  int iAttributeCount=0,iUniformCount=0;

  TiXmlDocument *XmlFile;
  TiXmlNode* pChild;
  TiXmlNode* pChild2;
  TiXmlNode* pChild3;
  TiXmlText* pText;
  int t;

  datei = fopen(lpcFileName,"r");
  if(datei == NULL){
    sprintf(meldung,"Konnte datei '%s' nicht ffnen!",lpcFileName);
    ThErrorMessage("XMLShader Error",meldung);
    return false;
  }
  fclose(datei);

  XmlFile = new TiXmlDocument(lpcFileName);
  XmlFile->SetCondenseWhiteSpace(false);
  if(!XmlFile->LoadFile()){
    ThErrorMessage("XMLShader Error",XmlFile->ErrorDesc());
    return false;
  }

  //Nach Shadername suchen
  for(pChild = XmlFile->FirstChild();pChild != 0;pChild = pChild->NextSibling()){
    if(pChild->Type() == TiXmlNode::ELEMENT){
      if(strcmp("Name",pChild->Value())==0){
        pChild2 = pChild->FirstChild();
        if(pChild2->Type() != TiXmlNode::TEXT){
          ThErrorMessage("XMLShader Error","Name Tag darf nur Text enthalten!");
          return false;
        }
        theShader = new Shader(pChild2->Value());
        break;
      }
    }
  }

  if(theShader == NULL){
    ThErrorMessage("XMLShader Error","Kein Name Tag fr Shader gefunden!");
    return false;
  }

  //Vertex Shader laden
  for(pChild = XmlFile->FirstChild();pChild != 0;pChild = pChild->NextSibling()){
    if(pChild->Type() == TiXmlNode::ELEMENT){
      if(strcmp("VertexShader",pChild->Value())==0){
        //printf("Found Vertex Shader Tag");
        if(!GenerateShaderSource(VertexShaderSource,pChild))
          return false;
      }
    }
  }

  //Geometry Shader Laden
  for(pChild = XmlFile->FirstChild();pChild != 0;pChild = pChild->NextSibling()){
    if(pChild->Type() == TiXmlNode::ELEMENT){
      if(strcmp("GeometryShader",pChild->Value())==0){
        if(!GenerateShaderSource(GeometryShaderSource,pChild))
          return false;

        for(pChild2 = pChild->FirstChild(); pChild2 != 0; pChild2 = pChild2->NextSibling()){
          if(pChild2->Type() == TiXmlNode::ELEMENT){
            //GS-Input
            if(strcmp("input",pChild2->Value())==0){
              pChild3 = pChild2->FirstChild();
              if(pChild3->Type() != TiXmlNode::TEXT){
                ThErrorMessage("XMLShader Error","Fehlerhafter Wert fr GeometryShader input");
                return false;
              }
              if(strcmp("points",pChild3->Value())==0)
                iGSin = GL_POINTS;
              else if(strcmp("lines",pChild3->Value())==0)
                iGSin = GL_LINES;
              else if(strcmp("lines_adjacency",pChild3->Value())==0)
                iGSin = GL_LINES_ADJACENCY_EXT;
              else if(strcmp("triangles",pChild3->Value())==0)
                iGSin = GL_TRIANGLES;
              else if(strcmp("triangles_adjacency",pChild3->Value())==0)
                iGSin = GL_TRIANGLES_ADJACENCY_EXT;
              else {
                ThErrorMessage("XMLShader Error","Unbekannter Wert fr GeometryShader input");
                return false;
              }
            }
            //GS-Output
            else if(strcmp("output",pChild2->Value())==0){
              pChild3 = pChild2->FirstChild();
              if(pChild3->Type() != TiXmlNode::TEXT){
                ThErrorMessage("XMLShader Error","Fehlerhafter Wert fr GeometryShader output");
                return false;
              }
              if(strcmp("points",pChild3->Value())==0)
                iGSout = GL_POINTS;
              else if(strcmp("line_strip",pChild3->Value())==0)
                iGSout = GL_LINE_STRIP;
              else if(strcmp("triangle_strip",pChild3->Value())==0)
                iGSout = GL_TRIANGLE_STRIP;
              else{
                ThErrorMessage("XMLShader Error","Unbekannter Wert fr GeometryShader output");
                return false;
              }
            }
            //GS-Num
            else if(strcmp("max_out",pChild2->Value())==0){
              pChild3 = pChild2->FirstChild();
              if(pChild3->Type() != TiXmlNode::TEXT){
                ThErrorMessage("XMLShader Error","Fehlerhafter Wert fr GeometryShader max_out");
                return false;
              }

              if(!ParseExpression(pChild3->Value(),&OptionStack,&VarStack,TempVar)){
                ThErrorMessage("XMLShader Error","Fehler in GeometryShader max_out expression");
                return false;
              }
              if(TempVar.type == TH_FLOAT)
                iGSnum = (int)TempVar.fData;
              else
                iGSnum = TempVar.iData;

              if(iGSnum <= 0){
                ThErrorMessage("XMLShader Error","GeometryShader max_out muss grer als 0 sein");
                return false;
              }
            }
          }
        }

      }
    }
  }

  //Fragment Shader Laden
  for(pChild = XmlFile->FirstChild();pChild != 0;pChild = pChild->NextSibling()){
    if(pChild->Type() == TiXmlNode::ELEMENT){
      if(strcmp("FragmentShader",pChild->Value())==0){
        if(!GenerateShaderSource(FragmentShaderSource,pChild))
          return false;
      }
    }
  }

  #ifdef TH_XML_SHADER_DEBUG
  sprintf(meldung,"%s-shader.log",theShader->GetName());
  datei = fopen(meldung,"w");
  if(VertexShaderSource.size() > 0)
    fprintf(datei,"Vertex Shader Source:\n%s\n\n",VertexShaderSource.c_str());
  if(GeometryShaderSource.size() > 0)
    fprintf(datei,"Geometry Shader Source:\n%s\n\n",GeometryShaderSource.c_str());
  if(FragmentShaderSource.size() > 0)
    fprintf(datei,"Fragment Shader Source:\n%s\n\n",FragmentShaderSource.c_str());
  fclose(datei);
  #endif

  if(VertexShaderSource.size() > 0)
    if(!theShader->LoadShaderSource(VertexShaderSource.c_str(),TH_VERTEX_SHADER)){
      ThErrorMessage("XMLShader Error","Fehler beim laden des Vertex Shaders");
      return false;
    }

  if(GeometryShaderSource.size() > 0){
    if(!GLEE_EXT_geometry_shader4 || !GLEE_EXT_gpu_shader4){
      ThErrorMessage("XMLShader Error","Kein Geometry Shader support!");
      return false;
    }
    if(!theShader->LoadShaderSource(GeometryShaderSource.c_str(),TH_GEOMETRY_SHADER)){
      ThErrorMessage("XMLShader Error","Fehler beim Laden des Geometry Shaders");
      return false;
    }
    if(iGSout == 0 || iGSin == 0 || iGSnum == 0){
      ThErrorMessage("XMLShader Error","Nicht alle ntigen GeomteryShader Parameter gesetzt!");
      return false;
    }
    theShader->SetGSParameter(TH_GS_INPUT,iGSin);
    theShader->SetGSParameter(TH_GS_OUTPUT,iGSout);
    theShader->SetGSParameter(TH_GS_OUTPUT_NUM,iGSnum);
  }

  if(FragmentShaderSource.size() > 0)
    if(!theShader->LoadShaderSource(FragmentShaderSource.c_str(),TH_FRAGMENT_SHADER)){
      ThErrorMessage("XMLShader Error","Fehler beim Laden des Fragment Shaders");
      return false;
    }

  if(!theShader->Link()){
    ThErrorMessage("XMLShader Error","Fehler beim Linken des Shaders");
    return false;
  }

  //Uniforms Laden
  for(pChild = XmlFile->FirstChild();pChild != 0;pChild = pChild->NextSibling()){
    if(pChild->Type() == TiXmlNode::ELEMENT){
      if(strcmp("Uniforms",pChild->Value())==0){
        if(!ProgressUniforms(pChild,Uniforms,UniformsInit,UniformBuffers,iUniformCount))
          return false;
      }
    }
  }

  //Attributes Laden
  for(pChild = XmlFile->FirstChild();pChild != 0;pChild = pChild->NextSibling()){
    if(pChild->Type() == TiXmlNode::ELEMENT){
      if(strcmp("Attributes",pChild->Value())==0){
        if(!ProgressAttributes(pChild,Attributes,iAttributeCount))
          return false;
      }
    }
  }

  if(!Uniforms.empty()){
    #ifdef TH_XML_SHADER_DEBUG
    printf("Uniforms: %s\n",Uniforms.c_str());
    #endif
    if(!theShader->LoadUniforms(Uniforms.c_str())){
      ThErrorMessage("XMLShader Error","Fehler beim laden der Uniforms");
      return false;
    }

    for(int i=0;i<UniformBuffers.size();i++){
      if(!theShader->GenBuffer(UniformBuffers[i].num,UniformBuffers[i].size,UniformBuffers[i].mode)){
        ThErrorMessage("XMLShader Error","Fehler beim erzeugen eines Bindable Uniform Buffers");
        return false;
      }
    }

    theShader->Use();

    for(int i=0;i<UniformsInit.size();i++){
      if(UniformsInit[i] != -1){
        theShader->SetUniform(i,UniformsInit[i]);
      }
    }

    UnloadShader();
  }

  if(!Attributes.empty()){
    #ifdef TH_XML_SHADER_DEBUG
    printf("Attributes: %s\n",Attributes.c_str());
    #endif
    if(!theShader->LoadAttributes(Attributes.c_str())){
      ThErrorMessage("XMLShader Error","Fehler beim laden der Attributes");
      return false;
    }
  }

  VarStack.clear();
  return true;
}

Shader *XMLShader::GetShader(){
  return theShader;
}
